header {* Syntax and Semantics of Assertions *}

theory Assertions imports Expressions begin


(*----------*)
(*- Syntax -*)
(*----------*)

datatype disj =
  D lit disj
| DFlse

datatype conj =
  C disj conj
| CTrue

datatype cnf =
  N conj ("\<forall>. _" [0] 999)


(*----------------*)
(*- Disjunctions -*)
(*----------------*)

fun domDisj :: "domain \<Rightarrow> domain \<Rightarrow> domain" (infixr "\<or>\<^sub>?" 75) where
  "True \<or>\<^sub>? _ = True"
| "False \<or>\<^sub>? b = b"
| "_ \<or>\<^sub>? True = True"
| "b \<or>\<^sub>? False = b"
| "_ \<or>\<^sub>? _ = \<bottom>"

theorem domDisj_classic [simp]:
  shows "a \<or>\<^sub>? b = (a \<or> b)"
  by (cases a) auto

theorem domDisj_false [simp]:
  shows "a \<or>\<^sub>? b = False \<longleftrightarrow> a = False \<and> b = False"
  by (cases "(a, b)" rule: domDisj.cases) auto

theorem domDisj_right_true [simp]:
  shows "b \<or>\<^sub>? True = True"
  by (cases b) auto

theorem domDisj_right_false [simp]:
  shows "b \<or>\<^sub>? False = b" 
  by (cases b) auto

theorem domDisj_true [simp]:
  shows "a \<or>\<^sub>? b = True \<longleftrightarrow> a = True \<or> b = True"
  by (cases "(a, b)" rule: domDisj.cases) auto

theorem domDisj_bot [simp]:
  assumes "a \<in> boolDom" and "b \<in> boolDom"
  shows "a \<or>\<^sub>? b = \<bottom> \<longleftrightarrow> a \<noteq> True \<and> b \<noteq> True \<and> (a = \<bottom> \<or> b = \<bottom>)"
  using assms by auto

theorem domDisj_commute:
  shows "a \<or>\<^sub>? b = b \<or>\<^sub>? a" 
  by (cases "(a, b)" rule: domDisj.cases) auto

theorem domDisj_assoc [simp]:
  shows "(a \<or>\<^sub>? b) \<or>\<^sub>? c = a \<or>\<^sub>? (b \<or>\<^sub>? c)"
  by (cases "(a, b)" rule: domDisj.cases, auto) (cases "(c, c)" rule: domDisj.cases, auto)+


(*----------------*)
(*- Disjunctions -*)
(*----------------*)

fun domConj :: "domain \<Rightarrow> domain \<Rightarrow> domain" (infixr "\<and>\<^sub>?" 80) where
  "False \<and>\<^sub>? _ = False"
| "True \<and>\<^sub>? b = b"
| "_ \<and>\<^sub>? False = False"
| "b \<and>\<^sub>? True = b"
| "_ \<and>\<^sub>? _ = \<bottom>"

theorem domConj_classic [simp]:
  shows "a \<and>\<^sub>? b = (a \<and> b)"
  by (cases a, auto)

theorem domConj_true [simp]:
  shows "a \<and>\<^sub>? b = True \<longleftrightarrow> a = True \<and> b = True"
  by (cases "(a, b)" rule: domConj.cases) auto

theorem domConj_right_true [simp]:
  shows "b \<and>\<^sub>? True = b"
  by (cases b) auto

theorem domConj_right_false [simp]:
  shows "b \<and>\<^sub>? False = False"
  by (cases b) auto

theorem domConj_false [simp]:
  shows "a \<and>\<^sub>? b = False \<longleftrightarrow> a = False \<or> b = False"
  by (cases "(a, b)" rule: domConj.cases) auto

theorem domConj_bot [simp]:
  assumes "a \<in> boolDom" and "b \<in> boolDom"
  shows "a \<and>\<^sub>? b = \<bottom> \<longleftrightarrow> a \<noteq> False \<and> b \<noteq> False \<and> (a = \<bottom> \<or> b = \<bottom>)"
  using assms by auto

theorem domConj_commute:
  shows "a \<and>\<^sub>? b = b \<and>\<^sub>? a"
  by (cases "(a, b)" rule: domConj.cases) auto

theorem domConj_assoc [simp]:
  shows "(a \<and>\<^sub>? b) \<and>\<^sub>? c = a \<and>\<^sub>? (b \<and>\<^sub>? c)"
  by (cases "(a, b)" rule: domConj.cases, auto) (cases "(c, c)" rule: domConj.cases, auto)+

theorem domConj_domDisj_deMorgan:
  shows "\<not>\<^sub>? (a \<and>\<^sub>? b) = (\<not>\<^sub>? a) \<or>\<^sub>? (\<not>\<^sub>? b)"
    and "\<not>\<^sub>? (a \<or>\<^sub>? b) = (\<not>\<^sub>? a) \<and>\<^sub>? (\<not>\<^sub>? b)"
  by (cases "(a, b)" rule: domConj.cases, auto)+


(* Semantics *)

instantiation disj and conj and cnf :: sem begin

  fun sem_disj :: "disj \<Rightarrow> env \<Rightarrow> domain" where
    "\<parallel>D l d\<parallel>\<^sub>E = \<parallel>l\<parallel>\<^sub>E \<or>\<^sub>? \<parallel>d\<parallel>\<^sub>E"
  | "\<parallel>DFlse\<parallel>\<^sub>E = False"

  fun sem_conj :: "conj \<Rightarrow> env \<Rightarrow> domain" where
    "\<parallel>C d c\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E \<and>\<^sub>? \<parallel>c\<parallel>\<^sub>E"
  | "\<parallel>CTrue\<parallel>\<^sub>E = True"

  fun sem_cnf_measure where
    "sem_cnf_measure (_, Env P f) = 1"
  | "sem_cnf_measure (_, Par P) = 0"

  function sem_cnf :: "cnf \<Rightarrow> env \<Rightarrow> domain" where
    "\<parallel>\<forall>. c\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<parallel>\<forall>. c\<parallel>\<^sub>R"
  | "\<parallel>\<forall>. c\<parallel>\<^sub>R = (if \<exists>f. \<parallel>c\<parallel>\<^sub>R\<^sub>+\<^sub>f = False then False else if \<forall>f. \<parallel>c\<parallel>\<^sub>R\<^sub>+\<^sub>f = True then True else \<bottom>)"
  by (pat_completeness) auto termination by (relation "measure sem_cnf_measure") auto

instance by intro_classes end

declare sem_cnf.simps(2)[simp del]

theorem sem_cnf_is_closed [simp]:
  fixes x :: cnf
  shows "P (\<parallel>x\<parallel>\<^sub>E\<^sub>+\<^sub>f) = P (\<parallel>x\<parallel>\<^sub>E)" 
  by (cases x) auto

theorem sem_cnf_closed:
  fixes x :: cnf assumes "P (\<parallel>x\<parallel>\<^sub>E\<^sub>R\<^sub>\<up>)"
  shows "P (\<parallel>x\<parallel>\<^sub>E)" 
  using assms by (cases x, cases E) auto


(* Boolean Expressions & Semantics *)

instantiation disj and conj and cnf :: boolExp begin

  definition bTrue_disj :: disj where
    "True.  \<equiv> D True. DFlse"
  definition bTrue_conj :: conj where
    "True.  \<equiv> CTrue"
  definition bTrue_cnf :: cnf where
    "True.  \<equiv> \<forall>. True."
  definition bFalse_disj :: disj where
    "False. \<equiv> DFlse"
  definition bFalse_conj :: conj where
    "False. \<equiv> C False. CTrue"
  definition bFalse_cnf :: cnf where
    "False. \<equiv> \<forall>. False."

instance by intro_classes end

declare bTrue_disj_def  [simp]
declare bTrue_conj_def  [simp]
declare bTrue_cnf_def   [simp]
declare bFalse_disj_def [simp]
declare bFalse_conj_def [simp]
declare bFalse_cnf_def  [simp]


(* Sub-Domain Theorems *)

theorem sem_disj_cases:
  fixes d :: disj
  shows "\<parallel>d\<parallel>\<^sub>E = \<bottom> \<or> \<parallel>d\<parallel>\<^sub>E = True \<or> \<parallel>d\<parallel>\<^sub>E = False"
  by (induct d) auto

corollary sdi [simp]:
  fixes d :: disj
  shows "\<parallel>d\<parallel>\<^sub>E \<in> boolDom"
  using sem_disj_cases[of d E] by auto

instantiation disj :: bool_sem begin instance
  by intro_classes auto end


theorem sem_conj_cases:
  fixes c :: conj
  shows "\<parallel>c\<parallel>\<^sub>E = \<bottom> \<or> \<parallel>c\<parallel>\<^sub>E = True \<or> \<parallel>c\<parallel>\<^sub>E = False"
  by (induct c) auto

corollary sco [simp]:
  fixes c :: conj
  shows "\<parallel>c\<parallel>\<^sub>E \<in> boolDom"
  using sem_conj_cases[of c E] by auto

instantiation conj :: bool_sem begin instance
  by intro_classes auto end


theorem sem_cnf_cases:
  fixes c :: cnf
  shows "\<parallel>c\<parallel>\<^sub>E = \<bottom> \<or> (\<exists>b. \<parallel>c\<parallel>\<^sub>E = b\<^sub>b)"
  by (cases "(c, E)" rule: sem_cnf.cases, auto, unfold sem_cnf.simps) metis+

corollary scn [simp]:
  fixes c :: cnf
  shows "\<parallel>c\<parallel>\<^sub>E \<in> boolDom"
  using sem_cnf_cases[of c E] by auto


(* CNF Semantics *)

corollary sem_cnf_falseD [dest!]:
  assumes "\<parallel>\<forall>. c\<parallel>\<^sub>E = False"
  shows "\<exists>f. \<parallel>c\<parallel>\<^sub>E\<^sub>R\<^sub>+\<^sub>f = False"
  using assms by (cases E, auto, unfold sem_cnf.simps) (split if_splits, auto)+

corollary sem_cnf_trueD [dest!]:
  assumes "\<parallel>\<forall>. c\<parallel>\<^sub>E = True"
  shows "\<forall>f. \<parallel>c\<parallel>\<^sub>E\<^sub>R\<^sub>+\<^sub>f = True"
  using assms by (cases E, auto, unfold sem_cnf.simps) (split if_splits, auto)+

corollary sem_cnf_botD [dest!]:
  assumes "\<parallel>\<forall>. c\<parallel>\<^sub>E = \<bottom>"
  shows "(\<exists>f. \<parallel>c\<parallel>\<^sub>E\<^sub>R\<^sub>+\<^sub>f = \<bottom>) \<and> (\<forall>f. \<parallel>c\<parallel>\<^sub>E\<^sub>R\<^sub>+\<^sub>f \<noteq> False)"
  using assms by (cases E, auto, unfold sem_cnf.simps) (metis sem_conj_cases, split if_splits, auto)+


corollary sem_cnf_falseI [intro!]:
  assumes "\<exists>f. \<parallel>c\<parallel>\<^sub>E\<^sub>R\<^sub>+\<^sub>f = False"
  shows "\<parallel>\<forall>. c\<parallel>\<^sub>E = False"
  using assms by - (rule sem_cnf_closed, unfold sem_cnf.simps, auto)

corollary sem_cnf_trueI [intro!]:
  assumes "\<And>f. \<parallel>c\<parallel>\<^sub>E\<^sub>R\<^sub>+\<^sub>f = True"
  shows "\<parallel>\<forall>. c\<parallel>\<^sub>E = True"
  using assms by - (rule sem_cnf_closed, unfold sem_cnf.simps, auto)

corollary sem_cnf_botI [intro!]:
  assumes "\<exists>f. \<parallel>c\<parallel>\<^sub>E\<^sub>R\<^sub>+\<^sub>f = \<bottom>" and "\<And>f. \<parallel>c\<parallel>\<^sub>E\<^sub>R\<^sub>+\<^sub>f \<noteq> False"
  shows "\<parallel>\<forall>. c\<parallel>\<^sub>E = \<bottom>"
  using assms by - (rule sem_cnf_closed, unfold sem_cnf.simps, auto, rule_tac x=f in exI, simp)

instantiation cnf :: bool_sem begin instance
  by intro_classes auto end


end
