header {* Code Equations, Data Type Refinement and Code Generation *}

theory Code imports Facts Story
  "~~/src/HOL/Library/Code_Integer"
  "~~/src/HOL/Library/Code_Char"
  "~~/src/HOL/Library/Code_Char_chr"
  "~~/src/HOL/Library/Efficient_Nat"
begin


(* Code Equations *)

(* Replacement constructor for datatype conj (C/CTrue): Conj, using lists *)

primrec Conj :: "disj list \<Rightarrow> conj" where
  "Conj []      = CTrue"
| "Conj (d # c) = C d (Conj c)"

code_datatype Conj

lemma conj_to_Conj [code]:
  "CTrue = Conj []"
  "C d (Conj c) = Conj (d # c)"
  by simp+


(* Replacement equations for functions on conj *)

(* Elementary *)

fun equal_Conj :: "disj list \<Rightarrow> disj list \<Rightarrow> bool" where
  "equal_Conj [] [] = True"
| "equal_Conj [] (_#_) = False"
| "equal_Conj (_#_) [] = False"
| "equal_Conj (a#t) (b#u) \<longleftrightarrow> a = b \<and> equal_Conj t u"

lemma Conj_inj:
  "inj Conj"
  by (rule injI, erule rev_mp) (induct_tac x y rule: equal_Conj.induct, auto)

lemma equal_Conj [code]:
   "equal_class.equal (Conj x) (Conj y) \<longleftrightarrow> x = y"
  using Conj_inj[THEN inj_eq] unfolding equal_conj_def .


(* Mapping *)

declare rev_conv_fold [code del]

lemma rev_foldl_general:
  "rev m @ n = foldl (\<lambda>l e. e # l) n m"
  by (induct m, auto) (metis fold_Cons_rev foldr_conv_fold foldr_conv_foldl rev_rev_ident)

lemma rev_foldl [code]:
  "rev = foldl (\<lambda>l e. e # l) []"
  by (rule ext, fold rev_foldl_general) simp

lemma xmap_Conj [code]:
  "xmap f (Conj l) = foldl (\<lambda>c d. xm f (C (xmap f d) c)) (xm f CTrue) (rev l)"
  by (induct l) auto

lemma xprop_Conj [code]:
  "xprop f (Conj (d # c)) \<longleftrightarrow> xp f (Conj (d # c)) \<and> xprop f d \<and> xprop f (Conj c)"
  "xprop f (Conj []) \<longleftrightarrow> xp f CTrue"
  by simp+


(* Sugar *)

lemma takeoutD [code]:
  "takeoutD d (Conj l) = Conj (filter (\<lambda>e. \<not> subsumes e d) l)"
  by (induct l) auto 

lemma Conj_conc [code]:
  "conj_conc (Conj (d # e)) c = sC d (conj_conc (Conj e) c)"
  "conj_conc (Conj []) c = c"
  by simp+

lemma disj_conc_Conj [code]:
  "disj_conc_conj d (Conj []) = CTrue"
  "disj_conc_conj d (Conj (e#c)) = (d \<or>. e) \<and>. disj_conc_conj d (Conj c)"
  by simp+


(* Conditionals *)

lemma pullC [code]:
  "pullC (Conj []) = CTrue"
  "pullC (Conj (d # c)) = pullD d \<and>. Conj c"
  by auto


(* Guards *)

lemma guard_Conj [code]:
  "sguard (Conj (d # c)) = sguard d \<and>. sguard (Conj c)"
  "sguard (Conj []) = CTrue"
  by auto


(* Simplifier *)

lemma inv_simp_Conj [code]:
  "inv_simp_conj (Conj []) x = x"
  "inv_simp_conj (Conj (d # c)) x = inv_simp_conj (Conj c) (takeoutD d x)"
 by simp+


(* Tail Recursive Map *)

fun rev_map where
  "rev_map f = foldl (\<lambda>l e. f e # l) []"

theorem rev_map_eq:
  "rev_map f l = rev (map f l)"
  by (induct l rule: rev_induct) auto

declare map.simps [code del]

lemma map_tail [code]:
  "map f l = rev (rev_map f l)"
  by (metis rev_map_eq rev_swap)


(* Fact Analysis Code Equations *)

lemma decideDC [code]:
  "decideD m (Conj [])   = CTrue"
  "decideD m (Conj (d#c)) = (
    if subsumes d m then decideD m (Conj c) else
    case decideD m (Conj c) of e \<Rightarrow>
    if e = False. then False. else
    C (case m of D l DFlse \<Rightarrow> decideL (\<not>. l) d | _ \<Rightarrow> d) e)"
  by simp+

lemma reducible [code]:
  "reducible (Conj (h#t)) x = (case h of
       D (L True (Equal a b)) DFlse \<Rightarrow> (if a = x \<and> notIn a b then b\<^sub>! else reducible (Conj t) x)
     | _ \<Rightarrow> reducible (Conj t) x)"
  "reducible (Conj []) x = None"
apply auto
apply (split disj.split, auto)
apply (split lit.split, auto)
apply (split bool.split, auto)
apply (split pred.split, auto)
apply (split disj.split, auto)
apply (split disj.split, auto)
apply (split disj.split, auto)
done

lemma rewrite_Conj [code]:
  "rewrite f x k (Conj [])   = CTrue"
  "rewrite f x k (Conj (d#c)) = (case rewrite f x k d of d' \<Rightarrow> sC d' (rewrite f x (f d' k) (Conj c)))"
  by simp+

lemma provable_fromC [code]:
  "provable_from (Conj (d # c)) l \<longleftrightarrow> d = False. \<or> d = toDisj l \<or> provable_from (Conj c) l \<or>
    option_case False (provable_from (Conj c)) (rewriteTrans (asFact d) l)"
  "provable_from (Conj []) l \<longleftrightarrow> False"
  by simp+

lemma smap_Conj [code]:
  "smap f (Conj (d#c)) = xm f (sC (smap f d) (smap f (Conj c)))"
  "smap f (Conj []) = xm f CTrue"
  by simp+


(* Taking care of a Char_ord issue *)

lemma [simp]:
  "m \<ge> 16 \<Longrightarrow> \<not> m \<le> nat_of_nibble n"
  by (metis less_le_trans nat_of_nibble_less_16 not_less)

lemma [simp]:
  "m \<ge> 16 \<Longrightarrow> \<not> m < nat_of_nibble n"
  by (metis less_asym less_le_trans nat_of_nibble_less_16)

lemma [code]:
  "c1 \<le> c2 \<longleftrightarrow> nat_of_char c1 \<le> nat_of_char c2"

apply auto
apply (cases c1, auto)
apply (cases c2, auto)
apply (simp add: nibble_less_def nat_of_char.simps)
apply ((case_tac nibble1, simp_all), (case_tac nibble2, simp_all)+)[1]
apply (simp add: nibble_less_eq_def nat_of_char.simps)

apply (cases c1, auto)
apply (cases c2, clarify)
apply (simp add: nibble_less_def nat_of_char.simps)
apply auto
apply (drule linorder_class.leI)
apply (case_tac nibble1, simp_all, ((case_tac nibble2, simp_all), (case_tac nibble1a, simp_all)+)+)[1]
apply (simp add: nibble_less_eq_def)
done

lemma [code]:
  "c1 < c2 \<longleftrightarrow> nat_of_char c1 < nat_of_char c2"

apply auto
apply (cases c1, auto)
apply (cases c2, auto)
apply (simp add: nat_of_char.simps nibble_less_def)
apply ((case_tac nibble1, simp_all), (case_tac nibble2, simp_all)+)[1]
apply (simp add: nibble_less_def nat_of_char.simps)

apply (cases c1, auto)
apply (cases c2, clarify)
apply (simp add: nibble_less_def nat_of_char.simps)
apply auto
apply (drule linorder_class.leI)
apply (case_tac nibble1, simp_all, ((case_tac nibble2, simp_all), (case_tac nibble1a, simp_all)+)+)[1]
done


(* Actual Code Generation *)

export_code wip inv_simp useEqs in Scala
  module_name XCend file "XCend.scala"


end

