header {* The Strictify Transformer *}

theory Conditionals imports Safe begin


(*---------------------------*)
(*- Conditional Expressions -*)
(*---------------------------*)

class condExp = xfold + kindOf +
fixes cond :: "lit \<Rightarrow> 'a :: {xfold, kindOf} \<Rightarrow> 'a \<Rightarrow> 'a"
  and sCond :: "lit \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> 'a" ("if. _ then _ else _ fi" [0, 0, 0] 1000)
  and bte :: "'a \<rightharpoonup> (lit \<times> 'a \<times> 'a)"
assumes bte_cond_def [simp]:
  "bte x = (b, t, e)\<^sub>! \<longleftrightarrow> x = cond b t e"
and cond_kind [simp]:
  "kindOf (cond b t e) = kindOf t"
and cond_conds [simp]:
  "conds (cond b t e) = 1 + conds b + conds t + conds e"

corollary bte_cond_inverse [simp]:
  shows "bte (cond b t e) = (b, t, e)\<^sub>!"
  using assms by simp

corollary cond_inject [simp]:
  shows "cond c u f = cond b t e \<longleftrightarrow> b = c \<and> t = u \<and> e = f"
  by auto (fold bte_cond_def, auto)


function topConds :: "'a :: condExp \<Rightarrow> nat" where
  "topConds x = (case bte x of (b, t, e)\<^sub>! \<Rightarrow> 1 + topConds t + topConds e | None \<Rightarrow> 0)"
  by (pat_completeness) auto termination by (relation "measure conds") auto

declare topConds.simps [simp del]


theorem topConds_bte_none [simp]:
  assumes "bte x = None"
  shows "topConds x = 0"
  using assms by (subst topConds.simps) simp

theorem topConds_cond [simp]:
  shows "topConds (cond b t e) = 1 + topConds t + topConds e"
  using assms by (subst topConds.simps) simp


theorem cond_induct [case_names None Cond]:
  assumes "\<And>x. bte x = None \<Longrightarrow> P x"
    and "\<And>b t e. \<lbrakk> P t; P e \<rbrakk> \<Longrightarrow> P (cond b t e)"
  shows "P y"
  using assms by - (rule topConds.induct, case_tac "bte x", metis, case_tac a, clarsimp, metis)


corollary topConds_le_conds [simp]:
  shows "topConds x \<le> conds x"
  by (induct x rule: cond_induct) auto


(* Instances *)

instantiation valueExp and pathExp and docExp and pred :: condExp begin

  definition cond_valueExp  where "cond \<equiv> VCond"
  definition cond_pathExp   where "cond \<equiv> PCond"
  definition cond_docExp    where "cond \<equiv> DCond"
  definition cond_pred      where "cond \<equiv> RCond"

  definition sCond_valueExp where "sCond \<equiv> sVCond"
  definition sCond_pathExp  where "sCond \<equiv> sPCond"
  definition sCond_docExp   where "sCond \<equiv> sDCond"
  definition sCond_pred     where "sCond \<equiv> sRCond"

  declare cond_valueExp_def [simp]
  declare cond_pathExp_def [simp]
  declare cond_docExp_def [simp]
  declare cond_pred_def [simp]

  declare sCond_valueExp_def [simp]
  declare sCond_pathExp_def [simp]
  declare sCond_docExp_def [simp]
  declare sCond_pred_def [simp]

  fun bte_valueExp where
    "bte (VCond b t e) = (b, t, e)\<^sub>!"
  | "bte_valueExp _ = None"

  fun bte_pathExp where
    "bte (PCond b t e) = (b, t, e)\<^sub>!"
  | "bte_pathExp _ = None"

  fun bte_docExp where
    "bte (DCond b t e) = (b, t, e)\<^sub>!"
  | "bte_docExp _ = None"

  fun bte_pred where
    "bte (RCond b t e) = (b, t, e)\<^sub>!"
  | "bte_pred _ = None"

instance by (intro_classes, auto) (case_tac x, auto)+ end


(* all conditionals are at the top *)

abbreviation ctop where
  "ctop x \<equiv> conds x = topConds x"

theorem conds_eq_ctop [dest!]:
  assumes "conds b + conds t + conds e = topConds t + topConds e"
  shows "conds b = 0 \<and> ctop t \<and> ctop e"
proof -
  from assms also have "... \<le> conds t + conds e" by (metis add_mono topConds_le_conds) finally
  have "conds b = 0" by auto moreover
  hence "conds t + conds e = topConds t + topConds e" using assms by (simp del: strict_conds)
  hence "conds t + conds e \<le> topConds t + conds e"
    and "conds t + conds e \<le> conds t + topConds e" by auto
  hence "conds t \<le> topConds t" and "conds e \<le> topConds e" by auto
  hence "ctop e" and "ctop t" by - (erule le_antisym, simp)+ ultimately
  show ?thesis by simp
qed


(* Semantics of Conditional Expressions *)

class cond_sem = condExp +
constrains x :: "'a :: {condExp, sem}"
assumes cond_sem [simp]:
  "\<parallel>cond b t e\<parallel>\<^sub>E = (if kindOf t = kindOf e then if\<^sub>? \<parallel>b\<parallel>\<^sub>E then \<parallel>t\<parallel>\<^sub>E else \<parallel>e\<parallel>\<^sub>E fi else \<bottom>)"

instantiation valueExp and pathExp and docExp and pred :: cond_sem begin
  instance by intro_classes auto end

lemmas cond_defs [simp del] = cond_valueExp_def cond_pathExp_def cond_docExp_def cond_pred_def
lemmas cond_folds [symmetric] = cond_defs


(* Summary of Assumptions for Conditionals *)

corollary cond_assumptions:
  fixes x :: "'a :: cond_sem"
  shows "bte x = (b, t, e)\<^sub>! \<longleftrightarrow> x = cond b t e"
    and "kindOf (cond b t e) = kindOf t"
    and "conds (cond b t e) = 1 + conds b + conds t + conds e"
    and "\<parallel>cond b t e\<parallel>\<^sub>E = (if kindOf t = kindOf e then if\<^sub>? \<parallel>b\<parallel>\<^sub>E then \<parallel>t\<parallel>\<^sub>E else \<parallel>e\<parallel>\<^sub>E fi else \<bottom>)"
  by simp+


(*---------------------------------------*)
(*- Pushing Functions into Conditionals -*)
(*---------------------------------------*)

function push :: "('a \<Rightarrow> 'b) \<Rightarrow> 'a :: condExp \<Rightarrow> 'b :: condExp" where
  "push f x = (case bte x of (b, t, e)\<^sub>! \<Rightarrow> cond b (push f t) (push f e) | None \<Rightarrow> f x)"
  by (pat_completeness) auto termination by (relation "measure (conds o snd)") auto

declare push.simps [simp del]

theorem push_bte_none [simp]:
  assumes "bte x = None"
  shows "push f x = f x"
  using assms by (subst push.simps) simp

theorem push_cond [simp]:
  shows "push f (cond b t e) = cond b (push f t) (push f e)"
  using assms by (subst push.simps) simp

definition push2 :: "('a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow> 'a :: condExp \<Rightarrow> 'b :: condExp \<Rightarrow> 'c :: condExp" where
  "push2 f x y \<equiv> push (\<lambda>a. push (f a) y) x"

definition push3 :: "('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd) \<Rightarrow> 'a :: condExp \<Rightarrow> 'b :: condExp \<Rightarrow> 'c :: condExp \<Rightarrow> 'd :: condExp" where
  "push3 f x y z \<equiv> push (\<lambda>a. push2 (f a) y z) x"


(* Kind Functions *)

definition kind_fun:  (* function is also a function on kinds, i.e., preserves the equivalence class *)
  "kind_fun f \<equiv> \<forall>a b. kindOf a = kindOf b \<longrightarrow> kindOf (f a) = kindOf (f b)"

theorem kind_fun_intro [intro!]:
  assumes "\<And>a b. kindOf a = kindOf b \<Longrightarrow> kindOf (f a) = kindOf (f b)"
  shows "kind_fun f"
  using assms by (metis kind_fun)

theorem kind_fun_apply [intro]:
  assumes "kind_fun f" and "kindOf a = kindOf b"
  shows "kindOf (f a) = kindOf (f b)"
  using assms by (metis kind_fun)


(* Kind of Push *)

theorem push_kind [simp]:
  assumes "kind_fun f"
  shows "kindOf (push f x) = kindOf (f x)"
  using assms by (induct x rule: cond_induct) auto

theorem push2_kind [simp]:
  assumes "\<And>x. kind_fun (f x)" and "\<And>y. kind_fun (\<lambda>x. f x y)"
  shows "kindOf (push2 f x y) = kindOf (f x y)"
  using assms unfolding push2_def by (subst push_kind, auto) (metis kind_fun_apply)


(* Semantics of Push *)

theorem push_sem [simp]:
  fixes f :: "'a :: cond_sem \<Rightarrow> 'b :: cond_sem"
  assumes "\<And>x. \<parallel>f x\<parallel>\<^sub>E = g (\<parallel>x\<parallel>\<^sub>E)" and "g \<bottom> = \<bottom>" and "kind_fun f" and "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>push f x\<parallel>\<^sub>E = g (\<parallel>x\<parallel>\<^sub>E)"
  using assms by (induct rule: cond_induct) (subst push.simps, auto)

theorem push_sem_kind [simp]:
  fixes f :: "'a :: cond_sem \<Rightarrow> 'b :: cond_sem"
  assumes "\<And>x. \<parallel>f x\<parallel>\<^sub>E = g (kindOf x) (\<parallel>x\<parallel>\<^sub>E)" and "\<And>x. g x \<bottom> = \<bottom>" and "kind_fun f" and "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>push f x\<parallel>\<^sub>E = g (kindOf x) (\<parallel>x\<parallel>\<^sub>E)"
  using assms by (induct rule: cond_induct) (subst push.simps, auto)

theorem push2_sem [simp]:
  fixes f :: "'a :: cond_sem \<Rightarrow> 'b :: cond_sem \<Rightarrow> 'c  :: cond_sem"
  assumes "\<And>x y. \<parallel>f x y\<parallel>\<^sub>E = g (\<parallel>x\<parallel>\<^sub>E) (\<parallel>y\<parallel>\<^sub>E)" and "\<And>y. g \<bottom> y = \<bottom>" "\<And>x. g x \<bottom> = \<bottom>"
    and "\<And>x. kind_fun (f x)" and "\<And>y. kind_fun (\<lambda>x. f x y)" and "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>" and "\<parallel>y\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>push2 f x y\<parallel>\<^sub>E = g (\<parallel>x\<parallel>\<^sub>E) (\<parallel>y\<parallel>\<^sub>E)"
  using assms unfolding push2_def by (subst push_sem, auto)+ (unfold kind_fun, metis)

theorem push2_sem_kind [simp]:
  fixes f :: "'a :: cond_sem \<Rightarrow> 'b :: cond_sem \<Rightarrow> 'c  :: cond_sem"
  assumes "\<And>x y. \<parallel>f x y\<parallel>\<^sub>E = g (kindOf x = kindOf y) (\<parallel>x\<parallel>\<^sub>E) (\<parallel>y\<parallel>\<^sub>E)"
    and "\<And>b y. g b \<bottom> y = \<bottom>" and "\<And>b x. g b x \<bottom> = \<bottom>"
    and "\<And>x. kind_fun (f x)" and "\<And>y. kind_fun (\<lambda>x. f x y)" and "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>" and "\<parallel>y\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>push2 f x y\<parallel>\<^sub>E = g (kindOf x = kindOf y) (\<parallel>x\<parallel>\<^sub>E) (\<parallel>y\<parallel>\<^sub>E)"
  using assms unfolding push2_def by (subst push_sem_kind, auto)+ (unfold kind_fun, metis)

theorem push3_sem [simp]:
  fixes f :: "'a :: cond_sem \<Rightarrow> 'b :: cond_sem \<Rightarrow> 'c  :: cond_sem \<Rightarrow> 'd :: cond_sem"
  assumes "\<And>x y z. \<parallel>f x y z\<parallel>\<^sub>E = g (\<parallel>x\<parallel>\<^sub>E) (\<parallel>y\<parallel>\<^sub>E) (\<parallel>z\<parallel>\<^sub>E)"
    and "\<And>y z. g \<bottom> y z = \<bottom>" and "\<And>x z. g x \<bottom> z = \<bottom>" and "\<And>x y. g x y \<bottom> = \<bottom>"
    and "\<And>x y. kind_fun (f x y)" and "\<And>y z. kind_fun (\<lambda>x. f x y z)" and "\<And>x z. kind_fun (\<lambda>y. f x y z)"
    and "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>" and "\<parallel>y\<parallel>\<^sub>E \<noteq> \<bottom>" and "\<parallel>z\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>push3 f x y z\<parallel>\<^sub>E = g (\<parallel>x\<parallel>\<^sub>E) (\<parallel>y\<parallel>\<^sub>E) (\<parallel>z\<parallel>\<^sub>E)"
  using assms unfolding push3_def by (subst push_sem, auto) (unfold kind_fun, metis) 


(* CTop for Pushs *)

theorem push_ctop [intro!]:
  assumes "\<And>x. strict x \<Longrightarrow> ctop (f x)" and "ctop x"
  shows "ctop (push f x)"
  using assms by (induct x rule: cond_induct) auto

theorem push2_ctop [intro!]:
  assumes "\<And>x y. \<lbrakk> strict x; strict y \<rbrakk> \<Longrightarrow> ctop (f x y)" and "ctop x" "ctop y"
  shows "ctop (push2 f x y)"
  using assms unfolding push2_def by auto

theorem push3_ctop [intro!]:
  assumes "\<And>x y z. \<lbrakk> strict x; strict y; strict z \<rbrakk> \<Longrightarrow> ctop (f x y z)" and "ctop x" "ctop y" "ctop z"
  shows "ctop (push3 f x y z)"
  using assms unfolding push3_def by auto


(*--------------------------------*)
(*- Pulling Conditionals Outside -*)
(*--------------------------------*)

fun theBool :: "lit \<Rightarrow> bool" where
  "theBool (L b p) = b"

fun thePred :: "lit \<Rightarrow> pred" where
  "thePred (L b p) = p"


fun pullVE :: "valueExp \<Rightarrow> valueExp" where
  "pullVE (Sum   m)     = push  Sum       m"
| "pullVE (Size  m)     = push  Size      m"
| "pullVE (Count m v)   = push2 Count     m v"
| "pullVE (Tally t m)   = push  (Tally t) m"
| "pullVE (Read  d p)   = push2 Read      d p"
| "pullVE (Scal  n v)   = push2 Scal      n v"
| "pullVE (VRep  n m)   = push2 VRep      n m"
| "pullVE (VPlus x y)   = push2 VPlus     x y"
| "pullVE (VMinus x y)  = push2 VMinus    x y"
| "pullVE (VCond b t e) = push (\<lambda>x. VCond (L (theBool b) x) t e) (thePred b)"
| "pullVE v = v"

fun pullPE :: "pathExp \<Rightarrow> pathExp" where
  "pullPE (Path  l p k) = push2 (Path l) p k"
| "pullPE (Kind  l p)   = push  (Kind l) p"
| "pullPE (PRep  n m)   = push2 PRep     n m"
| "pullPE (PPlus x y)   = push2 PPlus    x y"
| "pullPE (PMinus x y)  = push2 PMinus   x y"
| "pullPE (PCond b t e) = push (\<lambda>x. PCond (L (theBool b) x) t e) (thePred b)"
| "pullPE p = p"

fun pullDE :: "docExp \<Rightarrow> docExp" where
  "pullDE (Upd   d p v) = push3 Upd d p v"
| "pullDE (Del   d p)   = push2 Del d p"
| "pullDE (DCond b t e) = push (\<lambda>x. DCond (L (theBool b) x) t e) (thePred b)"
| "pullDE d = d"

fun pullRE :: "pred \<Rightarrow> pred" where
  "pullRE (Equal a b)   = push2 Equal a b"
| "pullRE (Less  a b)   = push2 Less  a b"
| "pullRE (RCond b t e) = push (\<lambda>x. RCond (L (theBool b) x) t e) (thePred b)"


(* Optimization by deciding some conditionals directly *)

definition decideCE :: "lit \<Rightarrow> 'a :: condExp \<Rightarrow> 'a :: condExp" where
  "decideCE l x \<equiv> (case bte x of
     (b, t, e)\<^sub>! \<Rightarrow> if kindOf t = kindOf e then if b = l then t else if b = \<not>. l then e else x else x | None \<Rightarrow> x)"

abbreviation decide where
  "decide l \<equiv> xmap (XMap (decideCE l) (decideCE l) (decideCE l) (decideCE l) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a))"

theorem decideCE_conds [simp]:
  "conds (decideCE l x) \<le> conds x"
  unfolding decideCE_def by (case_tac "bte x") auto

corollary decideCE_conds_leI [intro]:
  assumes "conds x \<le> n"
  shows "conds (decideCE l x) \<le> n"
  using assms by - (erule order_trans[rotated], simp)

corollary decideCE_conds_eqI [intro]:
  assumes "strict x"
  shows "conds (decideCE l x) = 0"
  using assms unfolding strict_conds[symmetric] by (metis decideCE_conds_leI le_0_eq)

theorem decide_conds_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "conds (decide x v) \<le> conds v"
    and "conds (decide x p) \<le> conds p"
    and "conds (decide x d) \<le> conds d"
    and "conds (decide x r) \<le> conds r"
    and "conds (decide x l) \<le> conds l"
  using assms by (induct v and p and d and r and l) (auto simp del: strict_conds)

theorem decide_conds_disj [simp]:
  fixes d :: disj
  shows "conds (decide l d) \<le> conds d"
  by (induct d) auto

corollary decide_conds_le_disjI [intro!]:
  fixes d :: disj
  assumes "conds d \<le> n"
  shows "conds (decide l d) \<le> n"
  using assms by - (erule order_trans[rotated], simp)

corollary decide_conds_lt_disjI [intro!]:
  fixes d :: disj
  assumes "conds d < n"
  shows "conds (decide l d) < n"
  using assms by - (erule xt1(7), simp)

corollary decide_conds_eq_disjI [intro!]:
  fixes d :: disj
  assumes "conds d = 0"
  shows "conds (decide l d) = 0"
  using assms by (metis decide_conds_disj less_nat_zero_code linorder_not_less nat_less_cases)


(* Remaining pull definitions *)

function (sequential) pullD :: "disj \<Rightarrow> conj" where
  "pullD (D (L b (RCond l t e)) d) = (l \<longrightarrow>. pullD (decide     l  (L b t \<or>. d)))
                                \<and>. (\<not>.l \<longrightarrow>. pullD (decide (\<not>. l) (L b e \<or>. d)))"
| "pullD (D l d) = \<not>.l \<longrightarrow>. pullD d"
| "pullD DFlse = False."
by pat_completeness auto
termination by (relation "measures [conds, size]") (auto intro!: le_imp_less_Suc)


fun pullC :: "conj \<Rightarrow> conj" where
  "pullC (C d c) = pullD d \<and>. c"
| "pullC CTrue = True."

abbreviation strictify where
  "strictify \<equiv> xmap (XMap pullVE pullPE pullDE pullRE (\<lambda>a. a) (\<lambda>a. a) pullC)"


(* Properties of decide *)

theorem decideCE_strict [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred
  shows "strict v \<Longrightarrow> strict (decideCE l v)"
    and "strict p \<Longrightarrow> strict (decideCE l p)"
    and "strict d \<Longrightarrow> strict (decideCE l d)"
    and "strict r \<Longrightarrow> strict (decideCE l r)"
  using assms unfolding decideCE_def
  by (case_tac "bte v", auto, fold strict_conds, auto, case_tac "bte p", auto, fold strict_conds, auto)
     (case_tac "bte d", auto, fold strict_conds, auto, case_tac "bte r", auto, fold strict_conds, auto)

theorem decide_strict_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "strict v \<Longrightarrow> strict (decide x v)"
    and "strict p \<Longrightarrow> strict (decide x p)"
    and "strict d \<Longrightarrow> strict (decide x d)"
    and "strict r \<Longrightarrow> strict (decide x r)"
    and "strict l \<Longrightarrow> strict (decide x l)"
  using assms by (induct v and p and d and r and l) auto

theorem decideCE_kind [simp]:
  shows "kindOf (decideCE l x) = kindOf x"
  unfolding decideCE_def by (case_tac "bte x") auto

theorem decide_kind [simp]:
  fixes p :: pathExp
  shows "kindOf (decide l p) = kindOf p"
  by (induct p) auto

theorem decideCE_ctop [intro!]:
  assumes "ctop x"
  shows "ctop (decideCE l x)"
  using assms unfolding decideCE_def by (case_tac "bte x") auto

theorem decide_ctop_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "ctop v \<Longrightarrow> ctop (decide x v)"
    and "ctop p \<Longrightarrow> ctop (decide x p)"
    and "ctop d \<Longrightarrow> ctop (decide x d)"
    and "ctop r \<Longrightarrow> ctop (decide x r)"
    and "ctop (thePred l) \<Longrightarrow> ctop (thePred (decide x l))"
  using assms by (induct v and p and d and r and l) (auto, unfold cond_folds, auto)

fun ctopD :: "disj \<Rightarrow> bool" where
  "ctopD (D l d) \<longleftrightarrow> ctop (thePred l) \<and> ctopD d"
| "ctopD DFlse \<longleftrightarrow> True"

theorem decide_ctop_disj [simp]:
  fixes d :: disj assumes "ctopD d"
  shows "ctopD (decide l d)"
  using assms by (induct d) auto


(* Semantics of decide *)

theorem decide_sem_exp_true [symmetric, simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  assumes "\<parallel>x\<parallel>\<^sub>E = True"
  shows "\<parallel>v\<parallel>\<^sub>E = \<parallel>decide x v\<parallel>\<^sub>E"
    and "\<parallel>p\<parallel>\<^sub>E = \<parallel>decide x p\<parallel>\<^sub>E"
    and "\<parallel>d\<parallel>\<^sub>E = \<parallel>decide x d\<parallel>\<^sub>E"
    and "\<parallel>r\<parallel>\<^sub>E = \<parallel>decide x r\<parallel>\<^sub>E"
    and "\<parallel>l\<parallel>\<^sub>E = \<parallel>decide x l\<parallel>\<^sub>E"
  using assms by (induct v and p and d and r and l) (auto, unfold decideCE_def, auto)

theorem decide_sem_exp_bot [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  assumes "\<parallel>x\<parallel>\<^sub>E = \<bottom>"
  shows "\<parallel>v\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>decide x v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E"
    and "\<parallel>p\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>decide x p\<parallel>\<^sub>E = \<parallel>p\<parallel>\<^sub>E"
    and "\<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>decide x d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
    and "\<parallel>r\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>decide x r\<parallel>\<^sub>E = \<parallel>r\<parallel>\<^sub>E"
    and "\<parallel>l\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>decide x l\<parallel>\<^sub>E = \<parallel>l\<parallel>\<^sub>E"
  using assms by (induct v and p and d and r and l, simp_all) (unfold decideCE_def, simp_all, safe, auto)

theorem decide_sem_disj_true [simp]:
  fixes d :: disj assumes "\<parallel>l\<parallel>\<^sub>E = True"
  shows "\<parallel>decide l d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
  using assms by (induct d) auto

theorem decide_sem_disj_bot [simp]:
  fixes d :: disj assumes "\<parallel>l\<parallel>\<^sub>E = \<bottom>" and "\<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>decide l d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
  using assms by (induct d) auto


(* Semantics of pullX and strictify *)

theorem pullVE_ide [simp]:
  shows "id\<^sub>E pullVE"
proof -
  { fix x :: valueExp and E assume dom: "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>" hence "\<parallel>pullVE x\<parallel>\<^sub>E = \<parallel>x\<parallel>\<^sub>E"
  proof (cases x)
    case Sum    thus ?thesis using dom by simp (subst push_sem, auto) next
    case Size   thus ?thesis using dom by simp (subst push_sem, auto) next
    case Count  thus ?thesis using dom by simp (subst push2_sem, auto) next
    case Tally  thus ?thesis using dom by simp (subst push_sem, auto) next
    case Read   thus ?thesis using dom by simp (subst push2_sem, auto) next
    case Scal   thus ?thesis using dom by simp (subst push2_sem, auto) next
    case VRep   thus ?thesis using dom by simp (subst push2_sem, auto) next
    case VPlus  thus ?thesis using dom by simp (subst push2_sem, auto) next
    case VMinus thus ?thesis using dom by simp (subst push2_sem, auto) next
    case (VCond b) thus ?thesis using dom by simp (subst push_sem, simp+, force, (cases b, force)+)
  qed simp+ }
  thus ?thesis unfolding id_dom_def by simp
qed

theorem pullPE_ide [simp]:
  shows "id\<^sub>E pullPE"
proof -
  { fix x :: pathExp and E assume dom: "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>" hence "\<parallel>pullPE x\<parallel>\<^sub>E = \<parallel>x\<parallel>\<^sub>E"
  proof (cases x)
    case Path   thus ?thesis using dom by simp (subst push2_sem, auto) next
    case Kind   thus ?thesis using dom by simp (subst push_sem, auto) next
    case PRep   thus ?thesis using dom by simp (subst push2_sem, auto) next
    case PPlus  thus ?thesis using dom by auto (subst push2_sem_kind, simp+, force, force, auto) next
    case PMinus thus ?thesis using dom by simp (subst push2_sem, auto) next
    case (PCond b) thus ?thesis using dom by auto (subst push_sem, simp+, force, (cases b, force)+)+
  qed simp+ }
  thus ?thesis unfolding id_dom_def by simp
qed

theorem pullPE_kind [simp]:
  shows "kindOf (pullPE p) = kindOf p"
proof (cases p)
  case Path   thus ?thesis by auto (subst push2_kind, auto) next
  case Kind   thus ?thesis by auto (subst push_kind, auto) next
  case PRep   thus ?thesis by auto (subst push2_kind, auto) next
  case PPlus  thus ?thesis by auto (subst push2_kind, auto) next
  case PMinus thus ?thesis by auto (subst push2_kind, auto) next
  case PCond  thus ?thesis by auto (subst push_kind, auto)
qed simp

theorem pullDE_ide [simp]:
  shows "id\<^sub>E pullDE"
proof -
  { fix x :: docExp and E assume dom: "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>" hence "\<parallel>pullDE x\<parallel>\<^sub>E = \<parallel>x\<parallel>\<^sub>E"
  proof (cases x)
    case Upd thus ?thesis using dom by simp (subst push3_sem, auto) next
    case Del thus ?thesis using dom by simp (subst push2_sem, auto) next
    case (DCond b) thus ?thesis using dom by simp (subst push_sem, simp+, force, (cases b, force)+)
  qed simp+ }
  thus ?thesis unfolding id_dom_def by simp
qed

theorem pullRE_ide [simp]:
  shows "id\<^sub>E pullRE"
proof -
  { fix x :: pred and E assume dom: "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>" hence "\<parallel>pullRE x\<parallel>\<^sub>E = \<parallel>x\<parallel>\<^sub>E"
  proof (cases x)
    case Equal thus ?thesis using dom by simp (subst push2_sem, auto) next
    case Less  thus ?thesis using dom by simp (subst push2_sem, auto) next
    case (RCond b) thus ?thesis using dom by simp (subst push_sem, simp+, force, (cases b, force)+)
  qed }
  thus ?thesis unfolding id_dom_def by simp
qed

theorem pullD_ide [simp]:
  assumes "\<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>pullD d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
  using assms by (induct d rule: pullD.induct) (rule_tac x=l and E=E in sbeI, auto)

theorem pullC_ide [simp]:
  shows "id\<^sub>E pullC"
  unfolding id_dom_def by clarify (case_tac x, auto)

corollary strictify_xmap_id:
  shows "xmap_id (XMap pullVE pullPE pullDE pullRE id id pullC)"
  by simp

theorem strictify_sem [simp]:
  fixes c :: cnf assumes "safe c"
  shows "\<parallel>strictify c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by - (erule sem_safe_xmap_id[rotated], fold id_def, simp)


(* Main Property *)

theorem ctopD_takeoutL [simp]:
  assumes "ctopD d"
  shows "ctopD (takeoutL l d)"
  using assms
  by (induct l d rule: takeoutL.induct) auto

theorem ctopD_sD [simp]:
  assumes "ctop (thePred l)" and "ctopD d"
  shows "ctopD (sD l d)"
  using assms unfolding sD_def by auto

theorem ctopD_disj_disj [simp]:
  assumes "ctopD (toDisj a)" and "ctopD (toDisj b)"
  shows "ctopD (a \<or>. b)"
  using assms unfolding disj_disj_def
  by - (erule rev_mp, erule rev_mp, rule disj_conc.induct[of _ "toDisj a" "toDisj b"], auto)

theorem pullD_strict [simp]:
  assumes "ctopD d"
  shows "strict (pullD d)"
  using assms by (induct d rule: pullD.induct, auto) (fold cond_pred_def, auto)


theorem lit_pred_xprop [simp]:
  shows "xprop p (thePred l) = xprop p l"
  by (cases l) auto

theorem lit_pred_xcount [simp]:
  shows "xcount c (thePred l) = xcount c l"
  by (cases l) auto

fun ctopL :: "lit \<Rightarrow> bool" where
  "ctopL (L b p) \<longleftrightarrow> ctop p"

lemma ctopL [simp]:
  "ctopL l \<longleftrightarrow> ctop (thePred l)"
  by (cases l) simp


theorem strictify_ctop_exp [simplified, simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "ctop (strictify v)"
    and "ctop (strictify p)"
    and "ctop (strictify d)"
    and "ctop (strictify r)"
    and "ctopL (strictify l)"
  by (induct v and p and d and r and l) (simp_all, safe, simp_all add: cond_folds)

theorem strictify_ctop_disj [simp]:
  fixes d :: disj
  shows "ctopD (strictify d)"
  using assms by (induct d) auto

theorem strictify_strict_conj [simp]:
  fixes c :: conj
  shows "strict (strictify c)"
  using assms by (induct c) auto

theorem strictify_strict [simp]:
  fixes c :: cnf
  shows "strict (strictify c)"
  using assms by (cases c) auto


declare cond_defs [simp]

end

