header {* The ElimDU Transformer *}

theory Elimination imports Intersection begin


(*------------------*)
(*- DU Elimination -*)
(*------------------*)

(* Unique Case *)

fun elimdu_uniq :: "valueExp \<Rightarrow> valueExp" where
  "elimdu_uniq (Read (Upd d p v) q) = if. q .\<in> p then v else elimdu_uniq (Read d q) fi"
| "elimdu_uniq (Read (Del d p) q) = if. q .\<in> (intersect q p) then {}. else elimdu_uniq (Read d q) fi"
| "elimdu_uniq v = v"

theorem elimdu_uniq_sem [simp]:
  assumes "uve v" and "\<parallel>v\<parallel>\<^sub>E = {w}\<^sub>*\<^sub>v" and "wform v"
  shows "\<parallel>elimdu_uniq v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E"
  using assms
proof (induct rule: elimdu_uniq.induct)
  case (1 de pe ve qe) then
    obtain d p q v and w where "\<parallel>de\<parallel>\<^sub>E = d\<^sub>d" "\<parallel>pe\<parallel>\<^sub>E = {p}\<^sub>*\<^sub>p" "\<parallel>qe\<parallel>\<^sub>E = {q}\<^sub>*\<^sub>p" "\<parallel>ve\<parallel>\<^sub>E = {v}\<^sub>*\<^sub>v"
      and "d[p \<rightarrow> v]({q}\<^sub>*)\<^sub>* = {w}\<^sub>*" by auto (drule_tac E=E in upe_sem, auto)
    thus ?case using 1 by (cases "p = q") auto next
  case (2 de pe qe) then
    obtain d p q and w where "\<parallel>de\<parallel>\<^sub>E = d\<^sub>d" "\<parallel>pe\<parallel>\<^sub>E = {p}\<^sub>*\<^sub>p" "\<parallel>qe\<parallel>\<^sub>E = {q}\<^sub>*\<^sub>p"
      and "d[p \<rightarrow> ]({q}\<^sub>*)\<^sub>* = {w}\<^sub>*" by auto (drule_tac E=E in upe_sem, auto)
    thus ?case using 2 by (cases "q \<in> desc p") auto
qed auto


(* General Case *)

fun rdus :: "valueExp \<Rightarrow> nat" where
  "rdus (Read (Upd d _ _) p) = Suc (rdus (Read d p))"
| "rdus (Read (Del d _) p)   = Suc (rdus (Read d p))"
| "rdus _ = 0"

lemma rdus_ignore_path [simp]:
  "rdus (Read d p) < Suc (rdus (Read d q))"
  by (induct d) simp+

function (sequential) elimdu_base :: "valueExp \<Rightarrow> valueExp" where
  "elimdu_base (Read (Upd d p v) q) =
    (elimdu_base (Read d (q .\<ominus> ((pcount q p) .\<otimes> p)))) \<oplus>. ((pcount q p) \<otimes>. v)"
| "elimdu_base (Read (Del d p) q) = elimdu_base (Read d (q .\<ominus> (intersect q p)))"
| "elimdu_base v = v"
by pat_completeness auto termination by (relation "measure rdus") auto

theorem elimdu_base_sem [simp]:
  assumes "\<parallel>v\<parallel>\<^sub>E \<noteq> \<bottom>" and "wform v"
  shows "\<parallel>elimdu_base v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E"
  using assms
proof (induct rule: elimdu_base.induct)
  case (1 de pe ve qe) then
    obtain d p q v where "\<parallel>de\<parallel>\<^sub>E = d\<^sub>d" "\<parallel>pe\<parallel>\<^sub>E = {p}\<^sub>*\<^sub>p" "\<parallel>qe\<parallel>\<^sub>E = q\<^sub>p" "\<parallel>ve\<parallel>\<^sub>E = {v}\<^sub>*\<^sub>v"
      by simp (safe, simp)+
    thus ?case using `wform (Read (Upd de pe ve) qe)` and "1.hyps"
      by auto (subst readM_upd, auto intro!: count_cong) next
  case (2 de pe qe) then
    obtain d p q where "\<parallel>de\<parallel>\<^sub>E = d\<^sub>d" "\<parallel>pe\<parallel>\<^sub>E = {p}\<^sub>*\<^sub>p" "\<parallel>qe\<parallel>\<^sub>E = q\<^sub>p"
      by simp (safe, simp)+
    thus ?case using `wform (Read (Del de pe) qe)` and "2.hyps"
      by auto (subst readM_del, auto)
qed auto


(* Generalization to Values & Paths *)

fun baseVE :: "bool \<Rightarrow> valueExp \<Rightarrow> valueExp"
and basePE :: "bool \<Rightarrow> pathExp  \<Rightarrow> pathExp" where

  "baseVE True  (Read d p) = elimdu_uniq (Read d (basePE True p))"
| "baseVE False (Read d p) = elimdu_base (Read d (basePE False p))"
| "baseVE b (Scal n m)     = n \<odot>. (baseVE b m)"
| "baseVE _ v              = v"

| "basePE b (Path l p k)   = (basePE b p)/l[baseVE b k]."
| "basePE _ p              = p"


(* Properties *)

theorem elimdu_base_wform [simp]:
  assumes "wform v"
  shows "wform (elimdu_base v)"
  using assms by (induct v rule: elimdu_base.induct) auto

theorem elimdu_uniq_wform_uve [simp]:
  assumes "wform v" and "uve v"
  shows "wform (elimdu_uniq v)" and "uve (elimdu_uniq v)"
  using assms by (induct v rule: elimdu_uniq.induct) auto


theorem baseXE_false_wform [simp]:
  shows "wform v \<Longrightarrow> wform (baseVE False v)"
    and "wform p \<Longrightarrow> wform (basePE False p)"
  by (induct v and p) auto

theorem baseXE_true_uxe_wform [simp]:
  shows "wform v \<Longrightarrow> uve v \<Longrightarrow> uve (baseVE True v)"
    and "wform v \<Longrightarrow> uve v \<Longrightarrow> wform (baseVE True v)"
    and "wform p \<Longrightarrow> upe p \<Longrightarrow> upe (basePE True p)"
    and "wform p \<Longrightarrow> upe p \<Longrightarrow> wform (basePE True p)"
  by (induct v and p) auto


theorem basePE_kind [simp]:
  shows "kindOf (basePE b p) = kindOf p"
  by (induct p) auto

theorem baseXE_false_sem [simp]:
  shows "wform v \<Longrightarrow> \<parallel>v\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>baseVE False v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E"
    and "wform p \<Longrightarrow> \<parallel>p\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>basePE False p\<parallel>\<^sub>E = \<parallel>p\<parallel>\<^sub>E"
  by (induct v and p) auto

theorem baseXE_true_sem_ex:
  shows "wform v \<Longrightarrow> uve v \<Longrightarrow> \<exists>w. \<parallel>v\<parallel>\<^sub>E = {w}\<^sub>*\<^sub>v \<Longrightarrow> \<parallel>baseVE True v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E"
    and "wform p \<Longrightarrow> upe p \<Longrightarrow> \<exists>q. \<parallel>p\<parallel>\<^sub>E = {q}\<^sub>*\<^sub>p \<Longrightarrow> \<parallel>basePE True p\<parallel>\<^sub>E = \<parallel>p\<parallel>\<^sub>E"
proof (induct _ True v and _ True p rule: baseVE_basePE.induct)
  case 1 thus ?case by auto (frule_tac E=E in upe_sem, auto) next
  case 3 thus ?case by auto (frule_tac E=E in uve_sem, auto) next
  case 5 thus ?case by auto (frule_tac E=E in uve_sem, auto)
qed auto

corollary baseVE_true_sem [simp]:
  assumes "wform v" and "uve v" and "\<parallel>v\<parallel>\<^sub>E = {w}\<^sub>*\<^sub>v"
  shows "\<parallel>baseVE True v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E"
  using assms by - (rule baseXE_true_sem_ex, simp+)

corollary basePE_true_sem [simp]:
  assumes "wform p" and "upe p" and "\<parallel>p\<parallel>\<^sub>E = {q}\<^sub>*\<^sub>p"
  shows "\<parallel>basePE True p\<parallel>\<^sub>E = \<parallel>p\<parallel>\<^sub>E"
  using assms by - (rule baseXE_true_sem_ex, simp+)


(* Global Elimination *)

fun elimVE :: "valueExp \<Rightarrow> valueExp" where
  "elimVE (Sum m)       = sum.   (baseVE False m)"
| "elimVE (Size m)      = size.  (baseVE False m)"
| "elimVE (Count m v)   = count. (baseVE False m) (baseVE True  v)"
| "elimVE (Tally t m)   = tally. t (baseVE False m)"
| "elimVE (Scal n m)    = (baseVE True  n) \<odot>. m"
| "elimVE (VRep a b)    = (baseVE True a)  \<otimes>. (baseVE False b)"
| "elimVE (VPlus a b)   = (baseVE False a) \<oplus>. (baseVE False b)"
| "elimVE (VMinus a b)  = (baseVE False a) \<ominus>. (baseVE False b)"
| "elimVE v             = v"

fun elimPE :: "pathExp \<Rightarrow> pathExp" where
  "elimPE (Kind l p)    = (basePE False p)/l."
| "elimPE (PPlus a b)   = (basePE False a) .\<oplus> (basePE False b)"
| "elimPE (PRep a b)    = (baseVE True  a) .\<otimes> (basePE False b)"
| "elimPE (PMinus a b)  = (basePE False a) .\<ominus> (basePE False b)"
| "elimPE p             = p"

fun elimDE :: "docExp \<Rightarrow> docExp" where
  "elimDE (Upd d p v)   = d[basePE True p \<rightarrow> baseVE True v]."
| "elimDE (Del d p)     = d[basePE True p \<rightarrow>]."
| "elimDE d             = d"

fun elimRE :: "pred \<Rightarrow> pred" where
  "elimRE (Equal a b)   = (baseVE True  a) =. (baseVE True  b)"
| "elimRE (Less a b)    = (baseVE True  a) <. (baseVE True  b)"
| "elimRE r             = r"

abbreviation elimdu where
  "elimdu \<equiv> xmap (XMap elimVE elimPE elimDE elimRE (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a))"


(* Properties *)

theorem elimdu_uxe [simp]:
  shows "uve v \<Longrightarrow> uve (elimdu v)"
    and "upe p \<Longrightarrow> upe (elimdu p)"
  by (induct v and p) auto

theorem elimdu_wform_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "wform v \<Longrightarrow> wform (elimdu v)"
    and "wform p \<Longrightarrow> wform (elimdu p)"
    and "wform d \<Longrightarrow> wform (elimdu d)"
    and "wform r \<Longrightarrow> wform (elimdu r)"
    and "wform l \<Longrightarrow> wform (elimdu l)"
  by (induct v and p and d and r and l) auto

theorem elimdu_wform_disj [simp]:
  fixes d :: disj assumes "wform d"
  shows "wform (elimdu d)"
  using assms by (induct d) auto

theorem elimdu_wform_conj [simp]:
  fixes c :: conj assumes "wform c"
  shows "wform (elimdu c)"
  using assms by (induct c) auto

theorem elimdu_wform [simp]:
  fixes c :: cnf assumes "wform c"
  shows "wform (elimdu c)"
  using assms by (cases c) simp+


(* Semantic Equivalence *)

theorem elimdu_kindOf [simp]:
  fixes p :: pathExp
  shows "kindOf (elimdu p) = kindOf p"
  by (induct p) auto

theorem elimdu_sem_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "wform v \<Longrightarrow> \<parallel>v\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>elimdu v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E"
    and "wform p \<Longrightarrow> \<parallel>p\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>elimdu p\<parallel>\<^sub>E = \<parallel>p\<parallel>\<^sub>E"
    and "wform d \<Longrightarrow> \<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>elimdu d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
    and "wform r \<Longrightarrow> \<parallel>r\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>elimdu r\<parallel>\<^sub>E = \<parallel>r\<parallel>\<^sub>E"
    and "wform l \<Longrightarrow> \<parallel>l\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>elimdu l\<parallel>\<^sub>E = \<parallel>l\<parallel>\<^sub>E"
  by (induct v and p and d and r and l) auto

theorem elimdu_sem_disj [simp]:
  fixes d :: disj assumes "wform d" and "\<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>elimdu d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
  using assms by (induct d) auto

theorem elimdu_sem_conj [simp]:
  fixes c :: conj assumes "wform c" and "\<parallel>c\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>elimdu c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by (induct c) auto

theorem elimdu_sem [simp]:
  fixes c :: cnf assumes "wform c" and "safe c"
  shows "\<parallel>elimdu c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by (cases c, auto) (rule sbeI, auto, rule_tac x=f in exI, simp)


(* Preserves Guarded *)

theorem elimdu_guarded_disj [simp]:
  fixes d :: disj assumes "guarded k d" and "wform d"
  shows "guarded k (elimdu d)"
  using assms by (induct d arbitrary: k) (auto, drule_tac x="\<not>. (elimdu lit) \<and>. k" in meta_spec,
    erule meta_mp, erule guarded_relax, auto)

theorem elimdu_guarded_conj [simp]:
  fixes c :: conj assumes "guarded k c" and "wform c"
  shows "guarded k (elimdu c)"
  using assms by (induct c arbitrary: k) (auto, drule_tac x="elimdu disj \<and>. k" in meta_spec,
    erule meta_mp, erule guarded_relax, auto, drule guarded_sem, auto)

theorem elimdu_guarded [simp]:
  fixes c :: cnf assumes "guarded k c" and "wform c"
  shows "guarded k (elimdu c)"
  using assms by (cases c) simp+


(* Main Property - dufree eliminates *)

theorem elimdu_base_dufree [simp]:
  assumes "dufree p" and "dutop (Read d p)"
  shows "dufree (elimdu_base (Read d p))"
  using assms
proof (induct d arbitrary: p rule: topDUs_docExp.induct)
  case (1 d q v) have "dutop d" using "1.prems" and nat_add_assoc by auto
    thus ?case using 1 by simp next
  case (2 d q) have "dutop d" using "2.prems" and nat_add_assoc by auto
    thus ?case using 2 by simp
qed auto

theorem elimdu_uniq_dufree [simp]:
  assumes "dufree p" and "dutop (Read d p)"
  shows "dufree (elimdu_uniq (Read d p))"
  using assms
proof (induct d arbitrary: p rule: topDUs_docExp.induct)
  case (1 d q v) have "dutop d" using "1.prems" and nat_add_assoc by auto
    thus ?case using 1 by simp next
  case (2 d q) have "dutop d" using "2.prems" and nat_add_assoc by auto
    thus ?case using 2 by simp
qed auto


theorem baseXE_dufree [simp]:
  shows "dutop v \<Longrightarrow> dufree (baseVE b v)"
    and "dutop p \<Longrightarrow> dufree (basePE b p)"
  using nat_add_assoc by (induct _ b v and _ b p rule: baseVE_basePE.induct) auto


theorem elimdu_dus_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "strict v \<Longrightarrow> dutop  (elimdu v)"
    and "strict p \<Longrightarrow> dutop  (elimdu p)"
    and "strict d \<Longrightarrow> dutop  (elimdu d)"
    and "strict r \<Longrightarrow> dufree (elimdu r)"
    and "strict l \<Longrightarrow> dufree (elimdu l)"
  by (induct v and p and d and r and l) auto
  
theorem elimdu_dufree_disj [simp]:
  fixes d :: disj assumes "strict d"
  shows "dufree (elimdu d)"
  using assms by (induct d) auto

theorem elimdu_dufree_conj [simp]:
  fixes c :: conj assumes "strict c"
  shows "dufree (elimdu c)"
  using assms by (induct c) auto

theorem elimdu_dufree [simp]:
  fixes c :: cnf assumes "strict c"
  shows "dufree (elimdu c)"
  using assms by (cases c) simp

end

