header {* Contextual Simplification (Facts and Rewriting) *}

theory Facts imports Safe
begin


(* Fact Identities *)

definition id_fact :: "conj \<Rightarrow> ('a :: sem \<Rightarrow> 'a) \<Rightarrow> bool" ("id\<^sub>F") where
  "id\<^sub>F k f \<equiv> \<forall>x E. \<parallel>k\<parallel>\<^sub>E = True \<longrightarrow> \<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom> \<longrightarrow> \<parallel>f x\<parallel>\<^sub>E = \<parallel>x\<parallel>\<^sub>E"

theorem sem_id_fact [simp]:
  assumes "\<parallel>k\<parallel>\<^sub>E = True" and "id\<^sub>F k f" and "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>f x\<parallel>\<^sub>E = \<parallel>x\<parallel>\<^sub>E"
  using assms unfolding id_fact_def by simp

corollary sem_id_factE:
  assumes "id\<^sub>F k f" and "\<parallel>k\<parallel>\<^sub>E = True" and "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>" and "\<parallel>x\<parallel>\<^sub>E = X"
  shows "\<parallel>f x\<parallel>\<^sub>E = X"
  using assms by simp

theorem id_fact_relax:
  assumes "id\<^sub>F k f" and "\<And>E. \<parallel>s\<parallel>\<^sub>E = True\<^sub>b \<Longrightarrow> \<parallel>k\<parallel>\<^sub>E = True\<^sub>b"
  shows "id\<^sub>F s f"
  using assms unfolding id_fact_def by auto

theorem id_id_fact [simp]:
  shows "id\<^sub>F k id"
  unfolding id_fact_def by simp


(* Basic Identities *)

definition id_basic :: "conj \<Rightarrow> ('a :: {sem, xfold} \<Rightarrow> 'a) \<Rightarrow> bool" ("id\<^sub>B") where
  "id\<^sub>B k f \<equiv> basic k \<longrightarrow> (\<forall>x. (strict x \<longrightarrow> strict (f x)) \<and> (straight x \<longrightarrow> straight (f x)))"

theorem id_basic_strict [simp]:
  assumes "id\<^sub>B k f" and "basic k" and "strict x"
  shows "strict (f x)"
  using assms unfolding id_basic_def by auto

theorem id_basic_straight [simp]:
  assumes "id\<^sub>B k f" and "basic k" and "straight x"
  shows "straight (f x)"
  using assms unfolding id_basic_def by auto

theorem id_id_basic [simp]:
  shows "id\<^sub>B k id" and "id\<^sub>B k (\<lambda>a. a)"
  unfolding id_basic_def by simp+


(* Guarded Identities *)

definition id_guarded :: "('a :: guarded \<Rightarrow> 'a) \<Rightarrow> bool" ("id\<^sub>G") where
  "id\<^sub>G f \<equiv> \<forall>x k. guarded k x \<longrightarrow> guarded k (f x)"

theorem id_guarded_guarded [simp]:
  assumes "guarded k x" and "id\<^sub>G f"
  shows "guarded k (f x)"
  using assms unfolding id_guarded_def by simp

theorem id_id_guarded [simp]:
  shows "id\<^sub>G id" and "id\<^sub>G (\<lambda>a. a)"
  unfolding id_guarded_def by simp+


(* Expression Mapping Identities *)

fun xmap_fid :: "conj \<Rightarrow> xmap \<Rightarrow> bool" where
  "xmap_fid k (XMap v p d r l e c) \<longleftrightarrow>
    id\<^sub>F k v \<and> id\<^sub>F k p \<and> id\<^sub>F k d \<and> id\<^sub>F k r \<and> id\<^sub>F k l \<and>
    id\<^sub>F k e \<and> id\<^sub>F k c \<and> (\<forall>x. kindOf (p x) = kindOf x \<and>
    id\<^sub>B k v \<and> id\<^sub>B k p \<and> id\<^sub>B k d \<and> id\<^sub>B k r \<and>
    id\<^sub>B k l \<and> id\<^sub>B k e \<and> id\<^sub>B k c \<and>
    id\<^sub>G e \<and> id\<^sub>G c)"

theorem sem_smap_fid_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  assumes "xmap_fid k f"
  shows "\<parallel>k\<parallel>\<^sub>E = True \<Longrightarrow> \<parallel>v\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>smap f v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E"
    and "\<parallel>k\<parallel>\<^sub>E = True \<Longrightarrow> \<parallel>p\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>smap f p\<parallel>\<^sub>E = \<parallel>p\<parallel>\<^sub>E" and "kindOf (smap f p) = kindOf p"
    and "\<parallel>k\<parallel>\<^sub>E = True \<Longrightarrow> \<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>smap f d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
    and "\<parallel>k\<parallel>\<^sub>E = True \<Longrightarrow> \<parallel>r\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>smap f r\<parallel>\<^sub>E = \<parallel>r\<parallel>\<^sub>E"
    and "\<parallel>k\<parallel>\<^sub>E = True \<Longrightarrow> \<parallel>l\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>smap f l\<parallel>\<^sub>E = \<parallel>l\<parallel>\<^sub>E"
  using assms and xmap.nchotomy[THEN spec, of f] by (induct v and p and d and r and l) auto

theorem sem_smap_fid_disj [simp]:
  fixes d :: disj assumes "xmap_fid k f" and "\<parallel>k\<parallel>\<^sub>E = True" and "\<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>smap f d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
  using assms 

apply (cases f)
apply clarsimp
apply (induct d)
apply auto
apply (rule_tac f=fun6 and k=k in sem_id_factE, auto)
apply (subst (asm) sem_smap_fid_exp, auto)+
apply (subst sem_smap_fid_exp, auto)+
apply (rule_tac f=fun6 and k=k in sem_id_factE, auto)
apply (subst (asm) sem_smap_fid_exp, auto)+
apply (subst sem_smap_fid_exp, auto)+
done

theorem sem_smap_fid_conj [simp]:
  fixes c :: conj assumes "xmap_fid k f" and "\<parallel>k\<parallel>\<^sub>E = True" and "\<parallel>c\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>smap f c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms

apply (cases f)
apply clarsimp
apply (induct c)
apply auto
apply (rule_tac f=fun7 and k=k in sem_id_factE, auto)
apply (subst (asm) sem_smap_fid_disj, auto)+
apply (subst sem_smap_fid_disj, auto)+
apply (rule_tac f=fun7 and k=k in sem_id_factE, auto)
apply (subst (asm) sem_smap_fid_disj, auto)+
apply (subst sem_smap_fid_disj, auto)+
done


theorem strict_smap_fid_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  assumes "xmap_fid k f" and "basic k"
  shows "strict v \<Longrightarrow> strict (smap f v)"
    and "strict p \<Longrightarrow> strict (smap f p)"
    and "strict d \<Longrightarrow> strict (smap f d)"
    and "strict r \<Longrightarrow> strict (smap f r)"
    and "strict l \<Longrightarrow> strict (smap f l)"
  using assms and xmap.nchotomy[THEN spec, of f] by (induct v and p and d and r and l) auto

theorem straight_smap_fid_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  assumes "xmap_fid k f" and "basic k"
  shows "straight v \<Longrightarrow> straight (smap f v)"
    and "straight p \<Longrightarrow> straight (smap f p)"
    and "straight d \<Longrightarrow> straight (smap f d)"
    and "straight r \<Longrightarrow> straight (smap f r)"
    and "straight l \<Longrightarrow> straight (smap f l)"
  using assms and xmap.nchotomy[THEN spec, of f] by (induct v and p and d and r and l) auto

theorem strict_smap_fid_disj [simp]:
  fixes d :: disj assumes "xmap_fid k f" and "strict d" and "basic k"
  shows "strict (smap f d)"
  using assms

apply (cases f, auto, induct d, auto)
apply (rule_tac f=fun6 in id_basic_strict, auto)
apply (rule sD_props.strict, auto)
apply (rule strict_smap_fid_exp, auto)
done

theorem straight_smap_fid_disj [simp]:
  fixes d :: disj assumes "xmap_fid k f" and "straight d" and "basic k"
  shows "straight (smap f d)"
  using assms

apply (cases f, auto, induct d, auto)
apply (rule_tac f=fun6 in id_basic_straight, auto)
apply (rule sD_props.straight, auto)
apply (rule straight_smap_fid_exp, auto)
done

theorem strict_smap_fid_conj [simp]:
  fixes c :: conj assumes "xmap_fid k f" and "strict c" and "basic k"
  shows "strict (smap f c)"
  using assms

apply (cases f, auto, induct c, auto)
apply (rule_tac f=fun7 in id_basic_strict, auto)
apply (rule sC_props.strict, auto)
apply (rule strict_smap_fid_disj, auto)
done

theorem straight_smap_fid_conj [simp]:
  fixes c :: conj assumes "xmap_fid k f" and "straight c" and "basic k"
  shows "straight (smap f c)"
  using assms

apply (cases f, auto, induct c, auto)
apply (rule_tac f=fun7 in id_basic_straight, auto)
apply (rule sC_props.straight, auto)
apply (rule straight_smap_fid_disj, auto)
done


(* Guarded Disj *)

fun decideL :: "lit \<Rightarrow> disj \<Rightarrow> disj" where
  "decideL m DFlse   = DFlse"
| "decideL m (D l d) = (
    if l = m    then decideL m d else
    if l = \<not>. m then True. else
    case decideL m d of e \<Rightarrow>
    if e = True. then True. else
    D l e)"

theorem sem_decideL [simp]:
  assumes "\<parallel>l\<parallel>\<^sub>E = False"
  shows "\<parallel>decideL l d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
  using assms by (induct l d rule: decideL.induct) (auto, rule sbeI, auto)

interpretation decideL_props:
  props "decideL l d" d
  by (unfold_locales, auto elim!: order_trans[rotated]) (induct l d rule: decideL.induct, auto)+

definition gD :: "lit \<Rightarrow> disj \<Rightarrow> disj" where
  "gD l e \<equiv>
    if l = False. then e else
    if l = True. then True. else
    case decideL l e of d \<Rightarrow>
    if d = True. then True. else D l d"

theorem gD_sem [simp]:
  assumes "\<parallel>l\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>gD l d\<parallel>\<^sub>E = \<parallel>D l d\<parallel>\<^sub>E"
  using assms unfolding gD_def by auto (subst sem_decideL[symmetric], auto)

interpretation gD_props:
  props "gD l d" "D l d"
  unfolding gD_def by unfold_locales (auto elim!: order_trans[rotated])


(* Guarded Conj *)

fun decideD :: "disj \<Rightarrow> conj \<Rightarrow> conj" where
  "decideD m CTrue   = CTrue"
| "decideD m (C d c) = (
    if subsumes d m then decideD m c else
    case decideD m c of e \<Rightarrow>
    if e = False. then False. else
    C (case m of D l DFlse \<Rightarrow> decideL (\<not>. l) d | _ \<Rightarrow> d) e)"

theorem sem_decideD [simp]:
  assumes "\<parallel>d\<parallel>\<^sub>E = True"
  shows "\<parallel>decideD d c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms
  by (induct d c rule: decideD.induct) (auto, rule sbeI, auto, (split disj.split, auto)+)

interpretation decideD_props:
  props "decideD d c" c
  by (unfold_locales, auto elim!: order_trans[rotated])
     (induct d c rule: decideD.induct, auto, (split disj.split, auto)+)+

definition gC :: "disj \<Rightarrow> conj \<Rightarrow> conj" where
  "gC d e \<equiv>
    if d = True. then e else
    if d = False. then False. else
    case decideD d e of c \<Rightarrow>
    if c = False. then False. else C d c"

theorem gC_sem [simp]:
  assumes "\<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>gC d c\<parallel>\<^sub>E = \<parallel>C d c\<parallel>\<^sub>E"
  using assms unfolding gC_def by auto (subst sem_decideD[symmetric], auto)

theorem gC_sem_true [dest!]:
  assumes "\<parallel>gC d c\<parallel>\<^sub>E = True"
  shows "\<parallel>C d c\<parallel>\<^sub>E = True"
  using assms unfolding gC_def
  by auto (split if_splits, auto)+

interpretation gC_props:
  props "gC d c" "C d c"
  unfolding gC_def by unfold_locales (auto elim!: order_trans[rotated])


(* Guarded Rewriting *)

class rewrite =
fixes rewrite :: "(disj \<Rightarrow> conj \<Rightarrow> conj) \<Rightarrow> (conj \<Rightarrow> xmap) \<Rightarrow> conj \<Rightarrow> 'a \<Rightarrow> 'a"

instantiation disj and conj and cnf :: rewrite begin

fun rewrite_disj :: "(disj \<Rightarrow> conj \<Rightarrow> conj) \<Rightarrow> (conj \<Rightarrow> xmap) \<Rightarrow> conj \<Rightarrow> disj \<Rightarrow> disj" where
  "rewrite f x k DFlse   = DFlse"
| "rewrite f x k (D l d) = (case smap (x k) l of l' \<Rightarrow>
    sD l' (rewrite f x (f (toDisj (\<not>. l')) k) d))"

fun rewrite_conj :: "(disj \<Rightarrow> conj \<Rightarrow> conj) \<Rightarrow> (conj \<Rightarrow> xmap) \<Rightarrow> conj \<Rightarrow> conj \<Rightarrow> conj" where
  "rewrite f x k CTrue   = CTrue"
| "rewrite f x k (C d c) = (case rewrite f x k d of d' \<Rightarrow>
    sC d' (rewrite f x (f d' k) c))"

fun rewrite_cnf :: "(disj \<Rightarrow> conj \<Rightarrow> conj) \<Rightarrow> (conj \<Rightarrow> xmap) \<Rightarrow> conj \<Rightarrow> cnf \<Rightarrow> cnf" where
  "rewrite f x k (\<forall>. c) = \<forall>. (rewrite f x k c)"

instance .. end


theorem decideL_guarded:
  fixes k :: conj assumes "guarded (\<not>. l \<and>. k) d"
  shows "guarded (\<not>. l \<and>. k) (decideL l d)"
  using assms

apply (induct l d arbitrary: k rule: decideL.induct)
apply auto
apply (drule_tac x=k in meta_spec, erule meta_mp, erule guarded_relax, auto)
apply (drule_tac x=k in meta_spec, erule meta_mp, erule guarded_relax, auto)
apply (rotate_tac 1)
apply (drule_tac x="\<not>. l \<and>. k" in meta_spec)
apply (drule meta_mp, auto)
done

theorem decideL_guarded_neg:
  fixes l :: lit and k :: conj assumes "guarded (l \<and>. k) d"
  shows "guarded (l \<and>. k) (decideL (\<not>. l) d)"
  using assms
proof -
  have "guarded (\<not>. (\<not>. l) \<and>. k) d" using assms by - (erule guarded_relax, auto)
  hence "guarded (\<not>. (\<not>. l) \<and>. k) (decideL (\<not>. l) d)" by - (erule decideL_guarded)
  thus ?thesis by - (erule guarded_relax, auto)
qed

theorem decideD_guarded:
  fixes k :: conj assumes "guarded (d \<and>. k) c"
  shows "guarded (d \<and>. k) (decideD d c)"
  using assms

apply (induct d c arbitrary: k rule: decideD.induct)
apply auto
apply (drule_tac x=k in meta_spec, erule meta_mp, erule guarded_relax, auto)
apply (split disj.split, auto)+
apply (rule_tac k="lit \<and>. k" in guarded_relax, auto)
apply (rule decideL_guarded_neg, auto)
apply (split disj.split, auto)+
apply (rotate_tac 1)
apply (drule_tac x="d \<and>. k" in meta_spec, drule meta_mp, auto)
apply (rotate_tac 1)
apply (drule_tac x="decideL (\<not>. lit) d \<and>. k" in meta_spec, drule meta_mp, auto)
apply (erule guarded_relax, auto)
apply (split disj.split, auto)+
apply (rotate_tac 1)
apply (drule_tac x="d \<and>. k" in meta_spec, drule meta_mp, auto)
apply (rotate_tac 1)
apply (drule_tac x="d \<and>. k" in meta_spec, drule meta_mp, auto)
done

theorem gC_guarded [simp]:
  assumes "guarded k (C d c)"
  shows "guarded k (gC d c)"
  unfolding gC_def using assms by auto (rule decideD_guarded, auto)


definition id_true :: "(disj \<Rightarrow> conj \<Rightarrow> conj) \<Rightarrow> bool" ("id\<^sub>T") where
  "id\<^sub>T g \<equiv> \<forall>d. basic d \<longrightarrow> (\<forall>c R f. (\<parallel>C d c\<parallel>\<^sub>R\<^sub>+\<^sub>f = True \<longrightarrow> \<parallel>g d c\<parallel>\<^sub>R\<^sub>+\<^sub>f = True) \<and>
     (guarded CTrue c \<longrightarrow> \<parallel>g d c\<parallel>\<^sub>R\<^sub>+\<^sub>f = True \<longrightarrow> \<parallel>C d c\<parallel>\<^sub>R\<^sub>+\<^sub>f = True) \<and>
     (guarded CTrue c \<longrightarrow> guarded c d \<longrightarrow> guarded CTrue (g d c)))"

theorem sem_id_true_imp [simp]:
  assumes "id\<^sub>T g" and "\<parallel>C d c\<parallel>\<^sub>R\<^sub>+\<^sub>f = True" and "basic d"
  shows "\<parallel>g d c\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
  using assms unfolding id_true_def by simp

theorem sem_id_true_rev_imp [simp]:
  assumes "id\<^sub>T g" and "guarded CTrue c" and "\<parallel>g d c\<parallel>\<^sub>R\<^sub>+\<^sub>f = True" and "basic d"
  shows "\<parallel>C d c\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
  using assms unfolding id_true_def by (drule_tac x=d in spec)
    (drule(1) mp, drule_tac x=c in spec, drule_tac x=R in spec, drule_tac x=f in spec, simp)

theorem guarded_id_true [simp]:
  assumes "id\<^sub>T f" and "guarded CTrue c" and "guarded c d" and "basic d"
  shows "guarded CTrue (f d c)"
  using assms unfolding id_true_def by simp


theorem conj_conj_id_true [simp]:
  shows "id\<^sub>T (\<lambda>d c. c \<and>. d)"
  unfolding id_true_def
  by auto (erule guarded_relax, auto)


theorem rewrite_disj_sem [simp]:
  fixes d :: disj
  assumes "guarded k d" and "\<And>k. xmap_fid k (x k)" and "\<parallel>k\<parallel>\<^sub>R\<^sub>+\<^sub>f = True" 
    and "id\<^sub>T g" and "\<And>d. id\<^sub>B (toConj d) (g d)" and "guarded CTrue k" and "basic d" and "basic k"
  shows "\<parallel>rewrite g x k d\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<parallel>d\<parallel>\<^sub>R\<^sub>+\<^sub>f"
  using assms

apply (induct d arbitrary: k)
apply auto
apply (drule_tac x="g (D (\<not>. (smap (x k) lit)) DFlse) k" in meta_spec)
apply (drule_tac x=k in meta_spec, auto)
apply (rule_tac x=lit and E="R\<^sub>+\<^sub>f" in sbeI, auto)
apply (drule meta_mp)
apply (erule guarded_relax)
apply (drule(2) sem_id_true_rev_imp, auto)
apply (drule meta_mp)
apply (rule_tac f=g in guarded_id_true, auto)
apply (erule meta_mp)
apply (drule_tac x="D (\<not>. (smap (x k) lit)) DFlse" in meta_spec, auto)
done

theorem rewrite_disj_guarded [simp]:
  fixes d :: disj
  assumes "guarded k d" and "\<And>k. xmap_fid k (x k)" and "guarded CTrue k"
    and "basic d" and "basic k" and "id\<^sub>T g" and "\<And>d. id\<^sub>B (toConj d) (g d)" 
  shows "guarded k (rewrite g x k d)"
  using assms

apply (induct d arbitrary: k)
apply auto
apply (drule_tac x=k in meta_spec)
apply (subst (asm) sem_smap_fid_exp, auto)
apply (drule_tac x="g (D (\<not>. (smap (x k) lit)) DFlse) k" in meta_spec)
apply (drule meta_mp)
defer
apply (drule meta_mp)
apply (drule_tac x=k in meta_spec)
apply (rule_tac f=g in guarded_id_true, auto)
apply (drule meta_mp)
apply (drule_tac x="D (\<not>. (smap (x k) lit)) DFlse" in meta_spec, auto)
apply (erule id_basic_strict, auto)
apply (rule lit_negation_props.strict, auto)
apply (rule strict_smap_fid_exp, auto)
apply (rule lit_negation_props.straight, auto)
apply (rule straight_smap_fid_exp, auto)
apply (erule id_basic_straight, auto)
apply (rule lit_negation_props.strict, auto)
apply (rule strict_smap_fid_exp, auto)
apply (rule lit_negation_props.straight, auto)
apply (rule straight_smap_fid_exp, auto)
apply (erule guarded_relax)
apply (drule_tac x=k in meta_spec)
apply (rule_tac g=g in sem_id_true_imp, auto)
apply (drule_tac x=k in meta_spec)
apply (erule guarded_relax)
apply (drule(2) sem_id_true_rev_imp, auto)
done

theorem rewrite_disj_strict [simp]:
  fixes d :: disj
  assumes "\<And>k. xmap_fid k (x k)" and "basic d" and "basic k" and "\<And>d. id\<^sub>B (toConj d) (g d)" 
  shows "strict (rewrite g x k d)"
  using assms

apply (induct d arbitrary: k)
apply auto
apply (rule sD_props.strict, auto)
apply (rule strict_smap_fid_exp, auto)
apply (drule_tac x="g (D (\<not>. (smap (x k) lit)) DFlse) k" in meta_spec)
apply (erule meta_mp)
apply (drule_tac x="D (\<not>. (smap (x k) lit)) DFlse" in meta_spec, auto)
apply (erule id_basic_strict, auto)
apply (rule lit_negation_props.strict, auto)
apply (rule strict_smap_fid_exp, auto)
apply (rule lit_negation_props.straight, auto)
apply (rule straight_smap_fid_exp, auto)
apply (erule id_basic_straight, auto)
apply (rule lit_negation_props.strict, auto)
apply (rule strict_smap_fid_exp, auto)
apply (rule lit_negation_props.straight, auto)
apply (rule straight_smap_fid_exp, auto)
done

theorem rewrite_disj_straight [simp]:
  fixes d :: disj
  assumes "\<And>k. xmap_fid k (x k)" and "basic d" and "basic k" and "\<And>d. id\<^sub>B (toConj d) (g d)" 
  shows "straight (rewrite g x k d)"
  using assms

apply (induct d arbitrary: k)
apply auto
apply (rule sD_props.straight, auto)
apply (rule straight_smap_fid_exp, auto)
apply (drule_tac x="g (D (\<not>. (smap (x k) lit)) DFlse) k" in meta_spec)
apply (erule meta_mp)
apply (drule_tac x="D (\<not>. (smap (x k) lit)) DFlse" in meta_spec, auto)
apply (erule id_basic_strict, auto)
apply (rule lit_negation_props.strict, auto)
apply (rule strict_smap_fid_exp, auto)
apply (rule lit_negation_props.straight, auto)
apply (rule straight_smap_fid_exp, auto)
apply (erule id_basic_straight, auto)
apply (rule lit_negation_props.strict, auto)
apply (rule strict_smap_fid_exp, auto)
apply (rule lit_negation_props.straight, auto)
apply (rule straight_smap_fid_exp, auto)
done


theorem rewrite_conj_sem [simp]:
  fixes c :: conj
  assumes "guarded k c" and "\<And>k. xmap_fid k (x k)" and "\<parallel>k\<parallel>\<^sub>R\<^sub>+\<^sub>f = True" and "id\<^sub>T g" and "basic c"
   and "guarded CTrue k" and "basic k" and "\<And>d. id\<^sub>B (toConj d) (g d)" 
  shows "\<parallel>rewrite g x k c\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<parallel>c\<parallel>\<^sub>R\<^sub>+\<^sub>f"
  using assms

apply (induct c arbitrary: k)
apply auto
apply (rule_tac x=disj and E="R\<^sub>+\<^sub>f" in sbeI, auto)
apply (drule(1) guarded_sem, auto)
apply (drule_tac x="g (rewrite g x k disj) k" in meta_spec)
apply (drule meta_mp)
apply (erule guarded_relax)
apply (frule (2) sem_id_true_rev_imp, auto)
apply (erule meta_mp, auto)
apply (rule_tac f="g (rewrite g x k disj)" in id_basic_strict, auto)
apply (rule_tac f="g (rewrite g x k disj)" in id_basic_straight, auto)
done

theorem rewrite_sem [simp]:
  fixes c :: cnf
  assumes "guarded CTrue c" and "\<And>k. xmap_fid k (x k)" and "id\<^sub>T g" and "basic c"
    and "\<And>d. id\<^sub>B (toConj d) (g d)" and "\<parallel>\<forall>. k\<parallel>\<^sub>E = True" and "guarded CTrue k" and "basic k"
  shows "\<parallel>rewrite g x k c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms

apply (cases c)
apply (auto, rule sbeI, auto)
apply (subst rewrite_conj_sem, auto)
apply (subst (asm) rewrite_conj_sem, auto)
apply (subst rewrite_conj_sem, auto)
apply (subst rewrite_conj_sem, auto)
done







(* Exploiting Knowledge *)

fun asFact :: "disj \<rightharpoonup> lit" where
  "asFact (D l DFlse) = l\<^sub>!"
| "asFact _ = None"

fun rewriteTrans :: "lit option \<Rightarrow> lit \<rightharpoonup> lit" where
  "rewriteTrans (L True (Equal a b))\<^sub>! (L e (Equal c d)) = (
    if a = c then (sL e (b =. d))\<^sub>! else
    if a = d then (sL e (c =. b))\<^sub>! else
    if b = c then (sL e (a =. d))\<^sub>! else
    if b = d then (sL e (c =. a))\<^sub>! else None)"
| "rewriteTrans (L True (Equal a b))\<^sub>! (L e (Less c d)) = (
    if a = c then (sL e (b <. d))\<^sub>! else
    if a = d then (sL e (c <. b))\<^sub>! else
    if b = c then (sL e (a <. d))\<^sub>! else
    if b = d then (sL e (c <. a))\<^sub>! else None)"
| "rewriteTrans _ _ = None"  (* should probably do some stuff with Less facts, too *)

theorem rewriteTrans_sem [dest]:
  assumes "rewriteTrans k\<^sub>! l = m\<^sub>!" and "\<parallel>k\<parallel>\<^sub>E = True" and "\<parallel>l\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>m\<parallel>\<^sub>E = \<parallel>l\<parallel>\<^sub>E"
  using assms by (cases "(k\<^sub>!, l)" rule: rewriteTrans.cases, auto) (split if_splits, auto)+


fun provable_from :: "conj \<Rightarrow> lit \<Rightarrow> bool" where
  "provable_from (C d c) l \<longleftrightarrow> d = False. \<or> d = toDisj l \<or> provable_from c l \<or>
    option_case False (provable_from c) (rewriteTrans (asFact d) l)"
| "provable_from CTrue l   \<longleftrightarrow> False"

theorem provable_from_sem:
  assumes "provable_from c l" and "\<parallel>c\<parallel>\<^sub>E = True" and "\<parallel>l\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>l\<parallel>\<^sub>E = True"
  using assms

apply (induct c l rule: provable_from.induct)
apply auto
apply (case_tac d rule: asFact.cases, auto)
apply (case_tac "rewriteTrans la\<^sub>! l", auto)
apply (drule_tac x=a in meta_spec, auto)
done

abbreviation notIn where
  "notIn v x \<equiv> xprop (XFold (op \<noteq> v) c\<^sub>T c\<^sub>T c\<^sub>T) x"

fun reducible :: "conj \<Rightarrow> valueExp \<rightharpoonup> valueExp" where
  "reducible (C (D (L True (Equal a b)) DFlse) c) x =
     (if a = x \<and> notIn a b then b\<^sub>! else reducible c x)"
| "reducible (C d c) x = reducible c x"
| "reducible _ _ = None"

theorem reducible_sem:
  assumes "reducible c a = b\<^sub>!" and "\<parallel>c\<parallel>\<^sub>E = True"
  shows "\<parallel>b\<parallel>\<^sub>E = \<parallel>a\<parallel>\<^sub>E"
  using assms by (induct c a rule: reducible.induct) (auto, split if_splits, auto)

theorem reducible_strict [simp]:
  assumes "strict k" and "reducible k x = a\<^sub>!"
  shows "strict a"
  using assms by (induct k x rule: reducible.induct) (auto, split if_splits, auto)

theorem reducible_straight [simp]:
  assumes "straight k" and "reducible k x = a\<^sub>!"
  shows "straight a"
  using assms by (induct k x rule: reducible.induct) (auto, split if_splits, auto)


(* Concrete Fact Simplification *)

fun trivExp :: "conj \<Rightarrow> valueExp \<Rightarrow> valueExp" where
  "trivExp k (Count a b) = (
     if (sUnique a \<or> provable_from k (size. a =. 1.)) \<and> provable_from k (a \<noteq>. b) then 0. else
     if provable_from k (size. a =. 0.) then 0. else (* seems unnecessary since size normalization *)
     Count a b)"
| "trivExp k (Read d p) = (
     if provable_from k (size. (Read d p) =. 0.) then {}. else Read d p)"
| "trivExp k v = v"  (*  if provable_from k (a =. b) then 1. else  *)

theorem trivExp_sem [simp]:
  assumes "\<parallel>k\<parallel>\<^sub>E = True" and "\<parallel>v\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>trivExp k v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E"
  using assms by (cases "(k, v)" rule: trivExp.cases) (auto dest!:sUnique_sem provable_from_sem)

corollary trivExp_fact_id [simp]:
  shows "id\<^sub>F k (trivExp k)"
  unfolding id_fact_def by simp

theorem trivExp_id_basic [simp]:
  shows "id\<^sub>B k (trivExp k)"
  unfolding id_basic_def
  by auto (case_tac "(k, x)" rule: trivExp.cases, auto)+


definition trivEq :: "conj \<Rightarrow> valueExp \<Rightarrow> valueExp" where
  "trivEq k a \<equiv> case reducible k a of Some b \<Rightarrow> b | None \<Rightarrow> a"

theorem trivEq_sem [simp]:
  assumes "\<parallel>k\<parallel>\<^sub>E = True"
  shows "\<parallel>trivEq k v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E"
  using assms unfolding trivEq_def
  by (auto split: option.split dest!: reducible_sem)

corollary trivEq_fact_id [simp]:
  shows "id\<^sub>F k (trivEq k)"
  unfolding id_fact_def by simp

theorem trivEq_id_basic [simp]:
  shows "id\<^sub>B k (trivEq k)"
  using assms unfolding id_basic_def and trivEq_def
  by auto (split option.split, auto)+


definition trivLit :: "conj \<Rightarrow> lit \<Rightarrow> lit" where
  "trivLit c l \<equiv>
    if provable_from c l then True. else
    if provable_from c (\<not>. l) then False. else l"

theorem trivLit_sem [simp]:
  assumes "\<parallel>k\<parallel>\<^sub>E = True" and "\<parallel>l\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>trivLit k l\<parallel>\<^sub>E = \<parallel>l\<parallel>\<^sub>E"
  unfolding trivLit_def using assms
  by (auto dest!: provable_from_sem)

corollary trivLit_fact_id [simp]:
  shows "id\<^sub>F k (trivLit k)"
  unfolding id_fact_def by simp

theorem triv_Lit_id_basic [simp]:
  shows "id\<^sub>B x (trivLit k)"
  unfolding id_basic_def and trivLit_def by auto


abbreviation factSimp where
  "factSimp k \<equiv> (XMap (trivEq k \<circ> trivExp k) id id id (trivLit k) id id)"

theorem fact_id_conc [simp]:
  assumes "id\<^sub>F k (a k)" an "id\<^sub>F k (b k)"
  shows "id\<^sub>F k (a k \<circ> b k)"
  using assms unfolding id_fact_def by simp

theorem basic_id_conc [simp]:
  assumes "id\<^sub>B k (a k)" and "id\<^sub>B k (b k)"
  shows "id\<^sub>B k (a k \<circ> b k)"
  using assms unfolding id_basic_def by simp

theorem factSimp_xmap_fid [simp]:
  shows "xmap_fid k (factSimp k)"
  using assms by auto



fun deriveK :: "disj \<Rightarrow> conj \<Rightarrow> conj" where
  "deriveK (D l DFlse) c = sguard l \<and>. (smap (factSimp (sguard l)) c)"
| "deriveK d c = c \<and>. d"

theorem deriveK_id_basic [simplified, simp]:
  shows "id\<^sub>B (toConj k) (deriveK k)"
  unfolding id_basic_def

apply auto
apply (case_tac "(k, x)" rule: deriveK.cases, auto)
apply (rule conj_conj_props.strict, auto)
apply (rule strict_smap_fid_conj, rule factSimp_xmap_fid, auto)
apply (case_tac "(k, x)" rule: deriveK.cases, auto)
apply (rule conj_conj_props.straight, auto)
apply (rule straight_smap_fid_conj, rule factSimp_xmap_fid, auto)
done



theorem guarded_smap_fid_disj [simp]:
  fixes d :: disj and x :: conj assumes "guarded (k \<and>. x) d"
  shows "guarded (k \<and>. x) (smap (factSimp k) d)"
  using assms
apply (induct d arbitrary: x)
apply auto
apply (subst (asm) sem_smap_fid_exp, rule factSimp_xmap_fid, auto)
apply (drule_tac x="\<not>. (smap (factSimp k) lit) \<and>. x" in meta_spec)
apply (drule meta_mp, auto)
apply (erule guarded_relax, auto)
apply (subst (asm) sem_smap_fid_exp, rule factSimp_xmap_fid, auto)
done

theorem guarded_smap_fid_conj [simp]:
  fixes c :: conj and x :: conj assumes "guarded (k \<and>. x) c"
  shows "guarded (k \<and>. x) (smap (factSimp k) c)"
  using assms

apply (induct c arbitrary: x)
apply auto
apply (drule_tac x="smap (factSimp k) disj \<and>. x" in meta_spec)
apply (drule meta_mp)
apply auto
apply (erule guarded_relax, auto)
apply (subst (asm) sem_smap_fid_disj, rule factSimp_xmap_fid, auto)
apply (drule guarded_sem, auto)
done


theorem deriveK_id_true [simp]:
  shows "id\<^sub>T deriveK"
  unfolding id_true_def

apply clarify
apply (case_tac "(d, c)" rule: deriveK.cases, auto)
apply auto
apply (subst sem_smap_fid_conj, rule factSimp_xmap_fid, auto)
apply (subst (asm) sem_smap_fid_conj, rule factSimp_xmap_fid, auto)
apply (drule_tac R=R and f=f in guarded_sem, auto)
apply (rule guarded_smap_fid_conj)
apply (erule guarded_relax, auto)
done

fun useEqs :: "cnf \<Rightarrow> cnf \<Rightarrow> cnf" where
  "useEqs (\<forall>. k) c = rewrite deriveK factSimp k c"

corollary useEqs_sem:
  assumes "guarded CTrue c" and "basic c" and "\<parallel>k\<parallel>\<^sub>E = True" and "guarded CTrue k" and "basic k"
  shows "\<parallel>useEqs k c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by (cases k) simp


(* Invariant Simplification *)

fun inv_simp_conj :: "conj \<Rightarrow> conj \<Rightarrow> conj" where
  "inv_simp_conj CTrue x = x"
| "inv_simp_conj (C d c) x = inv_simp_conj c (takeoutD d x)"

theorem inv_simp_conj_sem [simp]:
  assumes "\<parallel>i\<parallel>\<^sub>E = True"
  shows "\<parallel>inv_simp_conj i c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by (induct i c rule: inv_simp_conj.induct) auto

fun inv_simp where
  "inv_simp (\<forall>. i) (\<forall>. c) = \<forall>. (inv_simp_conj i c)"

theorem inv_simp_sem:
  assumes "\<parallel>i\<parallel>\<^sub>E = True"
  shows "\<parallel>inv_simp i c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by (cases c, cases i, cases E, auto) (rule sbeI, auto)+

end

