header {* Guarded Formulas *}

theory Guarded imports Sugar
begin


(* Concept of Guarded Formulas *)

class guarded = sem +
fixes guarded :: "conj \<Rightarrow> 'a \<Rightarrow> bool"
assumes guarded_sem [elim]:
  "guarded k x \<Longrightarrow> \<parallel>k\<parallel>\<^sub>R\<^sub>+\<^sub>f = True \<Longrightarrow> \<parallel>x\<parallel>\<^sub>R\<^sub>+\<^sub>f \<noteq> \<bottom>"
and guarded_relax [elim]:
  "guarded k x \<Longrightarrow> (\<And>R f. \<parallel>s\<parallel>\<^sub>R\<^sub>+\<^sub>f = True \<Longrightarrow> \<parallel>k\<parallel>\<^sub>R\<^sub>+\<^sub>f = True) \<Longrightarrow> guarded s x"


instantiation lit and disj and conj and cnf :: guarded begin
  fun guarded_lit :: "conj \<Rightarrow> lit \<Rightarrow> bool" where
    "guarded_lit k l   \<longleftrightarrow> (\<forall>R f. \<parallel>k\<parallel>\<^sub>R\<^sub>+\<^sub>f = True \<longrightarrow> \<parallel>l\<parallel>\<^sub>R\<^sub>+\<^sub>f \<noteq> \<bottom>)"

  fun guarded_disj :: "conj \<Rightarrow> disj \<Rightarrow> bool" where
    "guarded k DFlse   \<longleftrightarrow> True"
  | "guarded k (D l d) \<longleftrightarrow> guarded k l \<and> guarded (\<not>. l \<and>. k) d"

  fun guarded_conj :: "conj \<Rightarrow> conj \<Rightarrow> bool" where
    "guarded k CTrue   \<longleftrightarrow> True"
  | "guarded k (C d c) \<longleftrightarrow> guarded k d \<and> guarded (d \<and>. k) c"

  fun guarded_cnf :: "conj \<Rightarrow> cnf \<Rightarrow> bool" where
    "guarded k (\<forall>. c)  \<longleftrightarrow> guarded True. c"

  theorem guarded_disj_relax:
    fixes d :: disj
    assumes "guarded k d" and "\<And>R f. \<parallel>s\<parallel>\<^sub>R\<^sub>+\<^sub>f = True \<Longrightarrow> \<parallel>k\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
    shows "guarded s d"
    using assms
  proof (induct d arbitrary: s k)
    case (D l d) show ?case using D.hyps[of "\<not>. l \<and>. k" "\<not>. l \<and>. s"] and D.prems by simp
  qed simp

  theorem guarded_conj_relax:
    fixes c :: conj
    assumes "guarded k c" and "\<And>R f. \<parallel>s\<parallel>\<^sub>R\<^sub>+\<^sub>f = True \<Longrightarrow> \<parallel>k\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
    shows "guarded s c"
    using assms
  proof (induct c arbitrary: s k)
    case (C d c) show ?case using C.hyps[of "d \<and>. k" "d \<and>. s"] and C.prems
      by (auto elim: guarded_disj_relax)
  qed simp

  theorem guarded_disj_sem:
    fixes d :: disj
    assumes "guarded k d" and "\<parallel>k\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
    shows "\<parallel>d\<parallel>\<^sub>R\<^sub>+\<^sub>f \<noteq> \<bottom>"
    using assms
    by (induct k d rule: guarded_disj.induct)
       (auto, drule_tac x=R in spec, drule_tac x=f in spec, auto)

  theorem guarded_conj_sem:
    fixes c :: conj
    assumes "guarded k c" and "\<parallel>k\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
    shows "\<parallel>c\<parallel>\<^sub>R\<^sub>+\<^sub>f \<noteq> \<bottom>"
    using assms by (induct k c rule: guarded_conj.induct) (auto dest: guarded_disj_sem)

  theorem guarded_cnf_sem:
    fixes c :: cnf
    assumes "guarded k c" and "\<parallel>k\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
    shows "\<parallel>c\<parallel>\<^sub>R\<^sub>+\<^sub>f \<noteq> \<bottom>"
    using assms by (cases c) (auto, drule_tac R=R and f=fa in guarded_conj_sem, auto)

instance by intro_classes (safe elim!: guarded_disj_sem guarded_conj_sem  guarded_cnf_sem,
  auto elim!: guarded_disj_relax guarded_conj_relax, case_tac x, auto) end


(* Guarded Implications on Conjunctions *)

fun disj_imp_conj :: "disj \<Rightarrow> conj \<Rightarrow> conj" ("_ \<Longrightarrow>. _" [700, 700] 700) where
  "DFlse \<Longrightarrow>. c = CTrue"
| "D l d \<Longrightarrow>. c = (l \<longrightarrow>. c) \<and>. (\<not>.l \<longrightarrow>. (d \<Longrightarrow>. c))"

theorem sem_disj_imp_conj_general:
  assumes "guarded k d" and "\<parallel>k\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
  shows "\<parallel>d \<Longrightarrow>. c\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<not>\<^sub>? \<parallel>d\<parallel>\<^sub>R\<^sub>+\<^sub>f \<or>\<^sub>? \<parallel>c\<parallel>\<^sub>R\<^sub>+\<^sub>f"
  using assms
proof (induct d c arbitrary: k rule: disj_imp_conj.induct)
  case (2 l d) show ?case using "2.hyps"[of "\<not>. l \<and>. k"] and "2.prems"
    by auto (drule spec, drule spec, auto)
qed simp

corollary sem_disj_imp_conj [simp]:
  assumes "guarded CTrue d"
  shows "\<parallel>d \<Longrightarrow>. c\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<not>\<^sub>? \<parallel>d\<parallel>\<^sub>R\<^sub>+\<^sub>f \<or>\<^sub>? \<parallel>c\<parallel>\<^sub>R\<^sub>+\<^sub>f"
  using assms by (rule sem_disj_imp_conj_general) simp

interpretation disj_imp_conj_props:
  props_eq "d \<Longrightarrow>. c" "C d c"
  by unfold_locales (induct d c rule: disj_imp_conj.induct, auto)+
  

(* Behavior of Sugar Regarding Guarded *)

theorem takeoutL_guarded:
  fixes k :: conj
  assumes "guarded (\<not>. l \<and>. k) d"
  shows "guarded (\<not>. l \<and>. k) (takeoutL l d)"
  using assms

apply (induct l d arbitrary: k rule: takeoutL.induct)
apply auto
apply (drule_tac x=k in meta_spec)
apply (erule meta_mp, erule guarded_relax, auto)
apply (rotate_tac 1)
apply (drule_tac x="\<not>. l \<and>. k" in meta_spec)
apply (drule meta_mp, auto)
done

theorem sD_guarded [intro!]:
  assumes "guarded k (D l d)"
  shows "guarded k (sD l d)"
  using assms unfolding sD_def
  by auto (erule takeoutL_guarded)

theorem disj_conc_guarded:
  assumes "guarded k a" and "guarded (.\<not> a \<and>. k) b"
  shows "guarded k (disj_conc a b)"
  using assms

apply (induct a b arbitrary: k rule: disj_conc.induct)
apply auto
apply (drule_tac x="\<not>. l \<and>. k" in meta_spec)
apply auto
apply (erule meta_mp)
apply (drule_tac l=l in takeoutL_guarded)
apply (erule guarded_relax)
apply auto
done

corollary disj_disj_guarded [intro!]:
  assumes "guarded k (toDisj a)" and "guarded (.\<not> (toDisj a) \<and>. k) (toDisj b)"
  shows "guarded k (a \<or>. b)"
  using assms unfolding disj_disj_def by (rule disj_conc_guarded)


lemma takeoutD_guarded:
  fixes k :: conj
  assumes "guarded (d \<and>. k) c"
  shows "guarded (d \<and>. k) (takeoutD d c)"
  using assms

apply (induct d c arbitrary: k rule: takeoutD.induct)
apply auto
apply (drule_tac x=k in meta_spec)
apply (erule meta_mp, erule guarded_relax, auto)
apply (rotate_tac 1)
apply (drule_tac x="d \<and>. k" in meta_spec)
apply (drule_tac meta_mp, auto)
done

theorem sC_guarded [intro!]:
  assumes "guarded k (C d c)"
  shows "guarded k (sC d c)"
  using assms unfolding sC_def
  by auto (rule takeoutD_guarded, auto)

theorem conj_conc_guarded:
  assumes "guarded k a" and "guarded (a \<and>. k) b"
  shows "guarded k (conj_conc a b)"
  using assms
apply (induct a b arbitrary: k rule: conj_conc.induct)
apply auto
apply (drule_tac x="d \<and>. k" in meta_spec)
apply auto
apply (erule meta_mp)
apply (drule_tac d=d in takeoutD_guarded)
apply (erule guarded_relax)
apply auto
done

corollary conj_conj_guarded [intro!]:
  assumes "guarded k (toConj a)" and "guarded (toConj a \<and>. k) (toConj b)"
  shows "guarded k (a \<and>. b)"
  using assms by (subst conj_conj_def) (rule conj_conc_guarded)


theorem disj_negation_guarded [intro!]:
  assumes "guarded k d"
  shows "guarded k (.\<not> d)"
  using assms

apply (induct d arbitrary: k rule: disj_negation.induct)
apply auto
apply (drule_tac x="D (\<not>. l) DFlse \<and>. k" in meta_spec)
apply (erule meta_mp, auto)
done


theorem disj_conc_conj_guarded:
  assumes "guarded k d" and "guarded (.\<not> d \<and>. k) c"
  shows "guarded k (disj_conc_conj d c)"
  using assms

apply (induct d c arbitrary: k rule: disj_conc_conj.induct)
apply auto
apply (drule_tac x="C (d \<or>. e) CTrue \<and>. k" in meta_spec)
apply (drule meta_mp)
apply auto
apply (erule meta_mp)
apply (erule guarded_relax)
apply auto
done

corollary disj_disj_conj_guarded [intro!]:
  assumes "guarded k (toDisj d)" and "guarded (.\<not> (toDisj d) \<and>. k) c"
  shows "guarded k (d .\<or> c)"
  using assms unfolding disj_disj_conj_def by (rule disj_conc_conj_guarded)


theorem disj_imp_conj_guarded [intro!]:
  assumes "guarded k d" and "guarded (d \<and>. k) c"
  shows "guarded k (d \<Longrightarrow>. c)"
  using assms
apply (induct d c arbitrary: k rule: disj_imp_conj.induct)
apply auto
apply (erule guarded_relax, auto)
apply (drule_tac x="\<not>. l \<and>. k" in meta_spec, auto)
apply (rule_tac k="\<not>. l \<and>. k" in guarded_relax, auto)
apply (erule meta_mp, auto)
done

end

