header {* Path Count and Path Intersection *}

theory Intersection imports Safe Conditionals begin


(* Path Count & Containment *)

fun pcount :: "pathExp \<Rightarrow> pathExp \<Rightarrow> valueExp" where
  "pcount Root. Root.     = 1."
| "pcount (Kind m q)   (Path l p k) = (if l = m then pcount q p else 0.)"
| "pcount (Path m q r) (Path l p k) = (if l = m then (pcount q p) \<odot>. (count. r k) else 0.)"

| "pcount (PRep n q) p    = n \<odot>. (pcount q p)"
| "pcount (PPlus a b) p   = ((pcount a p) +. (pcount b p))"
| "pcount (PMinus a b) p  = ((pcount a p) -. (pcount b p))"

| "pcount (PCond b t e) p = VCond b (pcount t p) (pcount e p)"
| "pcount q (PCond b t e) = VCond b (pcount q t) (pcount q e)"

| "pcount _ _ = 0."


abbreviation lit_contained :: "pathExp \<Rightarrow> pathExp \<Rightarrow> pred" ("_ .\<in> _") where
  "p .\<in> q \<equiv> 0. <. pcount q p "

abbreviation lit_not_contained :: "pathExp \<Rightarrow> pathExp \<Rightarrow> pred" ("_ .\<notin> _") where
  "p .\<notin> q \<equiv> pcount q p =. 0."

interpretation pcount_prop_ve:
  prop_ve "pcount q p" "0."
  by unfold_locales (induct q p rule: pcount.induct, auto)

interpretation pcount_props:
  props_eq "pcount q p" "Count (Read $. q) (Read $. p)"
  by unfold_locales (induct q p rule: pcount.induct, auto)+


abbreviation domPCount ("pcount\<^sub>? _ _" [999, 999] 999) where
  "domPCount \<equiv> liftDom2 (semCount :: path mset \<Rightarrow> path mset \<Rightarrow> value mset option)"

theorem pcount_sem [simp]:
  assumes "upe p" and "\<parallel>p\<parallel>\<^sub>E = {r}\<^sub>*\<^sub>p" and "\<parallel>q\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>pcount q p\<parallel>\<^sub>E = pcount\<^sub>? \<parallel>q\<parallel>\<^sub>E \<parallel>p\<parallel>\<^sub>E"
  using assms
proof (induct q p arbitrary: r rule: pcount.induct)
  case 3 thus ?case by auto (drule_tac E=E in uve_sem, auto, metis int_mult) next
  case 4 thus ?case by auto (metis int_mult int_nat_eq) next
  case "8_1" thus ?case by auto (split if_splits, auto, rule sbeI, auto) next
  case "8_2" thus ?case by auto (split if_splits, auto, rule sbeI, auto) next
  case "8_3" thus ?case by auto (split if_splits, auto, rule sbeI, auto) next
qed auto


(* Intersecting Paths *)

fun intersect :: "pathExp \<Rightarrow> pathExp \<Rightarrow> pathExp" where
  "intersect Root. p   = (pcount Root. p) .\<otimes> Root."
| "intersect (Path l r k) p = (if kindOf (Path l r k) = kindOf p then
     (pcount r/l[k]. p) .\<otimes> p else (intersect r p)/l[k].)"
| "intersect (Kind l r) p   = (if kindOf (Kind l r)   = kindOf p then
     (pcount r/l. p)    .\<otimes> p else (intersect r p)/l.)"

| "intersect (PRep n r)    p = n .\<otimes> (intersect r p)"
| "intersect (PPlus r s)   p = (intersect r p) .\<oplus> (intersect s p)"
| "intersect (PMinus r s)  p = (intersect r p) .\<ominus> (intersect s p)"

| "intersect (PCond b t e) p = PCond b (intersect t p) (intersect e p)"

theorem intersect_kind [simp]:
  "kindOf (intersect q p) = kindOf q"
  by (induct q p rule: intersect.induct) auto

theorem intersect_sem [simp]:
  assumes "\<parallel>qe\<parallel>\<^sub>E = q\<^sub>p" and "\<parallel>pe\<parallel>\<^sub>E = {p}\<^sub>*\<^sub>p" and "upe pe"
  shows "\<parallel>intersect qe pe\<parallel>\<^sub>E = (q\<down>(desc p))\<^sub>p"
  using assms
proof (induct qe pe arbitrary: p q rule: intersect.induct)
  case 1 thus ?case by auto (case_tac "p = Root", auto) next

  case (2 l re ke pe) then obtain k where kpe: "kindOf pe = k\<^sub>!" by (cases "kindOf pe") auto show ?case
  proof (cases "kindOf re/l[ke]. = kindOf pe")
    case True thus ?thesis using 2 and kpe by auto (subst reduceM_desc_kind_eq, auto) next
    case False thus ?thesis using 2 and kpe by (case_tac "kindOf re", auto) (subst reduceM_desc_kind_neq, auto)
  qed next

  case (3 l re pe) then obtain k where kpe: "kindOf pe = k\<^sub>!" by (cases "kindOf pe") auto show ?case
  proof (cases "kindOf re/l. = kindOf pe")
    case True thus ?thesis using 3 and kpe by auto (subst reduceM_desc_kind_eq, auto) next
    case False thus ?thesis using 3 and kpe by (case_tac "kindOf re", auto) (subst reduceM_desc_kind_neq, auto)
  qed next

  case (5 a b c) thus ?case by auto (split if_splits, auto) next
  case 7 thus ?case by auto (rule sbeI, auto, split if_splits, auto)
qed auto

interpretation intersect_props:
  props_base  "intersect q p" "Count (Read $. q) (Read $. p)" +
  prop_dufree "intersect q p" "Count (Read $. q) (Read $. p)"
  by (unfold_locales, auto) (induct q p rule: intersect.induct, auto)+

end

