header {* Mapping and Folding Frameworks *}

theory MapFold imports Assertions begin


(* Mapping Functions *)

datatype xmap =
  XMap "valueExp \<Rightarrow> valueExp" "pathExp \<Rightarrow> pathExp" "docExp \<Rightarrow> docExp"
       "pred \<Rightarrow> pred" "lit \<Rightarrow> lit" "disj \<Rightarrow> disj" "conj \<Rightarrow> conj" 

class xmap =
fixes xmap :: "xmap \<Rightarrow> 'a \<Rightarrow> 'a"
  and xm   :: "xmap \<Rightarrow> 'a \<Rightarrow> 'a"
assumes xmap_id [simp]:
  "xmap (XMap id id id id id id id) x = x"


instantiation valueExp and pathExp and docExp and pred and lit :: xmap begin 

fun xm_valueExp :: "xmap \<Rightarrow> valueExp \<Rightarrow> valueExp"
and xm_pathExp  :: "xmap \<Rightarrow> pathExp  \<Rightarrow> pathExp"
and xm_docExp   :: "xmap \<Rightarrow> docExp   \<Rightarrow> docExp"
and xm_pred     :: "xmap \<Rightarrow> pred     \<Rightarrow> pred"
and xm_lit      :: "xmap \<Rightarrow> lit      \<Rightarrow> lit" where

  "xm (XMap f _ _ _ _ _ _) = f"
| "xm (XMap _ f _ _ _ _ _) = f"
| "xm (XMap _ _ f _ _ _ _) = f"
| "xm (XMap _ _ _ f _ _ _) = f"
| "xm (XMap _ _ _ _ f _ _) = f"

fun xmap_valueExp :: "xmap \<Rightarrow> valueExp \<Rightarrow> valueExp"
and xmap_pathExp  :: "xmap \<Rightarrow> pathExp  \<Rightarrow> pathExp"
and xmap_docExp   :: "xmap \<Rightarrow> docExp   \<Rightarrow> docExp"
and xmap_pred     :: "xmap \<Rightarrow> pred     \<Rightarrow> pred"
and xmap_lit      :: "xmap \<Rightarrow> lit      \<Rightarrow> lit" where

  "xmap f (Val v)       = xm f (Val v)"
| "xmap f (FVar n)      = xm f (FVar n)"
| "xmap f (PVar n)      = xm f (PVar n)"
| "xmap f (Sum m)       = xm f (Sum     (xmap f m))"
| "xmap f (Size m)      = xm f (Size    (xmap f m))"
| "xmap f (Count m v)   = xm f (Count   (xmap f m) (xmap f v))"
| "xmap f (Tally t m)   = xm f (Tally t (xmap f m))"
| "xmap f (Read d p)    = xm f (Read    (xmap f d) (xmap f p))"
| "xmap f (Scal n m)    = xm f (Scal    (xmap f n) (xmap f m))"
| "xmap f (VRep n m)    = xm f (VRep    (xmap f n) (xmap f m))"
| "xmap f (VPlus a b)   = xm f (VPlus   (xmap f a) (xmap f b))"
| "xmap f (VMinus a b)  = xm f (VMinus  (xmap f a) (xmap f b))"
| "xmap f (VCond l a b) = xm f (VCond   (xmap f l) (xmap f a) (xmap f b))"

| "xmap f PRoot         = xm f PRoot"
| "xmap f (Path l p v)  = xm f (Path l  (xmap f p) (xmap f v))"
| "xmap f (Kind l p)    = xm f (Kind l  (xmap f p))"
| "xmap f (PRep n m)    = xm f (PRep    (xmap f n) (xmap f m))"
| "xmap f (PPlus a b)   = xm f (PPlus   (xmap f a) (xmap f b))"
| "xmap f (PMinus a b)  = xm f (PMinus  (xmap f a) (xmap f b))"
| "xmap f (PCond l a b) = xm f (PCond   (xmap f l) (xmap f a) (xmap f b))"

| "xmap f Init          = xm f Init"
| "xmap f (Upd d p v)   = xm f (Upd     (xmap f d) (xmap f p) (xmap f v))"
| "xmap f (Del d p)     = xm f (Del     (xmap f d) (xmap f p))"
| "xmap f (DCond l a b) = xm f (DCond   (xmap f l) (xmap f a) (xmap f b))"

| "xmap f (Equal a b)   = xm f (Equal   (xmap f a) (xmap f b))"
| "xmap f (Less a b)    = xm f (Less    (xmap f a) (xmap f b))"
| "xmap f (RCond l a b) = xm f (RCond   (xmap f l) (xmap f a) (xmap f b))"

| "xmap f (L b r)       = xm f (L b (xmap f r))"

lemma xmap_sanity_id:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "xmap (XMap id id id id id id id) v = v"
    and "xmap (XMap id id id id id id id) p = p"
    and "xmap (XMap id id id id id id id) d = d"
    and "xmap (XMap id id id id id id id) r = r"
    and "xmap (XMap id id id id id id id) l = l"
  by (induct v and p and d and r and l) auto

instance by intro_classes (rule xmap_sanity_id)+ end


instantiation disj :: xmap begin

  fun xm_disj :: "xmap \<Rightarrow> disj \<Rightarrow> disj" where
    "xm (XMap _ _ _ _ _ f _) = f"

  fun xmap_disj :: "xmap \<Rightarrow> disj \<Rightarrow> disj" where
    "xmap f (D l d) = xm f (D (xmap f l) (xmap f d))"
  | "xmap f DFlse   = xm f DFlse"

instance by intro_classes (induct_tac x, auto, fold id_def, simp)+ end


instantiation conj :: xmap begin

  fun xm_conj :: "xmap \<Rightarrow> conj \<Rightarrow> conj" where
    "xm (XMap _ _ _ _ _ _ f) = f"

  fun xmap_conj :: "xmap \<Rightarrow> conj \<Rightarrow> conj" where
    "xmap f (C d c) = xm f (C (xmap f d) (xmap f c))"
  | "xmap f CTrue   = xm f CTrue"

instance by intro_classes (induct_tac x, auto, fold id_def, simp)+ end


instantiation cnf :: xmap begin

  fun xm_cnf :: "xmap \<Rightarrow> cnf \<Rightarrow> cnf" where
    "xm f = (id :: cnf \<Rightarrow> cnf)"

  fun xmap_cnf :: "xmap \<Rightarrow> cnf \<Rightarrow> cnf" where
    "xmap f (\<forall>. c) = \<forall>. (xmap f c)"

instance by intro_classes (induct_tac x, auto, fold id_def, simp)+ end


(* Folding Functions *)

datatype 'a xfold =
  XFold "valueExp \<Rightarrow> 'a" "pathExp \<Rightarrow> 'a" "docExp \<Rightarrow> 'a" "pred \<Rightarrow> 'a"

abbreviation const0 :: "'a \<Rightarrow> nat" ("c\<^sub>0") where
  "c\<^sub>0 \<equiv> \<lambda>x. 0"

abbreviation constT :: "'a \<Rightarrow> bool" ("c\<^sub>T") where
  "c\<^sub>T \<equiv> \<lambda>x. True"

abbreviation constE :: "'a \<Rightarrow> 'b set" ("c\<^sub>E") where
  "c\<^sub>E \<equiv> \<lambda>x. {}"

abbreviation constN :: "'a \<Rightarrow> 'b list" ("c\<^sub>L") where
  "c\<^sub>L \<equiv> \<lambda>x. []"

fun toXP :: "('a \<Rightarrow> nat) \<Rightarrow> ('a \<Rightarrow> bool)" where
  "toXP c x \<longleftrightarrow> c x = 0"

lemma toXP_c0_cT [simp]:
  shows "toXP c\<^sub>0 = c\<^sub>T"
  by (rule ext) simp


class xfold = 
fixes xprop  :: "bool     xfold \<Rightarrow> 'a \<Rightarrow> bool" 
  and xcount :: "nat      xfold \<Rightarrow> 'a \<Rightarrow> nat"
  and xvars  :: "name set xfold \<Rightarrow> 'a \<Rightarrow> name set"
  and xlits  :: "lit list xfold \<Rightarrow> 'a \<Rightarrow> lit list"
  and xp     :: "bool     xfold \<Rightarrow> 'a \<Rightarrow> bool"
  and xc     :: "nat      xfold \<Rightarrow> 'a \<Rightarrow> nat"
  and xv     :: "name set xfold \<Rightarrow> 'a \<Rightarrow> name set"
  and xl     :: "lit list xfold \<Rightarrow> 'a \<Rightarrow> lit list"
assumes xprop_xp [dest]:
  "xprop p x \<Longrightarrow> xp p x"
and xcount_xc [dest]:
  "xcount c x = 0 \<Longrightarrow> xc c x = 0"
and xvars_xv [simp]:
  "xv v x \<subseteq> xvars v x"
and xlits_xl [simp]:
  "set (xl l x) \<subseteq> set (xlits l x)"
and xprop_true [simp]:
  "xprop (XFold c\<^sub>T c\<^sub>T c\<^sub>T c\<^sub>T) x"
and xcount_zero [simp]:
  "xcount (XFold c\<^sub>0 c\<^sub>0 c\<^sub>0 c\<^sub>0) x = 0"
and xvars_empty [simp]:
  "xvars (XFold c\<^sub>E c\<^sub>E c\<^sub>E c\<^sub>E) x = {}"
and xlits_empty [simp]:
  "xlits (XFold c\<^sub>L c\<^sub>L c\<^sub>L c\<^sub>L) x = []"
and xcount_to_xprop:
  "toXP ac = ap \<Longrightarrow> toXP bc = bp \<Longrightarrow> toXP cc = cp \<Longrightarrow> toXP dc = dp \<Longrightarrow>
     xcount (XFold ac bc cc dc) x = 0 \<longleftrightarrow> xprop (XFold ap bp cp dp) x"


instantiation valueExp and pathExp and docExp and pred and lit :: xfold begin 

fun xp_valueExp :: "bool xfold \<Rightarrow> valueExp \<Rightarrow> bool"
and xp_pathExp  :: "bool xfold \<Rightarrow> pathExp  \<Rightarrow> bool"
and xp_docExp   :: "bool xfold \<Rightarrow> docExp   \<Rightarrow> bool"
and xp_pred     :: "bool xfold \<Rightarrow> pred     \<Rightarrow> bool"
and xp_lit      :: "bool xfold \<Rightarrow> lit      \<Rightarrow> bool" where

  "xp (XFold f _ _ _) = f"
| "xp (XFold _ f _ _) = f"
| "xp (XFold _ _ f _) = f"
| "xp (XFold _ _ _ f) = f"
| "xp_lit _ = c\<^sub>T"

fun xc_valueExp :: "nat xfold \<Rightarrow> valueExp \<Rightarrow> nat"
and xc_pathExp  :: "nat xfold \<Rightarrow> pathExp  \<Rightarrow> nat"
and xc_docExp   :: "nat xfold \<Rightarrow> docExp   \<Rightarrow> nat"
and xc_pred     :: "nat xfold \<Rightarrow> pred     \<Rightarrow> nat"
and xc_lit      :: "nat xfold \<Rightarrow> lit      \<Rightarrow> nat" where

  "xc (XFold f _ _ _) = f"
| "xc (XFold _ f _ _) = f"
| "xc (XFold _ _ f _) = f"
| "xc (XFold _ _ _ f) = f"
| "xc_lit _ = c\<^sub>0"

fun xv_valueExp :: "name set xfold \<Rightarrow> valueExp \<Rightarrow> name set"
and xv_pathExp  :: "name set xfold \<Rightarrow> pathExp  \<Rightarrow> name set"
and xv_docExp   :: "name set xfold \<Rightarrow> docExp   \<Rightarrow> name set"
and xv_pred     :: "name set xfold \<Rightarrow> pred     \<Rightarrow> name set"
and xv_lit      :: "name set xfold \<Rightarrow> lit      \<Rightarrow> name set" where

  "xv (XFold f _ _ _) = f"
| "xv (XFold _ f _ _) = f"
| "xv (XFold _ _ f _) = f"
| "xv (XFold _ _ _ f) = f"
| "xv_lit _ = c\<^sub>E"

fun xl_valueExp :: "lit list xfold \<Rightarrow> valueExp \<Rightarrow> lit list"
and xl_pathExp  :: "lit list xfold \<Rightarrow> pathExp  \<Rightarrow> lit list"
and xl_docExp   :: "lit list xfold \<Rightarrow> docExp   \<Rightarrow> lit list"
and xl_pred     :: "lit list xfold \<Rightarrow> pred     \<Rightarrow> lit list"
and xl_lit      :: "lit list xfold \<Rightarrow> lit      \<Rightarrow> lit list" where

  "xl (XFold f _ _ _) = f"
| "xl (XFold _ f _ _) = f"
| "xl (XFold _ _ f _) = f"
| "xl (XFold _ _ _ f) = f"
| "xl_lit _ = c\<^sub>L"

fun xprop_valueExp :: "bool xfold \<Rightarrow> valueExp \<Rightarrow> bool"
and xprop_pathExp  :: "bool xfold \<Rightarrow> pathExp  \<Rightarrow> bool"
and xprop_docExp   :: "bool xfold \<Rightarrow> docExp   \<Rightarrow> bool"
and xprop_pred     :: "bool xfold \<Rightarrow> pred     \<Rightarrow> bool"
and xprop_lit      :: "bool xfold \<Rightarrow> lit      \<Rightarrow> bool" where

  "xprop f (Val v)       \<longleftrightarrow> xp f (Val v)"
| "xprop f (FVar n)      \<longleftrightarrow> xp f (FVar n)"
| "xprop f (PVar n)      \<longleftrightarrow> xp f (PVar n)"
| "xprop f (Sum m)       \<longleftrightarrow> xp f (Sum m)       \<and> xprop f m"
| "xprop f (Size m)      \<longleftrightarrow> xp f (Size m)      \<and> xprop f m"
| "xprop f (Count m v)   \<longleftrightarrow> xp f (Count m v)   \<and> xprop f m \<and> xprop f v"
| "xprop f (Tally t m)   \<longleftrightarrow> xp f (Tally t m)   \<and> xprop f m"
| "xprop f (Read d p)    \<longleftrightarrow> xp f (Read d p)    \<and> xprop f d \<and> xprop f p"
| "xprop f (Scal n m)    \<longleftrightarrow> xp f (Scal n m)    \<and> xprop f n \<and> xprop f m"
| "xprop f (VRep n m)    \<longleftrightarrow> xp f (VRep n m)    \<and> xprop f n \<and> xprop f m"
| "xprop f (VPlus a b)   \<longleftrightarrow> xp f (VPlus a b)   \<and> xprop f a \<and> xprop f b"
| "xprop f (VMinus a b)  \<longleftrightarrow> xp f (VMinus a b)  \<and> xprop f a \<and> xprop f b"
| "xprop f (VCond l a b) \<longleftrightarrow> xp f (VCond l a b) \<and> xprop f l \<and> xprop f a \<and> xprop f b"

| "xprop f PRoot         \<longleftrightarrow> xp f PRoot"
| "xprop f (Path l p v)  \<longleftrightarrow> xp f (Path l p v)  \<and> xprop f p \<and> xprop f v"
| "xprop f (Kind l p)    \<longleftrightarrow> xp f (Kind l p)    \<and> xprop f p"
| "xprop f (PRep n m)    \<longleftrightarrow> xp f (PRep n m)    \<and> xprop f n \<and> xprop f m"
| "xprop f (PPlus a b)   \<longleftrightarrow> xp f (PPlus a b)   \<and> xprop f a \<and> xprop f b"
| "xprop f (PMinus a b)  \<longleftrightarrow> xp f (PMinus a b)  \<and> xprop f a \<and> xprop f b"
| "xprop f (PCond l a b) \<longleftrightarrow> xp f (PCond l a b) \<and> xprop f l \<and> xprop f a \<and> xprop f b"

| "xprop f Init          \<longleftrightarrow> xp f Init"
| "xprop f (Upd d p v)   \<longleftrightarrow> xp f (Upd d p v)   \<and> xprop f d \<and> xprop f p \<and> xprop f v"
| "xprop f (Del d p)     \<longleftrightarrow> xp f (Del d p)     \<and> xprop f d \<and> xprop f p"
| "xprop f (DCond l a b) \<longleftrightarrow> xp f (DCond l a b) \<and> xprop f l \<and> xprop f a \<and> xprop f b"

| "xprop f (Equal a b)   \<longleftrightarrow> xp f (Equal a b)   \<and> xprop f a \<and> xprop f b"
| "xprop f (Less a b)    \<longleftrightarrow> xp f (Less a b)    \<and> xprop f a \<and> xprop f b"
| "xprop f (RCond l a b) \<longleftrightarrow> xp f (RCond l a b) \<and> xprop f l \<and> xprop f a \<and> xprop f b"

| "xprop f (L b r)       \<longleftrightarrow> xprop f r"

theorem xprop_sanity_T:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "xprop (XFold c\<^sub>T c\<^sub>T c\<^sub>T c\<^sub>T) v"
    and "xprop (XFold c\<^sub>T c\<^sub>T c\<^sub>T c\<^sub>T) p"
    and "xprop (XFold c\<^sub>T c\<^sub>T c\<^sub>T c\<^sub>T) d"
    and "xprop (XFold c\<^sub>T c\<^sub>T c\<^sub>T c\<^sub>T) r"
    and "xprop (XFold c\<^sub>T c\<^sub>T c\<^sub>T c\<^sub>T) l"
  by (induct v and p and d and r and l) auto


fun xcount_valueExp :: "nat xfold \<Rightarrow> valueExp \<Rightarrow> nat"
and xcount_pathExp  :: "nat xfold \<Rightarrow> pathExp  \<Rightarrow> nat"
and xcount_docExp   :: "nat xfold \<Rightarrow> docExp   \<Rightarrow> nat"
and xcount_pred     :: "nat xfold \<Rightarrow> pred     \<Rightarrow> nat"
and xcount_lit      :: "nat xfold \<Rightarrow> lit      \<Rightarrow> nat" where

  "xcount f (Val v)       = xc f (Val v)"
| "xcount f (FVar n)      = xc f (FVar n)"
| "xcount f (PVar n)      = xc f (PVar n)"
| "xcount f (Sum m)       = xc f (Sum m)       + xcount f m"
| "xcount f (Size m)      = xc f (Size m)      + xcount f m"
| "xcount f (Count m v)   = xc f (Count m v)   + xcount f m + xcount f v"
| "xcount f (Tally t m)   = xc f (Tally t m)   + xcount f m"
| "xcount f (Read d p)    = xc f (Read d p)    + xcount f d + xcount f p"
| "xcount f (Scal n m)    = xc f (Scal n m)    + xcount f n + xcount f m"
| "xcount f (VRep n m)    = xc f (VRep n m)    + xcount f n + xcount f m"
| "xcount f (VPlus a b)   = xc f (VPlus a b)   + xcount f a + xcount f b"
| "xcount f (VMinus a b)  = xc f (VMinus a b)  + xcount f a + xcount f b"
| "xcount f (VCond l a b) = xc f (VCond l a b) + xcount f l + xcount f a + xcount f b"

| "xcount f PRoot         = xc f PRoot"
| "xcount f (Path l p v)  = xc f (Path l p v)  + xcount f p + xcount f v"
| "xcount f (Kind l p)    = xc f (Kind l p)    + xcount f p"
| "xcount f (PRep n m)    = xc f (PRep n m)    + xcount f n + xcount f m"
| "xcount f (PPlus a b)   = xc f (PPlus a b)   + xcount f a + xcount f b"
| "xcount f (PMinus a b)  = xc f (PMinus a b)  + xcount f a + xcount f b"
| "xcount f (PCond l a b) = xc f (PCond l a b) + xcount f l + xcount f a + xcount f b"

| "xcount f Init          = xc f Init"
| "xcount f (Upd d p v)   = xc f (Upd d p v)   + xcount f d + xcount f p + xcount f v"
| "xcount f (Del d p)     = xc f (Del d p)     + xcount f d + xcount f p"
| "xcount f (DCond l a b) = xc f (DCond l a b) + xcount f l + xcount f a + xcount f b"

| "xcount f (Equal a b)   = xc f (Equal a b)   + xcount f a + xcount f b"
| "xcount f (Less a b)    = xc f (Less a b)    + xcount f a + xcount f b"
| "xcount f (RCond l a b) = xc f (RCond l a b) + xcount f l + xcount f a + xcount f b"

| "xcount f (L b r)       = xcount f r"

theorem xcount_sanity_0:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "xcount (XFold c\<^sub>0 c\<^sub>0 c\<^sub>0 c\<^sub>0) v = 0"
    and "xcount (XFold c\<^sub>0 c\<^sub>0 c\<^sub>0 c\<^sub>0) p = 0"
    and "xcount (XFold c\<^sub>0 c\<^sub>0 c\<^sub>0 c\<^sub>0) d = 0"
    and "xcount (XFold c\<^sub>0 c\<^sub>0 c\<^sub>0 c\<^sub>0) r = 0"
    and "xcount (XFold c\<^sub>0 c\<^sub>0 c\<^sub>0 c\<^sub>0) l = 0"
  by (induct v and p and d and r and l) auto


fun xvars_valueExp :: "name set xfold \<Rightarrow> valueExp \<Rightarrow> name set"
and xvars_pathExp  :: "name set xfold \<Rightarrow> pathExp  \<Rightarrow> name set"
and xvars_docExp   :: "name set xfold \<Rightarrow> docExp   \<Rightarrow> name set"
and xvars_pred     :: "name set xfold \<Rightarrow> pred     \<Rightarrow> name set"
and xvars_lit      :: "name set xfold \<Rightarrow> lit      \<Rightarrow> name set" where

  "xvars f (Val v)       = xv f (Val v)"
| "xvars f (FVar n)      = xv f (FVar n)"
| "xvars f (PVar n)      = xv f (PVar n)"
| "xvars f (Sum m)       = xv f (Sum m)       \<union> xvars f m"
| "xvars f (Size m)      = xv f (Size m)      \<union> xvars f m"
| "xvars f (Count m v)   = xv f (Count m v)   \<union> xvars f m \<union> xvars f v"
| "xvars f (Tally t m)   = xv f (Tally t m)   \<union> xvars f m"
| "xvars f (Read d p)    = xv f (Read d p)    \<union> xvars f d \<union> xvars f p"
| "xvars f (Scal n m)    = xv f (Scal n m)    \<union> xvars f n \<union> xvars f m"
| "xvars f (VRep n m)    = xv f (VRep n m)    \<union> xvars f n \<union> xvars f m"
| "xvars f (VPlus a b)   = xv f (VPlus a b)   \<union> xvars f a \<union> xvars f b"
| "xvars f (VMinus a b)  = xv f (VMinus a b)  \<union> xvars f a \<union> xvars f b"
| "xvars f (VCond l a b) = xv f (VCond l a b) \<union> xvars f l \<union> xvars f a \<union> xvars f b"

| "xvars f PRoot         = xv f PRoot"
| "xvars f (Path l p v)  = xv f (Path l p v)  \<union> xvars f p \<union> xvars f v"
| "xvars f (Kind l p)    = xv f (Kind l p)    \<union> xvars f p"
| "xvars f (PRep n m)    = xv f (PRep n m)    \<union> xvars f n \<union> xvars f m"
| "xvars f (PPlus a b)   = xv f (PPlus a b)   \<union> xvars f a \<union> xvars f b"
| "xvars f (PMinus a b)  = xv f (PMinus a b)  \<union> xvars f a \<union> xvars f b"
| "xvars f (PCond l a b) = xv f (PCond l a b) \<union> xvars f l \<union> xvars f a \<union> xvars f b"

| "xvars f Init          = xv f Init"
| "xvars f (Upd d p v)   = xv f (Upd d p v)   \<union> xvars f d \<union> xvars f p \<union> xvars f v"
| "xvars f (Del d p)     = xv f (Del d p)     \<union> xvars f d \<union> xvars f p"
| "xvars f (DCond l a b) = xv f (DCond l a b) \<union> xvars f l \<union> xvars f a \<union> xvars f b"

| "xvars f (Equal a b)   = xv f (Equal a b)   \<union> xvars f a \<union> xvars f b"
| "xvars f (Less a b)    = xv f (Less a b)    \<union> xvars f a \<union> xvars f b"
| "xvars f (RCond l a b) = xv f (RCond l a b) \<union> xvars f l \<union> xvars f a \<union> xvars f b"

| "xvars f (L b r)       = xvars f r"

theorem xvars_sanity_E:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "xvars (XFold c\<^sub>E c\<^sub>E c\<^sub>E c\<^sub>E) v = {}"
    and "xvars (XFold c\<^sub>E c\<^sub>E c\<^sub>E c\<^sub>E) p = {}"
    and "xvars (XFold c\<^sub>E c\<^sub>E c\<^sub>E c\<^sub>E) d = {}"
    and "xvars (XFold c\<^sub>E c\<^sub>E c\<^sub>E c\<^sub>E) r = {}"
    and "xvars (XFold c\<^sub>E c\<^sub>E c\<^sub>E c\<^sub>E) l = {}"
  by (induct v and p and d and r and l) auto


fun xlits_valueExp :: "lit list xfold \<Rightarrow> valueExp \<Rightarrow> lit list"
and xlits_pathExp  :: "lit list xfold \<Rightarrow> pathExp  \<Rightarrow> lit list"
and xlits_docExp   :: "lit list xfold \<Rightarrow> docExp   \<Rightarrow> lit list"
and xlits_pred     :: "lit list xfold \<Rightarrow> pred     \<Rightarrow> lit list"
and xlits_lit      :: "lit list xfold \<Rightarrow> lit      \<Rightarrow> lit list" where

  "xlits f (Val v)       = xl f (Val v)"
| "xlits f (FVar n)      = xl f (FVar n)"
| "xlits f (PVar n)      = xl f (PVar n)"
| "xlits f (Sum m)       = xl f (Sum m)       @ xlits f m"
| "xlits f (Size m)      = xl f (Size m)      @ xlits f m"
| "xlits f (Count m v)   = xl f (Count m v)   @ xlits f m @ xlits f v"
| "xlits f (Tally t m)   = xl f (Tally t m)   @ xlits f m"
| "xlits f (Read d p)    = xl f (Read d p)    @ xlits f d @ xlits f p"
| "xlits f (Scal n m)    = xl f (Scal n m)    @ xlits f n @ xlits f m"
| "xlits f (VRep n m)    = xl f (VRep n m)    @ xlits f n @ xlits f m"
| "xlits f (VPlus a b)   = xl f (VPlus a b)   @ xlits f a @ xlits f b"
| "xlits f (VMinus a b)  = xl f (VMinus a b)  @ xlits f a @ xlits f b"
| "xlits f (VCond l a b) = xl f (VCond l a b) @ xlits f l @ xlits f a @ xlits f b"

| "xlits f PRoot         = xl f PRoot"
| "xlits f (Path l p v)  = xl f (Path l p v)  @ xlits f p @ xlits f v"
| "xlits f (Kind l p)    = xl f (Kind l p)    @ xlits f p"
| "xlits f (PRep n m)    = xl f (PRep n m)    @ xlits f n @ xlits f m"
| "xlits f (PPlus a b)   = xl f (PPlus a b)   @ xlits f a @ xlits f b"
| "xlits f (PMinus a b)  = xl f (PMinus a b)  @ xlits f a @ xlits f b"
| "xlits f (PCond l a b) = xl f (PCond l a b) @ xlits f l @ xlits f a @ xlits f b"

| "xlits f Init          = xl f Init"
| "xlits f (Upd d p v)   = xl f (Upd d p v)   @ xlits f d @ xlits f p @ xlits f v"
| "xlits f (Del d p)     = xl f (Del d p)     @ xlits f d @ xlits f p"
| "xlits f (DCond l a b) = xl f (DCond l a b) @ xlits f l @ xlits f a @ xlits f b"

| "xlits f (Equal a b)   = xl f (Equal a b)   @ xlits f a @ xlits f b"
| "xlits f (Less a b)    = xl f (Less a b)    @ xlits f a @ xlits f b"
| "xlits f (RCond l a b) = xl f (RCond l a b) @ xlits f l @ xlits f a @ xlits f b"

| "xlits f (L b r)       = xlits f r"

theorem xlits_sanity_E:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "xlits (XFold c\<^sub>L c\<^sub>L c\<^sub>L c\<^sub>L) v = []"
    and "xlits (XFold c\<^sub>L c\<^sub>L c\<^sub>L c\<^sub>L) p = []"
    and "xlits (XFold c\<^sub>L c\<^sub>L c\<^sub>L c\<^sub>L) d = []"
    and "xlits (XFold c\<^sub>L c\<^sub>L c\<^sub>L c\<^sub>L) r = []"
    and "xlits (XFold c\<^sub>L c\<^sub>L c\<^sub>L c\<^sub>L) l = []"
  by (induct v and p and d and r and l) auto


theorem xcount_xprop_connect:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  assumes "toXP ac = ap" and "toXP bc = bp" and "toXP cc = cp" and "toXP dc = dp"
  shows "xcount (XFold ac bc cc dc) v = 0 \<longleftrightarrow> xprop (XFold ap bp cp dp) v"
    and "xcount (XFold ac bc cc dc) p = 0 \<longleftrightarrow> xprop (XFold ap bp cp dp) p"
    and "xcount (XFold ac bc cc dc) d = 0 \<longleftrightarrow> xprop (XFold ap bp cp dp) d"
    and "xcount (XFold ac bc cc dc) r = 0 \<longleftrightarrow> xprop (XFold ap bp cp dp) r"
    and "xcount (XFold ac bc cc dc) l = 0 \<longleftrightarrow> xprop (XFold ap bp cp dp) l"
  using assms by (induct v and p and d and r and l) auto
  instance by intro_classes
    ((case_tac x, simp_all, case_tac x, simp_all, case_tac x, simp_all, (blast+)?, case_tac x, simp_all, (blast+)?)?,
    (rule xprop_sanity_T, rule xcount_sanity_0, rule xvars_sanity_E, rule xlits_sanity_E, erule (3) xcount_xprop_connect)+)+ end


instantiation disj :: xfold begin

  fun xp_disj :: "bool xfold \<Rightarrow> disj \<Rightarrow> bool" where
    "xp_disj _ = c\<^sub>T"

  fun xc_disj :: "nat xfold \<Rightarrow> disj \<Rightarrow> nat" where
    "xc_disj _ = c\<^sub>0"

  fun xv_disj :: "name set xfold \<Rightarrow> disj \<Rightarrow> name set" where
    "xv_disj _ = c\<^sub>E"

  fun xl_disj :: "lit list xfold \<Rightarrow> disj \<Rightarrow> lit list" where
    "xl_disj _ = c\<^sub>L"

  fun xprop_disj :: "bool xfold \<Rightarrow> disj \<Rightarrow> bool" where
    "xprop f (D l d) \<longleftrightarrow> xprop f l \<and> xprop f d"
  | "xprop f DFlse   \<longleftrightarrow> True"

  fun xcount_disj :: "nat xfold \<Rightarrow> disj \<Rightarrow> nat" where
    "xcount f (D l d) = xcount f l + xcount f d"
  | "xcount f DFlse   = 0"

  fun xvars_disj :: "name set xfold \<Rightarrow> disj \<Rightarrow> name set" where
    "xvars f (D l d) = xvars f l \<union> xvars f d"
  | "xvars f DFlse   = {}"

  fun xlits_disj :: "lit list xfold \<Rightarrow> disj \<Rightarrow> lit list" where
    "xlits f (D l d) = xlits f l @ xlits f d"
  | "xlits f DFlse   = []"

instance by (intro_classes, simp_all) (induct_tac x, (simp add: xcount_to_xprop)+)+ end


instantiation conj :: xfold begin

  fun xp_conj :: "bool xfold \<Rightarrow> conj \<Rightarrow> bool" where
    "xp_conj _ = c\<^sub>T"

  fun xc_conj :: "nat xfold \<Rightarrow> conj \<Rightarrow> nat" where
    "xc_conj _ = c\<^sub>0"

  fun xv_conj :: "name set xfold \<Rightarrow> conj \<Rightarrow> name set" where
    "xv_conj _ = c\<^sub>E"

  fun xl_conj :: "lit list xfold \<Rightarrow> conj \<Rightarrow> lit list" where
    "xl_conj _ = c\<^sub>L"

  fun xprop_conj :: "bool xfold \<Rightarrow> conj \<Rightarrow> bool" where
    "xprop f (C d c) \<longleftrightarrow> xprop f d \<and> xprop f c"
  | "xprop f CTrue   \<longleftrightarrow> True"

  fun xcount_conj :: "nat xfold \<Rightarrow> conj \<Rightarrow> nat" where
    "xcount f (C d c) = xcount f d + xcount f c"
  | "xcount f CTrue   = 0"

  fun xvars_conj :: "name set xfold \<Rightarrow> conj \<Rightarrow> name set" where
    "xvars f (C d c) = xvars f d \<union> xvars f c"
  | "xvars f CTrue   = {}"

  fun xlits_conj :: "lit list xfold \<Rightarrow> conj \<Rightarrow> lit list" where
    "xlits f (C d c) = xlits f d @ xlits f c"
  | "xlits f CTrue   = []"

instance by (intro_classes, simp_all) (induct_tac x, (simp add: xcount_to_xprop)+)+ end


instantiation cnf :: xfold begin

  fun xp_cnf :: "bool xfold \<Rightarrow> cnf \<Rightarrow> bool" where
    "xp_cnf f = c\<^sub>T"

  fun xc_cnf :: "nat xfold \<Rightarrow> cnf \<Rightarrow> nat" where
    "xc_cnf f = c\<^sub>0"

  fun xv_cnf :: "name set xfold \<Rightarrow> cnf \<Rightarrow> name set" where
    "xv_cnf f = c\<^sub>E"

  fun xl_cnf :: "lit list xfold \<Rightarrow> cnf \<Rightarrow> lit list" where
    "xl_cnf f = c\<^sub>L"

  fun xprop_cnf :: "bool xfold \<Rightarrow> cnf \<Rightarrow> bool" where
    "xprop f (\<forall>. c) \<longleftrightarrow> xprop f c"

  fun xcount_cnf :: "nat xfold \<Rightarrow> cnf \<Rightarrow> nat" where
    "xcount f (\<forall>. c) = xcount f c"

  fun xvars_cnf :: "name set xfold \<Rightarrow> cnf \<Rightarrow> name set" where
    "xvars f (\<forall>. c) = xvars f c"

  fun xlits_cnf :: "lit list xfold \<Rightarrow> cnf \<Rightarrow> lit list" where
    "xlits f (\<forall>. c) = xlits f c"

instance by intro_classes (case_tac x, auto simp: xcount_to_xprop)+ end


end

