header {* Infinite Multisets and Partial Mappings *}

theory Multisets imports Main begin

notation Some ("_\<^sub>!" [1000] 1000)


(*-----------------------------------*)
(*- MSet Definition, Count, Collect -*)
(*-----------------------------------*)

typedef 'a mset =
  "UNIV :: ('a \<Rightarrow> nat) set"
  by simp

definition count :: "'a mset \<Rightarrow> 'a \<Rightarrow> nat" where
  "count \<equiv> Rep_mset"

definition collect where
  "collect \<equiv> Abs_mset"


(*----------------------------------*)
(*- Core Theorems & Proof Patterns -*)
(*----------------------------------*)

theorem collect_count [simp]:
  shows "collect (count m) = m"
  unfolding count_def and collect_def by (rule Rep_mset_inverse)

theorem count_collect [simp]:
  shows "count (collect m) = m"
  unfolding count_def and collect_def by (rule Abs_mset_inverse) simp

theorem collect_inject [dest!]:
  assumes "collect a = collect b"
  shows "a = b"
  using assms by - (drule_tac f=count in arg_cong, simp)

theorem count_inject [dest!]:
  assumes "count a = count b"
  shows "a = b"
  using assms by - (drule_tac f=collect in arg_cong, simp)

theorem collect_cong [intro!]:
  assumes "\<And>x. p x = q x"
  shows "collect p = collect q"
  using assms unfolding collect_def by (subst Abs_mset_inject) auto

theorem count_cong [intro!]:
  assumes "\<And>e. count a e = count b e"
  shows "a = b"
  using assms by - (rule count_inject, rule ext, auto)


(*-----------------------------------*)
(*- Membership, Element Set & Subset-*)
(*-----------------------------------*)

abbreviation eleM :: "'a \<Rightarrow> 'a mset \<Rightarrow> bool" ("_ \<in>\<^sub>* _" [1000, 850] 100) where
  "e \<in>\<^sub>* m \<equiv> 0 < count m e"

abbreviation elemset :: "'a mset \<Rightarrow> 'a set" where
  "elemset m \<equiv> {e. e \<in>\<^sub>* m}"

abbreviation eleM_not :: "'a \<Rightarrow> 'a mset \<Rightarrow> bool" ("_ \<notin>\<^sub>* _" [1000, 850] 100) where
  "e \<notin>\<^sub>* m \<equiv> count m e = 0"

theorem elemset_norm [simp]:
  shows "0 = count m e \<longleftrightarrow> count m e = 0"
  by simp

abbreviation subsetM :: "'a mset \<Rightarrow> 'a mset \<Rightarrow> bool" ("_ \<subseteq>\<^sub>* _" [850, 850] 100) where
  "a \<subseteq>\<^sub>* b \<equiv> \<forall>e. count a e \<le> count b e"

theorem subsetM_elem [dest, elim?]:
  assumes "b \<subseteq>\<^sub>* a" and "x \<in>\<^sub>* b"
  shows "x \<in>\<^sub>* a"
  using assms by (metis gr0I le_0_eq)

theorem subsetM_elemset [dest, intro]:
  assumes "b \<subseteq>\<^sub>* a"
  shows "elemset b \<subseteq> elemset a"
  using assms by auto

theorem subsetM_int_distrib [simp]:
  assumes "b \<subseteq>\<^sub>* a"
  shows "int (count a x - count b x) = int (count a x) - int (count b x)"
  using assms by - (rule zdiff_int[symmetric], simp)


(*----------------------------------------*)
(*- Monoid and Minus Instances (0, +, -) -*)
(*----------------------------------------*)

instantiation mset :: (type) cancel_comm_monoid_add begin

  definition zeroM:
    "0 \<equiv> collect (\<lambda>e. 0)"

  definition plusM:
   "a + b \<equiv> collect (\<lambda>e. count a e + count b e)"

instance by (intro_classes, auto simp: plusM zeroM) (metis nat_add_left_cancel) end

instantiation mset :: (type) minus begin

  definition minusM:
    "a - b \<equiv> collect (\<lambda>e. count a e - count b e)"

instance by intro_classes end

abbreviation emptyM :: "'a mset" ("{}\<^sub>*") where
  "{}\<^sub>* \<equiv> 0"

theorem count_empty [simp]:
  shows "count {}\<^sub>* e = 0"
  unfolding zeroM by simp


abbreviation plusM_abbr :: "'a mset \<Rightarrow> 'a mset \<Rightarrow> 'a mset" ("_ \<oplus> _" [890, 895] 890) where
  "a \<oplus> b \<equiv> a + b"

theorem count_plus [simp]:
  shows "count (a \<oplus> b) e = count a e + count b e"
  unfolding plusM by simp


abbreviation minusM_abbr :: "'a mset \<Rightarrow> 'a mset \<Rightarrow> 'a mset" ("_ \<ominus> _" [890, 895] 890) where
  "a \<ominus> b \<equiv> a - b"

theorem count_minus [simp]:
  shows "count (a \<ominus> b) e = count a e - count b e"
  unfolding minusM by simp


(* Theorems *)

theorem zeroM_uninhab [simp]:
  shows "(\<forall>a. a \<notin>\<^sub>* m) \<longleftrightarrow> m = {}\<^sub>*"
  by auto

theorem zeroM_inhab_witness [dest!]:
  assumes "m \<noteq> {}\<^sub>*"
  shows "\<exists>e. e \<in>\<^sub>* m"
  using assms by auto

theorem elemset_zero:
  shows "elemset {}\<^sub>* = {}"
  by simp

theorem elemset_plus [simp]:
  shows "elemset (a \<oplus> b) = elemset a \<union> elemset b"
  by auto

theorem elemset_minus [simp]:
  shows "elemset (a \<ominus> b) = {e. count b e < count a e}"
  by auto

theorem elemset_minus_subset [simp]:
  shows "elemset (a \<ominus> b) \<subseteq> elemset a"
  unfolding minusM by auto

theorem plusM_zero_left [simp]:
  shows "{}\<^sub>* \<oplus> b = b"
  by simp

theorem plusM_zero_right [simp]:
  shows "a \<oplus> {}\<^sub>* = a"
  by simp

theorem plusM_zero_result [dest!]:
  assumes "a \<oplus> b = {}\<^sub>*"
  shows "a = {}\<^sub>* \<and> b = {}\<^sub>*"
  using assms unfolding plusM and zeroM by auto (metis add_is_0)+

theorem plusM_commute:
  shows "a \<oplus> b = b \<oplus> a"
  by auto

theorem plusM_assoc [simp]:
  shows "a \<oplus> (b \<oplus> c) = (a \<oplus> b) \<oplus> c"
  by auto

theorem plusM_commute_assoc:
  shows "a \<oplus> b \<oplus> c = a \<oplus> c \<oplus> b"
  by auto

theorem minusM_zero_left [simp]:
  shows "{}\<^sub>* \<ominus> b = {}\<^sub>*"
  unfolding minusM and zeroM by simp

theorem minusM_zero_right [simp]:
  shows "a \<ominus> {}\<^sub>* = a"
  unfolding minusM and zeroM by simp

theorem minusM_zero_result_subset [simp]:
  assumes "b \<subseteq>\<^sub>* a"
  shows "a \<ominus> b = {}\<^sub>* \<longleftrightarrow> a = b"
  using assms unfolding minusM and zeroM by auto (metis diff_diff_cancel minus_nat.diff_0)

theorem minusM_zero_result_general [dest]:
  assumes "a \<ominus> b = {}\<^sub>*"
  shows "a \<subseteq>\<^sub>* b"
  using assms unfolding minusM and zeroM by auto (metis diff_is_0_eq)


(*--------------*)
(*- Finiteness -*)
(*--------------*)

abbreviation finiteM where
  "finiteM m \<equiv> finite (elemset m)"


(* Theorems *)

theorem finiteM_subsetM [dest, elim?]:
  assumes "b \<subseteq>\<^sub>* a" and "finiteM a"
  shows "finiteM b"
  using assms by - (erule rev_finite_subset, auto)

theorem finiteM_minus_left [simp]:
  assumes "finiteM a"
  shows "finite {e. count b e < count a e}"
  using assms by - (erule rev_finite_subset, auto)


(*----------------------------------------------*)
(*- Partial Injections, Inversions & Constants -*)
(*----------------------------------------------*)

definition partial_inj :: "('a \<rightharpoonup> 'b) \<Rightarrow> bool" where
  "partial_inj f \<equiv> inj_on f (dom f)"

definition partial_inv where
  "partial_inv f \<equiv> inv f o Some"

definition partial_con:
  "partial_con f \<equiv> \<exists>y. \<forall>x. f x = y\<^sub>!"

definition partial_the:
  "partial_the f \<equiv> THE y. y \<in> ran f"


theorem partial_inj:
  shows "partial_inj f \<longleftrightarrow> (\<forall>x. \<forall>y. (\<exists>z. f x = z\<^sub>! \<and> f y = z\<^sub>!) \<longrightarrow> (x = y))"
  unfolding inj_on_def dom_def and partial_inj_def by auto

theorem partial_inv:
  shows "partial_inv f = (\<lambda>y. SOME x. f x = y\<^sub>!)"
  unfolding inv_def and partial_inv_def by auto

theorem partial_inj_restrict [simp, intro]:
  assumes "partial_inj f"
  shows "partial_inj (f |` m)"
  using assms unfolding partial_inj and restrict_map_def by auto

theorem partial_inj_singleton [simp]:
  assumes "partial_inj f" and "f x = y\<^sub>!"
  shows "{x. f x = y\<^sub>!} = {x}"
  using assms unfolding partial_inj by auto

theorem partial_inv_apply [simp]:
  assumes "x \<in> ran f"
  shows "f (partial_inv f x) = x\<^sub>!"
  using assms unfolding ran_def and partial_inv
  by (simp, subst some_eq_ex [of "\<lambda>y. f y = x\<^sub>!"], auto)

theorem partial_inj_inv [simp]:
  assumes "partial_inj f" and "f x = y\<^sub>!"
  shows "partial_inv f y = x"
  using assms unfolding partial_inj and partial_inv by - (rule some_equality, auto)

theorem partial_inv_inj [dest]:
  assumes "partial_inv f x = partial_inv f y" and "x \<in> ran f" "y \<in> ran f"
  shows "x = y"
  using assms by (metis partial_inv_apply the.simps)

theorem partial_con_apply [intro!]:
  assumes "partial_con f"
  shows "f x = (partial_the f)\<^sub>!"
  using assms unfolding partial_con partial_the and ran_def by auto

corollary partial_con_witness [elim]:
  assumes "f x = y\<^sub>!" and "partial_con f"
  shows "partial_the f = y"
  using assms by (subst (asm) partial_con_apply[of f]) simp+

corollary partial_con_not_the [elim]:
  assumes "partial_con f" and "partial_the f \<noteq> e"
  shows "f x \<noteq> e\<^sub>!"
  using assms by - (erule contrapos_nn, erule(1) partial_con_witness)

theorem partial_con_eq [elim]:
  assumes "partial_con f" and "x \<in> dom f" and "y \<in> dom f"
  shows "f x = f y"
proof -
  from partial_con_apply[of f x] and partial_con_apply[of f y] show ?thesis using assms by simp
qed


(*------------------------------------------*)
(*- Nat Setsums - setsum (count m) Toolbox -*)
(*------------------------------------------*)

lemma setsum_subtract_nat:
  fixes f :: "'a \<Rightarrow> nat"
  assumes "\<And>x. g x \<le> f x" and "finite A"
  shows "(\<Sum>x\<in>A. f x - g x) = setsum f A - setsum g A"
proof -
  have "(\<Sum>x\<in>A. (int \<circ> f) x - (int \<circ> g) x) = setsum (int \<circ> f) A - setsum (int \<circ> g) A"
    by - (rule setsum_subtractf)
  thus ?thesis using assms and zdiff_int by (auto simp: int_setsum[symmetric])
qed

lemma setsum_count_diff [simp]:
  shows "setsum (count m) (elemset m - {e}) = setsum (count m) (elemset m) - count m e"
  using assms by (cases "e \<in>\<^sub>* m", auto) (metis mem_Collect_eq setsum_diff1_nat)

lemma setsum_count_le [simp]:
  assumes "finiteM m"
  shows "count m e \<le> setsum (count m) (elemset m)"
  using assms by (cases "e \<in>\<^sub>* m", auto) (subst setsum_diff1', auto)


(*----------------------*)
(*- (Partial) Mappings -*)
(*----------------------*)

definition mapM :: "('a \<rightharpoonup> 'b) \<Rightarrow> 'a mset \<Rightarrow> 'b mset" where
  "mapM f m \<equiv> collect (\<lambda>e. setsum (count m) {x. x \<in>\<^sub>* m \<and> f x = e\<^sub>!})"

theorem map_sound_finite [intro]:
  assumes "finiteM m"
  shows "finite {x \<in> elemset m. P x}"
  using assms by simp

theorem map_sound_inject [intro]:
  assumes "partial_inj f"
  shows "finite {x. x \<in>\<^sub>* m \<and> f x = e\<^sub>!}"
  using assms by (rule_tac B="{partial_inv f e}" in finite_subset) auto

theorem map_sound_inject_total [intro]:
  assumes "partial_inj (Some \<circ> f)"
  shows "finite {x. x \<in>\<^sub>* m \<and> f x = e}"
  using assms by (rule_tac B="{partial_inv (Some \<circ> f) e}" in finite_subset) auto

theorem map_sound_finite_dom [intro]:
  assumes "finite (dom f)"
  shows "finite {x. x \<in>\<^sub>* m \<and> f x = e\<^sub>!}"
  using assms by (rule_tac B="dom f" in finite_subset, auto)

corollary map_sound:
  assumes "finiteM m \<or> partial_inj f \<or> finite (dom f)"
  shows "finite {x. x \<in>\<^sub>* m \<and> f x = e\<^sub>!}"
  using assms
proof safe
  assume "finiteM m"      thus ?thesis by auto next
  assume "partial_inj f"  thus ?thesis by (rule map_sound_inject) next
  assume "finite (dom f)" thus ?thesis by (rule map_sound_finite_dom)
qed

theorem count_map:
  shows "count (mapM f m) e = setsum (count m) {x. x \<in>\<^sub>* m \<and> f x = e\<^sub>!}"
  unfolding mapM_def by simp

theorem count_map_undefined [simp]:
  assumes "e \<notin> ran f"
  shows "count (mapM f m) e = 0"
  using assms unfolding mapM_def and ran_def by auto

theorem count_map_const_other [simp]:
  assumes "partial_con f" and "partial_the f \<noteq> e"
  shows "count (mapM f m) e = 0"
  using assms unfolding mapM_def by (simp, subst partial_con_not_the[of f]) auto

theorem count_map_inject_defined [simp]:
  assumes "partial_inj f" and "e \<in> ran f"
  shows "count (mapM f m) e = count m (partial_inv f e)"
  using assms
proof -
  from assms obtain x where x: "f x = e\<^sub>!" by (auto simp: ran_def)
  hence inverse: "partial_inv f e = x" using assms by simp
  show ?thesis
  proof (cases "x \<in>\<^sub>* m")
    case True with x have "{x. x \<in>\<^sub>* m \<and> f x = e\<^sub>!} = {x}" using assms by (auto simp: partial_inj)
      thus ?thesis using inverse unfolding mapM_def by auto next
    case False with x
      show ?thesis using assms unfolding mapM_def by auto (metis less_not_refl partial_inj)
  qed
qed


(* Theorems - Elem/Set *)

theorem eleM_map_dest [dest!]:
  assumes "y \<in>\<^sub>* mapM f m"
  shows "\<exists>x. x \<in>\<^sub>* m \<and> f x = y\<^sub>!"
  using assms unfolding mapM_def by auto (erule contrapos_pp, auto intro: setsum_0')

theorem eleM_map_elim [elim]:
  assumes "f x = y\<^sub>!" and "x \<in> elemset m" and "finite {x. x \<in>\<^sub>* m \<and> f x = y\<^sub>!}"
  shows "y \<in>\<^sub>* mapM f m"
  using assms unfolding mapM_def by auto (rule classical, auto, metis less_zeroE)

theorem eleM_map:
  assumes "\<forall>y. finite {x. x \<in>\<^sub>* m \<and> f x = y\<^sub>!}"
  shows "y \<in>\<^sub>* mapM f m \<longleftrightarrow> (\<exists>x. x \<in>\<^sub>* m \<and> f x = y\<^sub>!)"
proof
  assume "y \<in>\<^sub>* mapM f m" thus "\<exists>x. x \<in>\<^sub>* m \<and> f x = y\<^sub>!"
    using assms unfolding mapM_def by auto (erule contrapos_pp, auto) next
  assume "\<exists>x. x \<in>\<^sub>* m \<and> f x = y\<^sub>!" thus "y \<in>\<^sub>* mapM f m"
    using assms unfolding mapM_def by auto (rule classical, auto, metis less_zeroE)
qed


theorem elemset_map:
  assumes "\<And>y. finite {x. x \<in>\<^sub>* m \<and> f x = y\<^sub>!}"
  shows "elemset (mapM f m) = ran (f |` elemset m)"
  using assms unfolding ran_def and restrict_map_def by auto 

corollary elemset_map_finite [simp]:
  assumes "finiteM m"
  shows "elemset (mapM f m) = ran (f |` elemset m)"
  using assms by - (rule elemset_map, simp)

corollary elemset_map_finite_dom [simp]:
  assumes "finite (dom f)"
  shows "elemset (mapM f m) = ran (f |` elemset m)"
  using assms by - (rule elemset_map, rule map_sound_finite_dom)

corollary elemset_map_inject [simp]:
  assumes "partial_inj f"
  shows "elemset (mapM f m) = ran (f |` elemset m)"
  using assms by - (rule elemset_map, rule map_sound_inject)


(* Theorems - Finiteness *)

theorem finite_ran_restrict [simp, intro]:
  assumes "finite (ran f)"
  shows "finite (ran (f |` m))"
  using assms unfolding ran_def and restrict_map_def by - (erule rev_finite_subset, auto)

theorem finite_dom_ran [simp, intro]:
  assumes "finite (dom f)"
  shows "finite (ran f)"
proof -
  have "Some ` ran f = f ` dom f" unfolding dom_def ran_def and image_def by auto
  hence "finite (Some ` ran f)" using assms by simp
  thus ?thesis by - (erule finite_imageD, simp)
qed

corollary finite_map_inject [simp]:
  assumes "partial_inj f"
  shows "finite (ran f) \<longleftrightarrow> finite (dom f)"
proof -
  from assms have "dom f = partial_inv f ` (ran f)" unfolding image_def and ran_def by auto
  thus ?thesis by auto
qed


corollary elemset_map_const [dest]:
  assumes "partial_con f" and "finiteM m"
  shows "ran (f |` elemset m) \<subseteq> {partial_the f}"
  using assms by auto (metis partial_con_witness ran_restrictD)

corollary elemset_map_const_not_empty [simp]:
  assumes "partial_con f" and "finiteM m" and "m \<noteq> {}\<^sub>*"
  shows "ran (f |` elemset m) = {partial_the f}"
  using partial_con_apply [of f] and assms by (auto simp: ran_def restrict_map_def)


theorem mapM_elem_ran [dest]:
  assumes "x \<in> ran (f |` elemset m)"
  shows "x \<in> ran f"
  using assms unfolding restrict_map_def and ran_def by auto (case_tac "a \<in>\<^sub>* m", auto)


(* Theorems - Subset *)

theorem subsetM_map:
  assumes sub: "b \<subseteq>\<^sub>* a" and sound: "\<And>y. finite {x \<in> elemset a. f x = y\<^sub>!}"
  shows "mapM f b \<subseteq>\<^sub>* mapM f a"
proof
  fix e show "count (mapM f b) e \<le> count (mapM f a) e" unfolding mapM_def
  using sub and sound[of e] by auto (rule setsum_le_included, auto elim: rev_finite_subset)
qed

corollary subsetM_map_finite [simp, elim]:
  assumes "b \<subseteq>\<^sub>* a" and "finiteM a"
  shows "mapM f b \<subseteq>\<^sub>* mapM f a"
  using assms by - (erule subsetM_map, simp)

corollary subsetM_map_finite_dom [simp, elim]:
  assumes "b \<subseteq>\<^sub>* a" and "finite (dom f)"
  shows "mapM f b \<subseteq>\<^sub>* mapM f a"
  using assms by - (erule subsetM_map, auto, rule map_sound_finite_dom[simplified])

corollary subsetM_map_inject [simp, elim]:
  assumes "b \<subseteq>\<^sub>* a" and "partial_inj f"
  shows "mapM f b \<subseteq>\<^sub>* mapM f a"
  using assms by - (erule subsetM_map, auto, rule map_sound_inject[simplified])


(* Theorems - Zero/Plus/Minus *)

theorem zeroM_map [simp]:
  shows "mapM f {}\<^sub>* = {}\<^sub>*"
  unfolding mapM_def by auto

theorem zeroM_map_result [intro!]:
  assumes "\<And>x. x \<in>\<^sub>* m \<Longrightarrow> f x = None"
  shows "mapM f m = {}\<^sub>*"
proof
  fix e have "{x. x \<in>\<^sub>* m \<and> f x = e\<^sub>!} = {}" using assms by auto
  hence "setsum (count m) {x. x \<in>\<^sub>* m \<and> f x = e\<^sub>!} = setsum (count m) {}" by - (erule arg_cong)
  thus "count (mapM f m) e = count {}\<^sub>* e" unfolding mapM_def by simp
qed


theorem plusM_map_homomorphism:
  assumes "\<And>y. finite {x. x \<in>\<^sub>* a \<oplus> b \<and> f x = y\<^sub>!}"
  shows "mapM f (a \<oplus> b) = mapM f a \<oplus> mapM f b"
proof
  fix e show "count (mapM f (a \<oplus> b)) e = count (mapM f a \<oplus> mapM f b) e"
  proof -
    have "setsum (count a) {x \<in> elemset a. f x = e\<^sub>!} =
          setsum (count a) {x \<in> elemset (a \<oplus> b). f x = e\<^sub>!}"
     and "setsum (count b) {x \<in> elemset b. f x = e\<^sub>!} =
          setsum (count b) {x \<in> elemset (a \<oplus> b). f x = e\<^sub>!}"
    using assms[of e] by - (rule setsum_mono_zero_left, auto)+
    thus ?thesis unfolding mapM_def by clarsimp (rule setsum_addf)
  qed
qed

corollary plusM_map_homomorphism_finite [simp]:
  assumes "finiteM a" and "finiteM b"
  shows "mapM f (a \<oplus> b) = mapM f a \<oplus> mapM f b"
  using assms by - (rule plusM_map_homomorphism, simp)

corollary plusM_map_homomorphism_finite_dom [simp]:
  assumes "finite (dom f)"
  shows "mapM f (a \<oplus> b) = mapM f a \<oplus> mapM f b"
  using assms by - (rule plusM_map_homomorphism, rule map_sound_finite_dom)

corollary plusM_map_homomorphism_inject [simp]:
  assumes "partial_inj f"
  shows "mapM f (a \<oplus> b) = mapM f a \<oplus> mapM f b"
  using assms by - (rule plusM_map_homomorphism, rule map_sound_inject)


theorem minusM_map_homomorphism:
  assumes sub: "b \<subseteq>\<^sub>* a" and sound: "\<And>y. finite {x. x \<in>\<^sub>* a \<and> f x = y\<^sub>!}"
  shows "mapM f (a \<ominus> b) = mapM f a \<ominus> mapM f b"
proof
  fix e show "count (mapM f (a \<ominus> b)) e = count (mapM f a \<ominus> mapM f b) e"
  proof -
    have "setsum (\<lambda>x. count a x - count b x) {x. count b x < count a x \<and> f x = e\<^sub>!} =
          setsum (\<lambda>x. count a x - count b x) {x \<in> elemset a. f x = e\<^sub>!}"
     and "setsum (count b) {x. x \<in>\<^sub>* b \<and> f x = e\<^sub>!} = setsum (count b) {x. x \<in>\<^sub>* a \<and> f x = e\<^sub>!}"
    using sub and sound[of e] by - (rule setsum_mono_zero_left, auto)+
    thus ?thesis using sub and sound[of e] unfolding mapM_def
      by clarsimp (rule setsum_subtract_nat, auto)
  qed
qed

corollary minusM_map_homomorphism_finite [simp]:
  assumes "b \<subseteq>\<^sub>* a" and "finiteM a"
  shows "mapM f (a \<ominus> b) = mapM f a \<ominus> mapM f b"
  using assms by - (erule minusM_map_homomorphism, simp)

corollary minusM_map_homomorphism_finite_dom [simp]:
  assumes "b \<subseteq>\<^sub>* a" and "finite (dom f)"
  shows "mapM f (a \<ominus> b) = mapM f a \<ominus> mapM f b"
  using assms by - (erule minusM_map_homomorphism, rule map_sound_finite_dom)

theorem minusM_map_homomorphism_inject [simp]:  (* we can drop an assumption for injections *)
  assumes "partial_inj f"
  shows "mapM f (a \<ominus> b) = mapM f a \<ominus> mapM f b"
  using assms by clarify (case_tac "e \<in> ran f", simp+)


(*--------------*)
(*- Reductions -*)
(*--------------*)

abbreviation reduceM :: "'a mset \<Rightarrow> 'a set \<Rightarrow> 'a mset" ("_\<down>_" [1000, 1000] 1000) where
  "m\<down>s \<equiv> mapM (Some |` s) m"

theorem reduceM_inject_inj [simp]:
  shows "partial_inj (Some |` s)"
  unfolding partial_inj and restrict_map_def by simp

theorem reduceM_inject_inv [simp]:
  assumes "e \<in> s"
  shows "partial_inv (Some |` s) e = e"
  using assms by simp

theorem reduceM_ran [simp]:
  shows "ran (Some |` s) = s"
  unfolding ran_def and restrict_map_def by auto

theorem count_reduce:
  shows "count (m\<down>s) e = (if e \<in> s then count m e else 0)"
  by simp

lemma restrict_some_eq [dest!]:
  assumes "(Some |` s) x = y\<^sub>!"
  shows "x = y \<and> y \<in> s"
  using assms unfolding restrict_map_def by (split if_splits) auto

lemma restrict_none_eq [simp]:
  shows "(Some |` s) x = None \<longleftrightarrow> x \<notin> s"
  unfolding restrict_map_def by simp


lemma Some_inject_inj [simp]:
  shows "partial_inj Some"
  unfolding partial_inj by simp

lemma Some_ran [simp]:
  shows "ran Some = UNIV"
  unfolding ran_def by simp

lemma mapM_some_id [simp]:
  shows "mapM Some m = m"
  by auto


(* Theorems *)

theorem reduceM_twice [simp]:
  shows "(m\<down>s)\<down>s = m\<down>s"
  by clarify (case_tac "e \<in> s", simp+)

theorem eleM_reduce [simp]:
  shows "x \<in>\<^sub>* m\<down>s \<longleftrightarrow> x \<in>\<^sub>* m \<and> x \<in> s"
  by auto

theorem eleM_reduce_not [simp]:
  shows "x \<notin>\<^sub>* m\<down>s \<longleftrightarrow> x \<notin>\<^sub>* m \<or> x \<notin> s"
  by (metis count_reduce)

theorem reduceM_elemset [simp]:
  shows "m\<down>(elemset m) = m"
  by clarify (case_tac "e \<in>\<^sub>* m", simp+)

theorem reduceM_superset [simp]:
  assumes "elemset m \<subseteq> s"
  shows "m\<down>s = m"
  using assms by clarify (case_tac "e \<in>\<^sub>* m", auto)

theorem reduceM_inter [simp]:
  shows "m\<down>(r \<inter> s) = (m\<down>r)\<down>s"
  by clarify (metis count_reduce Int_iff)

theorem reduceM_subsetM [simp]:
  shows "m\<down>s \<subseteq>\<^sub>* m"
  by clarify (metis count_reduce le0 order_refl)

theorem reduceM_empty [simp]:
  shows "m\<down>{} = {}\<^sub>*"
  by auto


(*--------*)
(*- Size -*)
(*--------*)

definition sizeM :: "'a mset \<Rightarrow> nat" where  (* sizeM is sound within finiteM *)
  "sizeM m \<equiv> setsum (count m) (elemset m)"

theorem sizeM_map_def:
  "sizeM m = count (mapM (\<lambda>e. ()\<^sub>!) m) ()"
  unfolding sizeM_def and mapM_def by simp

theorem sizeM_finite_eq:
  assumes "m \<noteq> {}\<^sub>*"
  shows "finiteM m \<longleftrightarrow> 0 < sizeM m"
  using assms unfolding sizeM_def by (case_tac "finiteM m", auto) (erule contrapos_pp, force)

theorem sizeM_finite [dest]:
  assumes "sizeM m = n" and "n \<noteq> 0"
  shows "finite (elemset m)"
  using assms unfolding sizeM_def by - (rule classical, simp)

theorem sizeM_normalize [simp]:
  shows "0 < sizeM m \<longleftrightarrow> m \<noteq> {}\<^sub>* \<and> finiteM m"
  unfolding sizeM_def
  by (metis elemset_zero linorder_neq_iff sizeM_def sizeM_finite_eq setsum.empty) 

theorem sizeM_reduce_setsum:
  assumes "finite s"
  shows "sizeM m\<down>s = setsum (count m) s"
  using assms unfolding sizeM_def by simp (rule setsum_mono_zero_left, auto)

theorem count_map_const [simp]:
  assumes "partial_con f" and "partial_the f = e"
  shows "count (mapM f m) e = sizeM m"
  using assms unfolding sizeM_def and mapM_def by auto (rule setsum_cong, auto)


(* Theorems *)

theorem subsetM_size [intro, simp]:
  assumes "b \<subseteq>\<^sub>* a" and "finiteM a"
  shows "sizeM b \<le> sizeM a"
  using assms unfolding sizeM_def by - (rule setsum_le_included, auto)

theorem zeroM_size [simp]:
  shows "sizeM {}\<^sub>* = 0"
  unfolding sizeM_def by simp

theorem sizeM_eq_0 [simp]:
  shows "sizeM m = 0 \<longleftrightarrow> m = {}\<^sub>* \<or> \<not> finiteM m"
  unfolding sizeM_def by auto

theorem plusM_size_homomorphism [simp]:
  assumes "finiteM a" and "finiteM b"
  shows "sizeM (a \<oplus> b) = sizeM a + sizeM b"
  using assms unfolding sizeM_map_def by simp

theorem minusM_size_homomorphism [simp]:
  assumes "b \<subseteq>\<^sub>* a" and "finiteM a"
  shows "sizeM (a \<ominus> b) = sizeM a - sizeM b"
  using assms unfolding sizeM_map_def by simp

theorem subsetM_size_eq [elim]:
  assumes sub: "b \<subseteq>\<^sub>* a" and size: "sizeM b = sizeM a" and fin: "finiteM a"
  shows "a = b"
proof (rule ccontr)
  assume "a \<noteq> b" then obtain e where "count a e \<noteq> count b e" by auto
  hence "count b e < count a e" using sub by (metis le_neq_implies_less)
  hence "setsum (count b) (elemset b - {e}) > setsum (count a) (elemset a - {e})"
    using size and fin unfolding sizeM_def
    by simp (rule diff_less_mono2, auto, erule less_le_trans, auto)
  moreover have "setsum (count b) (elemset b - {e}) \<le> setsum (count a) (elemset a - {e})"
    using sub and fin by - (rule setsum_le_included, auto)
  ultimately show False by simp
qed

theorem mapM_size_constant [simp]:
  assumes "partial_con f"
  shows "sizeM (mapM f m) = sizeM m"
  using assms unfolding sizeM_def
  by (cases "0 < sizeM m", auto intro!: setsum_0', metis sizeM_def)
     (case_tac "partial_the f = a", auto)

theorem mapM_size_inject:
  assumes "partial_inj f"
  shows "sizeM (mapM f m) = sizeM m\<down>(dom f)"
proof -
  have count: "\<And>x. x \<in> ran f \<longrightarrow> count (mapM f m) x = count m\<down>(dom f) (partial_inv f x)"
    using assms by (metis count_map_inject_defined count_reduce domI partial_inv_apply)
  have "inj_on (partial_inv f) (elemset (mapM f m))"
    using assms unfolding inj_on_def by auto moreover
  have "dom f \<inter> elemset m = partial_inv f ` elemset (mapM f m)"
    using assms unfolding image_def by (auto simp: restrict_map_def ran_def) ultimately
  show ?thesis unfolding sizeM_def
    by (subst setsum_reindex_cong[of "partial_inv f"], auto simp del: eleM_reduce) 
       (subst count, auto simp: ran_def)
qed


theorem reduceM_size_not_less [simp]:
  assumes "finiteM m"
  shows "\<not> sizeM m\<down>f < sizeM m \<longleftrightarrow> sizeM m\<down>f = sizeM m"
proof -
  have "sizeM m\<down>f \<le> sizeM m" using assms by (metis reduceM_subsetM subsetM_size)
  thus ?thesis by simp
qed

theorem reduceM_size_neq:
  assumes "sizeM m \<noteq> sizeM m\<down>s"
  shows "\<exists>e. e \<in>\<^sub>* m \<and> e \<notin> s"
proof (rule ccontr, clarsimp)
  assume "\<forall>e. e \<notin>\<^sub>* m \<or> e \<in> s"
  hence "{e. e \<in>\<^sub>* m \<and> e \<in> s} = elemset m" by auto (metis less_zeroE) 
  hence "sizeM m = sizeM m\<down>s" unfolding sizeM_def
    by simp (rule setsum_cong2, subst count_reduce, auto)
  thus False using assms by simp
qed

theorem reduceM_size_eq:
  assumes "sizeM m = sizeM m\<down>s" and "finiteM m"
  shows "\<forall>e. e \<in>\<^sub>* m \<longrightarrow> e \<in> s"
proof (rule ccontr, clarsimp)
  fix e assume "e \<in>\<^sub>* m" and "e \<notin> s"
  hence "setsum (count m\<down>s) (elemset m\<down>s) = setsum (count m\<down>s) (elemset m\<down>s - {e})"
    using `finiteM m` by - (rule setsum_mono_zero_right, auto) also
  have "... \<le> setsum (count m) (elemset m - {e})"
    using `finiteM m` by - (rule setsum_le_included, auto) also
  have "... < setsum (count m) (elemset m)"
    using `finiteM m` and `e \<in>\<^sub>* m`
    by auto (metis diff_less gr0I le_0_eq setsum_count_le) finally
  have "sizeM m\<down>s < sizeM m" unfolding sizeM_def .
  thus False using assms by simp
qed

corollary reduceM_size [simp]:
  assumes "finiteM m"
  shows "sizeM m = sizeM m\<down>s \<longleftrightarrow> (\<forall>e. e \<in>\<^sub>* m \<longrightarrow> e \<in> s)"
proof
  assume "sizeM m = sizeM m\<down>s"
    thus "\<forall>e. e \<in>\<^sub>* m \<longrightarrow> e \<in> s" using assms by - (erule reduceM_size_eq) next
  assume "\<forall>e. e \<in>\<^sub>* m \<longrightarrow> e \<in> s"
    thus "sizeM m = sizeM m\<down>s" by - (erule contrapos_pp, drule reduceM_size_neq, auto)
qed

corollary reduceM_size_sym [simp]:
  assumes "finiteM m"
  shows "sizeM m\<down>s = sizeM m \<longleftrightarrow> (\<forall>e. e \<in>\<^sub>* m \<longrightarrow> e \<in> s)"
  by (subst reduceM_size[symmetric, OF assms]) auto


(*--------------*)
(*- Singletons -*)
(*--------------*)

abbreviation unique :: "'a mset \<Rightarrow> bool" where
  "unique m \<equiv> sizeM m = Suc 0"

abbreviation proper :: "'a mset \<Rightarrow> bool" where
  "proper m \<equiv> sizeM m \<noteq> Suc 0"

definition selectM :: "'a mset \<Rightarrow> 'a" ("_\<^sub>1" [1000] 1000) where
  "m\<^sub>1 \<equiv> THE e. count m e = 1"

definition singleM :: "'a \<Rightarrow> 'a mset" ("{_}\<^sub>*") where
  "{e}\<^sub>* \<equiv> collect (\<lambda>x. if x = e then 1 else 0)"

theorem count_single:
  shows "count {x}\<^sub>* e = (if e = x then 1 else 0)"
  unfolding singleM_def by simp

theorem count_single_same [simp]:
  shows "count {e}\<^sub>* e = 1"
  unfolding singleM_def by simp

theorem count_single_other [simp]:
  assumes "e \<noteq> f"
  shows "count {e}\<^sub>* f = 0"
  using assms unfolding singleM_def by simp

theorem uniqueM_single [simp]:
  shows "unique {e}\<^sub>*"
  unfolding singleM_def and sizeM_def by simp

theorem uniqueM_single_exists [dest!]:
  assumes "unique m"
  shows "\<exists>e. m = {e}\<^sub>*"
proof -
  from assms have "finiteM m" by auto
  then obtain e where "count m e = 1" and "\<forall>f. e \<noteq> f \<longrightarrow> count m f = 0"
    using assms unfolding sizeM_def by (subst (asm) setsum_eq_Suc0_iff, auto)
  hence "m = {e}\<^sub>*" unfolding singleM_def by auto
  thus ?thesis ..
qed

theorem singleM_cong [simp]:
  shows "{a}\<^sub>* = {b}\<^sub>* \<longleftrightarrow> a = b"
  unfolding singleM_def by auto (metis one_neq_zero)


(* Theorems *)

theorem eleM_single [simp]:
  shows "e \<in>\<^sub>* {f}\<^sub>* \<longleftrightarrow> e = f"
  unfolding singleM_def by simp

theorem subsetM_single_left [simp]:
  shows "{e}\<^sub>* \<subseteq>\<^sub>* m \<longleftrightarrow> e \<in>\<^sub>* m"
  unfolding singleM_def by auto

theorem subsetM_single_right [dest]:
  assumes "m \<subseteq>\<^sub>* {e}\<^sub>*"
  shows "m = {e}\<^sub>* \<or> m = {}\<^sub>*"
proof (rule disjCI)
  assume "m \<noteq> {}\<^sub>*" hence "count m e = 1" using assms unfolding singleM_def
    by auto (drule_tac x=ea in spec, case_tac "e = ea", auto) moreover
  have "\<forall>f. e \<noteq> f \<longrightarrow> count m f = 0" using assms by auto ultimately
  show "m = {e}\<^sub>*" unfolding singleM_def by auto
qed


theorem singleM_select [simp]:
  shows "{e}\<^sub>*\<^sub>1 = e"
  unfolding singleM_def and selectM_def by simp

theorem selectM_single [simp]:
  assumes "unique p"
  shows "{p\<^sub>1}\<^sub>* = p"
  using assms by auto

theorem selectM_unique [simp]:
  assumes "unique m" and "p \<in>\<^sub>* m"
  shows "m\<^sub>1 = p"
  using assms by auto

theorem selectM_unique_other [simp]:
  assumes "unique m" and "p \<notin>\<^sub>* m"
  shows "m\<^sub>1 \<noteq> p"
  using assms by auto


theorem singleM_zero [simp]:
  shows "{e}\<^sub>* \<noteq> {}\<^sub>*"
  unfolding singleM_def and zeroM by auto (metis one_neq_zero)


theorem singleM_plus_result [dest!]:
  assumes "a \<oplus> b = {e}\<^sub>*"
  shows "a = {e}\<^sub>* \<and> b = {}\<^sub>* \<or> a = {}\<^sub>* \<and> b = {e}\<^sub>*"
proof -
  have "count a e + count b e = 1" and "\<forall>f. f \<noteq> e \<longrightarrow> count a f + count b f = 0"
    using assms unfolding plusM and singleM_def by (auto, metis One_nat_def) (metis add_is_0)+
  hence "count a e = 1 \<and> count b e = 0 \<or> count a e = 0 \<and> count b e = 1"
    and "\<forall>f. f \<noteq> e \<longrightarrow> count a f = 0" and "\<forall>f. f \<noteq> e \<longrightarrow> count b f = 0" by auto
  thus ?thesis using assms by (metis plusM_zero_left plusM_zero_right zeroM_uninhab)
qed


theorem mapM_single_defined [simp]:
  assumes "f x = y\<^sub>!"
  shows "mapM f {x}\<^sub>* = {y}\<^sub>*"
  using assms unfolding mapM_def and singleM_def by auto

theorem mapM_single_undefined [simp]:
  assumes "f x = None"
  shows "mapM f {x}\<^sub>* = {}\<^sub>*"
  using assms unfolding mapM_def and singleM_def by auto

theorem mapM_single_single_result [dest!]:
  assumes "mapM f {x}\<^sub>* = {y}\<^sub>*"
  shows "f x = y\<^sub>!"
proof (cases "f x = None")
  case True thus ?thesis using assms by simp next
  case False thus ?thesis using assms by auto
qed

theorem mapM_unique [simp]:
  assumes "unique m"
  shows "unique (mapM f m) \<longleftrightarrow> f m\<^sub>1 \<noteq> None"
  using assms by auto


(*-------------*)
(*- Universes -*)
(*-------------*)

definition msuniv :: "'a mset" ("\<infinity>") where
  "\<infinity> \<equiv> collect (\<lambda>e. 1)"

theorem count_univ [simp]:
  shows "count \<infinity> e = 1"
  unfolding msuniv_def by simp


(* Theorems *)

theorem elemset_univ [simp]:
  shows "elemset \<infinity> = UNIV"
  unfolding msuniv_def by simp

theorem sizeM_univ [simp]:
  shows "sizeM (\<infinity> :: 'a mset) = card (UNIV :: 'a set)"
  unfolding msuniv_def and sizeM_def by simp


(*---------------*)
(*- Replication -*)
(*---------------*)

definition repM :: "nat \<Rightarrow> 'a mset \<Rightarrow> 'a mset" ("_ \<otimes> _" [101, 895] 895) where
  "n \<otimes> m \<equiv> collect (\<lambda>e. n * count m e)"

theorem count_rep [simp]:
  shows "count (n \<otimes> m) e = n * count m e"
  unfolding repM_def by simp


(* Theorems *)

theorem zeroM_rep_left [simp]:
  shows "0 \<otimes> b = {}\<^sub>*"
  unfolding repM_def by auto

theorem zeroM_rep_right [simp]:
  shows "n \<otimes> {}\<^sub>* = {}\<^sub>*"
  unfolding repM_def by auto

theorem zeroM_rep_result [dest!]:
  assumes "n \<otimes> m = {}\<^sub>*"
  shows "n = 0 \<or> m = {}\<^sub>*"
  using assms unfolding zeroM and repM_def by auto (metis mult_is_0 nat_neq_iff)

theorem plusM_rep_distrib [simp]:
  shows "n \<otimes> (a \<oplus> b) = (n \<otimes> a) \<oplus> (n \<otimes> b)"
  unfolding repM_def by auto (rule add_mult_distrib2)

theorem minusM_rep_distrib [simp]:
  shows "n \<otimes> (a \<ominus> b) = (n \<otimes> a) \<ominus> (n \<otimes> b)"
  unfolding repM_def by auto (rule diff_mult_distrib2)


theorem sizeM_rep_homomorphism [simp]:
  shows "sizeM (n \<otimes> m) = n * sizeM m"
  unfolding repM_def and sizeM_def
  by auto (cases "n = 0", auto, rule setsum_right_distrib[symmetric])

theorem mapM_rep_homomorphism [simp]:
  shows "mapM f (n \<otimes> m) = n \<otimes> mapM f m"
  unfolding repM_def and mapM_def
  by auto (cases "n = 0", auto, rule setsum_right_distrib[symmetric])


theorem repM_one [simplified, simp]:
  shows "1 \<otimes> m = m"
  unfolding repM_def by simp

theorem singleM_rep_result [dest!]:
  assumes "n \<otimes> m = {e}\<^sub>*"
  shows "n = 1 \<and> m = {e}\<^sub>*"
proof -
  have "n = 1" using assms unfolding repM_def and singleM_def
    by auto (drule_tac x=e in fun_cong, force)
  thus ?thesis using assms and repM_one[of m] by simp
qed

theorem singleM_rep_count_reduce [simp]:
  shows "count m e \<otimes> {e}\<^sub>* = m\<down>{e}"
  by auto (case_tac "e = ea", auto)

corollary reduceM_single:
  shows "m\<down>{e} = count m e \<otimes> {e}\<^sub>*"
  by simp


(*------------------*)
(*- Set Operations -*)
(*------------------*)

definition interM :: "'a mset \<Rightarrow> 'a mset \<Rightarrow> 'a mset" ("_ \<inter>\<^sub>* _" [890, 895] 890) where (* hm \<dots> *)
  "a \<inter>\<^sub>* b \<equiv> collect (\<lambda>e. min (count a e) (count b e)) "

theorem count_inter [simp]:
  shows "count (a \<inter>\<^sub>* b) e = min (count a e) (count b e)"
  unfolding interM_def by simp

definition unionM :: "'a mset \<Rightarrow> 'a mset \<Rightarrow> 'a mset" ("_ \<union>\<^sub>* _" [880, 890] 880) where
  "a \<union>\<^sub>* b \<equiv> collect (\<lambda>e. max (count a e) (count b e)) "

theorem count_union [simp]:
  shows "count (a \<union>\<^sub>* b) e = max (count a e) (count b e)"
  unfolding unionM_def by simp


(* Theorems *)

theorem zeroM_inter_left [simp]:
  shows "{}\<^sub>* \<inter>\<^sub>* b = {}\<^sub>*"
  unfolding interM_def by auto

theorem zeroM_inter_right [simp]:
  shows "a \<inter>\<^sub>* {}\<^sub>* = {}\<^sub>*"
  unfolding interM_def by auto

theorem zeroM_union_left [simp]:
  shows "{}\<^sub>* \<union>\<^sub>* b = b"
  unfolding unionM_def by auto

theorem zeroM_union_right [simp]:
  shows "a \<union>\<^sub>* {}\<^sub>* = a"
  unfolding unionM_def by auto


declare count_cong [rule del]  (* from now on we want multisets to hold as abstraction *)

end

