header {* A Total Ordering for Terms *}

theory Ordering imports Assertions
  "~~/src/HOL/Library/Char_ord"
  "~~/src/HOL/Library/List_lexord"
begin


datatype ord = LT | EQ | GT

fun thenComp :: "ord \<Rightarrow> ord \<Rightarrow> ord" ("_ \<rhd> _" [450, 500] 450) where
  "EQ \<rhd> b = b"
| "a  \<rhd> _ = a"


class comp =
fixes comp :: "'a \<Rightarrow> 'a \<Rightarrow> ord" ("_ \<prec> _" [550, 550] 500)

fun fromOrder where
  "fromOrder a b = (if a < b then LT else if a = b then EQ else GT)"


instantiation label and key and token and name :: comp begin
  fun comp_label :: "label \<Rightarrow> label \<Rightarrow> ord" where
    "Label a \<prec> Label b = fromOrder a b"

  fun comp_key :: "key \<Rightarrow> key \<Rightarrow> ord" where
    "Key a \<prec> Key b = fromOrder a b"

  fun comp_token :: "token \<Rightarrow> token \<Rightarrow> ord" where
    "Token a \<prec> Token b = fromOrder a b"

  fun comp_name :: "name \<Rightarrow> name \<Rightarrow> ord" where
    "Name a \<prec> Name b = fromOrder a b"
instance .. end

instantiation bool and nat and int :: comp begin
  fun comp_bool :: "bool \<Rightarrow> bool \<Rightarrow> ord" where
    "comp_bool a b = fromOrder a b"

  fun comp_nat :: "nat \<Rightarrow> nat \<Rightarrow> ord" where
    "comp_nat a b = fromOrder a b"

  fun comp_int :: "int \<Rightarrow> int \<Rightarrow> ord" where
    "comp_int a b = fromOrder a b"
instance .. end


class ordSize =
fixes ordSize :: "'a \<Rightarrow> nat"

fun oc :: "'a :: ordSize \<Rightarrow> 'a \<Rightarrow> ord" where
  "oc a b = ordSize a \<prec> ordSize b"


instantiation "value" and type :: ordSize begin
  fun ordSize_value :: "value \<Rightarrow> nat" where
    "ordSize (VInt _)   = 0"
  | "ordSize (VToken _) = 1"
  | "ordSize (VKey _)   = 2"
  | "ordSize Unit       = 3"

  fun ordSize_type :: "type \<Rightarrow> nat" where
    "ordSize TInt   = 0"
  | "ordSize TToken = 1"
  | "ordSize TKey   = 2"
  | "ordSize TUnit  = 3"
instance .. end

instantiation "value" and type :: comp begin
  fun comp_value :: "value \<Rightarrow> value \<Rightarrow> ord" where
    "(VInt a)   \<prec> v = (case v of VInt b   \<Rightarrow> a \<prec> b | _ \<Rightarrow> oc (VInt a) v)"
  | "(VToken a) \<prec> v = (case v of VToken b \<Rightarrow> a \<prec> b | _ \<Rightarrow> oc (VToken a) v)"
  | "(VKey a)   \<prec> v = (case v of VKey b   \<Rightarrow> a \<prec> b | _ \<Rightarrow> oc (VKey a) v)"
  | "Unit       \<prec> v = (if v = Unit then EQ else oc Unit v)"

  fun comp_type :: "type \<Rightarrow> type \<Rightarrow> ord" where
    "comp_type a b = oc a b"
instance .. end


instantiation valueExp and pathExp and docExp and pred :: ordSize begin

fun ordSize_valueExp :: "valueExp \<Rightarrow> nat" where
  "ordSize (VCond _ _ _) = 0"

| "ordSize (Val _)       = 1"

| "ordSize (VRep _ _)    = 2"
| "ordSize (VPlus _ _)   = 3"
| "ordSize (VMinus _ _)  = 4"

| "ordSize (Scal _ _)    = 5"

| "ordSize (Sum _)       = 6"
| "ordSize (Count _ _)   = 7"
| "ordSize (Tally _ _)   = 8"
| "ordSize (Size _)      = 9"

| "ordSize (Read _ _)    = 10"

| "ordSize (PVar _)      = 11"
| "ordSize (FVar _)      = 12"

fun ordSize_pathExp :: "pathExp \<Rightarrow> nat" where
  "ordSize (PCond _ _ _) = 0"

| "ordSize Root.         = 1"
| "ordSize (Path _ _ _)  = 2"
| "ordSize (Kind _ _)    = 3"

| "ordSize (PRep _ _)    = 4"
| "ordSize (PPlus _ _)   = 5"
| "ordSize (PMinus _ _)  = 6"

fun ordSize_docExp :: "docExp \<Rightarrow> nat" where
  "ordSize (DCond _ _ _) = 0"

| "ordSize $.            = 1"
| "ordSize (Upd _ _ _)   = 2"
| "ordSize (Del _ _)     = 3"

fun ordSize_pred :: "pred \<Rightarrow> nat" where
  "ordSize (RCond _ _ _) = 0"

| "ordSize (Equal _ _)   = 1"
| "ordSize (Less _ _)    = 2"

instance .. end


instantiation valueExp and pathExp and docExp and pred and lit :: comp begin

fun comp_valueExp :: "valueExp \<Rightarrow> valueExp \<Rightarrow> ord"
and comp_pathExp  :: "pathExp  \<Rightarrow> pathExp  \<Rightarrow> ord"
and comp_docExp   :: "docExp   \<Rightarrow> docExp   \<Rightarrow> ord"
and comp_pred     :: "pred     \<Rightarrow> pred     \<Rightarrow> ord"
and comp_lit      :: "lit      \<Rightarrow> lit      \<Rightarrow> ord" where

  "Val a       \<prec> v = (case v of Val b       \<Rightarrow> a \<prec> b                   | _ \<Rightarrow> oc (Val a) v)"
| "PVar a      \<prec> v = (case v of PVar b      \<Rightarrow> a \<prec> b                   | _ \<Rightarrow> oc (PVar a) v)" 

| "Tally t a   \<prec> v = (case v of Tally u b   \<Rightarrow> a \<prec> b \<rhd> t \<prec> u          | _ \<Rightarrow> oc (Tally t a) v)"
| "Count a b   \<prec> v = (case v of Count c d   \<Rightarrow> a \<prec> c \<rhd> b \<prec> d          | _ \<Rightarrow> oc (Count a b) v)"
| "Size a      \<prec> v = (case v of Size b      \<Rightarrow> a \<prec> b                   | _ \<Rightarrow> oc (Size a) v)"
| "Sum a       \<prec> v = (case v of Sum b       \<Rightarrow> a \<prec> b                   | _ \<Rightarrow> oc (Sum a) v)"

| "Scal a b    \<prec> v = (case v of Scal c d    \<Rightarrow> b \<prec> d \<rhd> a \<prec> c          | _ \<Rightarrow> oc (Scal a b) v)"
| "VRep a b    \<prec> v = (case v of VRep c d    \<Rightarrow> a \<prec> c \<rhd> b \<prec> d          | _ \<Rightarrow> oc (VRep a b) v)"
| "VPlus a b   \<prec> v = (case v of VPlus c d   \<Rightarrow> a \<prec> c \<rhd> b \<prec> d          | _ \<Rightarrow> oc (VPlus a b) v)"
| "VMinus a b  \<prec> v = (case v of VMinus c d  \<Rightarrow> a \<prec> c \<rhd> b \<prec> d          | _ \<Rightarrow> oc (VMinus a b) v)"

| "Read a b    \<prec> v = (case v of Read c d    \<Rightarrow> a \<prec> c \<rhd> b \<prec> d          | _ \<Rightarrow> oc (Read a b) v)"

| "FVar a      \<prec> v = (case v of FVar b      \<Rightarrow> a \<prec> b                   | _ \<Rightarrow> oc (FVar a) v)"

| "Root.       \<prec> p = (if p = Root. then EQ else oc Root. p)"
| "Path a b c  \<prec> p = (case p of Path d e f  \<Rightarrow> b \<prec> e \<rhd> a \<prec> d \<rhd> c \<prec> f | _ \<Rightarrow> oc (Path a b c) p)"
| "Kind a b    \<prec> p = (case p of Kind d e    \<Rightarrow> b \<prec> e \<rhd> a \<prec> d          | _ \<Rightarrow> oc (Kind a b) p)"

| "PRep a b    \<prec> p = (case p of PRep c d    \<Rightarrow> a \<prec> c \<rhd> b \<prec> d          | _ \<Rightarrow> oc (PRep a b) p)"
| "PPlus a b   \<prec> p = (case p of PPlus c d   \<Rightarrow> a \<prec> c \<rhd> b \<prec> d          | _ \<Rightarrow> oc (PPlus a b) p)"
| "PMinus a b  \<prec> p = (case p of PMinus c d  \<Rightarrow> a \<prec> c \<rhd> b \<prec> d          | _ \<Rightarrow> oc (PMinus a b) p)"

| "$.          \<prec> d = (if d = $. then EQ else oc $. d)"
| "Upd a b c   \<prec> d = (case d of Upd d e f   \<Rightarrow> a \<prec> d \<rhd> b \<prec> e \<rhd> c \<prec> f | _ \<Rightarrow> oc (Upd a b c) d)"
| "Del a b     \<prec> d = (case d of Del d e     \<Rightarrow> a \<prec> d \<rhd> b \<prec> e          | _ \<Rightarrow> oc (Del a b) d)"

| "Equal a b   \<prec> r = (case r of Equal c d   \<Rightarrow> a \<prec> c \<rhd> b \<prec> d          | _ \<Rightarrow> oc (Equal a b) r)" 
| "Less a b    \<prec> r = (case r of Less c d    \<Rightarrow> a \<prec> c \<rhd> b \<prec> d          | _ \<Rightarrow> oc (Less a b) r)" 

| "L a b       \<prec> L c d = a \<prec> c \<rhd> b \<prec> d"

| "VCond a b c \<prec> v = (case v of VCond d e f \<Rightarrow> b \<prec> e \<rhd> c \<prec> f \<rhd> a \<prec> d | _ \<Rightarrow> oc (VCond a b c) v)"
| "PCond a b c \<prec> p = (case p of PCond d e f \<Rightarrow> b \<prec> e \<rhd> c \<prec> f \<rhd> a \<prec> d | _ \<Rightarrow> oc (PCond a b c) p)"
| "DCond a b c \<prec> d = (case d of DCond d e f \<Rightarrow> b \<prec> e \<rhd> c \<prec> f \<rhd> a \<prec> d | _ \<Rightarrow> oc (DCond a b c) d)"
| "RCond a b c \<prec> r = (case r of RCond d e f \<Rightarrow> b \<prec> e \<rhd> c \<prec> f \<rhd> a \<prec> d | _ \<Rightarrow> oc (RCond a b c) r)"

instance .. end

end
