header {* Property Preservation of Strictify *}

theory Preservation imports Conditionals begin


(*-----------------------------------*)
(*- Strictify Preserves Well-Formed -*)
(*-----------------------------------*)

theorem push_sum_uve [simp]:
  shows "uve (push Sum v)"
  by (induct v, auto) (subst push.simps, auto)

theorem push_sum_wform [simp]:
  assumes "wform v"
  shows "wform (push Sum v)"
  using assms by (induct v, auto) (subst push.simps, auto)

theorem push_size_uve [simp]:
  shows "uve (push Size v)"
  by (induct v, auto) (subst push.simps, auto)

theorem push_size_wform [simp]:
  assumes "wform v"
  shows "wform (push Size v)"
  using assms by (induct v, auto) (subst push.simps, auto)

theorem push_count_uve [simp]:
  shows "uve (push (Count v) w)"
  by (induct w, auto) (subst push.simps, auto)

theorem push_count_wform [simp]:
  assumes "wform w" and "wform v" and "uve w"
  shows "wform (push (Count v) w)"
  using assms by (induct w, auto) (subst push.simps, auto)

theorem push2_count_uve [simp]:
  shows "uve (push2 Count v w)"
  unfolding push2_def by (induct v, auto) (subst push.simps, auto)

theorem push2_count_wform [simp]:
  assumes "wform w" and "wform v" and "uve w"
  shows "wform (push2 Count v w)"
  using assms unfolding push2_def by (induct v, auto) (subst push.simps, auto)

theorem push_tally_uve [simp]:
  shows "uve (push (Tally t) v)"
  by (induct v, auto) (subst push.simps, auto)

theorem push_tally_wform [simp]:
  assumes "wform v"
  shows "wform (push (Tally t) v)"
  using assms by (induct v, auto) (subst push.simps, auto)


theorem push_read_uve [simp]:
  assumes "upe p"
  shows "uve (push (Read d) p)"
  using assms by (induct p rule: peDomain.induct, auto) (subst push.simps, auto)

theorem push_read_wform [simp]:
  assumes "wform p" and "wform d"
  shows "wform (push (Read d) p)"
  using assms by (induct p rule: peDomain.induct, auto) (subst push.simps, auto)

theorem push2_read_uve [simp]:
  assumes "upe p"
  shows "uve (push2 Read d p)"
  using assms unfolding push2_def by (induct d rule: deDomain.induct, auto) (subst push.simps, auto)

theorem push2_read_wform [simp]:
  assumes "wform p" and "wform d"
  shows "wform (push2 Read d p)"
  using assms unfolding push2_def by (induct d rule: deDomain.induct, auto) (subst push.simps, auto)


theorem push_scal_uve [simp]:
  assumes "uve w"
  shows "uve (push (Scal v) w)"
  using assms by (induct w, auto) (subst push.simps, auto)

theorem push_scal_wform [simp]:
  assumes "wform w" and "wform v" and "uve v"
  shows "wform (push (Scal v) w)"
  using assms by (induct w, auto) (subst push.simps, auto)

theorem push2_scal_uve [simp]:
  assumes "uve w"
  shows "uve (push2 Scal v w)"
  using assms unfolding push2_def by (induct v, auto) (subst (2) push.simps, auto)

theorem push2_scal_wform [simp]:
  assumes "wform w" and "wform v" and "uve v"
  shows "wform (push2 Scal v w)"
  using assms unfolding push2_def by (induct v, auto) (subst (2) push.simps, auto)

theorem push_vrep_wform [simp]:
  assumes "wform a" and "wform b" and "uve a"
  shows "wform (push (VRep a) b)"
  using assms by (induct b, auto) (subst push.simps, auto)

theorem push2_vrep_wform [simp]:
  assumes "wform a" and "wform b" and "uve a"
  shows "wform (push2 VRep a b)"
  using assms unfolding push2_def by (induct a, auto) (subst (2) push.simps, auto)

theorem push_vplus_wform [simp]:
  assumes "wform a" and "wform b"
  shows "wform (push (VPlus a) b)"
  using assms by (induct b, auto) (subst push.simps, auto)

theorem push2_vplus_wform [simp]:
  assumes "wform a" and "wform b"
  shows "wform (push2 VPlus a b)"
  using assms unfolding push2_def by (induct a, auto) (subst (2) push.simps, auto)

theorem push_vminus_wform [simp]:
  assumes "wform a" and "wform b"
  shows "wform (push (VMinus a) b)"
  using assms by (induct b, auto) (subst push.simps, auto)

theorem push2_vminus_wform [simp]:
  assumes "wform a" and "wform b"
  shows "wform (push2 VMinus a b)"
  using assms unfolding push2_def by (induct a, auto) (subst (2) push.simps, auto)


theorem push_path_upe [simp]:
  assumes "upe p" and "uve k"
  shows "upe (push (Path l p) k)"
  using assms by (induct k) (auto, subst push.simps, auto)

theorem push_path_wform [simp]:
  assumes "wform p" and "wform k"
  shows "wform (push (Path l p) k)"
  using assms by (induct k, auto) (subst push.simps, auto)

theorem push2_path_upe [simp]:
  assumes "upe p" and "uve k"
  shows "upe (push2 (Path l) p k)"
  using assms unfolding push2_def
  by (induct p rule: peDomain.induct) (auto, subst push.simps, auto)

theorem push2_path_wform [simp]:
  assumes "wform p" and "wform k"
  shows "wform (push2 (Path l) p k)"
  using assms unfolding push2_def by (induct p rule: peDomain.induct, auto) (subst push.simps, auto)

theorem push_kind_wform [simp]:
  assumes "wform p"
  shows "wform (push (Kind l) p)"
  using assms by (induct p rule: peDomain.induct) (auto, subst push.simps, simp)


theorem push_prep_wform [simp]:
  assumes "wform a" and "wform b" and "uve a"
  shows "wform (push (PRep a) b)"
  using assms
  by (induct b rule: peDomain.induct) (auto, subst push.simps, auto)

theorem push2_prep_wform [simp]:
  assumes "wform a" and "wform b" and "uve a"
  shows "wform (push2 PRep a b)"
  using assms unfolding push2_def by (induct a, auto) (subst push.simps, auto)

theorem push_pplus_wform [simp]:
  assumes "wform a" and "wform b"
  shows "wform (push (PPlus a) b)"
  using assms
  by (induct b rule: peDomain.induct) (auto, subst push.simps, auto)

theorem push2_pplus_wform [simp]:
  assumes "wform a" and "wform b"
  shows "wform (push2 PPlus a b)"
  using assms unfolding push2_def
  by (induct a rule: peDomain.induct) (auto, subst push.simps, auto)

theorem push_pminus_wform [simp]:
  assumes "wform a" and "wform b"
  shows "wform (push (PMinus a) b)"
  using assms by (induct b rule: peDomain.induct) (auto, subst push.simps, auto)

theorem push2_pminus_wform [simp]:
  assumes "wform a" and "wform b"
  shows "wform (push2 PMinus a b)"
  using assms unfolding push2_def by (induct a rule: peDomain.induct) (auto, subst push.simps, auto)


theorem push_upd_wform [simp]:
  assumes "wform d" and "wform p" and "wform v" and "upe p" and "uve v"
  shows "wform (push (Upd d p) v)"
  using assms by (induct v) (auto, subst push.simps, auto)

theorem push2_upd_wform [simp]:
  assumes "wform d" and "wform p" and "wform v" and "upe p" and "uve v"
  shows "wform (push2 (Upd d) p v)"
  using assms unfolding push2_def by (induct p rule: peDomain.induct) (auto, subst push.simps, auto)

theorem push3_upd_wform [simp]:
  assumes "wform d" and "wform p" and "wform v" and "upe p" and "uve v"
  shows "wform (push3 Upd d p v)"
  using assms unfolding push3_def by (induct d rule: deDomain.induct, auto) (subst push.simps, auto)

theorem push_del_wform [simp]:
  assumes "wform d" and "wform p" and "upe p"
  shows "wform (push (Del d) p)"
  using assms by (induct p rule: peDomain.induct) (auto, subst push.simps, auto)

theorem push2_del_wform [simp]:
  assumes "wform d" and "wform p" and "upe p"
  shows "wform (push2 Del d p)"
  using assms unfolding push2_def by (induct d rule: deDomain.induct, auto) (subst push.simps, auto)


theorem push_equal_wform [simp]:
  assumes "wform a" and "wform b" and "uve a" and "uve b"
  shows "wform (push (Equal a) b)"
  using assms by (induct b) (auto, subst push.simps, auto)

theorem push2_equal_wform [simp]:
  assumes "wform a" and "wform b" and "uve a" and "uve b"
  shows "wform (push2 Equal a b)"
  using assms unfolding push2_def by (induct a) (auto, subst (2) push.simps, auto)

theorem push_less_wform [simp]:
  assumes "wform a" and "wform b" and "uve a" and "uve b"
  shows "wform (push (Less a) b)"
  using assms by (induct b) (auto, subst push.simps, auto)

theorem push2_less_wform [simp]:
  assumes "wform a" and "wform b" and "uve a" and "uve b"
  shows "wform (push2 Less a b)"
  using assms unfolding push2_def by (induct a) (auto, subst (2) push.simps, auto)


theorem push_vcond_uve [simp]:
  assumes "uve t" and "uve e"
  shows "uve (push (\<lambda>x. VCond (L b x) t e) r)"
  using assms by (induct r) (auto, subst push.simps, auto)

theorem push_pcond_upe [simp]:
  assumes "upe t" and "upe e"
  shows "upe (push (\<lambda>x. PCond (L b x) t e) r)"
  using assms by (induct r) (auto, subst push.simps, auto)

theorem strictify_uxe [simp]:
  shows "uve v \<Longrightarrow> uve (strictify v)"
    and "upe p \<Longrightarrow> upe (strictify p)"
  by (induct v and p) auto

theorem push_vcond_wform [simp]:
  assumes "wform t" and "wform e" and "wform r"
  shows "wform (push (\<lambda>x. VCond (L b x) t e) r)"
  using assms by (induct r, auto) (subst push.simps, auto)

theorem push_pcond_wform [simp]:
  assumes "wform t" and "wform e" and "wform r"
  shows "wform (push (\<lambda>x. PCond (L b x) t e) r)"
  using assms by (induct r, auto) (subst push.simps, auto)

theorem push_dcond_wform [simp]:
  assumes "wform t" and "wform e" and "wform r"
  shows "wform (push (\<lambda>x. DCond (L b x) t e) r)"
  using assms by (induct r, auto) (subst push.simps, auto)

theorem push_rcond_wform [simp]:
  assumes "wform t" and "wform e" and "wform r"
  shows "wform (push (\<lambda>x. RCond (L b x) t e) r)"
  using assms by (induct r, auto) (subst push.simps, auto)


theorem strictify_wform_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "wform v \<Longrightarrow> wform (strictify v)"
    and "wform p \<Longrightarrow> wform (strictify p)"
    and "wform d \<Longrightarrow> wform (strictify d)"
    and "wform r \<Longrightarrow> wform (strictify r)"
    and "wform l \<Longrightarrow> wform (strictify l)"
  by (induct v and p and d and r and l) auto

theorem strictify_wform_disj [simp]:
  fixes d :: disj assumes "wform d"
  shows "wform (strictify d)"
  using assms by (induct d) auto

theorem decideCE_wform [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred
  shows "wform v \<Longrightarrow> wform (decideCE l v)"
    and "wform p \<Longrightarrow> wform (decideCE l p)"
    and "wform d \<Longrightarrow> wform (decideCE l d)"
    and "wform r \<Longrightarrow> wform (decideCE l r)"
  using assms unfolding decideCE_def
  by (case_tac "bte v", auto, case_tac "bte p", auto, case_tac "bte d", auto, case_tac "bte r", auto)

theorem decideCE_uxe [simp]:
  fixes v :: valueExp and p :: pathExp
  shows "uve v \<Longrightarrow> uve (decideCE l v)"
    and "upe p \<Longrightarrow> upe (decideCE l p)"
  using assms unfolding decideCE_def by (case_tac "bte v", auto) (case_tac "bte p", auto)

theorem decide_uxe [simp]:
  fixes v :: valueExp and p :: pathExp
  shows "uve v \<Longrightarrow> uve (decide x v)"
    and "upe p \<Longrightarrow> upe (decide x p)"
  by (induct v and p rule: uve_upe.induct) auto

theorem decide_wform_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "wform v \<Longrightarrow> wform (decide x v)"
    and "wform p \<Longrightarrow> wform (decide x p)"
    and "wform d \<Longrightarrow> wform (decide x d)"
    and "wform r \<Longrightarrow> wform (decide x r)"
    and "wform l \<Longrightarrow> wform (decide x l)"
  by (induct v and p and d and r and l) auto

theorem decide_wform_disj [simp]:
  fixes d :: disj assumes "wform d"
  shows "wform (decide l d)"
  using assms by (induct d) auto

theorem pullD_wform [simp]:
  assumes "wform d"
  shows "wform (pullD d)"
  using assms by (induct d rule: pullD.induct) auto

theorem strictify_wform_conj [simp]:
  fixes c :: conj assumes "wform c"
  shows "wform (strictify c)"
  using assms by (induct c) auto

theorem strictify_wform [simp]:
  fixes c :: cnf assumes "wform c"
  shows "wform (strictify c)"
  using assms by (cases c) simp+


(*------------------------------*)
(*- Strictify Preserves DUFree -*)
(*------------------------------*)

theorem push_sum_dufree [simp]:
  assumes "dufree v"
  shows "dufree (push Sum v)"
  using assms by (induct v, auto) (subst push.simps, auto)

theorem push_size_dufree [simp]:
  assumes "dufree v"
  shows "dufree (push Size v)"
  using assms by (induct v, auto) (subst push.simps, auto)

theorem push_count_dufree [simp]:
  assumes "dufree v" and "dufree w"
  shows "dufree (push (Count v) w)"
  using assms by (induct w, auto) (subst push.simps, auto)

theorem push2_count_dufree [simp]:
  assumes "dufree v" and "dufree w"
  shows "dufree (push2 Count v w)"
  using assms unfolding push2_def by (induct v, auto) (subst push.simps, auto)

theorem push_tally_dufree [simp]:
  assumes "dufree v"
  shows "dufree (push (Tally t) v)"
  using assms by (induct v, auto) (subst push.simps, auto)


theorem push_read_dufree [simp]:
  assumes "dufree d" and "dufree p"
  shows "dufree (push (Read d) p)"
  using assms by (induct p rule: peDomain.induct) (auto, subst push.simps, auto)

theorem push2_read_dufree [simp]:
  assumes "dufree d" and "dufree p"
  shows "dufree (push2 Read d p)"
  using assms unfolding push2_def
 by (induct d rule: deDomain.induct) (auto, subst (2) push.simps, auto)


theorem push_scal_dufree [simp]:
  assumes "dufree v" and "dufree w"
  shows "dufree (push (Scal v) w)"
  using assms by (induct w, auto) (subst push.simps, auto)

theorem push2_scal_dufree [simp]:
  assumes "dufree v" and "dufree w"
  shows "dufree (push2 Scal v w)"
  using assms unfolding push2_def by (induct v, auto) (subst push.simps, auto)

theorem push_vrep_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push (VRep a) b)"
  using assms by (induct b, auto) (subst push.simps, auto)

theorem push2_vrep_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push2 VRep a b)"
  using assms unfolding push2_def by (induct a, auto) (subst push.simps, auto)

theorem push_vplus_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push (VPlus a) b)"
  using assms by (induct b, auto) (subst push.simps, auto)

theorem push2_vplus_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push2 VPlus a b)"
  using assms unfolding push2_def by (induct a, auto) (subst push.simps, auto)

theorem push_vminus_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push (VMinus a) b)"
  using assms by (induct b, auto) (subst push.simps, auto)

theorem push2_vminus_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push2 VMinus a b)"
  using assms unfolding push2_def by (induct a, auto) (subst push.simps, auto)


theorem push_path_dufree [simp]:
  assumes "dufree p" and "dufree k"
  shows "dufree (push (Path l p) k)"
  using assms by (induct k, auto) (subst push.simps, auto)

theorem push2_path_dufree [simp]:
  assumes "dufree p" and "dufree k"
  shows "dufree (push2 (Path l) p k)"
  using assms unfolding push2_def by (induct p rule: peDomain.induct) (auto, subst push.simps, auto)

theorem push_kind_dufree [simp]:
  assumes "dufree p"
  shows "dufree (push (Kind l) p)"
  using assms by (induct p) (auto, subst push.simps, auto)


theorem push_prep_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push (PRep a) b)"
  using assms by (induct b) (auto, subst push.simps, auto)

theorem push2_prep_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push2 PRep a b)"
  using assms unfolding push2_def by (induct a, auto) (subst (2) push.simps, auto)

theorem push_pplus_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push (PPlus a) b)"
  using assms by (induct b) (auto, subst push.simps, auto) 

theorem push2_pplus_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push2 PPlus a b)"
  using assms unfolding push2_def by (induct a) (auto, subst (2) push.simps, auto)

theorem push_pminus_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push (PMinus a) b)"
  using assms by (induct b) (auto, subst push.simps, auto)

theorem push2_pminus_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push2 PMinus a b)"
  using assms unfolding push2_def by (induct a) (auto, subst (2) push.simps, auto)


theorem push_equal_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push (Equal a) b)"
  using assms by (induct b, auto) (subst push.simps, auto)

theorem push2_equal_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push2 Equal a b)"
  using assms unfolding push2_def by (induct a, auto) (subst push.simps, auto)

theorem push_less_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push (Less a) b)"
  using assms by (induct b, auto) (subst push.simps, auto)

theorem push2_less_dufree [simp]:
  assumes "dufree a" and "dufree b"
  shows "dufree (push2 Less a b)"
  using assms unfolding push2_def by (induct a, auto) (subst push.simps, auto)


theorem push_vcond_dufree [simp]:
  assumes "dufree r" and "dufree t" and "dufree e"
  shows "dufree (push (\<lambda>x. VCond (L b x) t e) r)"
  using assms by (induct r, auto) (subst push.simps, auto)

theorem push_pcond_dufree [simp]:
  assumes "dufree r" and "dufree t" and "dufree e"
  shows "dufree (push (\<lambda>x. PCond (L b x) t e) r)"
  using assms by (induct r, auto) (subst push.simps, auto)

theorem push_dcond_dufree [simp]:
  assumes "dufree r" and "dufree t" and "dufree e"
  shows "dufree (push (\<lambda>x. DCond (L b x) t e) r)"
  using assms by (induct r, auto) (subst push.simps, auto)

theorem push_rcond_dufree [simp]:
  assumes "dufree r" and "dufree t" and "dufree e"
  shows "dufree (push (\<lambda>x. RCond (L b x) t e) r)"
  using assms by (induct r, auto) (subst push.simps, auto)


theorem strictify_dufree_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "dufree v \<Longrightarrow> dufree (strictify v)"
    and "dufree p \<Longrightarrow> dufree (strictify p)"
    and "dufree d \<Longrightarrow> dufree (strictify d)"
    and "dufree r \<Longrightarrow> dufree (strictify r)"
    and "dufree l \<Longrightarrow> dufree (strictify l)"
  by (induct v and p and d and r and l) auto

theorem strictify_dufree_disj [simp]:
  fixes d :: disj assumes "dufree d"
  shows "dufree (strictify d)"
  using assms by (induct d) auto

theorem decideCE_dufree [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred
  shows "dufree v \<Longrightarrow> dufree (decideCE l v)"
    and "dufree p \<Longrightarrow> dufree (decideCE l p)"
    and "dufree d \<Longrightarrow> dufree (decideCE l d)"
    and "dufree r \<Longrightarrow> dufree (decideCE l r)"
  using assms unfolding decideCE_def
  by (case_tac "bte v", auto, case_tac "bte p", auto, case_tac "bte d", auto, case_tac "bte r", auto)

theorem decide_dufree_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "dufree v \<Longrightarrow> dufree (decide x v)"
    and "dufree p \<Longrightarrow> dufree (decide x p)"
    and "dufree d \<Longrightarrow> dufree (decide x d)"
    and "dufree r \<Longrightarrow> dufree (decide x r)"
    and "dufree l \<Longrightarrow> dufree (decide x l)"
  by (induct v and p and d and r and l) auto

theorem decide_dufree_disj [simp]:
  fixes d :: disj assumes "dufree d"
  shows "dufree (decide l d)"
  using assms by (induct d) auto

theorem pullD_dufree [simp]:
  assumes "dufree d"
  shows "dufree (pullD d)"
  using assms by (induct d rule: pullD.induct) auto

theorem strictify_dufree_conj [simp]:
  fixes c :: conj assumes "dufree c"
  shows "dufree (strictify c)"
  using assms by (induct c) auto

theorem strictify_dufree [simp]:
  fixes c :: cnf assumes "dufree c"
  shows "dufree (strictify c)"
  using assms by (cases c) simp+



(*-------------------------------*)
(*- Strictify Preserves Guarded -*)
(*-------------------------------*)

theorem decide_guarded_disj [intro!]:
  fixes d :: disj and k :: conj assumes "guarded (l \<and>. k) d"
  shows "guarded (l \<and>. k) (decide l d)"
  using assms by (induct d arbitrary: k) (auto, drule_tac x="\<not>. (decide l lit) \<and>. k" in meta_spec,
    drule meta_mp, (erule guarded_relax, auto)+)

theorem strictify_guarded_disj [simp]:
  fixes d :: disj assumes "guarded k d"
  shows "guarded k (strictify d)"
  using assms by (induct d arbitrary: k) (auto, drule_tac x="\<not>. (strictify lit) \<and>. k" in meta_spec,
    erule meta_mp, erule guarded_relax, auto)

theorem pullD_guarded [simp]:
  assumes "guarded k d"
  shows "guarded k (pullD d)"
  using assms

apply (induct d arbitrary: k rule: pullD.induct)
apply auto[1]
apply (drule_tac x=R in spec, drule_tac x=f in spec, clarsimp)
apply (drule_tac x="C (D (\<not>. (\<not>. l)) DFlse) CTrue \<and>. k" in meta_spec)
apply (erule meta_mp)
apply (rule_tac k="l \<and>. k" in guarded_relax)
apply auto[2]
apply (drule_tac x=R in spec, drule_tac x=f in spec, clarsimp)
apply (erule guarded_relax)
apply auto[1]
apply (drule_tac x=R in spec, drule_tac x=f in spec, clarsimp)
apply (rotate_tac 1)
apply (drule_tac x="C (D (\<not>. (\<not>. (\<not>. l))) DFlse) CTrue \<and>. l \<longrightarrow>. pullD (decide l (L b t \<or>. d)) \<and>. k" in meta_spec)
apply (erule meta_mp)
apply (rule_tac k="\<not>. l \<and>. k" in guarded_relax)
apply auto[2]
apply (drule_tac x=R in spec, drule_tac x=f in spec, clarsimp)
apply (erule guarded_relax)
apply auto[1]
apply simp
apply safe
apply auto[1]
apply auto[1]
apply (drule_tac x="C (D (\<not>. (\<not>. (\<not>. (L v (Equal vb vc))))) DFlse) CTrue \<and>. k" in meta_spec)
apply (erule meta_mp)
apply (erule guarded_relax)
apply clarsimp
apply simp
apply safe
apply auto[1]
apply auto[1]
apply (drule_tac x="C (D (\<not>. (\<not>. (\<not>. (L v (Less vb vc))))) DFlse) CTrue \<and>. k" in meta_spec)
apply (erule meta_mp)
apply (erule guarded_relax)
apply clarsimp
apply simp
done

theorem strictify_guarded_conj [simp]:
  fixes c :: conj assumes "guarded k c"
  shows "guarded k (strictify c)"
  using assms by (induct c arbitrary: k) (auto, drule_tac x="pullD (strictify disj) \<and>. k" in meta_spec,
    erule meta_mp, erule guarded_relax, auto, drule guarded_sem, auto)

theorem strictify_guarded [simp]:
  fixes c :: cnf assumes "guarded k c"
  shows "guarded k (strictify c)"
  using assms by (cases c) simp+

end

