header {* Syntax and Semantics of Statements and the WP Transformer *}

theory Procedures imports Substitutions Safe begin


(* Syntax *)

datatype stm =
  Seq stm stm
| Ite lit stm stm
| Assign name valueExp
| Assert conj

| Insert pathExp label valueExp
| Update pathExp valueExp
| Delete pathExp

abbreviation Skip where
  "Skip \<equiv> Assert CTrue"

instantiation stm :: xfold begin

  definition xp_stm :: "bool xfold \<Rightarrow> stm \<Rightarrow> bool" where
    "xp_stm _ \<equiv> c\<^sub>T"

  definition xc_stm :: "nat xfold \<Rightarrow> stm \<Rightarrow> nat" where
    "xc_stm _ \<equiv> c\<^sub>0"

  definition xv_stm :: "name set xfold \<Rightarrow> stm \<Rightarrow> name set" where
    "xv_stm _ \<equiv> c\<^sub>E"

  definition xl_stm :: "lit list xfold \<Rightarrow> stm \<Rightarrow> lit list" where
    "xl_stm _ \<equiv> c\<^sub>L"

  fun xprop_stm :: "bool xfold \<Rightarrow> stm \<Rightarrow> bool" where
    "xprop f (Seq a b)      \<longleftrightarrow> xprop f a \<and> xprop f b"
  | "xprop f (Ite b t e)    \<longleftrightarrow> xprop f b \<and> xprop f t \<and> xprop f e"
  | "xprop f (Assign n v)   \<longleftrightarrow> xprop f v"
  | "xprop f (Assert c)     \<longleftrightarrow> xprop f c"
  | "xprop f (Insert p l k) \<longleftrightarrow> xprop f p \<and> xprop f k"
  | "xprop f (Update p v)   \<longleftrightarrow> xprop f p \<and> xprop f v"
  | "xprop f (Delete p)     \<longleftrightarrow> xprop f p"

  fun xcount_stm :: "nat xfold \<Rightarrow> stm \<Rightarrow> nat" where
    "xcount f (Seq a b)      = xcount f a + xcount f b"
  | "xcount f (Ite b t e)    = xcount f b + xcount f t + xcount f e"
  | "xcount f (Assign n v)   = xcount f v"
  | "xcount f (Assert c)     = xcount f c"
  | "xcount f (Insert p l k) = xcount f p + xcount f k"
  | "xcount f (Update p v)   = xcount f p + xcount f v"
  | "xcount f (Delete p)     = xcount f p"

  fun xvars_stm :: "name set xfold \<Rightarrow> stm \<Rightarrow> name set" where
    "xvars f (Seq a b)      = xvars f a \<union> xvars f b"
  | "xvars f (Ite b t e)    = xvars f b \<union> xvars f t \<union> xvars f e"
  | "xvars f (Assign n v)   = xvars f v"
  | "xvars f (Assert c)     = xvars f c"
  | "xvars f (Insert p l k) = xvars f p \<union> xvars f k"
  | "xvars f (Update p v)   = xvars f p \<union> xvars f v"
  | "xvars f (Delete p)     = xvars f p"

  fun xlits_stm :: "lit list xfold \<Rightarrow> stm \<Rightarrow> lit list" where
    "xlits f (Seq a b)      = xlits f a @ xlits f b"
  | "xlits f (Ite b t e)    = xlits f b @ xlits f t @ xlits f e"
  | "xlits f (Assign n v)   = xlits f v"
  | "xlits f (Assert c)     = xlits f c"
  | "xlits f (Insert p l k) = xlits f p @ xlits f k"
  | "xlits f (Update p v)   = xlits f p @ xlits f v"
  | "xlits f (Delete p)     = xlits f p"

  lemmas xx_stm_defs [simp] = xp_stm_def xc_stm_def xv_stm_def xl_stm_def

instance by (intro_classes, simp+) (induct_tac x, simp_all add: xcount_to_xprop)+ end


(* Semantics *)

inductive sem_stm :: "stm \<Rightarrow> param \<Rightarrow> param \<Rightarrow> bool" ("_, _ \<leadsto> _" [1000, 0, 1000] 100) where
  Seq:    "\<lbrakk> a, R \<leadsto> S; b, S \<leadsto> T \<rbrakk> \<Longrightarrow> (Seq a b), R \<leadsto> T"
| Ite:    "\<lbrakk> \<parallel>b\<parallel>\<^sub>R = c\<^sub>b; c \<longrightarrow> t, R \<leadsto> S; \<not>c \<longrightarrow> e, R \<leadsto> S \<rbrakk> \<Longrightarrow> (Ite b t e), R \<leadsto> S"
| Assign: "\<parallel>w\<parallel>\<^sub>R = {v}\<^sub>*\<^sub>v \<Longrightarrow> (Assign n w), R \<leadsto> R(n \<mapsto>\<^sub>c v)"
| Assert: "\<parallel>c\<parallel>\<^sub>R = True\<^sub>b \<Longrightarrow> (Assert c), R \<leadsto> R"
| Insert: "\<lbrakk> \<parallel>q\<parallel>\<^sub>R = {p}\<^sub>*\<^sub>p; \<parallel>j\<parallel>\<^sub>R = {k\<^sub>k}\<^sub>*\<^sub>v; p/l[k] \<notin>\<^sub>1 R($); p \<in>\<^sub>1 R($) \<rbrakk> \<Longrightarrow> (Insert q l j), R \<leadsto> R($ \<mapsto> R($)[p/l[k] \<rightarrow> Unit])"
| Update: "\<lbrakk> \<parallel>q\<parallel>\<^sub>R = {p}\<^sub>*\<^sub>p; \<parallel>w\<parallel>\<^sub>R = {v}\<^sub>*\<^sub>v; p \<in>\<^sub>1 R($) \<rbrakk> \<Longrightarrow> (Update q w), R \<leadsto> R($ \<mapsto> R($)[p \<rightarrow> v])"
| Delete: "\<lbrakk> \<parallel>q\<parallel>\<^sub>R = {p}\<^sub>*\<^sub>p; p \<in>\<^sub>1 R($) \<rbrakk> \<Longrightarrow> (Delete q), R \<leadsto> R($ \<mapsto> R($)[p \<rightarrow>])"

inductive_cases sem_stm_Seq:       "(Seq s1 s2), R \<leadsto> S"
inductive_cases sem_stm_Ite:       "(Ite b t e), R \<leadsto> S"
inductive_cases sem_stm_Assign:   "(Assign n v), R \<leadsto> S"
inductive_cases sem_stm_Assert:     "(Assert Q), R \<leadsto> S"
inductive_cases sem_stm_Insert: "(Insert p l i), R \<leadsto> S"
inductive_cases sem_stm_Update:   "(Update p v), R \<leadsto> S"
inductive_cases sem_stm_Delete:     "(Delete p), R \<leadsto> S"


theorem sem_stm_document [simp]:
  assumes "c, R \<leadsto> S" and "document R($)"
  shows "document S($)"
  using assms by (induct S) auto

theorem sem_stm_deterministic [rule_format, elim]:
  assumes "c, R \<leadsto> S"
  shows "\<forall>T. c, R \<leadsto> T \<longrightarrow> T = S"
  using assms
proof induct
  case Seq    thus ?case by safe (erule sem_stm_Seq, simp) next
  case Ite    thus ?case by safe (erule sem_stm_Ite, (drule spec)+, force)+ next
  case Assign thus ?case by safe (erule sem_stm_Assign, simp) next
  case Assert thus ?case by safe (erule sem_stm_Assert, simp) next
  case Delete thus ?case by safe (erule sem_stm_Delete, simp) next
  case Insert thus ?case by safe (erule sem_stm_Insert, simp) next
  case Update thus ?case by safe (erule sem_stm_Update, simp)
qed


(* Hoare Logic and WP *)

definition sem_Hoare :: "cnf \<Rightarrow> stm \<Rightarrow> cnf \<Rightarrow> bool" ("\<Turnstile> {_} _ {_}") where
  "\<Turnstile> {P} s {Q} \<equiv> \<forall>R. \<parallel>P\<parallel>\<^sub>R = True \<longrightarrow> (\<exists>S. (s, R \<leadsto> S) \<and> \<parallel>Q\<parallel>\<^sub>S = True)"

fun wpc :: "stm \<Rightarrow> conj \<Rightarrow> conj" where
  "wpc (Seq a b)      P = wpc a (wpc b P)"
| "wpc (Ite b t e)    P = guard b \<and>. (b \<longrightarrow>. wpc t P) \<and>. (\<not>. b \<longrightarrow>. wpc e P)"
| "wpc (Assign n x)   P = guard x \<and>. unique. x \<and>. P[ x / PVar n ]\<^sub>v"
| "wpc (Assert Q)     P = Q \<and>. P"
| "wpc (Insert p l k) P = guard k \<and>. unique. k \<and>. k is. TKey \<and>. guard p \<and>. p \<in>\<^sub>1.$ \<and>. p/l[k]. \<notin>\<^sub>1.$ \<and>. P[ $.[p/l[k]. \<rightarrow> Val Unit]. / $. ]\<^sub>d"
| "wpc (Update p v)   P = guard v \<and>. unique. v \<and>. guard p \<and>. p \<in>\<^sub>1.$ \<and>. P[ $.[p \<rightarrow> v]. / $. ]\<^sub>d"
| "wpc (Delete p)     P = guard p \<and>. p \<in>\<^sub>1.$ \<and>. P[ $.[p \<rightarrow>]. / $. ]\<^sub>d"

fun wp :: "stm \<Rightarrow> cnf \<Rightarrow> cnf" where
  "wp s (\<forall>. c) = \<forall>. (wpc s c)"


(* Syntactic Properties *)

theorem wpc_basic [simp]:
  assumes "basic s" and "basic P"
  shows "strict (wpc s P)" and "straight (wpc s P)"
  using assms by (induct arbitrary: P) auto

corollary wp_basic [simp]:
  assumes "basic s" and "basic P"
  shows "strict (wp s P)" and "straight (wp s P)"
  using assms by (cases P, simp)+

fun wform_stm :: "stm \<Rightarrow> bool" where
  "wform_stm (Seq a b)      \<longleftrightarrow> wform_stm a \<and> wform_stm b"
| "wform_stm (Ite b t e)    \<longleftrightarrow> wform b \<and> wform_stm t \<and> wform_stm e"
| "wform_stm (Assign n v)   \<longleftrightarrow> uve v \<and> wform v"
| "wform_stm (Assert P)     \<longleftrightarrow> wform P"
| "wform_stm (Insert p l k) \<longleftrightarrow> upe p \<and> uve k \<and> wform p \<and> wform k"
| "wform_stm (Update p v)   \<longleftrightarrow> upe p \<and> uve v \<and> wform p \<and> wform v"
| "wform_stm (Delete p)     \<longleftrightarrow> upe p \<and> wform p"

theorem wp_wform [simp]:
  assumes "wform P" and "basic P" and "wform_stm s" and "basic s"
  shows "wform (wp s P)"
  using assms by (cases P, auto) (induct, auto)


(* Semantic Properties *)

theorem wp_sound_conj:
  assumes "closed s" and "basic s"
  shows "\<Turnstile> {wp s (\<forall>. c)} s {\<forall>. c}"
  using assms unfolding sem_Hoare_def
proof (induct arbitrary: c)
  case (Ite b t e) thus ?case proof clarify
    fix R assume "\<parallel>wp (Ite b t e) (\<forall>. c)\<parallel>\<^sub>R\<^sub>\<up> = True" then
    obtain x and S where "\<parallel>b\<parallel>\<^sub>R = x" and "x \<longrightarrow> t, R \<leadsto> S"
        and "\<not> x \<longrightarrow> e, R \<leadsto> S" and "\<forall>f. \<parallel>c\<parallel>\<^sub>S\<^sub>+\<^sub>f = True" using Ite
      by - (clarsimp, erule sbeE, simp_all, (metis par.simps(2) sem_cnf_trueD)+)
    thus "\<exists>S. (Ite b t e), R \<leadsto> S \<and> \<parallel>\<forall>. c\<parallel>\<^sub>S\<^sub>\<up> = True"
      using Ite by (auto intro: sem_stm.Ite)
  qed next

  case Assign thus ?case by auto (metis par.simps(2) sem_cnf_trueI sem_stm.Assign) next
  case Assert thus ?case by (auto intro: sem_stm.Assert) next

  case (Seq a b) thus ?case proof clarify
    have ca: "closed a" "basic a" and cb: "closed b" "basic b" using Seq by simp+
    fix R assume "\<parallel>wp (Seq a b) (\<forall>. c)\<parallel>\<^sub>R\<^sub>\<up> = True" then
    obtain T where "a, R \<leadsto> T" and "\<parallel>\<forall>. (wpc b c)\<parallel>\<^sub>T\<^sub>\<up> = True"
      using ca and Seq(1)[of "wpc b c", THEN spec, of R] by auto moreover then
    obtain S where "b, T \<leadsto> S" and "\<parallel>\<forall>. c\<parallel>\<^sub>S\<^sub>\<up> = True"
      using cb and Seq(2)[of c, THEN spec, of T] by auto ultimately
    show "\<exists>S. (Seq a b), R \<leadsto> S \<and> \<parallel>\<forall>. c\<parallel>\<^sub>S\<^sub>\<up> = True" by (auto intro: sem_stm.Seq)
  qed next

  case (Insert p l k) thus ?case proof clarify
    fix R assume "\<parallel>wp (Insert p l k) (\<forall>. c)\<parallel>\<^sub>R\<^sub>\<up> = True" then
    obtain q and j where "\<parallel>p\<parallel>\<^sub>R = {q}\<^sub>*\<^sub>p" "\<parallel>k\<parallel>\<^sub>R = {j\<^sub>k}\<^sub>*\<^sub>v" and "q/l[j] \<notin>\<^sub>1 R($)" "q \<in>\<^sub>1 R($)"
      and "\<forall>f. \<parallel>c[ $.[p/l[k]. \<rightarrow> Val Unit]. / $. ]\<^sub>d\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
      using Insert by clarsimp (subst (asm) sem_conj_doc_subst, auto, rule doc_upd_document, auto)
    thus "\<exists>S. (Insert p l k), R \<leadsto> S \<and> \<parallel>\<forall>. c\<parallel>\<^sub>S\<^sub>\<up> = True"
      using Insert by (rule_tac x="R($ \<mapsto> R($)[q/l[j] \<rightarrow> Unit])" in exI) (auto intro: sem_stm.Insert)
  qed next

  case (Update p v) thus ?case proof clarify
    fix R assume "\<parallel>wp (Update p v) (\<forall>. c)\<parallel>\<^sub>R\<^sub>\<up> = True" then
    obtain q and x where "\<parallel>p\<parallel>\<^sub>R = {q}\<^sub>*\<^sub>p" and "\<parallel>v\<parallel>\<^sub>R = {x}\<^sub>*\<^sub>v" and "q \<in>\<^sub>1 R($)"
      and "\<forall>f. \<parallel>c[ $.[p \<rightarrow> v]. / $. ]\<^sub>d\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
      using Update by clarsimp
        (subst (asm) sem_conj_doc_subst, auto, rule doc_upd_document, rule document_parent_dom, auto)
    thus "\<exists>S. (Update p v), R \<leadsto> S \<and> \<parallel>\<forall>. c\<parallel>\<^sub>S\<^sub>\<up> = True"
      using Update by (rule_tac x="R($ \<mapsto> R($)[q \<rightarrow> x])" in exI) (auto intro: sem_stm.Update)
  qed next

  case (Delete p) thus ?case proof clarify
    fix R assume "\<parallel>wp (Delete p) (\<forall>. c)\<parallel>\<^sub>R\<^sub>\<up> = True" then
    obtain q where "\<parallel>p\<parallel>\<^sub>R = {q}\<^sub>*\<^sub>p" and "q \<in>\<^sub>1 R($)" and "\<forall>f. \<parallel>c[ $.[p \<rightarrow>]. / $. ]\<^sub>d\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
      using Delete by clarsimp blast
    thus "\<exists>S. (Delete p), R \<leadsto> S \<and> \<parallel>\<forall>. c\<parallel>\<^sub>S\<^sub>\<up> = True"
      using Delete by (rule_tac x="R($ \<mapsto> R($)[q \<rightarrow> ])" in exI) (auto intro: sem_stm.Delete)
  qed
qed

corollary wp_sound [simp]:
  assumes "closed s" and "basic s"
  shows "\<Turnstile> {wp s Q} s {Q}"
  using assms and wp_sound_conj by (cases Q) simp


theorem wp_weakest_general:
  assumes "closed s" and "basic s" and "s, R \<leadsto> S" and "\<parallel>\<forall>. Q\<parallel>\<^sub>S\<^sub>\<up> = True"
  shows "\<parallel>\<forall>. wpc s Q\<parallel>\<^sub>R\<^sub>\<up> = True"
  using assms
proof (induct arbitrary: Q R S)
  case Seq    thus ?case by - (erule sem_stm_Seq, simp) next
  case Ite    thus ?case by - (erule sem_stm_Ite, auto) next
  case Assign thus ?case by - (erule sem_stm_Assign, auto) next
  case Assert thus ?case by - (erule sem_stm_Assert, simp) next
  case Delete thus ?case by - (erule sem_stm_Delete, force) next
  case Insert thus ?case by - (erule sem_stm_Insert, auto) next
  case Update thus ?case by - (erule sem_stm_Update, auto)
qed

theorem wp_weakest:
  assumes "closed s" and "basic s" and "\<Turnstile> {P} s {Q}" and "\<parallel>P\<parallel>\<^sub>E = True"
  shows "\<parallel>wp s Q\<parallel>\<^sub>E = True"
  using assms unfolding sem_Hoare_def by (cases Q, simp, cases E, simp) (metis wp_weakest_general)+


end

