header {* Syntactic Properties of Assertions *}

theory Properties imports MapFold begin


(* Document Manipulations *)

fun countDU :: "docExp \<Rightarrow> nat" where
  "countDU (Upd _ _  _) = 1"
| "countDU (Del _ _)    = 1"
| "countDU _            = 0"

fun noDU :: "docExp \<Rightarrow> bool" where
  "noDU (Upd _ _ _) = False"
| "noDU (Del _ _)   = False"
| "noDU _           = True"

abbreviation dus where
  "dus \<equiv> xcount (XFold c\<^sub>0 c\<^sub>0 countDU c\<^sub>0)"

abbreviation dufree where
  "dufree \<equiv> xprop (XFold c\<^sub>T c\<^sub>T noDU c\<^sub>T)"

lemma toXP_DU [simp]:
  shows "toXP countDU = noDU"
  by (rule ext, case_tac x) simp+

corollary dufree_dus [simp]:
  shows "dus x = 0 \<longleftrightarrow> dufree x"
  by (rule xcount_to_xprop) simp+


(* Document Manipulations on the Top-Level *)

class topDUs = xfold +
fixes topDUs :: "'a :: xfold \<Rightarrow> nat"
assumes topDUs_dus [simp]:
  "topDUs x \<le> dus x"

instantiation valueExp and pathExp and docExp :: topDUs begin

  fun topDUs_docExp :: "docExp \<Rightarrow> nat" where
    "topDUs (Upd d _ _) = Suc (topDUs d)"
  | "topDUs (Del d _)   = Suc (topDUs d)"
  | "topDUs_docExp _    = 0"

  fun topDUs_valueExp :: "valueExp \<Rightarrow> nat"
  and topDUs_pathExp :: "pathExp \<Rightarrow> nat" where
    "topDUs (Read d p) = topDUs d + topDUs p"
  | "topDUs (Scal n m) = topDUs m"
  | "topDUs_valueExp _ = 0"

  | "topDUs (Path l p k) = topDUs p + topDUs k"
  | "topDUs_pathExp _    = 0"

  theorem topDUs_dus_exp:
    fixes v :: valueExp and p :: pathExp and d :: docExp
    shows "topDUs v \<le> dus v"
      and "topDUs p \<le> dus p"
      and "topDUs d \<le> dus d"
    by (induct v and p and d) auto

instance by intro_classes (rule topDUs_dus_exp)+ end


corollary dufree_topDUs_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp
  shows "dufree v \<Longrightarrow> topDUs v = 0"
    and "dufree p \<Longrightarrow> topDUs p = 0"
    and "dufree d \<Longrightarrow> topDUs d = 0"
  unfolding dufree_dus[symmetric] by (metis topDUs_dus le_0_eq)+

theorem topDUs_diff:
  shows "\<exists>x. dus d = topDUs d + x"
  by (metis le_iff_add topDUs_dus)

theorem topDUs_diffD:
  assumes "P (dus d)"
  shows "\<exists>x. P (topDUs d + x)"
  using assms by (metis topDUs_diff)

abbreviation dutop where
  "dutop x \<equiv> dus x = topDUs x"


(* Strict Expressions *)

fun countVC :: "valueExp \<Rightarrow> nat" where
  "countVC (VCond _ _ _) = 1"
| "countVC _ = 0"

fun countPC :: "pathExp \<Rightarrow> nat" where
  "countPC (PCond _ _ _) = 1"
| "countPC _ = 0"

fun countDC :: "docExp \<Rightarrow> nat" where
  "countDC (DCond _ _ _) = 1"
| "countDC _ = 0"

fun countRC :: "pred \<Rightarrow> nat" where
  "countRC (RCond _ _ _) = 1"
| "countRC _ = 0"

fun noVC :: "valueExp \<Rightarrow> bool" where
  "noVC (VCond _ _ _) = False"
| "noVC _ = True"

fun noPC :: "pathExp \<Rightarrow> bool" where
  "noPC (PCond _ _ _) = False"
| "noPC _ = True"

fun noDC :: "docExp \<Rightarrow> bool" where
  "noDC (DCond _ _ _) = False"
| "noDC _ = True"

fun noRC :: "pred \<Rightarrow> bool" where
  "noRC (RCond _ _ _) = False"
| "noRC _ = True"

abbreviation conds where
  "conds \<equiv> xcount (XFold countVC countPC countDC countRC)"

abbreviation strict where
  "strict \<equiv> xprop (XFold noVC noPC noDC noRC)"

lemma toXP_Cond [simp]:
  shows "toXP countVC = noVC"
    and "toXP countPC = noPC"
    and "toXP countDC = noDC"
    and "toXP countRC = noRC"
  by (rule ext, case_tac x, simp+)+

corollary strict_conds [simp]:
  shows "conds x = 0 \<longleftrightarrow> strict x"
  by (rule xcount_to_xprop) simp+


(* Check for Non-Triviality *)

fun xs where
  "xs x = (if size x > 0 then 1 else 0)"

abbreviation xsize where
  "xsize \<equiv> xcount (XFold xs xs xs xs)"

theorem xcount_non_triv_exp:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "xsize v + conds v = size v"
    and "xsize p + conds p = size p"
    and "xsize d + conds d = size d"
    and "xsize r + conds r = size r"
    and "xsize l + conds l + 1 = size l"
  by (induct v and p and d and r and l) auto


(* Straight Paths *)

fun countPOps :: "pathExp \<Rightarrow> nat" where
  "countPOps (PRep _ _)   = 1"
| "countPOps (PPlus _ _)  = 1"
| "countPOps (PMinus _ _) = 1"
| "countPOps _            = 0"

fun countVOps :: "valueExp \<Rightarrow> nat" where
  "countVOps (VRep _ _)   = 1"
| "countVOps (VMinus _ _) = 1"
| "countVOps _            = 0"

fun noPOps :: "pathExp \<Rightarrow> bool" where
  "noPOps (PRep _ _)   = False"
| "noPOps (PPlus _ _)  = False"
| "noPOps (PMinus _ _) = False"
| "noPOps _            = True"

fun noVOps :: "valueExp \<Rightarrow> bool" where
  "noVOps (VRep _ _)   = False"
| "noVOps (VMinus _ _) = False"
| "noVOps _            = True"

abbreviation ops where
  "ops \<equiv> xcount (XFold countVOps countPOps c\<^sub>0 c\<^sub>0)"

abbreviation straight where
  "straight \<equiv> xprop (XFold noVOps noPOps c\<^sub>T c\<^sub>T)"

abbreviation basic where
  "basic x \<equiv> strict x \<and> straight x"

lemma toXP_POps [simp]:
  "toXP countPOps = noPOps"
  by (rule ext, case_tac x) simp+

lemma toXP_VOps [simp]:
  "toXP countVOps = noVOps"
  by (rule ext, case_tac x) simp+

corollary straight_ops [simp]:
  "ops x = 0 \<longleftrightarrow> straight x"
  by (rule xcount_to_xprop) simp+


(* Multiset Operations on the Top-Level *)

class topOps = xfold +
fixes topOps :: "'a :: xfold \<Rightarrow> nat"
assumes topOps_ops [simp]:
  "topOps x \<le> ops x"


instantiation valueExp and pathExp and docExp :: topOps begin

  fun topOps_pathExp :: "pathExp \<Rightarrow> nat" where
    "topOps (PRep n p)   = Suc (topOps p)"
  | "topOps (PPlus a b)  = Suc (topOps a + topOps b)"
  | "topOps (PMinus a b) = Suc (topOps a + topOps b)"
  | "topOps_pathExp _    = 0"

  fun topOps_valueExp :: "valueExp \<Rightarrow> nat" where
    "topOps (Scal n m)   = topOps m"
  | "topOps (VRep n m)   = Suc (topOps m)"
  | "topOps (VPlus a b)  = topOps a + topOps b"
  | "topOps (VMinus a b) = Suc (topOps a + topOps b)"
  | "topOps_valueExp _   = 0"

  fun topOps_docExp :: "docExp \<Rightarrow> nat" where
    "topOps_docExp _ = 0"

  theorem topOps_ops_pathExp:
    fixes p :: pathExp
    shows "topOps p \<le> ops p"
    by (induct p) auto

  theorem topOps_ops_valueExp:
    fixes v :: valueExp
    shows "topOps v \<le> ops v"
    by (induct v) auto

instance by intro_classes (rule topOps_ops_valueExp, rule topOps_ops_pathExp, simp) end


corollary straight_topOps_exp [simp]:
  fixes v :: valueExp and p :: pathExp
  shows "straight v \<Longrightarrow> topOps v = 0"
    and "straight p \<Longrightarrow> topOps p = 0"
  unfolding straight_ops[symmetric] by (metis topOps_ops le_0_eq)+

theorem topOps_diff:
  shows "\<exists>x. ops d = topOps d + x"
  by (metis le_iff_add topOps_ops)

theorem topOps_diffD:
  assumes "P (ops d)"
  shows "\<exists>x. P (topOps d + x)"
  using assms by (metis topOps_diff)

abbreviation otop where
  "otop p \<equiv> ops p = topOps p"


(* Auxiliary Lemmas for Proving Properties using xcount *)

lemma plus_le_zero [dest]:
  fixes a :: nat assumes "a + b = c" and "c \<le> a"
  shows "a = c \<and> b = 0"
  using assms by arith

lemma plus_le_zero_sym [dest]:
  fixes a :: nat assumes "b + a = c" and "c \<le> a"
  shows "a = c \<and> b = 0"
  using assms by arith

lemma plus_le_twice [dest]:
  fixes a :: nat assumes "a + b = c + d" and "c \<le> a" and "d \<le> b"
  shows "a = c \<and> b = d"
  using assms by arith

lemma plus_le_twice_twice [dest]:
  fixes a :: nat assumes "a + (b + c) = d + e" and "d \<le> b" and "e \<le> c"
  shows "a = 0 \<and> b = d \<and> c = e"
  using assms by auto

lemma plus_twice_le [intro]:
  fixes a :: nat assumes "a \<le> c" and "b \<le> d"
  shows "a + b \<le> c + d"
  using assms by arith


(* Statically Unique Expressions (Conservative) *)

fun uve :: "valueExp \<Rightarrow> bool"
and upe :: "pathExp \<Rightarrow> bool" where

  "uve (Val _)       \<longleftrightarrow> True"
| "uve (FVar _)      \<longleftrightarrow> True"
| "uve (PVar _)      \<longleftrightarrow> True"
| "uve (Sum _)       \<longleftrightarrow> True"
| "uve (Size _)      \<longleftrightarrow> True"
| "uve (Count _ _)   \<longleftrightarrow> True"
| "uve (Tally _ _)   \<longleftrightarrow> True"
| "uve (Read d p)    \<longleftrightarrow> upe p"
| "uve (Scal n m)    \<longleftrightarrow> uve m" (* could be narrowed, but why bother... *)
| "uve (VRep _ _)    \<longleftrightarrow> False" (* same as for scal *)
| "uve (VPlus _ _)   \<longleftrightarrow> False" (* conservative *)
| "uve (VMinus _ _)  \<longleftrightarrow> False"
| "uve (VCond _ t e) \<longleftrightarrow> uve t \<and> uve e"

| "upe Root.         \<longleftrightarrow> True"
| "upe (Path l p k)  \<longleftrightarrow> upe p \<and> uve k"
| "upe (Kind l p)    \<longleftrightarrow> False" (* either empty or infinite, but not single, COULD BE EMPTY... *)
| "upe (PRep _ _)    \<longleftrightarrow> False" (* same as for vrep *)
| "upe (PPlus _ _)   \<longleftrightarrow> False"
| "upe (PMinus _ _)  \<longleftrightarrow> False"
| "upe (PCond _ t e) \<longleftrightarrow> upe t \<and> upe e"


theorem uxe_sem:
  shows "uve ve \<Longrightarrow> \<parallel>ve\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>ve\<parallel>\<^sub>E = {}\<^sub>*\<^sub>v \<or> (\<exists>v. \<parallel>ve\<parallel>\<^sub>E = {v}\<^sub>*\<^sub>v)"
    and "upe pe \<Longrightarrow> \<parallel>pe\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>pe\<parallel>\<^sub>E = {}\<^sub>*\<^sub>p \<or> (\<exists>p. \<parallel>pe\<parallel>\<^sub>E = {p}\<^sub>*\<^sub>p)"
  by (induct rule: uve_upe.induct, auto)
     (case_tac "x(pa)\<^sub>1", auto, case_tac v, auto, split if_splits, auto)

corollary uve_sem:
  assumes "uve v" and "\<parallel>v\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>v\<parallel>\<^sub>E = {}\<^sub>*\<^sub>v \<or> (\<exists>w. \<parallel>v\<parallel>\<^sub>E = {w}\<^sub>*\<^sub>v)"
  using assms by - (drule uxe_sem[of _ E], auto)

corollary upe_sem:
  assumes "upe p" and "\<parallel>p\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>p\<parallel>\<^sub>E = {}\<^sub>*\<^sub>p \<or> (\<exists>q. \<parallel>p\<parallel>\<^sub>E = {q}\<^sub>*\<^sub>p)"
  using assms by - (drule uxe_sem[of _ E], auto)


(* Well-Formed Formulas *)

fun wfVE :: "valueExp \<Rightarrow> bool" where
  "wfVE (Count m v) \<longleftrightarrow> uve v"
| "wfVE (Scal n m)  \<longleftrightarrow> uve n"
| "wfVE (VRep n m)  \<longleftrightarrow> uve n"
| "wfVE _           \<longleftrightarrow> True"

fun wfPE :: "pathExp \<Rightarrow> bool" where
  "wfPE (PRep n m)  \<longleftrightarrow> uve n"
| "wfPE _           \<longleftrightarrow> True"

fun wfDE :: "docExp \<Rightarrow> bool" where
  "wfDE (Upd d p v) \<longleftrightarrow> upe p \<and> uve v"
| "wfDE (Del d p)   \<longleftrightarrow> upe p"
| "wfDE _           \<longleftrightarrow> True"

fun wfRE :: "pred \<Rightarrow> bool" where
  "wfRE (Equal a b) \<longleftrightarrow> uve a \<and> uve b"
| "wfRE (Less a b)  \<longleftrightarrow> uve a \<and> uve b"
| "wfRE _           \<longleftrightarrow> True"

abbreviation wform where
  "wform \<equiv> xprop (XFold wfVE wfPE wfDE wfRE)"


(* Quantifier-Free Expressions and Formulas *)

fun noFVar :: "valueExp \<Rightarrow> bool" where
  "noFVar (FVar _) = False"
| "noFVar _ = True"

abbreviation closed where
  "closed \<equiv> xprop (XFold noFVar c\<^sub>T c\<^sub>T c\<^sub>T)"

class closed = xfold + sem +
assumes sem_closed [simp]:
  "xprop (XFold noFVar c\<^sub>T c\<^sub>T c\<^sub>T) x \<Longrightarrow> \<parallel>x\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<parallel>x\<parallel>\<^sub>R"

instantiation cnf :: closed begin
  instance by intro_classes (rule sem_cnf_is_closed) end  (* trivial instance, as cnfs are always closed *)

theorem sem_closed_exp:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "closed v \<Longrightarrow> \<parallel>v\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<parallel>v\<parallel>\<^sub>R"
    and "closed p \<Longrightarrow> \<parallel>p\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<parallel>p\<parallel>\<^sub>R"
    and "closed d \<Longrightarrow> \<parallel>d\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<parallel>d\<parallel>\<^sub>R"
    and "closed r \<Longrightarrow> \<parallel>r\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<parallel>r\<parallel>\<^sub>R"
    and "closed l \<Longrightarrow> \<parallel>l\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<parallel>l\<parallel>\<^sub>R"
  by (induct v and p and d and r and l) auto

instantiation valueExp and pathExp and docExp and pred and lit :: closed begin
  instance by intro_classes (erule sem_closed_exp)+ end


theorem sem_closed_disj:
  fixes d :: disj assumes "closed d"
  shows "\<parallel>d\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<parallel>d\<parallel>\<^sub>R"
  using assms by (induct d) auto

instantiation disj :: closed begin
  instance by intro_classes (erule sem_closed_disj) end


theorem sem_closed_conj:
  fixes c :: conj assumes "closed c"
  shows "\<parallel>c\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<parallel>c\<parallel>\<^sub>R"
  using assms by (induct c) auto

instantiation conj :: closed begin
  instance by intro_classes (erule sem_closed_conj) end


(* Closed Semantics and Formulas *)

theorem sem_not_closed_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "\<not> closed v \<Longrightarrow> strict v \<Longrightarrow> \<parallel>v\<parallel>\<^sub>R\<^sub>\<up> = \<bottom>"
    and "\<not> closed p \<Longrightarrow> strict p \<Longrightarrow> \<parallel>p\<parallel>\<^sub>R\<^sub>\<up> = \<bottom>"
    and "\<not> closed d \<Longrightarrow> strict d \<Longrightarrow> \<parallel>d\<parallel>\<^sub>R\<^sub>\<up> = \<bottom>"
    and "\<not> closed r \<Longrightarrow> strict r \<Longrightarrow> \<parallel>r\<parallel>\<^sub>R\<^sub>\<up> = \<bottom>"
    and "\<not> closed l \<Longrightarrow> strict l \<Longrightarrow> \<parallel>l\<parallel>\<^sub>R\<^sub>\<up> = \<bottom>"
  by (induct v and p and d and r and l) auto

(* disj and conj are not strict, so there are no similar theorems *)

theorem sem_closed_cnf_closed:
  assumes "closed c"
  shows "\<parallel>\<forall>. c\<parallel>\<^sub>R\<^sub>\<up> = \<parallel>c\<parallel>\<^sub>R"
  using assms by - (rule sbeI, auto)

corollary sem_closed_cnf [simp]:
  assumes "closed c"
  shows "\<parallel>\<forall>. c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by (cases E) (simp, erule sem_closed_cnf_closed)+

end

