header {* Safety and the Safeguard Transformer *}

theory Safe imports Guarded Simplify begin


(* Safe Expressions and Assertions *)

abbreviation safe :: "'a :: sem \<Rightarrow> bool" where
  "safe x \<equiv> \<forall>E. \<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>"

corollary guarded_cnf_safe:
  fixes c :: cnf assumes "guarded k c"
  shows "safe c"
  using assms by (cases c) (auto, drule_tac R="E\<^sub>R" and f=f in guarded_conj_sem, auto)


(* Domain Identities *)

definition id_dom :: "('a :: sem \<Rightarrow> 'a) \<Rightarrow> bool" ("id\<^sub>E") where
  "id\<^sub>E f \<equiv> \<forall>x E. \<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom> \<longrightarrow> \<parallel>f x\<parallel>\<^sub>E = \<parallel>x\<parallel>\<^sub>E"

theorem sem_id_dom [simp]:
  assumes "id\<^sub>E f" and "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>f x\<parallel>\<^sub>E = \<parallel>x\<parallel>\<^sub>E"
  using assms unfolding id_dom_def by simp

theorem id_id_dom [simp]:
  shows "id\<^sub>E id" and "id\<^sub>E (\<lambda>a. a)"
  unfolding id_dom_def by simp+


(* Expression Mapping Identities *)

fun xmap_id :: "xmap \<Rightarrow> bool" where
  "xmap_id (XMap v p d r l e c) \<longleftrightarrow>
    id\<^sub>E v \<and> id\<^sub>E p \<and> id\<^sub>E d \<and> id\<^sub>E r \<and> id\<^sub>E l \<and> id\<^sub>E e \<and> id\<^sub>E c \<and> (\<forall>x. kindOf (p x) = kindOf x)"

theorem sem_xmap_id_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  assumes "xmap_id f"
  shows "\<parallel>v\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>xmap f v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E"
    and "\<parallel>p\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>xmap f p\<parallel>\<^sub>E = \<parallel>p\<parallel>\<^sub>E" and "kindOf (xmap f p) = kindOf p"
    and "\<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>xmap f d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
    and "\<parallel>r\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>xmap f r\<parallel>\<^sub>E = \<parallel>r\<parallel>\<^sub>E"
    and "\<parallel>l\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>xmap f l\<parallel>\<^sub>E = \<parallel>l\<parallel>\<^sub>E"
 using assms and xmap.nchotomy[THEN spec, of f] by (induct v and p and d and r and l) auto

theorem sem_smap_id_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  assumes "xmap_id f"
  shows "\<parallel>v\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>smap f v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E"
    and "\<parallel>p\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>smap f p\<parallel>\<^sub>E = \<parallel>p\<parallel>\<^sub>E" and "kindOf (smap f p) = kindOf p"
    and "\<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>smap f d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
    and "\<parallel>r\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>smap f r\<parallel>\<^sub>E = \<parallel>r\<parallel>\<^sub>E"
    and "\<parallel>l\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>smap f l\<parallel>\<^sub>E = \<parallel>l\<parallel>\<^sub>E"
  using assms and xmap.nchotomy[THEN spec, of f]
  by (induct v and p and d and r and l) (simp_all, safe, simp_all, auto)


theorem sem_xmap_id_disj [simp]:
  fixes d :: disj assumes "xmap_id f" and "\<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>xmap f d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
  using assms by (cases f, clarsimp) (induct d, auto)

theorem sem_smap_id_disj [simp]:
  fixes d :: disj assumes "xmap_id f" and "\<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>smap f d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
  using assms by (cases f, clarsimp) (induct d, auto)


theorem sem_xmap_id_conj [simp]:
  fixes c :: conj assumes "xmap_id f" and "\<parallel>c\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>xmap f c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by (cases f, clarsimp) (induct c, auto)

theorem sem_smap_id_conj [simp]:
  fixes c :: conj assumes "xmap_id f" and "\<parallel>c\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>smap f c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by (cases f, clarsimp) (induct c, auto)


(* Simplify Preserves Guarded *)

theorem smap_simplify_guarded_disj [simp]:
  fixes d :: disj assumes "guarded k d"
  shows "guarded k (smap simplify d)"
  using assms by (induct d arbitrary: k) (auto, drule_tac x="\<not>. (smap simplify lit) \<and>. k" in meta_spec,
    erule meta_mp, erule guarded_relax, auto)

theorem smap_simplify_guarded_conj [simp]:
  fixes c :: conj assumes "guarded k c"
  shows "guarded k (smap simplify c)"
  using assms by (induct c arbitrary: k) (auto, drule_tac x="smap simplify disj \<and>. k" in meta_spec,
    erule meta_mp, erule guarded_relax, auto, drule guarded_sem, auto)

theorem smap_simplify_guarded [simp]:
  fixes c :: cnf assumes "guarded k c"
  shows "guarded k (smap simplify c)"
  using assms by (cases c) simp+


(* Core Theorem of Semantically Equal Syntactic Transformations *)

theorem sem_safe_xmap_id [simp]:
  fixes c :: cnf assumes "xmap_id f" and "safe c"
  shows "\<parallel>xmap f c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by - (cases c, auto, rule sbeI, auto, rule_tac x=fa in exI, auto)

theorem sem_safe_smap_id [simp]:
  fixes c :: cnf assumes "xmap_id f" and "safe c"
  shows "\<parallel>smap f c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by (cases c) (auto, rule sbeI, auto, rule_tac x=fa in exI, auto)


(* Characterization of Failure *)

class fails = sem + xfold +
constrains x :: "'a :: {sem, xfold}"
fixes fails :: "'a \<Rightarrow> disj"
assumes fails_sem [simp]:
  "basic x \<Longrightarrow> \<parallel>fails x\<parallel>\<^sub>R\<^sub>+\<^sub>f = (\<parallel>x\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<bottom>)"
and fails_basic [simp]:
  "basic x \<Longrightarrow> strict (fails x)"
  "basic x \<Longrightarrow> straight (fails x)"
and fails_wform [simp]:
  "basic x \<Longrightarrow> wform x \<Longrightarrow> wform (fails x)"  (* strict not necessary for totality, but easier to prove *)
and fails_closed [simp]:
  "basic x \<Longrightarrow> closed x \<Longrightarrow> closed (fails x)"

abbreviation guard :: "'a :: fails \<Rightarrow> conj" where
  "guard x \<equiv> .\<not> (fails x)"

corollary fails_total [simp]:
  fixes x :: "'a :: {fails, closed}" assumes "basic x"
  shows "\<parallel>fails x\<parallel>\<^sub>R\<^sub>+\<^sub>f \<noteq> \<bottom>"
  using assms by simp

corollary fails_dom_closed [simp]:
  fixes x :: "'a :: {fails, closed}" assumes "closed x" and "basic x"
  shows "\<parallel>fails x\<parallel>\<^sub>E = (\<parallel>x\<parallel>\<^sub>E = \<bottom>)"
  using assms by (cases E) (auto simp add: sem_closed[symmetric, of _ _ f] simp del: sem_closed)

corollary fails_total_closed [simp]:
  fixes x :: "'a :: {fails, closed}" assumes "closed x" and "basic x"
  shows "\<parallel>fails x\<parallel>\<^sub>E \<noteq> \<bottom>"
  using assms by (cases E) (auto simp add: sem_closed[symmetric, of _ _ f] simp del: sem_closed)


instantiation valueExp and pathExp and docExp and pred and lit :: fails begin

fun fails_valueExp :: "valueExp \<Rightarrow> disj"
and fails_pathExp  :: "pathExp  \<Rightarrow> disj"
and fails_docExp   :: "docExp   \<Rightarrow> disj"
and fails_pred     :: "pred     \<Rightarrow> disj"
and fails_lit      :: "lit      \<Rightarrow> disj" where

  "fails (Val _) = False."
| "fails (FVar _) = False."
| "fails (PVar _) = False."

| "fails (Sum m) = fails m"
| "fails (Size m) = fails m"
| "fails (Count m v) = fails m \<or>. fails v \<or>. proper. v"
| "fails (Tally t m) = fails m"

| "fails (Read d p) = fails d \<or>. fails p"

| "fails (Scal n m) = fails n \<or>. proper. n \<or>. n is.not TInt \<or>. fails m"
| "fails (VPlus a b) = fails a \<or>. fails b"

| "fails Root. = False."
| "fails (Path l p k) = fails p \<or>. fails k"
| "fails (Kind l p) = fails p"

| "fails $. = False."
| "fails (Upd d p v) = fails d \<or>. fails p \<or>. .proper p \<or>. fails v \<or>. proper. v"
| "fails (Del d p) = fails d \<or>. fails p \<or>. .proper p"

| "fails (Equal a b) = fails a \<or>. proper. a \<or>. fails b \<or>. proper. b"
| "fails (Less a b) = fails a \<or>. proper. a \<or>. a is.not TInt \<or>.
                      fails b \<or>. proper. b \<or>. b is.not TInt"

| "fails (L b p) = fails p"

| "fails_valueExp _ = undefined"
| "fails_pathExp  _ = undefined"
| "fails_docExp   _ = undefined"
| "fails_pred     _ = undefined"

(* Non-Basic constructs and their potential characterizations

| "fails (v \<otimes>. m) = unique. v \<rightarrow>! v is. TInt \<rightarrow>! v \<ge>. 0. \<rightarrow>! (fails v \<or>. fails m)"
| "fails (a \<ominus>. b) = False." (* b not subset of a *)

| "fails (v .\<otimes> m) = unique. v \<rightarrow>! v is. TInt \<rightarrow>! v \<ge>. 0. \<rightarrow>! (fails v \<or>. fails m)"
| "fails (a .\<oplus> b) = False." (* kindOf a \<noteq> kindOf b  or  fails a  or  fails b *)
| "fails (a .\<ominus> b) = False." (* b not subset of a *)

| "fails (VCond b t e) = False." (* fails b  or  b \<and> fails t  or  \<not>b \<and> fails e *)
| "fails (PCond b t e) = False." (* kindOf t \<noteq> kindOf e  or  fails b  or  b \<and> fails t  or  \<not>b \<and> fails e *)
| "fails (DCond b t e) = False." (* fails b  or  b \<and> fails t  or  \<not>b \<and> fails e *)
| "fails (RCond b t e) = False." (* fails b  or  b \<and> fails t  or  \<not>b \<and> fails e *)

*)

theorem fails_basic:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "basic v \<Longrightarrow> basic (fails v)"
    and "basic p \<Longrightarrow> basic (fails p)"
    and "basic d \<Longrightarrow> basic (fails d)"
    and "basic r \<Longrightarrow> basic (fails r)"
    and "basic l \<Longrightarrow> basic (fails l)"
  by (induct v and p and d and r and l) auto

theorem fails_wform:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "wform v \<Longrightarrow> basic v \<Longrightarrow> wform (fails v)"
    and "wform p \<Longrightarrow> basic p \<Longrightarrow> wform (fails p)"
    and "wform d \<Longrightarrow> basic d \<Longrightarrow> wform (fails d)"
    and "wform r \<Longrightarrow> basic r \<Longrightarrow> wform (fails r)"
    and "wform l \<Longrightarrow> basic l \<Longrightarrow> wform (fails l)"
  by (induct v and p and d and r and l) auto

theorem fails_closed:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "closed v \<Longrightarrow> basic v \<Longrightarrow> closed (fails v)"
    and "closed p \<Longrightarrow> basic p \<Longrightarrow> closed (fails p)"
    and "closed d \<Longrightarrow> basic d \<Longrightarrow> closed (fails d)"
    and "closed r \<Longrightarrow> basic r \<Longrightarrow> closed (fails r)"
    and "closed l \<Longrightarrow> basic l \<Longrightarrow> closed (fails l)"
  by (induct v and p and d and r and l) auto

theorem fails_dom_exp_false:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "\<parallel>v\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> basic v \<Longrightarrow> \<parallel>fails v\<parallel>\<^sub>E = False"
    and "\<parallel>p\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> basic p \<Longrightarrow> \<parallel>fails p\<parallel>\<^sub>E = False"
    and "\<parallel>d\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> basic d \<Longrightarrow> \<parallel>fails d\<parallel>\<^sub>E = False"
    and "\<parallel>r\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> basic r \<Longrightarrow> \<parallel>fails r\<parallel>\<^sub>E = False"
    and "\<parallel>l\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> basic l \<Longrightarrow> \<parallel>fails l\<parallel>\<^sub>E = False"
  by (induct v and p and d and r and l) auto

theorem fails_dom_exp_true:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "\<parallel>v\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<bottom> \<Longrightarrow> basic v \<Longrightarrow> \<parallel>fails v\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
    and "\<parallel>p\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<bottom> \<Longrightarrow> basic p \<Longrightarrow> \<parallel>fails p\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
    and "\<parallel>d\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<bottom> \<Longrightarrow> basic d \<Longrightarrow> \<parallel>fails d\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
    and "\<parallel>r\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<bottom> \<Longrightarrow> basic r \<Longrightarrow> \<parallel>fails r\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
    and "\<parallel>l\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<bottom> \<Longrightarrow> basic l \<Longrightarrow> \<parallel>fails l\<parallel>\<^sub>R\<^sub>+\<^sub>f = True"
  by (induct v and p and d and r and l) (simp_all, safe, auto, drule_tac x=ea in spec, simp)

instance by (intro_classes, auto simp: fails_basic fails_wform fails_closed)
            (case_tac "\<parallel>x\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<bottom>", auto simp: fails_dom_exp_false fails_dom_exp_true)+ end


(* Fails is Guarded *)

theorem proper_guarded [intro!]:
  assumes "basic p" and "\<And>R f. \<parallel>k\<parallel>\<^sub>R\<^sub>+\<^sub>f = True \<Longrightarrow> \<parallel>p\<parallel>\<^sub>R\<^sub>+\<^sub>f \<noteq> \<bottom>"
  shows "guarded k (.proper p)"
  using assms
apply (induct p arbitrary: k rule: disj_proper.induct)
apply simp_all
apply safe
apply simp_all
apply auto
apply (drule_tac x=R in meta_spec)
apply (drule_tac x=f in meta_spec)
apply auto
apply (drule_tac x=R in meta_spec)
apply (drule_tac x=f in meta_spec)
apply auto
apply (drule_tac x=ka in meta_spec)
apply (drule meta_mp)
apply auto
apply (drule_tac x=R in meta_spec)
apply (drule_tac x=f in meta_spec)
apply auto
done


theorem fails_guarded [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "basic v \<Longrightarrow> guarded k (fails v)"
    and "basic p \<Longrightarrow> guarded k (fails p)"
    and "basic d \<Longrightarrow> guarded k (fails d)"
    and "basic r \<Longrightarrow> guarded k (fails r)"
    and "basic l \<Longrightarrow> guarded k (fails l)"
  by (induct v and p and d and r and l) auto+


(* Domain Guards *)

class sguard = bool_sem + xfold +
constrains x :: "'a :: {bool_sem, xfold}"
fixes sguard :: "'a \<Rightarrow> conj"
assumes sguard_sem [simp]:
  "basic x \<Longrightarrow> \<parallel>sguard x\<parallel>\<^sub>R\<^sub>+\<^sub>f = (\<parallel>x\<parallel>\<^sub>R\<^sub>+\<^sub>f \<noteq> \<bottom>) \<and>\<^sub>? \<parallel>x\<parallel>\<^sub>R\<^sub>+\<^sub>f"
and sguard_wform [simp]:
  "basic x \<Longrightarrow> wform x \<Longrightarrow> wform (sguard x)"
and sguard_basic [simp]:
  "basic x \<Longrightarrow> strict (sguard x)"
  "basic x \<Longrightarrow> straight (sguard x)"

corollary sguard_sem_dom:
  assumes "basic x" and "\<parallel>x\<parallel>\<^sub>R\<^sub>+\<^sub>f \<noteq> \<bottom>"
  shows "\<parallel>sguard x\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<parallel>x\<parallel>\<^sub>R\<^sub>+\<^sub>f"
  using assms by simp

corollary sguard_sem_bot:
  assumes "basic x" and "\<parallel>x\<parallel>\<^sub>R\<^sub>+\<^sub>f = \<bottom>"
  shows "\<parallel>sguard x\<parallel>\<^sub>R\<^sub>+\<^sub>f = False"
  using assms by simp


instantiation lit :: sguard begin
  definition sguard_lit :: "lit \<Rightarrow> conj" where
    "sguard_lit l \<equiv> guard l \<and>. l"
instance by intro_classes (unfold sguard_lit_def, auto) end


instantiation disj :: sguard begin
  fun sguard_disj :: "disj \<Rightarrow> conj" where
    "sguard (D l d) = (fails l \<or>. \<not>. l) \<Longrightarrow>. sguard d"
  | "sguard DFlse = False."

instance by intro_classes
  ((erule rev_mp)+, induct_tac x, auto, (subst sem_disj_imp_conj)?, ((rule sbeI, auto)+)?)+ end

instantiation conj :: sguard begin
  fun sguard_conj :: "conj \<Rightarrow> conj" where
    "sguard (C d c) = sguard d \<and>. sguard c"
  | "sguard CTrue = True."
instance by intro_classes ((erule rev_mp)+, induct_tac x, auto, ((rule sbeI, auto)+)?)+ end


(* sguard guards *)

theorem sguard_lit_guarded [simp]:
  fixes l :: lit assumes "basic l"
  shows "guarded CTrue (sguard l)"
  using assms unfolding sguard_lit_def by auto

theorem sguard_disj_guarded [simp]:
  fixes d :: disj assumes "basic d"
  shows "guarded CTrue (sguard d)"
  using assms by (induct d) auto

theorem sguard_conj_guarded [simp]:
  fixes c :: conj assumes "basic c"
  shows "guarded CTrue (sguard c)"
  using assms by (induct c) (auto, erule guarded_relax, auto)


(* Securing Unsafe Formulas *)

fun safeguard :: "cnf \<Rightarrow> cnf" where
  "safeguard (\<forall>. c) = \<forall>. (sguard c)"

theorem safeguard_safe [simp]:
  assumes "basic x"
  shows "safe (safeguard x)"
  using assms by (cases x) auto

theorem safeguard_sem_dom [simp]:
  assumes "basic x" and "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>safeguard x\<parallel>\<^sub>E = \<parallel>x\<parallel>\<^sub>E"
  using assms by (cases E, auto) (cases x, auto)+

theorem safeguard_sem_bot [simp]:
  assumes "basic x" and "\<parallel>x\<parallel>\<^sub>E = \<bottom>"
  shows "\<parallel>safeguard x\<parallel>\<^sub>E = False"
  using assms by (cases E, auto) (cases x, auto)+


(* Safeguard Properties *)

theorem safeguard_wform [simp]:
  assumes "wform c" and "basic c"
  shows "wform (safeguard c)"
  using assms by (cases c) simp+

theorem safeguard_basic [simp]:
  assumes "basic c"
  shows "strict (safeguard c)" and "straight (safeguard c)"
  using assms by (cases c, simp)+

theorem safeguard_guarded [simp]:
  assumes "basic c"
  shows "guarded k (safeguard c)"
  using assms by (cases c) simp

end

