theory Simplify imports Syntax
begin


(* Mapping for the Simplifying Syntax *)

class smap = xmap +
fixes smap :: "xmap \<Rightarrow> 'a \<Rightarrow> 'a"

instantiation valueExp and pathExp and docExp and pred and lit :: smap begin 

fun smap_valueExp :: "xmap \<Rightarrow> valueExp \<Rightarrow> valueExp"
and smap_pathExp  :: "xmap \<Rightarrow> pathExp  \<Rightarrow> pathExp"
and smap_docExp   :: "xmap \<Rightarrow> docExp   \<Rightarrow> docExp"
and smap_pred     :: "xmap \<Rightarrow> pred     \<Rightarrow> pred"
and smap_lit      :: "xmap \<Rightarrow> lit      \<Rightarrow> lit" where

  "smap f (Val v)       = xm f (Val v)"
| "smap f (FVar n)      = xm f (FVar n)"
| "smap f (PVar n)      = xm f (PVar n)"
| "smap f (Sum m)       = xm f (sSum     (smap f m))"
| "smap f (Size m)      = xm f (sSize    (smap f m))"
| "smap f (Count m v)   = xm f (sCount   (smap f m) (smap f v))"
| "smap f (Tally t m)   = xm f (sTally t (smap f m))"
| "smap f (Read d p)    = xm f (sRead    (smap f d) (smap f p))"
| "smap f (Scal n m)    = xm f (sScal    (smap f n) (smap f m))"
| "smap f (VRep n m)    = xm f (sVRep    (smap f n) (smap f m))"
| "smap f (VPlus a b)   = xm f (sVPlus   (smap f a) (smap f b))"
| "smap f (VMinus a b)  = xm f (sVMinus  (smap f a) (smap f b))"
| "smap f (VCond l a b) = xm f (sVCond   (smap f l) (smap f a) (smap f b))"

| "smap f PRoot         = xm f PRoot"
| "smap f (Path l p v)  = xm f (sPath    (smap f p) l (smap f v))"
| "smap f (Kind l p)    = xm f (sKind    (smap f p) l)"
| "smap f (PRep n m)    = xm f (sPRep    (smap f n) (smap f m))"
| "smap f (PPlus a b)   = xm f (sPPlus   (smap f a) (smap f b))"
| "smap f (PMinus a b)  = xm f (sPMinus  (smap f a) (smap f b))"
| "smap f (PCond l a b) = xm f (sPCond   (smap f l) (smap f a) (smap f b))"

| "smap f Init          = xm f Init"
| "smap f (Upd d p v)   = xm f (sUpd     (smap f d) (smap f p) (smap f v))"
| "smap f (Del d p)     = xm f (sDel     (smap f d) (smap f p))"
| "smap f (DCond l a b) = xm f (sDCond   (smap f l) (smap f a) (smap f b))"

| "smap f (Equal a b)   = xm f (sEqual   (smap f a) (smap f b))"
| "smap f (Less a b)    = xm f (sLess    (smap f a) (smap f b))"
| "smap f (RCond l a b) = xm f (sRCond   (smap f l) (smap f a) (smap f b))"

| "smap f (L b r)       = xm f (sL b     (smap f r))"

instance by intro_classes end

instantiation disj :: smap begin
  fun smap_disj :: "xmap \<Rightarrow> disj \<Rightarrow> disj" where
    "smap f (D l d) = xm f (sD (smap f l) (smap f d))"
  | "smap f DFlse   = xm f DFlse"
instance by intro_classes end

instantiation conj :: smap begin
  fun smap_conj :: "xmap \<Rightarrow> conj \<Rightarrow> conj" where
    "smap f (C d c) = xm f (sC (smap f d) (smap f c))"
  | "smap f CTrue   = xm f CTrue"
instance by intro_classes end

instantiation cnf :: smap begin
  fun smap_cnf :: "xmap \<Rightarrow> cnf \<Rightarrow> cnf" where
    "smap f (\<forall>. c) = \<forall>. (smap f c)"
instance by intro_classes end


(* Using Identities to Simplify *)

abbreviation simplify :: xmap where
  "simplify \<equiv> XMap (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a)"


(* Properties *)

theorem smap_simplify_uxe [simp]:
  shows "uve v \<Longrightarrow> uve (smap simplify v)"
    and "upe p \<Longrightarrow> upe (smap simplify p)"
  by (induct v and p rule: uve_upe.induct) auto 

theorem smap_simplify_wform_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "wform v \<Longrightarrow> wform (smap simplify v)"
    and "wform p \<Longrightarrow> wform (smap simplify p)"
    and "wform d \<Longrightarrow> wform (smap simplify d)"
    and "wform r \<Longrightarrow> wform (smap simplify r)"
    and "wform l \<Longrightarrow> wform (smap simplify l)"
  using assms by (induct v and p and d and r and l) auto

theorem smap_simplify_wform_disj [simp]:
  fixes d :: disj assumes "wform d"
  shows "wform (smap simplify d)"
  using assms by (induct d) auto

theorem smap_simplify_wform_conj [simp]:
  fixes c :: conj assumes "wform c"
  shows "wform (smap simplify c)"
  using assms by (induct c) auto

theorem smap_simplify_wform [simp]:
  fixes c :: cnf assumes "wform c"
  shows "wform (smap simplify c)"
  using assms by (cases c) simp+


theorem smap_simplify_strict_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "strict v \<Longrightarrow> strict (smap simplify v)"
    and "strict p \<Longrightarrow> strict (smap simplify p)"
    and "strict d \<Longrightarrow> strict (smap simplify d)"
    and "strict r \<Longrightarrow> strict (smap simplify r)"
    and "strict l \<Longrightarrow> strict (smap simplify l)"
  using assms by (induct v and p and d and r and l) auto

theorem smap_simplify_strict_disj [simp]:
  fixes d :: disj assumes "strict d"
  shows "strict (smap simplify d)"
  using assms by (induct d) auto

theorem smap_simplify_strict_conj [simp]:
  fixes c :: conj assumes "strict c"
  shows "strict (smap simplify c)"
  using assms by (induct c) auto

theorem smap_simplify_strict [simp]:
  fixes c :: cnf assumes "strict c"
  shows "strict (smap simplify c)"
  using assms by (cases c) simp+


theorem smap_simplify_straight_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "straight v \<Longrightarrow> straight (smap simplify v)"
    and "straight p \<Longrightarrow> straight (smap simplify p)"
    and "straight d \<Longrightarrow> straight (smap simplify d)"
    and "straight r \<Longrightarrow> straight (smap simplify r)"
    and "straight l \<Longrightarrow> straight (smap simplify l)"
  using assms by (induct v and p and d and r and l) auto

theorem smap_simplify_straight_disj [simp]:
  fixes d :: disj assumes "straight d"
  shows "straight (smap simplify d)"
  using assms by (induct d) auto

theorem smap_simplify_straight_conj [simp]:
  fixes c :: conj assumes "straight c"
  shows "straight (smap simplify c)"
  using assms by (induct c) auto

theorem smap_simplify_straight [simp]:
  fixes c :: cnf assumes "straight c"
  shows "straight (smap simplify c)"
  using assms by (cases c) simp+


theorem smap_simplify_dufree_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "dufree v \<Longrightarrow> dufree (smap simplify v)"
    and "dufree p \<Longrightarrow> dufree (smap simplify p)"
    and "dufree d \<Longrightarrow> dufree (smap simplify d)"
    and "dufree r \<Longrightarrow> dufree (smap simplify r)"
    and "dufree l \<Longrightarrow> dufree (smap simplify l)"
  using assms by (induct v and p and d and r and l) auto

theorem smap_simplify_dufree_disj [simp]:
  fixes d :: disj assumes "dufree d"
  shows "dufree (smap simplify d)"
  using assms by (induct d) auto

theorem smap_simplify_dufree_conj [simp]:
  fixes c :: conj assumes "dufree c"
  shows "dufree (smap simplify c)"
  using assms by (induct c) auto

theorem smap_simplify_dufree [simp]:
  fixes c :: cnf assumes "dufree c"
  shows "dufree (smap simplify c)"
  using assms by (cases c) simp+

end

