header {* The Combined WIP Transformer *}

theory Story imports Procedures Elimination Straighten Preservation begin


(* The complete Story *)

locale nice_formula =
fixes Q :: cnf
assumes wformQ [simp]:
  "wform Q"
and strictQ [simp]:
  "strict Q"
and straightQ [simp]:
  "straight Q"
and dufreeQ [simp]:
  "dufree Q"
and guardedQ [simp]:
  "guarded True. Q"

locale nice_program = nice_formula +
fixes s :: stm
assumes wformS [simp]:
  "wform_stm s"
and strictS [simp]:
  "strict s"
and straightS [simp]:
  "straight s"
and closedS [simp]:
  "closed s"


(* Big Picture *)

definition wip :: "stm \<Rightarrow> cnf \<Rightarrow> cnf" where
  "wip s \<equiv> straighten \<circ> strictify \<circ> smap simplify \<circ> elimdu \<circ> safeguard \<circ> wp s"


theorem (in nice_program) wip_sound:
  shows "\<Turnstile> {wip s Q} s {Q}"
  using wp_sound[of s Q] unfolding sem_Hoare_def and wip_def
  by auto (drule_tac x=R in spec, erule mp, rule sbeI, auto)

theorem (in nice_program) wip_nice:
  shows "nice_formula (wip s Q)"
  unfolding nice_formula_def and wip_def by auto


abbreviation core where
  "core x \<equiv> wform x \<and> strict x \<and> straight x \<and> dufree x \<and> guarded True. x"

corollary wip_props:
  assumes "wform_stm s" and "basic s" and "closed s" and "core Q"
  shows "\<Turnstile> {wip s Q} s {Q}" and "core (wip s Q)"
  using assms by auto (rule nice_program.wip_sound, unfold_locales, auto, unfold wip_def, auto)

theorem normalize_core:
  fixes c :: cnf assumes "wform c"
  shows "core (straighten (strictify (elimdu (safeguard (straighten (strictify c))))))"
  using assms by simp

end

