header {* The Straighten Transformer *}

theory Straighten imports Safe begin

declare plus_twice_le [rule del, intro!]


locale props_straighten = props_base + prop_strict + prop_dufree
locale pprops_straighten = props_straighten + prop_pe
locale vprops_straighten = props_straighten + prop_ve


(*-----------------*)
(*- Normalization -*)
(*-----------------*)

fun nSum :: "valueExp \<Rightarrow> valueExp" where
  "nSum (Scal n m)   = n \<odot>. (nSum m)"
| "nSum (VRep n m)   = n \<odot>. (nSum m)"
| "nSum (VPlus a b)  = nSum a +. nSum b"
| "nSum (VMinus a b) = nSum a -. nSum b"

| "nSum v = sum. v"

interpretation nSum_sem:
  semantics "nSum m" "Sum m"
  by unfold_locales (induct m rule: nSum.induct, auto)

interpretation nSum_props:
  props_eq "nSum m" "Sum m"
  by unfold_locales (induct m rule: nSum.induct, auto)+

interpretation nSum_prop_ve:
  prop_ve "nSum m" "Sum m"
  by unfold_locales (induct m rule: nSum.induct, auto)


fun nTally :: "type \<Rightarrow> valueExp \<Rightarrow> valueExp" where
  "nTally t (Scal n m)   = (if t = TInt then nTally t m else 0.)"
| "nTally t (VRep n m)   = n \<odot>. (nTally t m)"
| "nTally t (VPlus a b)  = nTally t a +. nTally t b"
| "nTally t (VMinus a b) = nTally t a -. nTally t b"

| "nTally t v = tally. t v"

lemma plus_minus_one_times [simp]:
  fixes a :: int
  shows "a + -1 * b = a - b"
  by arith

interpretation nTally_sem:
  semantics "nTally t m" "Tally t m"
  by unfold_locales (induct m rule: nTally.induct, simp_all, safe, simp_all, auto, 
    unfold int_mult, auto, subst zdiff_int, auto)

interpretation nTally_props:
  props_eq "nTally t m" "Tally t m"
  by (unfold_locales, auto elim!: order_trans[rotated])
     (induct t m rule: nTally.induct, auto intro!: le_SucI)+

interpretation nTally_prop_ve:
  prop_ve "nTally t m" "Tally t m"
  by unfold_locales (induct m rule: nTally.induct, auto)


fun nSize :: "valueExp \<Rightarrow> valueExp" where
  "nSize (Scal n m)   = nTally TInt m"
| "nSize (VRep n m)   = n \<odot>. (nSize m)"
| "nSize (VPlus a b)  = nSize a +. nSize b"
| "nSize (VMinus a b) = nSize a -. nSize b"

| "nSize v = size. v"

interpretation nSize_sem:
  semantics "nSize m" "Size m"
  using nTally_sem.domain
  by unfold_locales (induct m rule: nSize.induct, simp_all, safe, simp_all, auto,
    unfold int_mult, auto, subst zdiff_int, auto)

interpretation nSize_props:
  props_eq "nSize m" "Size m"
  by (unfold_locales, auto elim!: order_trans[rotated])
     (induct m rule: nSize.induct, auto intro!: le_SucI)+

interpretation nSize_prop_ve:
  prop_ve "nSize m" "Size m"
  by unfold_locales (induct m rule: nSize.induct, auto)

corollary nAggregate_sem:
  assumes "\<parallel>m\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>nSum m\<parallel>\<^sub>E = \<parallel>Sum m\<parallel>\<^sub>E" and "\<parallel>nTally t m\<parallel>\<^sub>E = \<parallel>Tally t m\<parallel>\<^sub>E" and "\<parallel>nSize m\<parallel>\<^sub>E = \<parallel>Size m\<parallel>\<^sub>E"
  using assms by -
    (subst nSum_sem.domain, auto, subst nTally_sem.domain, auto, subst nSize_sem.domain, auto)


fun nCount :: "valueExp option \<Rightarrow> valueExp \<Rightarrow> valueExp \<Rightarrow> valueExp" where
  "nCount None (Scal n m) v = nCount n\<^sub>! m v"
| "nCount i\<^sub>! (Scal n m) v   = nCount (i \<odot>. n)\<^sub>! m v"
| "nCount i (VRep n m) v    = n \<odot>. (nCount i m v)"
| "nCount i (VPlus a b) v   = nCount i a v +. nCount i b v"
| "nCount i (VMinus a b) v  = nCount i a v -. nCount i b v"

| "nCount None m v = count. m v"
| "nCount i\<^sub>! m v = count. (i \<odot>. m) v"

theorem nCount_sem_general:
  assumes "\<parallel>Count (Scal (the n) m) p\<parallel>\<^sub>E \<noteq> \<bottom>" and "n \<noteq> None"
  shows "\<parallel>nCount n m p\<parallel>\<^sub>E = \<parallel>Count (Scal (the n) m) p\<parallel>\<^sub>E"
  using assms 
proof (induct n m p rule: nCount.induct)
  case 3 thus ?case using int_mult by auto next
  case ("7_9" n b) thus ?case by auto (rule sbeI[of b E], auto) next
qed auto

corollary nCount_sem_general_fast:
  assumes "\<parallel>p\<parallel>\<^sub>E = {a}\<^sub>*\<^sub>v" and "\<parallel>the n\<parallel>\<^sub>E = {b}\<^sub>i\<^sub>v" and "\<parallel>m\<parallel>\<^sub>E = c\<^sub>v" and "n \<noteq> None"
  shows "\<parallel>nCount n m p\<parallel>\<^sub>E = \<parallel>Count (Scal (the n) m) p\<parallel>\<^sub>E"
  using assms by - (rule nCount_sem_general, simp+)

theorem nCount_sem:
  assumes "\<parallel>Count m p\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>nCount None m p\<parallel>\<^sub>E = \<parallel>Count m p\<parallel>\<^sub>E"
  using assms and nCount_sem_general_fast
proof (induct "None :: valueExp option" m p rule: nCount.induct)
  case 3 thus ?case using int_mult by auto next
qed auto

interpretation nCount_sem:
  semantics "nCount None m p" "Count m p"
  by unfold_locales (erule nCount_sem)

interpretation nCount_general_props:
  props_eq "nCount n\<^sub>! m p" "Count (Scal n m) p"
  by unfold_locales (induct "n\<^sub>!" m p arbitrary: n rule: nCount.induct, auto)+

interpretation nCount_props:
  props_eq "nCount None m p" "Count m p"
  by unfold_locales (induct "None :: valueExp option" m p rule: nCount.induct, auto)+

interpretation nCount_general_prop_ve:
  prop_ve "nCount n\<^sub>! m p" "Count (Scal n m) p"
  by unfold_locales (induct "n\<^sub>!" m p arbitrary: n rule: nCount.induct, auto)+

interpretation nCount_prop_ve:
  prop_ve "nCount None m p" "Count m p"
  by unfold_locales (induct "None :: valueExp option" m p rule: nCount.induct, auto)+


fun nRead :: "docExp \<Rightarrow> pathExp \<Rightarrow> valueExp" where
  "nRead d (PRep n p) = n \<otimes>. (nRead d p)"
| "nRead d (PPlus a b) = (nRead d a) \<oplus>. (nRead d b)"
| "nRead d (PMinus a b) = (nRead d a) \<ominus>. (nRead d b)"

| "nRead d p = d(p)."

interpretation nRead_sem:
  semantics "nRead d p" "Read d p"
  by unfold_locales (induct d p rule: nRead.induct, auto, cut_tac x=v and E=E in sbeI, auto)

interpretation nRead_props:
  vprops_eq "nRead d p" "Read d p"
  by unfold_locales (induct d p rule: nRead.induct, auto)+


fun nvPath :: "label \<Rightarrow> pathExp \<Rightarrow> valueExp \<Rightarrow> pathExp" where
  "nvPath l p (Scal n m)   = p/l[{}.]."
| "nvPath l p (VRep n m)   = n .\<otimes> (nvPath l p m)"
| "nvPath l p (VPlus a b)  = (nvPath l p a) .\<oplus> (nvPath l p b)"
| "nvPath l p (VMinus a b) = (nvPath l p a) .\<ominus> (nvPath l p b)"

| "nvPath l p k = p/l[k]."

interpretation nvPath_props:
  pprops_straighten "nvPath l p k" "Path l p k"
  by unfold_locales (induct l p k rule: nvPath.induct, auto)+

theorem nvPath_sem:
  assumes "\<parallel>p\<parallel>\<^sub>E \<noteq> \<bottom>" and "\<parallel>k\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>nvPath l p k\<parallel>\<^sub>E = \<parallel>Path l p k\<parallel>\<^sub>E"
proof -
  obtain q where "\<parallel>p\<parallel>\<^sub>E = q\<^sub>p" using assms and speC[of p E] by force
  thus ?thesis using assms by (induct l p k rule: nvPath.induct) auto
qed

interpretation nvPath_sem:
  semantics "nvPath l p k" "Path l p k"
  by unfold_locales (rule nvPath_sem, auto)


fun npPath :: "label \<Rightarrow> pathExp \<Rightarrow> valueExp \<Rightarrow> pathExp" where
  "npPath l (PRep n p) k   = n .\<otimes> (npPath l p k)"
| "npPath l (PPlus a b) k  = (npPath l a k) .\<oplus> (npPath l b k)"
| "npPath l (PMinus a b) k = (npPath l a k) .\<ominus> (npPath l b k)"

| "npPath l p k = nvPath l p k"  (* now serving nvPath (!) *)

interpretation npPath_props:
  pprops_straighten "npPath l p k" "Path l p k"
  by unfold_locales (induct l p k rule: npPath.induct, auto)+

theorem npPath_sem:
  assumes "\<parallel>p\<parallel>\<^sub>E \<noteq> \<bottom>" and "\<parallel>k\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>npPath l p k\<parallel>\<^sub>E = \<parallel>Path l p k\<parallel>\<^sub>E"
proof -
  obtain r where "\<parallel>k\<parallel>\<^sub>E = r\<^sub>v" using assms and sveC[of k E] by force
  thus ?thesis using assms and nvPath_sem by (induct l p k rule: npPath.induct) auto
qed

interpretation npPath_sem:
  semantics "npPath l p k" "Path l p k"
  by unfold_locales (rule npPath_sem, auto)


fun nKind :: "label \<Rightarrow> pathExp \<Rightarrow> pathExp" where
  "nKind l (PRep n p)   = n .\<otimes> (nKind l p)"
| "nKind l (PPlus a b)  = (nKind l a) .\<oplus> (nKind l b)"
| "nKind l (PMinus a b) = (nKind l a) .\<ominus> (nKind l b)"

| "nKind l p = p/l."

interpretation nKind_props:
  pprops_eq "nKind l p" "Kind l p"
  by unfold_locales (induct l p rule: nKind.induct, auto)+

interpretation nKind_sem:
  semantics "nKind l p" "Kind l p"
  by unfold_locales (induct l p rule: nKind.induct, auto)


corollary nStep_kind:
  shows "kindOf (nvPath l p k) = kindOf p/l[k]."
    and "kindOf (npPath l p k) = kindOf p/l[k]."
    and "kindOf (nKind  l p)   = kindOf p/l."
  by simp+

corollary nPath_sem:
  assumes "\<parallel>p\<parallel>\<^sub>E \<noteq> \<bottom>" and "\<parallel>k\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>nvPath l p k\<parallel>\<^sub>E = \<parallel>Path l p k\<parallel>\<^sub>E" and "\<parallel>npPath l p k\<parallel>\<^sub>E = \<parallel>Path l p k\<parallel>\<^sub>E"
  using assms by - (erule(1) nvPath_sem, erule(1) npPath_sem)


corollary nHomomorphism_uxe:
  shows "uve (nSum m)"  and "uve (nCount i m v)"
    and "uve (nSize m)" and "uve (nTally t m)"
    and "upe p \<Longrightarrow> uve (nRead d p)"
    and "upe (Path l p k) \<Longrightarrow> upe (nvPath l p k)"
    and "upe (Path l p k) \<Longrightarrow> upe (npPath l p k)"
  by simp_all (cases i, simp+)


(* Generalization to Values & Paths *)

fun normVE :: "valueExp \<Rightarrow> valueExp" where
  "normVE (Sum m)     = nSum m"
| "normVE (Tally t m) = nTally t m"
| "normVE (Size m)    = nSize m"
| "normVE (Count m v) = nCount None m v"
| "normVE (Read d p)  = nRead d p"
| "normVE v = v"

fun normPE :: "pathExp \<Rightarrow> pathExp" where
  "normPE (Path l p k) = npPath l p k"
| "normPE (Kind l p)   = nKind l p"
| "normPE p = p"

theorem normVE_ide [simp]:
  shows "id\<^sub>E normVE"
  unfolding id_dom_def using nAggregate_sem nCount_sem nRead_sem.domain
  by auto (case_tac x, auto)

theorem normPE_ide [simp]:
  shows "id\<^sub>E normPE"
  unfolding id_dom_def using npPath_sem nKind_sem.domain
  by auto (case_tac x, auto)

interpretation normVE_props:
  vprops_eq "normVE v" v
  by unfold_locales (cases v, auto)+

interpretation normPE_props:
  pprops_straighten "normPE p" p
  by unfold_locales (cases p, auto)+

corollary normXE_ide:
  shows "id\<^sub>E normVE" and "id\<^sub>E normPE"
    and "kindOf (normPE x) = kindOf x"
  by simp+


(* Global Elimination *)

abbreviation straighten where
  "straighten \<equiv> xmap (XMap normVE normPE (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a))"

theorem straighten_sem [simp]:
  fixes c :: cnf assumes "safe c"
  shows "\<parallel>straighten c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by - (erule sem_safe_xmap_id[rotated], fold id_def, simp)


(* Preservation Properties *)

theorem straighten_kind [simp]:
  fixes p :: pathExp
  shows "kindOf (straighten p) = kindOf p"
  by (induct p) auto

theorem straighten_uxe [simp]:
  shows "uve v \<Longrightarrow> uve (straighten v)"
    and "upe p \<Longrightarrow> upe (straighten p)"
  by (induct _ v and _ p rule: uve_upe.induct) auto

theorem straighten_wform_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "wform v \<Longrightarrow> wform (straighten v)"
    and "wform p \<Longrightarrow> wform (straighten p)"
    and "wform d \<Longrightarrow> wform (straighten d)"
    and "wform r \<Longrightarrow> wform (straighten r)"
    and "wform l \<Longrightarrow> wform (straighten l)"
  using straighten_uxe by (induct v and p and d and r and l) auto

theorem straighten_closed_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "closed v \<Longrightarrow> closed (straighten v)"
    and "closed p \<Longrightarrow> closed (straighten p)"
    and "closed d \<Longrightarrow> closed (straighten d)"
    and "closed r \<Longrightarrow> closed (straighten r)"
    and "closed l \<Longrightarrow> closed (straighten l)"
  by (induct v and p and d and r and l) auto

theorem straighten_strict_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "strict v \<Longrightarrow> strict (straighten v)"
    and "strict p \<Longrightarrow> strict (straighten p)"
    and "strict d \<Longrightarrow> strict (straighten d)"
    and "strict r \<Longrightarrow> strict (straighten r)"
    and "strict l \<Longrightarrow> strict (straighten l)"
  by (induct v and p and d and r and l) auto

theorem straighten_dufree_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "dufree v \<Longrightarrow> dufree (straighten v)"
    and "dufree p \<Longrightarrow> dufree (straighten p)"
    and "dufree d \<Longrightarrow> dufree (straighten d)"
    and "dufree r \<Longrightarrow> dufree (straighten r)"
    and "dufree l \<Longrightarrow> dufree (straighten l)"
  by (induct v and p and d and r and l) auto

interpretation straighten_disj_props:
  props_straighten "straighten d" d for d :: disj
  by unfold_locales (induct d, auto)+

interpretation straighten_conj_props:
  props_straighten "straighten c" c for c :: conj
  by unfold_locales (induct c, auto)+

interpretation straighten_props:
  props_straighten "straighten c" c for c :: cnf
  by unfold_locales (cases c, simp)+


(* Preserves Guarded *)

theorem straighten_guarded_disj [simp]:
  fixes d :: disj assumes "guarded k d"
  shows "guarded k (straighten d)"
  using assms by (induct d arbitrary: k) (auto, drule_tac x="\<not>. (straighten lit) \<and>. k" in meta_spec,
    erule meta_mp, erule guarded_relax, auto)

theorem straighten_guarded_conj [simp]:
  fixes c :: conj assumes "guarded k c"
  shows "guarded k (straighten c)"
  using assms by (induct c arbitrary: k) (auto, drule_tac x="straighten disj \<and>. k" in meta_spec,
    erule meta_mp, erule guarded_relax, auto, drule guarded_sem, auto)

theorem straighten_guarded [simp]:
  fixes c :: cnf assumes "guarded k c"
  shows "guarded k (straighten c)"
  using assms by (cases c) simp+


(* Main Property - Preserves & Establishes Straight *)

theorem nSum_straight [simp]:
  assumes "otop m"
  shows "straight (nSum m)"
  using assms by (induct m rule: nSum.induct) auto

theorem nTally_straight [simp]:
  assumes "otop m"
  shows "straight (nTally t m)"
  using assms by (induct t m rule: nTally.induct) auto

theorem nSize_straight [simp]:
  assumes "otop m"
  shows "straight (nSize m)"
  using assms by (induct m rule: nSize.induct) auto

corollary nAggregate_straight:
  assumes "otop m"
  shows "straight (nSum m)"
    and "straight (nSize m)"
    and "straight (nTally t m)"
  using assms by simp+

theorem nCount_straight_general [simp]:
  assumes "otop m" and "straight v" and "straight n"
  shows "straight (nCount n\<^sub>! m v)"
  using assms by (induct "n\<^sub>!" m v arbitrary: n rule: nCount.induct) auto

theorem nCount_straight [simp]:
  assumes "otop m" and "straight v"
  shows "straight (nCount None m v)"
  using assms by (induct "None :: valueExp option" m v rule: nCount.induct) auto

theorem nRead_otop [simp]:
  assumes "straight d" and "otop p"
  shows "otop (nRead d p)"
  using assms by (induct d p rule: nRead.induct) auto

theorem nvPath_otop [simp]:
  assumes "straight p" and "otop k"
  shows "otop (nvPath l p k)"
  using assms by (induct l p k rule: nvPath.induct) auto

theorem npPath_otop [simp]:
  assumes "otop p" and "otop k"
  shows "otop (npPath l p k)"
  using assms by (induct l p k rule: npPath.induct) auto

theorem nKind_otop [simp]:
  assumes "otop p"
  shows "otop (nKind l p)"
  using assms by (induct l p rule: nKind.induct) auto


theorem uxe_topOps [simp]:
  shows "uve v \<Longrightarrow> topOps v = 0"
    and "upe p \<Longrightarrow> topOps p = 0"
  by (induct v and p) auto

theorem straighten_straight_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "wform v \<Longrightarrow> strict v \<Longrightarrow> otop (straighten v)"
    and "wform p \<Longrightarrow> strict p \<Longrightarrow> otop (straighten p)"
    and "wform d \<Longrightarrow> strict d \<Longrightarrow> straight (straighten d)"
    and "wform r \<Longrightarrow> strict r \<Longrightarrow> straight (straighten r)"
    and "wform l \<Longrightarrow> strict l \<Longrightarrow> straight (straighten l)"
  by (induct v and p and d and r and l) auto

theorem straighten_straight_disj [simp]:
  fixes d :: disj assumes "wform d" and "strict d"
  shows "straight (straighten d)"
  using assms by (induct d) auto

theorem straighten_straight_conj [simp]:
  fixes c :: conj assumes "wform c" and "strict c"
  shows "straight (straighten c)"
  using assms by (induct c) auto

theorem straighten_straight [simp]:
  fixes c :: cnf assumes "wform c" and "strict c"
  shows "straight (straighten c)"
  using assms by (cases c) simp+

end

