header {* Syntax and Semantics of Substitutions *}

theory Substitutions imports Properties begin


(* Definitions *)

abbreviation substValue ::
  "'a \<Rightarrow> valueExp \<Rightarrow> valueExp \<Rightarrow> ('a :: xmap)" ("_[ _ '/ _ ]\<^sub>v"  [700, 0, 0] 700) where
  "x[ v' / v ]\<^sub>v \<equiv> xmap (XMap (\<lambda>a. if a = v then v' else a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a)) x"

abbreviation substPath ::
  "'a \<Rightarrow> pathExp \<Rightarrow> pathExp \<Rightarrow> ('a :: xmap)" ("_[ _ '/ _ ]\<^sub>p"  [700, 0, 0] 700) where
  "x[ p' / p ]\<^sub>p \<equiv> xmap (XMap (\<lambda>a. a) (\<lambda>a. if a = p then p' else a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a)) x"

abbreviation substDoc ::
  "'a \<Rightarrow> docExp \<Rightarrow> docExp \<Rightarrow> ('a :: xmap)" ("_[ _ '/ _ ]\<^sub>d"  [700, 0, 0] 700) where
  "x[ d' / d ]\<^sub>d \<equiv> xmap (XMap (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. if a = d then d' else a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a) (\<lambda>a. a)) x"

theorem substValue_def:
  "x[ v' / v ]\<^sub>v = xmap (XMap (id(v := v')) id id id id id id) x"
  unfolding id_def fun_upd_def ..

theorem substPath_def:
  "x[ p' / p ]\<^sub>p = xmap (XMap id (id(p := p')) id id id id id) x"
  unfolding id_def fun_upd_def ..

theorem substDoc_def:
  "x[ d' / d ]\<^sub>d = xmap (XMap id id (id(d := d')) id id id id) x"
  unfolding id_def fun_upd_def ..


(* Current Document Substitution Semantics *)

theorem sem_doc_subst [simp]:
  fixes v :: valueExp and p :: pathExp and e :: docExp and r :: pred and l :: lit
  assumes "\<parallel>de\<parallel>\<^sub>E = d\<^sub>d" and "document d"
  shows "\<parallel>v[ de / $. ]\<^sub>d\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E($ \<mapsto> d)"
    and "\<parallel>p[ de / $. ]\<^sub>d\<parallel>\<^sub>E = \<parallel>p\<parallel>\<^sub>E($ \<mapsto> d)" and "kindOf (p[ de / $. ]\<^sub>d) = kindOf p"
    and "\<parallel>e[ de / $. ]\<^sub>d\<parallel>\<^sub>E = \<parallel>e\<parallel>\<^sub>E($ \<mapsto> d)"
    and "\<parallel>r[ de / $. ]\<^sub>d\<parallel>\<^sub>E = \<parallel>r\<parallel>\<^sub>E($ \<mapsto> d)"
    and "\<parallel>l[ de / $. ]\<^sub>d\<parallel>\<^sub>E = \<parallel>l\<parallel>\<^sub>E($ \<mapsto> d)"
  using assms by (induct v and p and e and r and l) auto

theorem sem_disj_doc_subst [simp]:
  fixes j :: disj assumes "\<parallel>d\<parallel>\<^sub>E = e\<^sub>d" and "document e"
  shows "\<parallel>j[ d / $. ]\<^sub>d\<parallel>\<^sub>E = \<parallel>j\<parallel>\<^sub>E($ \<mapsto> e)"
  using assms by (induct j) auto

theorem sem_conj_doc_subst [simp]:
  fixes c :: conj assumes "\<parallel>d\<parallel>\<^sub>E = e\<^sub>d" and "document e"
  shows "\<parallel>c[ d / $. ]\<^sub>d\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E($ \<mapsto> e)"
  using assms by (induct c) auto

theorem sem_cnf_doc_subst [simp]:
  fixes c :: cnf assumes "\<parallel>d\<parallel>\<^sub>E\<^sub>R = e\<^sub>d" and "document e" and "strict d"
  shows "\<parallel>c[ d / $. ]\<^sub>d\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E($ \<mapsto> e)"
  using assms by (cases "closed d", auto, cases c, auto, cases E, auto) (rule sbeI, auto)+

corollary sem_cnf_doc_subst_conj [simp]:
  assumes "\<parallel>d\<parallel>\<^sub>E\<^sub>R = e\<^sub>d" and "document e" and "strict d"
  shows "\<parallel>\<forall>. (c[ d / $. ]\<^sub>d)\<parallel>\<^sub>E = \<parallel>\<forall>. c\<parallel>\<^sub>E($ \<mapsto> e)"
  using assms by (subst sem_cnf_doc_subst[symmetric]) auto


(* Value Constant Substitution Semantics *)

theorem sem_value_subst [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  assumes "\<parallel>ve\<parallel>\<^sub>E = {x}\<^sub>*\<^sub>v"
  shows "\<parallel>v[ ve / PVar n ]\<^sub>v\<parallel>\<^sub>E = \<parallel>v\<parallel>\<^sub>E(n \<mapsto>\<^sub>c x)"
    and "\<parallel>p[ ve / PVar n ]\<^sub>v\<parallel>\<^sub>E = \<parallel>p\<parallel>\<^sub>E(n \<mapsto>\<^sub>c x)" and "kindOf (p[ ve / PVar n ]\<^sub>v) = kindOf p"
    and "\<parallel>d[ ve / PVar n ]\<^sub>v\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E(n \<mapsto>\<^sub>c x)"
    and "\<parallel>r[ ve / PVar n ]\<^sub>v\<parallel>\<^sub>E = \<parallel>r\<parallel>\<^sub>E(n \<mapsto>\<^sub>c x)"
    and "\<parallel>l[ ve / PVar n ]\<^sub>v\<parallel>\<^sub>E = \<parallel>l\<parallel>\<^sub>E(n \<mapsto>\<^sub>c x)"
  using assms by (induct v and p and d and r and l) auto 

theorem sem_disj_value_subst [simp]:
  fixes j :: disj assumes "\<parallel>v\<parallel>\<^sub>E = {x}\<^sub>*\<^sub>v"
  shows "\<parallel>j[ v / PVar n ]\<^sub>v\<parallel>\<^sub>E = \<parallel>j\<parallel>\<^sub>E(n \<mapsto>\<^sub>c x)"
  using assms by (induct j) auto

theorem sem_conj_value_subst [simp]:
  fixes c :: conj assumes "\<parallel>v\<parallel>\<^sub>E = {x}\<^sub>*\<^sub>v"
  shows "\<parallel>c[ v / PVar n ]\<^sub>v\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E(n \<mapsto>\<^sub>c x)"
  using assms by (induct c) auto

theorem sem_cnf_value_subst [simp]:
  fixes c :: cnf assumes "\<parallel>v\<parallel>\<^sub>E\<^sub>R = {x}\<^sub>*\<^sub>v" and "strict v"
  shows "\<parallel>c[ v / PVar n ]\<^sub>v\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E(n \<mapsto>\<^sub>c x)"
  using assms by (cases "closed v", auto, cases c, auto, cases E, auto) (rule sbeI, auto)+

corollary sem_cnf_value_subst_conj [simp]:
  assumes "\<parallel>v\<parallel>\<^sub>E\<^sub>R = {x}\<^sub>*\<^sub>v" and "strict v"
  shows "\<parallel>\<forall>. (c[ v / PVar n ]\<^sub>v)\<parallel>\<^sub>E = \<parallel>\<forall>. c\<parallel>\<^sub>E(n \<mapsto>\<^sub>c x)"
  using assms by (subst sem_cnf_value_subst[symmetric]) auto


(* Well-Formed Substitutions *)

theorem subst_doc_uxe [simp]:
  shows "uve v \<Longrightarrow> uve (v[ x / $. ]\<^sub>d)"
    and "upe p \<Longrightarrow> upe (p[ x / $. ]\<^sub>d)"
  by (induct _ v and _ p rule: uve_upe.induct) auto

theorem subst_doc_wform_exp [rule_format, simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "wform v \<longrightarrow> wform x \<longrightarrow> wform (v[ x / $. ]\<^sub>d)"
    and "wform p \<longrightarrow> wform x \<longrightarrow> wform (p[ x / $. ]\<^sub>d)"
    and "wform d \<longrightarrow> wform x \<longrightarrow> wform (d[ x / $. ]\<^sub>d)"
    and "wform r \<longrightarrow> wform x \<longrightarrow> wform (r[ x / $. ]\<^sub>d)"
    and "wform l \<longrightarrow> wform x \<longrightarrow> wform (l[ x / $. ]\<^sub>d)"
  by (induct v and p and d and r and l) auto

theorem subst_doc_wform_disj [simp]:
  fixes d :: disj assumes "wform d" and "wform x"
  shows "wform (d[ x / $. ]\<^sub>d)"
  using assms by (induct d) auto

theorem subst_doc_wform_conj [simp]:
  fixes c :: conj assumes "wform c" and "wform x"
  shows "wform (c[ x / $. ]\<^sub>d)"
  using assms by (induct c) auto

theorem subst_val_uxe [simp]:
  shows "uve v \<Longrightarrow> uve x \<Longrightarrow> uve (v[ x / y ]\<^sub>v)"
    and "upe p \<Longrightarrow> uve x \<Longrightarrow> upe (p[ x / y ]\<^sub>v)"
  by (induct _ v and _ p rule: uve_upe.induct) auto

theorem subst_val_wform_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "wform v \<Longrightarrow> wform x \<Longrightarrow> uve x \<Longrightarrow> wform (v[ x / y ]\<^sub>v)"
    and "wform p \<Longrightarrow> wform x \<Longrightarrow> uve x \<Longrightarrow> wform (p[ x / y ]\<^sub>v)"
    and "wform d \<Longrightarrow> wform x \<Longrightarrow> uve x \<Longrightarrow> wform (d[ x / y ]\<^sub>v)"
    and "wform r \<Longrightarrow> wform x \<Longrightarrow> uve x \<Longrightarrow> wform (r[ x / y ]\<^sub>v)"
    and "wform l \<Longrightarrow> wform x \<Longrightarrow> uve x \<Longrightarrow> wform (l[ x / y ]\<^sub>v)"
  by (induct v and p and d and r and l) auto

theorem subst_val_wform_disj [simp]:
  fixes d :: disj assumes "wform d" and "wform x" and "uve x"
  shows "wform (d[ x / y ]\<^sub>v)"
  using assms by (induct d) auto

theorem subst_val_wform_conj [simp]:
  fixes c :: conj assumes "wform c" and "wform x" and "uve x"
  shows "wform (c[ x / y ]\<^sub>v)"
  using assms by (induct c) auto


(* Closed Substitutions *)

theorem subst_val_closed_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "closed v \<Longrightarrow> closed x \<Longrightarrow> closed (v[ x / y ]\<^sub>v)"
    and "closed p \<Longrightarrow> closed x \<Longrightarrow> closed (p[ x / y ]\<^sub>v)"
    and "closed d \<Longrightarrow> closed x \<Longrightarrow> closed (d[ x / y ]\<^sub>v)"
    and "closed r \<Longrightarrow> closed x \<Longrightarrow> closed (r[ x / y ]\<^sub>v)"
    and "closed l \<Longrightarrow> closed x \<Longrightarrow> closed (l[ x / y ]\<^sub>v)"
  by (induct v and p and d and r and l) auto

theorem subst_val_closed_disj [simp]:
  fixes d :: disj assumes "closed d" and "closed x"
  shows "closed (d[ x / y ]\<^sub>v)"
  using assms by (induct d) auto

theorem subst_val_closed_conj [simp]:
  fixes c :: conj assumes "closed c" and "closed x"
  shows "closed (c[ x / y ]\<^sub>v)"
  using assms by (induct c) auto

theorem subst_doc_closed_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "closed v \<Longrightarrow> closed x \<Longrightarrow> closed (v[ x / $. ]\<^sub>d)"
    and "closed p \<Longrightarrow> closed x \<Longrightarrow> closed (p[ x / $. ]\<^sub>d)"
    and "closed d \<Longrightarrow> closed x \<Longrightarrow> closed (d[ x / $. ]\<^sub>d)"
    and "closed r \<Longrightarrow> closed x \<Longrightarrow> closed (r[ x / $. ]\<^sub>d)"
    and "closed l \<Longrightarrow> closed x \<Longrightarrow> closed (l[ x / $. ]\<^sub>d)"
  by (induct v and p and d and r and l) auto

theorem subst_doc_closed_disj [simp]:
  fixes d :: disj assumes "closed d" and "closed x"
  shows "closed (d[ x / $. ]\<^sub>d)"
  using assms by (induct d) auto

theorem subst_doc_closed_conj [simp]:
  fixes c :: conj assumes "closed c" and "closed x"
  shows "closed (c[ x / $. ]\<^sub>d)"
  using assms by (induct c) auto


(* Strict Substitutions *)

theorem subst_val_strict_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "strict v \<Longrightarrow> strict x \<Longrightarrow> strict (v[ x / y ]\<^sub>v)"
    and "strict p \<Longrightarrow> strict x \<Longrightarrow> strict (p[ x / y ]\<^sub>v)"
    and "strict d \<Longrightarrow> strict x \<Longrightarrow> strict (d[ x / y ]\<^sub>v)"
    and "strict r \<Longrightarrow> strict x \<Longrightarrow> strict (r[ x / y ]\<^sub>v)"
    and "strict l \<Longrightarrow> strict x \<Longrightarrow> strict (l[ x / y ]\<^sub>v)"
  by (induct v and p and d and r and l) auto

theorem subst_val_strict_disj [simp]:
  fixes d :: disj assumes "strict d" and "strict x"
  shows "strict (d[ x / y ]\<^sub>v)"
  using assms by (induct d) auto

theorem subst_val_strict_conj [simp]:
  fixes c :: conj assumes "strict c" and "strict x"
  shows "strict (c[ x / y ]\<^sub>v)"
  using assms by (induct c) auto

theorem subst_doc_strict_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "strict v \<Longrightarrow> strict x \<Longrightarrow> strict (v[ x / $. ]\<^sub>d)"
    and "strict p \<Longrightarrow> strict x \<Longrightarrow> strict (p[ x / $. ]\<^sub>d)"
    and "strict d \<Longrightarrow> strict x \<Longrightarrow> strict (d[ x / $. ]\<^sub>d)"
    and "strict r \<Longrightarrow> strict x \<Longrightarrow> strict (r[ x / $. ]\<^sub>d)"
    and "strict l \<Longrightarrow> strict x \<Longrightarrow> strict (l[ x / $. ]\<^sub>d)"
  by (induct v and p and d and r and l) auto

theorem subst_doc_strict_disj [simp]:
  fixes d :: disj assumes "strict d" and "strict x"
  shows "strict (d[ x / $. ]\<^sub>d)"
  using assms by (induct d) auto

theorem subst_doc_strict_conj [simp]:
  fixes c :: conj assumes "strict c" and "strict x"
  shows "strict (c[ x / $. ]\<^sub>d)"
  using assms by (induct c) auto


(* Straight Substitutions *)

theorem subst_val_straight_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "straight v \<Longrightarrow> straight x \<Longrightarrow> straight (v[ x / y ]\<^sub>v)"
    and "straight p \<Longrightarrow> straight x \<Longrightarrow> straight (p[ x / y ]\<^sub>v)"
    and "straight d \<Longrightarrow> straight x \<Longrightarrow> straight (d[ x / y ]\<^sub>v)"
    and "straight r \<Longrightarrow> straight x \<Longrightarrow> straight (r[ x / y ]\<^sub>v)"
    and "straight l \<Longrightarrow> straight x \<Longrightarrow> straight (l[ x / y ]\<^sub>v)"
  by (induct v and p and d and r and l) auto

theorem subst_val_straight_disj [simp]:
  fixes d :: disj assumes "straight d" and "straight x"
  shows "straight (d[ x / y ]\<^sub>v)"
  using assms by (induct d) auto

theorem subst_val_straight_conj [simp]:
  fixes c :: conj assumes "straight c" and "straight x"
  shows "straight (c[ x / y ]\<^sub>v)"
  using assms by (induct c) auto

theorem subst_doc_straight_exp [simp]:
  fixes v :: valueExp and p :: pathExp and d :: docExp and r :: pred and l :: lit
  shows "straight v \<Longrightarrow> straight x \<Longrightarrow> straight (v[ x / $. ]\<^sub>d)"
    and "straight p \<Longrightarrow> straight x \<Longrightarrow> straight (p[ x / $. ]\<^sub>d)"
    and "straight d \<Longrightarrow> straight x \<Longrightarrow> straight (d[ x / $. ]\<^sub>d)"
    and "straight r \<Longrightarrow> straight x \<Longrightarrow> straight (r[ x / $. ]\<^sub>d)"
    and "straight l \<Longrightarrow> straight x \<Longrightarrow> straight (l[ x / $. ]\<^sub>d)"
  by (induct v and p and d and r and l) auto

theorem subst_doc_straight_disj [simp]:
  fixes d :: disj assumes "straight d" and "straight x"
  shows "straight (d[ x / $. ]\<^sub>d)"
  using assms by (induct d) auto

theorem subst_doc_straight_conj [simp]:
  fixes c :: conj assumes "straight c" and "straight x"
  shows "straight (c[ x / $. ]\<^sub>d)"
  using assms by (induct c) auto


end
