header {* Concrete Syntax Definitions (Simplifying Sugar) *}

theory Syntax imports Properties Ordering
begin


(* Semantics Preservation *)

locale semantics = fixes b a
assumes domain:  "\<parallel>a\<parallel>\<^sub>E \<noteq> \<bottom> \<Longrightarrow> \<parallel>b\<parallel>\<^sub>E = \<parallel>a\<parallel>\<^sub>E"

theorem (in semantics) undef [dest!]:  (* unsafe in general! *)
  assumes "\<parallel>b\<parallel>\<^sub>E = \<bottom>"
  shows "\<parallel>a\<parallel>\<^sub>E = \<bottom>"
  using assms and domain by - (erule contrapos_pp, simp)


(* Property Preservation *)

locale props_base = fixes y x
assumes wform    [simp]: "wform  x \<Longrightarrow> wform  y" 
assumes closed   [simp]: "closed x \<Longrightarrow> closed y"

locale prop_strict = fixes y x
assumes strict   [simp]: "strict   x \<Longrightarrow> strict   y" 

locale prop_straight = fixes y x
assumes straight [simp]: "straight x \<Longrightarrow> straight y" 

locale prop_basic = prop_strict + prop_straight

locale prop_dufree = fixes y x
assumes dufree   [simp]: "dufree   x \<Longrightarrow> dufree   y" 

locale props_eq = props_base + prop_basic + prop_dufree

locale props_mono = fixes y x
assumes dus    [intro!]: "dus   x \<le> n \<Longrightarrow> dus   y \<le> n"
assumes conds  [intro!]: "conds x \<le> n \<Longrightarrow> conds y \<le> n"
assumes ops    [intro!]: "ops   x \<le> n \<Longrightarrow> ops   y \<le> n"

locale props = props_eq + props_mono

locale props_top = props_mono +
assumes dutop    [simp]: "dutop x \<Longrightarrow> dutop y"
assumes otop     [simp]: "otop  x \<Longrightarrow> otop  y"

locale prop_ve = fixes y x
assumes uve      [simp]: "uve x \<Longrightarrow> uve y"

locale prop_pe = fixes y x
assumes upe      [simp]: "upe x \<Longrightarrow> upe y"
assumes kind     [simp]: "kindOf y = kindOf x"

locale vprops = props + props_top + prop_ve
locale pprops = props + props_top + prop_pe
locale dprops = props + props_top

locale vprops_eq = props_eq + prop_ve
locale pprops_eq = props_eq + prop_pe


declare bexp_normalize [simp del]
declare bexp_defs [simp]

interpretation bTrue_pred_props:  props "True.  :: pred" "Root." by unfold_locales simp+
interpretation bFalse_pred_props: props "False. :: pred" "Root." by unfold_locales simp+
interpretation bTrue_lit_props:   props "True.  :: lit"  "Root." by unfold_locales simp+
interpretation bFalse_lit_props:  props "False. :: lit"  "Root." by unfold_locales simp+
interpretation bTrue_disj_props:  props "True.  :: disj" "Root." by unfold_locales simp+
interpretation bFalse_disj_props: props "False. :: disj" "Root." by unfold_locales simp+
interpretation bTrue_conj_props:  props "True.  :: conj" "Root." by unfold_locales simp+
interpretation bFalse_conj_props: props "False. :: conj" "Root." by unfold_locales simp+

declare bexp_normalize [simp]
declare bexp_defs [simp del]


(* Static Analysis *)

fun sType :: "valueExp \<rightharpoonup> type" where
  "sType (Val v)      = (typeOf v)\<^sub>!"
| "sType (FVar _)     = TKey\<^sub>!"
| "sType (Sum _)      = TInt\<^sub>!"
| "sType (Size _)     = TInt\<^sub>!"
| "sType (Count _ _)  = TInt\<^sub>!"
| "sType (Tally _ _)  = TInt\<^sub>!"
| "sType (Scal _ _)   = TInt\<^sub>!"
| "sType (VRep _ m)   = sType m"
| "sType (VPlus a b)  = (if sType a = sType b then sType a else None)"
| "sType (VMinus a _) = sType a"
| "sType _            = None"

theorem sType_sem:
  assumes "sType v = t\<^sub>!" and "\<parallel>v\<parallel>\<^sub>E = m\<^sub>v"
  shows "hasTypeM t m"
  using assms by (induct v arbitrary: m, auto)
    (cases E, auto, split if_splits, auto, split if_splits, auto)


fun sUnique :: "valueExp \<Rightarrow> bool" where
  "sUnique (Val _)     \<longleftrightarrow> True"
| "sUnique (FVar _)    \<longleftrightarrow> True"
| "sUnique (PVar _)    \<longleftrightarrow> True"
| "sUnique (Sum _)     \<longleftrightarrow> True"
| "sUnique (Size _)    \<longleftrightarrow> True"
| "sUnique (Count _ _) \<longleftrightarrow> True"
| "sUnique (Tally _ _) \<longleftrightarrow> True"
| "sUnique (Scal _ m)  \<longleftrightarrow> sType m = TInt\<^sub>! \<and> sUnique m"
(* | "sUnique (VRep n m)  \<longleftrightarrow> n = 1. \<and> sUnique m"  does not play well with uve *)
| "sUnique _           \<longleftrightarrow> False"

theorem sUnique_sem:
  assumes "sUnique v" and "\<parallel>v\<parallel>\<^sub>E = m\<^sub>v"
  shows "unique m"
  using assms
proof (induct v arbitrary: m)
  case Scal thus ?case by (auto dest!: sType_sem) (drule_tac x=y in meta_spec, force) next
  case FVar thus ?case by (cases E) auto
qed auto

theorem sUnique_uve [simp]:
  assumes "sUnique v"
  shows "uve v"
  using assms by (induct v rule: sUnique.induct) auto

theorem sUnique_topDUs [simp]:
  assumes "sUnique v"
  shows "topDUs v = 0"
  using assms by (induct v rule: sUnique.induct) auto

theorem sUnique_topOps [simp]:
  assumes "sUnique v"
  shows "topOps v = 0"
  using assms by (induct v rule: sUnique.induct) auto


fun sNat :: "valueExp \<Rightarrow> bool" where
  "sNat (Val (VInt n)) \<longleftrightarrow> n \<ge> 0"
| "sNat (Size _)    \<longleftrightarrow> True"
| "sNat (Count _ _) \<longleftrightarrow> True"
| "sNat (Tally _ _) \<longleftrightarrow> True"
| "sNat (Scal n m)  \<longleftrightarrow> sNat n \<and> sNat m"
| "sNat _           \<longleftrightarrow> False"

theorem sNat_sem:
  assumes "sNat v" and "\<parallel>v\<parallel>\<^sub>E = m\<^sub>v"
  shows "\<exists>n. m = {n}\<^sub>i \<and> n \<ge> 0"
  using assms by (induct v arbitrary: m rule: sNat.induct, auto)
    (drule_tac x="{xa}\<^sub>i" in meta_spec, drule_tac x=y in meta_spec, auto, metis zero_le_mult_iff)


fun sEmpty :: "pathExp \<Rightarrow> bool" where
  "sEmpty Root.         \<longleftrightarrow> False"
| "sEmpty (Path l p k)  \<longleftrightarrow> k = {}. \<or> sEmpty p"
| "sEmpty (Kind l p)    \<longleftrightarrow> sEmpty p"
| "sEmpty (PRep n p)    \<longleftrightarrow> n = 0. \<or> sEmpty p"
| "sEmpty (PPlus a b)   \<longleftrightarrow> sEmpty a \<and> sEmpty b"
| "sEmpty (PMinus a b)  \<longleftrightarrow> sEmpty a \<or> a = b"
| "sEmpty (PCond b t e) \<longleftrightarrow> sEmpty t \<and> sEmpty e"

theorem sEmpty_sem:
  assumes "sEmpty p" and "\<parallel>p\<parallel>\<^sub>E = q\<^sub>p"
  shows "q = {}\<^sub>*"
  using assms

apply (induct p arbitrary: q rule: sEmpty.induct)
apply auto
apply (drule_tac x=x in meta_spec, simp)
apply (drule_tac x=x in meta_spec, simp)
apply (drule_tac x=y in meta_spec, simp)
apply (split if_splits, auto)
apply (drule_tac x=x in meta_spec, simp)
apply (drule_tac x=x in meta_spec, simp)
apply (split if_splits, auto)
apply (erule sbeE, auto)
done


corollary sProp_sem:
  assumes "\<parallel>v\<parallel>\<^sub>E = m\<^sub>v"
  shows "sType v = t\<^sub>! \<Longrightarrow> hasTypeM t m"
    and "sUnique v \<Longrightarrow> unique m"
    and "sNat v \<Longrightarrow> \<exists>n. m = {n}\<^sub>i \<and> n \<ge> 0"
  using assms by auto (metis sType_sem, metis sUnique_sem, metis sNat_sem)


(*----------*)
(*-  Size  -*)
(*----------*)

definition sSize :: "valueExp \<Rightarrow> valueExp" ("size. _" [1000] 999) where
  "size. m \<equiv>
    if m = {}. then 0. else
    if sUnique m then 1. else
    Size m"

interpretation sSize_sem:
  semantics "size. m" "Size m"
  unfolding sSize_def by unfold_locales (auto dest!: sUnique_sem)

interpretation sSize_prop:
  vprops "size. m" "Size m"
  unfolding sSize_def by unfold_locales simp+

corollary sSize_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>v" 
  shows "\<parallel>size. m\<parallel>\<^sub>E = \<parallel>Size m\<parallel>\<^sub>E"
  using assms by (subst sSize_sem.domain) auto


(*-----------*)
(*-  Count  -*)
(*-----------*)

definition sCount :: "valueExp \<Rightarrow> valueExp \<Rightarrow> valueExp" ("count. _ _" [1000, 1000] 999) where
  "count. m n \<equiv> 
    if m = {}. then 0. else
    if m = n then 1. else
    if (case (sType m, sType n) of (t1\<^sub>!, t2\<^sub>!) \<Rightarrow> t1 \<noteq> t2 | _ \<Rightarrow> False) then 0. else
    if sUnique m \<and> n \<prec> m = GT then Count n m else
    Count m n"

interpretation sCount_sem:
  semantics "count. m n" "Count m n"
  unfolding sCount_def by unfold_locales
    (auto split: option.split, auto dest!: sType_sem sUnique_sem, (case_tac "e = ea", auto)+)

interpretation sCount_props:
  vprops "count. m n" "Count m n"
  unfolding sCount_def by unfold_locales simp+

corollary sCount_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>v" and "\<parallel>n\<parallel>\<^sub>E = {v}\<^sub>*\<^sub>v"
  shows "\<parallel>count. m n\<parallel>\<^sub>E = \<parallel>Count m n\<parallel>\<^sub>E"
  using assms by (subst sCount_sem.domain) auto


(*-----------*)
(*-  Tally  -*)
(*-----------*)

definition sTally :: "type \<Rightarrow> valueExp \<Rightarrow> valueExp" ("tally. _ _" [1000, 1000] 999) where
  "tally. t m \<equiv>
     if sType m = t\<^sub>! then size. m else
     if sType m \<noteq> None then 0. else
     Tally t m"

interpretation sTally_sem:
  semantics "tally. t m" "Tally t m"
  using sSize_sem.domain unfolding sTally_def by unfold_locales (auto, auto dest!: sType_sem)

interpretation sTally_props:
  vprops "tally. t m" "Tally t m"
  unfolding sTally_def by unfold_locales auto

corollary sTally_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>v" 
  shows "\<parallel>tally. t m\<parallel>\<^sub>E = \<parallel>Tally t m\<parallel>\<^sub>E"
  using assms by (subst sTally_sem.domain) auto


(*----------*)
(*-  VRep  -*)
(*----------*)

definition sVRep :: "valueExp \<Rightarrow> valueExp \<Rightarrow> valueExp" ("_ \<otimes>. _" [1000, 1000] 999) where
  "n \<otimes>. m \<equiv>
    if n = 1. then m else
    if n = 0. then {}. else
    if m = {}. then {}. else
    VRep n m"

interpretation sVRep_sem:
  semantics "n \<otimes>. m" "VRep n m"
  unfolding sVRep_def by unfold_locales auto

interpretation sVRep_props:
  vprops "n \<otimes>. m" "VRep n m"
  unfolding sVRep_def by unfold_locales auto

corollary sVRep_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>v" and "\<parallel>n\<parallel>\<^sub>E = {v}\<^sub>i\<^sub>v" and "v \<ge> 0"
  shows "\<parallel>n \<otimes>. m\<parallel>\<^sub>E = \<parallel>VRep n m\<parallel>\<^sub>E"
  using assms by (subst sVRep_sem.domain) auto


(*-----------*)
(*-  VPlus  -*)
(*-----------*)

definition sVPlus :: "valueExp \<Rightarrow> valueExp \<Rightarrow> valueExp" ("_ \<oplus>. _" [1000, 1000] 999) where
  "a \<oplus>. b \<equiv>
    if a = {}. then b else
    if b = {}. then a else
    VPlus a b"

interpretation sVPlus_sem:
  semantics "a \<oplus>. b" "VPlus a b"
  unfolding sVPlus_def by unfold_locales auto

interpretation sVPlus_props:
  vprops "a \<oplus>. b" "VPlus a b"
  unfolding sVPlus_def by unfold_locales auto

corollary sVPlus_sem_simplified [simp]:
  assumes "\<parallel>a\<parallel>\<^sub>E = x\<^sub>v" and "\<parallel>b\<parallel>\<^sub>E = y\<^sub>v"
  shows "\<parallel>a \<oplus>. b\<parallel>\<^sub>E = \<parallel>VPlus a b\<parallel>\<^sub>E"
  using assms by (subst sVPlus_sem.domain) auto


(*------------*)
(*-  VMinus  -*)
(*------------*)

definition sVMinus :: "valueExp \<Rightarrow> valueExp \<Rightarrow> valueExp" ("_ \<ominus>. _" [1000, 1000] 999) where
  "a \<ominus>. b \<equiv>
    if a = {}. then {}. else
    if b = {}. then a else
    VMinus a b"

interpretation sVMinus_sem:
  semantics "a \<ominus>. b" "VMinus a b"
  unfolding sVMinus_def by unfold_locales auto

interpretation sVMinus_props:
  vprops "a \<ominus>. b" "VMinus a b"
  unfolding sVMinus_def by unfold_locales auto

corollary sVMinus_sem_simplified [simp]:
  assumes "\<parallel>a\<parallel>\<^sub>E = x\<^sub>v" and "\<parallel>b\<parallel>\<^sub>E = y\<^sub>v" and "y \<subseteq>\<^sub>* x"
  shows "\<parallel>a \<ominus>. b\<parallel>\<^sub>E = \<parallel>VMinus a b\<parallel>\<^sub>E"
  using assms by (subst sVMinus_sem.domain) auto


(*----------*)
(*-  Scal  -*)
(*----------*)

definition scal :: "valueExp \<Rightarrow> valueExp \<Rightarrow> valueExp" where
  "scal a b \<equiv>
    if a = null. then \<bottom>. else
    if b = null. then {}. else
    if b = 1. \<and> uve a then a else
    if b = 0. then 0. else
    if sType b = TInt\<^sub>! then
      if a = 1. then b else
      if a = 0. \<and> sUnique b then 0. else
      if sUnique b \<and> uve a \<and> b \<prec> a = LT then Scal b a else Scal a b else
    Scal a b"

interpretation scal_sem:
  semantics "scal a b" "Scal a b"
  unfolding scal_def by unfold_locales (auto dest!: sType_sem sUnique_sem)

interpretation scal_props:
  vprops "scal a b" "Scal a b"
  unfolding scal_def by unfold_locales auto

corollary scal_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>v" and "\<parallel>n\<parallel>\<^sub>E = {v}\<^sub>i\<^sub>v"
  shows "\<parallel>scal n m\<parallel>\<^sub>E = \<parallel>Scal n m\<parallel>\<^sub>E"
  using assms by (subst scal_sem.domain) auto

fun isIV :: "valueExp \<Rightarrow> bool" where
  "isIV (Val n\<^sub>i) = True"
| "isIV v = False"

fun theIV :: "valueExp \<Rightarrow> int" where
  "theIV (Val n) = n"
| "theIV v = undefined"

theorem isIVD [dest!]:
  assumes "isIV v"
  shows "\<exists>n. v = Val n\<^sub>i"
  using assms by (cases v rule: isIV.cases) auto

fun isSum :: "valueExp \<Rightarrow> bool" where
  "isSum (Sum m) = True"
| "isSum v = False"

fun sScal :: "valueExp \<Rightarrow> valueExp \<Rightarrow> valueExp"  ("_ \<odot>. _" [1000, 1000] 999) where
  "a \<odot>. (Scal b v) = (
    if isIV a \<and> isIV b then scal (Val (theIV a * theIV b)) v else
    if b \<prec> a = LT then scal b (a \<odot>. v) else
    scal a (Scal b v))"
| "a \<odot>. (Val b) = (
    if isIV a then Val (theIV a * b) else
    if uve a then scal (Val b) a else
    scal a (Val b))"
| "a \<odot>. (VPlus b c) = (a \<odot>. b) \<oplus>. (a \<odot>. c)"
| "a \<odot>. (VMinus b c) = (a \<odot>. b) \<ominus>. (a \<odot>. c)"
| "a \<odot>. (Sum m) = Sum (a \<odot>. m)"
| "a \<odot>. b = scal a b"

abbreviation exp_neg :: "valueExp \<Rightarrow> valueExp" ("~.") where
  "~. x \<equiv> ~1. \<odot>. x"

interpretation sScal_sem:
  semantics "a \<odot>. b" "Scal a b"
  by (unfold_locales, induct a b rule: sScal.induct, simp_all)
     ((safe, simp_all)+, (subst mult_commute, simp)+)

theorem sScal_uve [simp]:
  assumes "uve (Scal n m)"
  shows "uve (n \<odot>. m)"
  using assms by (induct n m rule: sScal.induct) auto

interpretation sScal_props:
  vprops_eq "a \<odot>. b" "Scal a b"
  by (unfold_locales, auto elim!: order_trans[rotated])
     (induct a b rule: sScal.induct, auto simp: nat_add_assoc[symmetric])+

theorem sScal_dutop [simp]: (* special theorem for elimdu transformer *)
  assumes "dufree n" and "dutop m"
  shows "dutop (n \<odot>. m)"
  using assms by (induct n m rule: sScal.induct) auto

declare sScal.simps [simp del]

corollary sScal_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>v" and "\<parallel>n\<parallel>\<^sub>E = {v}\<^sub>i\<^sub>v"
  shows "\<parallel>n \<odot>. m\<parallel>\<^sub>E = \<parallel>Scal n m\<parallel>\<^sub>E"
  using assms by (subst sScal_sem.domain) auto


(*----------*)
(*-  Read  -*)
(*----------*)

definition sRead :: "docExp \<Rightarrow> pathExp \<Rightarrow> valueExp"  ("_'(_')." [1000, 0] 1000) where
  "d(p). \<equiv>
    if sEmpty p then {}. else
    Read d p"

interpretation sRead_sem:
  semantics "d(p)." "Read d p"
  unfolding sRead_def by unfold_locales (auto dest: sEmpty_sem)

interpretation sRead_props:
  vprops "d(p)." "Read d p"
  unfolding sRead_def by unfold_locales simp+

corollary sRead_sem_simplified [simp]:
  assumes "\<parallel>d\<parallel>\<^sub>E = x\<^sub>d" and "\<parallel>p\<parallel>\<^sub>E = y\<^sub>p"
  shows "\<parallel>d(p).\<parallel>\<^sub>E = \<parallel>Read d p\<parallel>\<^sub>E"
  using assms by (subst sRead_sem.domain) auto


(*---------*)
(*-  Sum  -*)
(*---------*)

(* Linearize *)

fun toList :: "valueExp \<Rightarrow> valueExp list \<Rightarrow> valueExp list" where
  "toList (Sum m)     l = toList m l"
| "toList (VPlus a b) l = toList a (toList b l)"
| "toList (Scal n m)  l = (
    if n = 0. then l else
    if n = 1. then toList m l else
    Scal n m # l)"
| "toList v           l = v # l"

fun toPlus :: "valueExp list \<Rightarrow> valueExp" where
  "toPlus []      = {}."
| "toPlus (a # b) = (
    if a = 0. then toPlus b else
    a \<oplus>. (toPlus b))"

definition flatten :: "valueExp \<Rightarrow> valueExp" where
  "flatten v \<equiv> toPlus (toList v [])"

interpretation toPlus_sem:
  semantics "Sum (toPlus (h # t))" "Sum (VPlus h (toPlus t))"
  by unfold_locales auto

theorem finite_single_value [simp]:
  fixes E :: env
  shows "finite {e. e\<^sub>i = x}"
  by (cases x) auto

theorem env_fvar_sem [dest]:
  assumes "E(n)\<^sub>w = v\<^sub>v"
  shows "\<exists>k. v = {k\<^sub>k}\<^sub>*"
  using assms by (cases E) auto

interpretation toPlus_toList_sem_general:
  semantics "Sum (toPlus (toList v l))" "Sum (VPlus v (toPlus l))"

apply (unfold_locales, auto)
apply (induct v l rule: toList.induct)
apply auto
apply (drule_tac x=E in meta_spec)+
apply (drule_tac x=ya in meta_spec, drule_tac x=x in meta_spec)
apply (drule_tac x=y in meta_spec, auto)
apply (subst plusM_sum_homomorphism, auto)
apply (rule_tac B="elemset (asIntM x)" in rev_finite_subset, auto)
apply (rule_tac x=va and E=E in sbeI, auto)
done

interpretation toPlus_toList_sem:
  semantics "Sum (toPlus (toList v []))" "Sum v"
  by unfold_locales (subst toPlus_toList_sem_general.domain, auto)

interpretation flatten_sem:
  semantics "Sum (flatten v)" "Sum v"
  unfolding flatten_def ..

interpretation toPlus_props:
  vprops "Sum (toPlus (h # t))" "Sum (VPlus h (toPlus t))"
  by unfold_locales (cases t, auto)+

lemma toPlus_toList_dufree [simp]:
  assumes "dufree (toPlus l)" and "dufree v"
  shows "dufree (toPlus (toList v l))"
  using assms by (induct v l rule: toList.induct) auto

lemma toPlus_toList_otop [simp]:
  assumes "otop (toPlus l)" and "otop v"
  shows "otop (toPlus (toList v l))"
  using assms by (induct v l rule: toList.induct) auto

interpretation toPlus_toList_props_general:
  vprops "toPlus (toList v l)" "VPlus v (toPlus l)"
  by (unfold_locales, auto elim!: order_trans[rotated]) (induct v l rule: toList.induct, auto)+

interpretation toPlus_toList_props:
  vprops "Sum (toPlus (toList v []))" "Sum v"
  by (unfold_locales, auto)

interpretation flatten_props:
  vprops "Sum (flatten v)" "Sum v"
  unfolding flatten_def ..

lemmas flatten_props [intro!] = 
  flatten_props.closed flatten_props.wform flatten_props.uve
  flatten_props.strict flatten_props.straight
  flatten_props.dufree flatten_props.dutop flatten_props.otop


(* Cancellation *)

fun neutralize :: "valueExp \<Rightarrow> valueExp \<Rightarrow> valueExp" where
  "neutralize a (VPlus b s) = (
     if a = ~. b then s else case neutralize a s of s' \<Rightarrow>
     if s' = a \<oplus>. s then a \<oplus>. (b \<oplus>. s) else b \<oplus>. s')"
| "neutralize a b = (if a = ~. b then {}. else a \<oplus>. b)"

declare mult_minus1 [simp del]

interpretation neutralize_sem:
  semantics "Sum (neutralize a b)" "Sum (VPlus a b)"

apply unfold_locales
apply auto
apply (induct a b rule: neutralize.induct)
apply auto
apply (drule_tac x=E in meta_spec)+
apply (drule_tac x=xa in meta_spec, drule_tac x=xa in meta_spec)
apply (drule_tac x=ya in meta_spec, drule_tac x=ya in meta_spec)
apply auto
apply (subst plusM_sum_homomorphism, auto)
apply (rule_tac B="elemset (asIntM x)" in rev_finite_subset, auto)
apply (rule_tac x=v and E=E in sbeI, auto)
done

interpretation neutralize_props:
  vprops "Sum (neutralize a b)" "Sum (VPlus a b)"
  by (unfold_locales, auto elim!: order_trans[rotated])
     (induct a b rule: neutralize.induct, auto)+  

lemmas neutralize_props [intro!] = 
  neutralize_props.closed neutralize_props.wform neutralize_props.uve
  neutralize_props.strict neutralize_props.straight
  neutralize_props.dufree neutralize_props.dutop neutralize_props.otop

fun cancel :: "valueExp \<Rightarrow> valueExp" where
  "cancel (VPlus a b) = neutralize a (cancel b)"
| "cancel v = v"

interpretation cancel_sem:
  semantics "Sum (cancel v)" "Sum v"
  by (unfold_locales, induct v rule: cancel.induct, auto)
     (drule_tac x=E in meta_spec, subst neutralize_sem.domain[simplified], auto)

interpretation cancel_props:
  vprops "Sum (cancel v)" "Sum v"
  by unfold_locales (induct v rule: cancel.induct, simp only: cancel.simps, safe, force+)+

lemmas cancel_props [intro!] = 
  cancel_props.closed cancel_props.wform cancel_props.uve
  cancel_props.strict cancel_props.straight
  cancel_props.dufree cancel_props.dutop cancel_props.otop


(* Ordering *)

fun insertOperand :: "valueExp \<Rightarrow> valueExp \<Rightarrow> valueExp" where
  "insertOperand a (VPlus b v) = (
    if isIV a \<and> isIV b then insertOperand (Val (theIV a + theIV b)) v else
    if b \<prec> a = LT then b \<oplus>. (insertOperand a v) else
    a \<oplus>. (b \<oplus>. v))"
| "insertOperand a (Val b) = (
    if isIV a then Val (theIV a + b) else
    (Val b) \<oplus>. a)"
| "insertOperand a b = (
    if b \<prec> a = LT then b \<oplus>. a else
    a \<oplus>. b)"

interpretation insertOperand_sem:
  semantics "Sum (insertOperand a b)" "Sum (VPlus a b)"

apply unfold_locales
apply (induct a b rule: insertOperand.induct)
apply simp_all
apply (safe, simp_all)+
apply auto
apply (drule_tac x=E in meta_spec)+
apply auto
apply (drule_tac x=E in meta_spec)+
apply auto
apply (subst plusM_sum_homomorphism, auto)
apply (rule_tac B="elemset (asIntM xb)" in rev_finite_subset, auto)
apply (subst plusM_sum_homomorphism, auto)
apply (rule_tac B="elemset (asIntM xb)" in rev_finite_subset, auto)
done

interpretation insertOperand_props:
  vprops "Sum (insertOperand a b)" "Sum (VPlus a b)"
  by (unfold_locales, auto elim!: order_trans[rotated])
     (induct a b rule: insertOperand.induct, auto simp: nat_add_assoc[symmetric])+

lemmas insertOperand_props [intro!] = 
  insertOperand_props.closed insertOperand_props.wform insertOperand_props.uve
  insertOperand_props.strict insertOperand_props.straight
  insertOperand_props.dufree insertOperand_props.dutop insertOperand_props.otop

fun order :: "valueExp \<Rightarrow> valueExp" where
  "order (VPlus a b) = insertOperand a (order b)"
| "order v = v"

interpretation order_sem:
  semantics "Sum (order v)" "Sum v"
  by (unfold_locales, induct v rule: order.induct, auto)
     (drule_tac x=E in meta_spec, subst insertOperand_sem.domain[simplified], auto)

interpretation order_props:
  vprops "Sum (order v)" "Sum v"
  by unfold_locales (induct v rule: order.induct, simp only: order.simps, safe, force+)+

lemmas order_props [intro!] = 
  order_props.closed order_props.wform order_props.uve
  order_props.strict order_props.straight
  order_props.dufree order_props.dutop order_props.otop


(* Remaining Trivials *)

definition sumTriv :: "valueExp \<Rightarrow> valueExp"   where
  "sumTriv m \<equiv>
    if m = {}. then 0. else
    if sUnique m \<and> sType m = TInt\<^sub>! then m else
    if sType m \<noteq> None \<and> sType m \<noteq> TInt\<^sub>! then 0. else
    Sum m"

lemma sumM_not_int [intro]:
  assumes "hasTypeM t m" and "t \<noteq> TInt"
  shows "sumM (asIntM m) = 0"
  using assms unfolding sumM_def by auto (subgoal_tac "{e. e\<^sub>i \<in>\<^sub>* m} = {}", auto)

interpretation sumTriv_sem:
  semantics "sumTriv m" "Sum m"
  unfolding sumTriv_def
  by unfold_locales (auto, auto dest!: sType_sem sUnique_sem)

interpretation sumTriv_props:
  vprops "sumTriv m" "Sum m"
  unfolding sumTriv_def by unfold_locales auto

lemmas sum_props [intro!] = 
  sumTriv_props.closed sumTriv_props.wform sumTriv_props.uve
  sumTriv_props.strict sumTriv_props.straight
  sumTriv_props.dufree sumTriv_props.dutop sumTriv_props.otop


(* Final Syntax Definition *)

definition sSum :: "valueExp \<Rightarrow> valueExp" ("sum. _" [1000] 999) where
  "sum. m \<equiv> sumTriv (order (cancel (flatten m)))"

abbreviation exp_plus :: "valueExp \<Rightarrow> valueExp \<Rightarrow> valueExp" ("_ +. _" [999, 999] 999) where
  "a +. b \<equiv> sum. (a \<oplus>. b)"

abbreviation exp_minus :: "valueExp \<Rightarrow> valueExp \<Rightarrow> valueExp" ("_ -. _" [999, 999] 999) where
  "a -. b \<equiv> a +. (~. b)"

interpretation sSum_sem:
  semantics "sum. m" "Sum m"
  unfolding sSum_def

apply unfold_locales
apply (subst sumTriv_sem.domain, simp)
apply (subst order_sem.domain[simplified])
apply (subst cancel_sem.domain[simplified])
apply (subst flatten_sem.domain[simplified])
apply auto
apply (subst order_sem.domain[simplified])
apply (subst cancel_sem.domain[simplified])
apply (subst flatten_sem.domain[simplified])
apply auto
apply (subst cancel_sem.domain[simplified])
apply (subst flatten_sem.domain[simplified])
apply auto
apply (subst flatten_sem.domain[simplified])
apply auto
done

interpretation sSum_props:
  vprops "sum. m" "Sum m"
  unfolding sSum_def by unfold_locales auto

corollary sSum_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>v" 
  shows "\<parallel>sum. m\<parallel>\<^sub>E = \<parallel>Sum m\<parallel>\<^sub>E"
  using assms by (subst sSum_sem.domain) auto


(*----------*)
(*-  PRep  -*)
(*----------*)

definition sPRep :: "valueExp \<Rightarrow> pathExp \<Rightarrow> pathExp" ("_ .\<otimes> _" [1000, 1000] 999) where
  "n .\<otimes> m \<equiv>
    if n = 1. then m else
    if sEmpty m then m else
    PRep n m"

interpretation sPRep_sem:
  semantics "n .\<otimes> m" "PRep n m"
  unfolding sPRep_def by unfold_locales (auto dest: sEmpty_sem)

interpretation sPRep_props:
  pprops "n .\<otimes> m" "PRep n m"
  unfolding sPRep_def by unfold_locales auto

corollary sPRep_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>p" and "\<parallel>n\<parallel>\<^sub>E = {v}\<^sub>i\<^sub>v" and "v \<ge> 0"
  shows "\<parallel>n .\<otimes> m\<parallel>\<^sub>E = \<parallel>PRep n m\<parallel>\<^sub>E"
  using assms by (subst sPRep_sem.domain) auto


(*-----------*)
(*-  PPlus  -*)
(*-----------*)

definition sPPlus :: "pathExp \<Rightarrow> pathExp \<Rightarrow> pathExp" ("_ .\<oplus> _" [1000, 1000] 999) where
  "a .\<oplus> b \<equiv>
    if sEmpty b then a else
    if sEmpty a \<and> kindOf a = kindOf b then b else
    PPlus a b"

interpretation sPPlus_sem:
  semantics "a .\<oplus> b" "PPlus a b"
  unfolding sPPlus_def by (unfold_locales, auto) ((drule(1) sEmpty_sem)+, auto)+

interpretation sPPlus_props:
  pprops "a .\<oplus> b" "PPlus a b"
  unfolding sPPlus_def by unfold_locales auto

corollary sPPlus_sem_simplified [simp]:
  assumes "\<parallel>a\<parallel>\<^sub>E = x\<^sub>p" and "\<parallel>b\<parallel>\<^sub>E = y\<^sub>p" and "kindOf a = kindOf b"
  shows "\<parallel>a .\<oplus> b\<parallel>\<^sub>E = \<parallel>PPlus a b\<parallel>\<^sub>E"
  using assms by (subst sPPlus_sem.domain) auto


(*------------*)
(*-  PMinus  -*)
(*------------*)

definition sPMinus :: "pathExp \<Rightarrow> pathExp \<Rightarrow> pathExp" ("_ .\<ominus> _" [1000, 1000] 999) where
  "a .\<ominus> b \<equiv>
    if sEmpty a then a else
    if sEmpty b then a else
    PMinus a b"

interpretation sPMinus_sem:
  semantics "a .\<ominus> b" "PMinus a b"
  unfolding sPMinus_def by unfold_locales (auto dest: sEmpty_sem)

interpretation sPMinus_props:
  pprops "a .\<ominus> b" "PMinus a b"
  unfolding sPMinus_def by unfold_locales auto

corollary sPMinus_sem_simplified [simp]:
  assumes "\<parallel>a\<parallel>\<^sub>E = x\<^sub>p" and "\<parallel>b\<parallel>\<^sub>E = y\<^sub>p" and "y \<subseteq>\<^sub>* x"
  shows "\<parallel>a .\<ominus> b\<parallel>\<^sub>E = \<parallel>PMinus a b\<parallel>\<^sub>E"
  using assms by (subst sPMinus_sem.domain) auto


(*---------*)
(*-  Upd  -*)
(*---------*)

definition sUpd :: "docExp \<Rightarrow> pathExp \<Rightarrow> valueExp \<Rightarrow> docExp" ("_[_ \<rightarrow> _]." [1000, 0, 0] 1000) where
  "d[p \<rightarrow> v]. \<equiv> Upd d p v"

interpretation sUpd_sem:
  semantics "d[p \<rightarrow> v]." "Upd d p v"
  unfolding sUpd_def by unfold_locales simp

interpretation sUpd_props:
  dprops "d[p \<rightarrow> v]." "Upd d p v"
  unfolding sUpd_def by unfold_locales

corollary sUpd_sem_simplified [simp]:
  assumes "\<parallel>d\<parallel>\<^sub>E = x\<^sub>d" and "\<parallel>p\<parallel>\<^sub>E = {q}\<^sub>*\<^sub>p" and "\<parallel>v\<parallel>\<^sub>E = {w}\<^sub>*\<^sub>v"
  shows "\<parallel>d[p \<rightarrow> v].\<parallel>\<^sub>E = \<parallel>Upd d p v\<parallel>\<^sub>E"
  using assms by (subst sUpd_sem.domain) auto


(*---------*)
(*-  Del  -*)
(*---------*)

definition sDel :: "docExp \<Rightarrow> pathExp \<Rightarrow> docExp" ("_[_ \<rightarrow> ]." [1000, 0] 1000) where
  "d[p \<rightarrow> ]. \<equiv> Del d p"

interpretation sDel_sem:
  semantics "d[p \<rightarrow> ]." "Del d p"
  unfolding sDel_def by unfold_locales simp

interpretation sDel_props:
  dprops "d[p \<rightarrow> ]." "Del d p"
  unfolding sDel_def by unfold_locales

corollary sDel_sem_simplified [simp]:
  assumes "\<parallel>d\<parallel>\<^sub>E = x\<^sub>d" and "\<parallel>p\<parallel>\<^sub>E = {q}\<^sub>*\<^sub>p"
  shows "\<parallel>d[p \<rightarrow> ].\<parallel>\<^sub>E = \<parallel>Del d p\<parallel>\<^sub>E"
  using assms by (subst sDel_sem.domain) auto


(*----------*)
(*-  Path  -*)
(*----------*)

definition sPath :: "pathExp \<Rightarrow> label \<Rightarrow> valueExp \<Rightarrow> pathExp" ("_'/_[_]." [1000, 0, 0] 1000) where
  "p/l[k]. \<equiv> Path l p k"

interpretation sPath_sem:
  semantics "p/l[k]." "Path l p k"
  unfolding sPath_def by unfold_locales auto

interpretation sPath_props:
  pprops "p/l[k]." "Path l p k"
  unfolding sPath_def by unfold_locales auto

corollary sPath_sem_simplified [simp]:
  assumes "\<parallel>p\<parallel>\<^sub>E = x\<^sub>p" and "\<parallel>k\<parallel>\<^sub>E = y\<^sub>v"
  shows "\<parallel>p/l[k].\<parallel>\<^sub>E = \<parallel>Path l p k\<parallel>\<^sub>E"
  using assms by (subst sPath_sem.domain) auto


(*----------*)
(*-  Kind  -*)
(*----------*)

definition sKind :: "pathExp \<Rightarrow> label \<Rightarrow> pathExp" ("_'/_." [1000, 0] 1000) where
  "p/l. \<equiv> Kind l p"

interpretation sKind_sem:
  semantics "p/l." "Kind l p"
  unfolding sKind_def by unfold_locales auto

interpretation sKind_props:
  pprops "p/l." "Kind l p"
  unfolding sKind_def by unfold_locales auto

corollary sKind_sem_simplified [simp]:
  assumes "\<parallel>p\<parallel>\<^sub>E = x\<^sub>p"
  shows "\<parallel>p/l.\<parallel>\<^sub>E = \<parallel>Kind l p\<parallel>\<^sub>E"
  using assms by (subst sKind_sem.domain) auto


(*-----------*)
(*-  Equal  -*)
(*-----------*)

fun isFVar :: "valueExp \<Rightarrow> bool" where
  "isFVar (FVar _) = True"
| "isFVar _ = False"

fun equal_values :: "valueExp \<Rightarrow> valueExp \<Rightarrow> pred" where
  "equal_values (Val a) (Val b) = (if a = b then True. else False.)"
| "equal_values a b = Equal a b"

theorem equal_values_Sem [simp]:
  shows "\<parallel>equal_values a b\<parallel>\<^sub>E = \<parallel>Equal a b\<parallel>\<^sub>E"
  by unfold_locales (cases "(a, b)" rule: equal_values.cases, simp+)

interpretation equal_values_props:
  props "equal_values a b" "Equal a b"
  by unfold_locales (cases "(a, b)" rule: equal_values.cases, auto)+


fun isNegated :: "valueExp \<Rightarrow> bool" where
  "isNegated (Val n\<^sub>i) \<longleftrightarrow> n \<le> 0"
| "isNegated (Scal (Val n\<^sub>i) m) \<longleftrightarrow> n < 0"
| "isNegated (Sum m) \<longleftrightarrow> isNegated m"
| "isNegated _ \<longleftrightarrow> False"

(*
fun balanceEqual :: "valueExp \<Rightarrow> valueExp \<Rightarrow> pred" where
  "balanceEqual (Sum (VPlus a m)) b = (
    if isIV a \<or> isNegated a then balanceEqual (Sum m) (b -. a) else Equal (Sum (VPlus a m)) b)"
| "balanceEqual a b = (
    if isIV a \<or> isNegated a then equal_values 0. (b -. a) else equal_values (sum. a) b)"

theorem balanceEqual_sem [simp]:
  assumes "\<parallel>a\<parallel>\<^sub>E = {x}\<^sub>i\<^sub>v" and "\<parallel>b\<parallel>\<^sub>E = {y}\<^sub>i\<^sub>v"
  shows "\<parallel>balanceEqual a b\<parallel>\<^sub>E = \<parallel>Equal a b\<parallel>\<^sub>E"
  using assms by (induct a b arbitrary: x y rule: balanceEqual.induct) auto

interpretation balanceEqual_props:
  props_eq "balanceEqual a b" "Equal a b"
  by unfold_locales (induct a b rule: balanceEqual.induct, auto)+
*)

fun balanceEqual where
  "balanceEqual (Equal (Sum (VPlus a b)) c) = (
    if sType c = TInt\<^sub>! \<and> isIV a then
      if isNegated (sum. b) \<and> isNegated (c -. a) then
        Equal (a -. c) (sum. (~. b)) else
        Equal (sum. b) (c -. a) else
    Equal (Sum (VPlus a b)) c)"
| "balanceEqual x = x"

interpretation balanceEqual_sem:
  semantics "balanceEqual e" e
  by unfold_locales (cases e rule: balanceEqual.cases, auto dest!: sType_sem, subst distrib_left, auto)

interpretation balanceEqual_props:
  props_eq "balanceEqual e" e
  by unfold_locales (cases e rule: balanceEqual.cases, auto)+


fun isSize :: "valueExp \<Rightarrow> bool" where
  "isSize (Size _) = True"
| "isSize _ = False"

fun equalSize :: "valueExp \<Rightarrow> valueExp \<Rightarrow> pred" where
  "equalSize (Size m) (Val n\<^sub>i) = (
    if uve m \<and> n > 1 then False. else
    Equal (Size m) (Val n\<^sub>i))"
| "equalSize m n = Equal m n"

interpretation equalSize_sem:
  semantics "equalSize m n" "Equal m n"
  by unfold_locales (cases "(m, n)" rule: equalSize.cases, auto, drule_tac E=E in uve_sem, auto)

interpretation equalSize_props:
  props "equalSize m n" "Equal m n"
  by unfold_locales (cases "(m, n)" rule: equalSize.cases, auto)+

definition sEqual :: "valueExp \<Rightarrow> valueExp \<Rightarrow> pred" ("_ =. _" [999, 999] 999) where
  "a =. b \<equiv>
    if a = b then True. else
    if sNat a \<and> isIV b \<and> theIV b < 0 then False. else
    if isSize a \<and> isIV b then equalSize a b else
    if b \<prec> a = GT then balanceEqual (equal_values b a) else balanceEqual (equal_values a b)"

(*    if sType a = TInt\<^sub>! \<and> sType b = TInt\<^sub>! then balanceEqual (a -. b) 0. else *)

interpretation sEqual_sem:
  semantics "a =. b" "Equal a b"
  unfolding sEqual_def using balanceEqual_sem.domain and equalSize_sem.domain
  by unfold_locales (auto, cases a, auto, cases a, auto, cases a, auto,
     cases a, auto dest!: sNat_sem, drule mult_nonneg_nonneg, auto)

interpretation sEqual_props:
  props_eq "a =. b" "Equal a b"
  unfolding sEqual_def by unfold_locales (auto elim!: order_trans[rotated])

corollary sEqual_sem_simplified [simp]:
  assumes "\<parallel>a\<parallel>\<^sub>E = {x}\<^sub>*\<^sub>v" and "\<parallel>b\<parallel>\<^sub>E = {y}\<^sub>*\<^sub>v"
  shows "\<parallel>a =. b\<parallel>\<^sub>E = \<parallel>Equal a b\<parallel>\<^sub>E"
  using assms by (subst sEqual_sem.domain) auto


(*----------*)
(*-  Less  -*)
(*----------*)

fun less_values :: "valueExp \<Rightarrow> valueExp \<Rightarrow> pred" where
  "less_values (Val a) (Val b) = (if a < b then True. else False.)"
| "less_values a b = Less a b"

interpretation less_values_sem:
  semantics "less_values a b" "Less a b"
  by unfold_locales (cases "(a, b)" rule: less_values.cases, simp+)

interpretation less_values_props:
  props "less_values a b" "Less a b"
  by unfold_locales (cases "(a, b)" rule: less_values.cases, auto)+


fun balanceLeft where
  "balanceLeft (Less (Sum (VPlus a b)) c) = (
    if isIV a then
      if isNegated (sum. b) \<and> isNegated (c -. a) then
        Less (a -. c) (sum. (~. b)) else
        Less (sum. b) (c -. a) else
    Less (Sum (VPlus a b)) c)"
| "balanceLeft x = x"

interpretation balanceLeft_sem:
  semantics "balanceLeft e" e
  by unfold_locales (cases e rule: balanceLeft.cases, auto)

interpretation balanceLeft_props:
  props_eq "balanceLeft e" e
  by unfold_locales (cases e rule: balanceLeft.cases, auto)+

fun balanceRight where
  "balanceRight (Less c (Sum (VPlus a b))) = (
    if isIV a then
      if isNegated (sum. b) \<and> isNegated (c -. a) then
        Less (sum. (~. b)) (a -. c) else
        Less (c -. a) (sum. b) else
    Less c (Sum (VPlus a b)))"
| "balanceRight x = x"

interpretation balanceRight_sem:
  semantics "balanceRight e" e
  by unfold_locales (cases e rule: balanceRight.cases, auto)

interpretation balanceRight_props:
  props_eq "balanceRight e" e
  by unfold_locales (cases e rule: balanceRight.cases, auto)+

definition balanceLess where
  "balanceLess \<equiv> balanceLeft \<circ> balanceRight"

interpretation balanceLess_sem:
  semantics "balanceLess e" e
  using balanceLeft_sem.domain and balanceRight_sem.domain
  unfolding balanceLess_def by unfold_locales simp

interpretation balanceLess_props:
  props_eq "balanceLess e" e
  unfolding balanceLess_def by unfold_locales simp+


fun isCount :: "valueExp \<Rightarrow> bool" where
  "isCount (Count _ _) = True"
| "isCount _ = False"

fun countLess :: "valueExp \<Rightarrow> pred" where
  "countLess (Count a b) = (
    if sUnique a then a =. b else Less 0. (Count a b))"
| "countLess v = Less 0. v"

interpretation countLess_sem:
  semantics "countLess v" "Less 0. v"
  by unfold_locales (auto, cases v, auto dest!: sUnique_sem)

interpretation countLess_props:
  props_eq "countLess v" "Less 0. v"
  by unfold_locales (cases v, auto)+


fun sizeLess :: "valueExp \<Rightarrow> pred" where
  "sizeLess (Size m) = (
    if uve m then size. m =. 1. else Less 0. (Size m))"
| "sizeLess v = Less 0. v"

interpretation sizeLess_sem:
  semantics "sizeLess v" "Less 0. v"
  by unfold_locales (auto, cases v, auto, drule_tac E=E in uve_sem, auto)

interpretation sizeLess_props:
  props_eq "sizeLess v" "Less 0. v"
  by unfold_locales (cases v, auto)+


fun isTally :: "valueExp \<Rightarrow> bool" where
  "isTally (Tally _ _) = True"
| "isTally _ = False"

fun tallyLess :: "valueExp \<Rightarrow> pred" where
  "tallyLess (Tally t m) = (
    if uve m then tally. t m =. 1. else Less 0. (Tally t m))"
| "tallyLess v = Less 0. v"

interpretation tallyLess_sem:
  semantics "tallyLess v" "Less 0. v"
  by unfold_locales (auto, cases v, auto, drule_tac E=E in uve_sem, auto)

interpretation tallyLess_props:
  props_eq "tallyLess v" "Less 0. v"
  by unfold_locales (cases v, auto)+


definition sLess :: "valueExp \<Rightarrow> valueExp \<Rightarrow> pred" ("_ <. _" [999, 999] 999) where
  "a <. b \<equiv>
    if a = b then False. else
    if sNat a then
      if b = 0. then False. else
      if b = 1. then a =. 0. else
      if a = 0. then
        if isCount b then countLess b else
        if isSize b then sizeLess b else
        if isTally b then tallyLess b else
        balanceLess (less_values a b)
      else balanceLess (less_values a b)
    else balanceLess (less_values a b)"


interpretation sLess_sem:
  semantics "a <. b" "Less a b"
  unfolding sLess_def

apply unfold_locales
apply (split if_splits)+
apply (cut_tac ve=b and E=E in sveC)
apply (auto simp: countLess_sem.domain sizeLess_sem.domain tallyLess_sem.domain
       less_values_sem.domain balanceLess_sem.domain)
apply (auto dest!: sNat_sem)
done

interpretation sLess_props:
  props_eq "a <. b" "Less a b"
  unfolding sLess_def by unfold_locales auto

corollary sLess_sem_simplified [simp]:
  assumes "\<parallel>a\<parallel>\<^sub>E = {x}\<^sub>i\<^sub>v" and "\<parallel>b\<parallel>\<^sub>E = {y}\<^sub>i\<^sub>v"
  shows "\<parallel>a <. b\<parallel>\<^sub>E = \<parallel>Less a b\<parallel>\<^sub>E"
  using assms by (subst sLess_sem.domain) auto


(*--------------*)
(*-  Negation  -*)
(*--------------*)

fun sNegate :: "lit \<Rightarrow> lit" where
  "sNegate (L True (Equal (Size m) n)) = (
    if uve m then
      if n = 0. then L True (Equal (Size m) 1.) else
      if n = 1. then L True (Equal (Size m) 0.) else
      L False (Equal (Size m) n) 
    else L False (Equal (Size m) n))"
| "sNegate (L True (Equal (Tally t m) n)) = (
    if uve m then
      if n = 0. then L True (Equal (Tally t m) 1.) else
      if n = 1. then L True (Equal (Tally t m) 0.) else
      L False (Equal (Tally t m) n) 
    else L False (Equal (Tally t m) n))"
| "sNegate (L True (Less a b)) = (
    if isIV b then L True (Less (Val (theIV b - 1)) a) else
    if isIV a then L True (Less b (Val (theIV a + 1)))
    else L False (Less a b))"
| "sNegate (L True p) = (
    if p = True. then False. else
    if p = False. then True. else
    L False p)"
| "sNegate (L False p) = (
    if p = True. then True. else
    if p = False. then False. else
    L True p)"

(* There should be other normalizing cases. *)


theorem sNegate_sem_defined:
  assumes "\<parallel>l\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>sNegate l\<parallel>\<^sub>E = \<not>\<^sub>? \<parallel>l\<parallel>\<^sub>E"
  using assms

apply (cases l rule: sNegate.cases)
apply simp_all
apply (safe, simp_all)+
apply (auto simp del: bexp_normalize simp add: bexp_defs)
apply (drule_tac E=E in uve_sem, auto)+
done

theorem sNegate_sem_bot:
  assumes "\<parallel>l\<parallel>\<^sub>E = \<bottom>"
  shows "\<parallel>sNegate l\<parallel>\<^sub>E = \<bottom>"
  using assms

apply (cases l rule: sNegate.cases)
apply simp_all
apply (safe, simp_all)+
apply (auto simp del: bexp_normalize simp add: bexp_defs)
done

corollary sNegate_sem [simp]:
  shows "\<parallel>sNegate l\<parallel>\<^sub>E = \<not>\<^sub>? \<parallel>l\<parallel>\<^sub>E"
  by (case_tac "\<parallel>l\<parallel>\<^sub>E = \<bottom>", simp) (erule sNegate_sem_bot, erule sNegate_sem_defined)

interpretation sNegate_props:
  props "sNegate l" "l"
  by unfold_locales (cases l rule: sNegate.cases, auto)+

declare [[coercion "L True"]]


(*--------------------*)
(*-  Not Equal/Less  -*)
(*--------------------*)

definition sNotEqual :: "valueExp \<Rightarrow> valueExp \<Rightarrow> lit" ("_ \<noteq>. _" [999, 999] 999) where
  "a \<noteq>. b \<equiv> sNegate (a =. b)"

interpretation sNotEqual_sem:
  semantics "a \<noteq>. b" "L False (Equal a b)"
  unfolding sNotEqual_def using sEqual_sem.domain by (unfold_locales, auto)

interpretation sNotEqual_props:
  props_eq "a \<noteq>. b" "L False (Equal a b)"
  unfolding sNotEqual_def by unfold_locales auto

corollary sNotEqual_sem_simplified [simp]:
  assumes "\<parallel>a\<parallel>\<^sub>E = {x}\<^sub>*\<^sub>v" and "\<parallel>b\<parallel>\<^sub>E = {y}\<^sub>*\<^sub>v"
  shows "\<parallel>a \<noteq>. b\<parallel>\<^sub>E = \<parallel>L False (Equal a b)\<parallel>\<^sub>E"
  using assms by (subst sNotEqual_sem.domain) auto

(* might be too restrictive for future enhancements of sEqual

lemma domEq_commute:
  shows "a =\<^sub>? b = b =\<^sub>? a"
  by (cases a, auto, cases b, auto)  (case_tac "unique mset", auto, case_tac "unique mseta", auto)

corollary sEqual_sem_eq [simp]:
  assumes "a \<noteq> b"
  shows "\<parallel>a =. b\<parallel>\<^sub>E = \<parallel>Equal a b\<parallel>\<^sub>E"
  unfolding sEqual_def using assms by auto (rule domEq_commute)

corollary sNotEqual_sem_eq [simp]:
  assumes "a \<noteq> b"
  shows "\<parallel>a \<noteq>. b\<parallel>\<^sub>E = \<parallel>L False (Equal a b)\<parallel>\<^sub>E"
  unfolding sNotEqual_def using assms
  by (auto simp del: sNegate.simps) (case_tac "\<parallel>a\<parallel>\<^sub>E =\<^sub>? \<parallel>b\<parallel>\<^sub>E", auto)
*)


definition sNotLess :: "valueExp \<Rightarrow> valueExp \<Rightarrow> lit" ("_ \<ge>. _" [999, 999] 999) where
  "a \<ge>. b \<equiv> sNegate (a <. b)"

interpretation sNotLess_sem:
  semantics "a \<ge>. b" "L False (Less a b)"
  using sLess_sem.domain unfolding sNotLess_def by unfold_locales auto

interpretation sNotLess_props:
  props_eq "a \<ge>. b" "L False (Less a b)"
  unfolding sNotLess_def by unfold_locales auto

corollary sNotLess_sem_simplified [simp]:
  assumes "\<parallel>a\<parallel>\<^sub>E = {x}\<^sub>i\<^sub>v" and "\<parallel>b\<parallel>\<^sub>E = {y}\<^sub>i\<^sub>v"
  shows "\<parallel>a \<ge>. b\<parallel>\<^sub>E = \<parallel>L False (Less a b)\<parallel>\<^sub>E"
  using assms by (subst sNotLess_sem.domain) auto


(*-----------*)
(*-  Sugar  -*)
(*-----------*)

abbreviation lit_is_type :: "valueExp \<Rightarrow> type \<Rightarrow> pred" ("_ is. _" [999, 1000] 999) where
  "v is. t \<equiv> tally. t v =. size. v"
abbreviation lit_not_type :: "valueExp \<Rightarrow> type \<Rightarrow> lit" ("_ is.not _" [999, 1000] 999) where
  "v is.not t \<equiv> tally. t v \<noteq>. size. v"

abbreviation lit_unique :: "valueExp \<Rightarrow> pred" ("unique. _" [1000] 999) where
  "unique. v \<equiv> size. v =. 1."
abbreviation lit_proper :: "valueExp \<Rightarrow> lit" ("proper. _" [1000] 999) where
  "proper. v \<equiv> size. v \<noteq>. 1."
abbreviation lit_empty :: "valueExp \<Rightarrow> pred" ("empty. _" [1000] 999) where
  "empty. v \<equiv> size. v =. 0."


theorem lit_is_type_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>v" 
  shows "\<parallel>m is. t\<parallel>\<^sub>E = \<parallel>Equal (Tally t m) (Size m)\<parallel>\<^sub>E"
  using assms by (subst sEqual_sem_simplified) auto

theorem lit_is_not_type_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>v" 
  shows "\<parallel>m is.not t\<parallel>\<^sub>E = \<parallel>L False (Equal (Tally t m) (Size m))\<parallel>\<^sub>E"
  using assms by (subst sNotEqual_sem_simplified) auto


theorem lit_unique_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>v" 
  shows "\<parallel>unique. m\<parallel>\<^sub>E = \<parallel>Equal (Size m) 1.\<parallel>\<^sub>E"
  using assms by (subst sEqual_sem_simplified) auto

theorem lit_proper_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>v" 
  shows "\<parallel>proper. m\<parallel>\<^sub>E = \<parallel>L False (Equal (Size m) 1.)\<parallel>\<^sub>E"
  using assms by (subst sNotEqual_sem_simplified) auto

theorem lit_empty_sem_simplified [simp]:
  assumes "\<parallel>m\<parallel>\<^sub>E = x\<^sub>v" 
  shows "\<parallel>empty. m\<parallel>\<^sub>E = \<parallel>Equal (Size m) 0.\<parallel>\<^sub>E"
  using assms by (subst sEqual_sem_simplified) auto


(*-----------*)
(*-  Conds  -*)
(*-----------*)

definition sVCond :: "lit \<Rightarrow> valueExp \<Rightarrow> valueExp \<Rightarrow> valueExp" where
  "sVCond b t e \<equiv>
    if b = True.  then t else
    if b = False. then e else
    VCond b t e"

interpretation sVCond_sem:
  semantics "sVCond b t e" "VCond b t e"
  unfolding sVCond_def by unfold_locales auto

interpretation sVCond_props:
  vprops "sVCond b t e" "VCond b t e"
  unfolding sVCond_def by unfold_locales auto

definition sPCond :: "lit \<Rightarrow> pathExp \<Rightarrow> pathExp \<Rightarrow> pathExp" where
  "sPCond b t e \<equiv>
    if b = True.  then t else
    PCond b t e"

interpretation sPCond_sem:
  semantics "sPCond b t e" "PCond b t e"
  unfolding sPCond_def by unfold_locales auto

interpretation sPCond_props:
  pprops "sPCond b t e" "PCond b t e"
  unfolding sPCond_def by unfold_locales auto


definition sDCond :: "lit \<Rightarrow> docExp \<Rightarrow> docExp \<Rightarrow> docExp" where
  "sDCond b t e \<equiv>
    if b = True.  then t else
    if b = False. then e else
    DCond b t e"

interpretation sDCond_sem:
  semantics "sDCond b t e" "DCond b t e"
  unfolding sDCond_def by unfold_locales auto

interpretation sDCond_props:
  props "sDCond b t e" "DCond b t e"
  unfolding sDCond_def by unfold_locales auto


definition sRCond :: "lit \<Rightarrow> pred \<Rightarrow> pred \<Rightarrow> pred" where
  "sRCond b t e \<equiv>
    if b = True.  then t else
    if b = False. then e else
    RCond b t e"

interpretation sRCond_sem:
  semantics "sRCond b t e" "RCond b t e"
  unfolding sRCond_def by unfold_locales auto

interpretation sRCond_props:
  props "sRCond b t e" "RCond b t e"
  unfolding sRCond_def by unfold_locales auto


theorem sVCond_sem_simplified [simp]:
  shows "\<parallel>b\<parallel>\<^sub>E = True  \<Longrightarrow> \<parallel>sVCond b t e\<parallel>\<^sub>E = \<parallel>t\<parallel>\<^sub>E"
    and "\<parallel>b\<parallel>\<^sub>E = False \<Longrightarrow> \<parallel>sVCond b t e\<parallel>\<^sub>E = \<parallel>e\<parallel>\<^sub>E"
    and "\<parallel>b\<parallel>\<^sub>E = \<bottom> \<Longrightarrow> \<parallel>sVCond b t e\<parallel>\<^sub>E = \<bottom>"
  using assms unfolding sVCond_def by auto

theorem sPCond_sem_simplified [simp]:
  assumes "kindOf t = kindOf e"
  shows "\<parallel>b\<parallel>\<^sub>E = True  \<Longrightarrow> \<parallel>sPCond b t e\<parallel>\<^sub>E = \<parallel>t\<parallel>\<^sub>E"
    and "\<parallel>b\<parallel>\<^sub>E = False \<Longrightarrow> \<parallel>sPCond b t e\<parallel>\<^sub>E = \<parallel>e\<parallel>\<^sub>E"
    and "\<parallel>b\<parallel>\<^sub>E = \<bottom> \<Longrightarrow> \<parallel>sPCond b t e\<parallel>\<^sub>E = \<bottom>"
  using assms unfolding sPCond_def by auto

theorem sDCond_sem_simplified [simp]:
  shows "\<parallel>b\<parallel>\<^sub>E = True  \<Longrightarrow> \<parallel>sDCond b t e\<parallel>\<^sub>E = \<parallel>t\<parallel>\<^sub>E"
    and "\<parallel>b\<parallel>\<^sub>E = False \<Longrightarrow> \<parallel>sDCond b t e\<parallel>\<^sub>E = \<parallel>e\<parallel>\<^sub>E"
    and "\<parallel>b\<parallel>\<^sub>E = \<bottom> \<Longrightarrow> \<parallel>sDCond b t e\<parallel>\<^sub>E = \<bottom>"
  using assms unfolding sDCond_def by auto

theorem sRCond_sem_simplified [simp]:
  shows "\<parallel>b\<parallel>\<^sub>E = True  \<Longrightarrow> \<parallel>sRCond b t e\<parallel>\<^sub>E = \<parallel>t\<parallel>\<^sub>E"
    and "\<parallel>b\<parallel>\<^sub>E = False \<Longrightarrow> \<parallel>sRCond b t e\<parallel>\<^sub>E = \<parallel>e\<parallel>\<^sub>E"
    and "\<parallel>b\<parallel>\<^sub>E = \<bottom> \<Longrightarrow> \<parallel>sRCond b t e\<parallel>\<^sub>E = \<bottom>"
  using assms unfolding sRCond_def by auto


(*---------*)
(*-  Lit  -*)
(*---------*)

definition sL :: "bool \<Rightarrow> pred \<Rightarrow> lit" where
  "sL b p \<equiv> if b then p else sNegate p"

theorem sL_sem [simp]:
  shows "\<parallel>sL b p\<parallel>\<^sub>E = \<parallel>L b p\<parallel>\<^sub>E"
  unfolding sL_def by simp

interpretation sL_props:
  props "sL b p" "L b p"
  unfolding sL_def by unfold_locales auto


(*----------*)
(*-  Disj  -*)
(*----------*)

fun takeoutL :: "lit \<Rightarrow> disj \<Rightarrow> disj" where
  "takeoutL m DFlse   = DFlse"
| "takeoutL m (D l d) = (if l = m then takeoutL m d else D l (takeoutL m d))"

theorem sem_takeoutL_eq_false [simp]:
  assumes "\<parallel>l\<parallel>\<^sub>E = False"
  shows "\<parallel>takeoutL l d\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E"
  using assms by (induct l d rule: takeoutL.induct) auto

theorem sem_takeoutL_eq_true [simp]:
  assumes "\<parallel>d\<parallel>\<^sub>E = True" and "\<parallel>l\<parallel>\<^sub>E \<noteq> True"
  shows "\<parallel>takeoutL l d\<parallel>\<^sub>E = True"
  using assms by (induct l d rule: takeoutL.induct) auto

theorem sem_takeoutL_true [dest]:
  assumes "\<parallel>takeoutL l d\<parallel>\<^sub>E = True"
  shows "\<parallel>d\<parallel>\<^sub>E = True"
  using assms by (induct l d rule: takeoutL.induct) (auto, split if_splits, auto)

interpretation takeoutL_props:
  props "takeoutL l d" d
  by (unfold_locales, auto elim!: order_trans[rotated]) (induct l d rule: takeoutL.induct, auto)+

definition sD :: "lit \<Rightarrow> disj \<Rightarrow> disj" where
  "sD l d \<equiv>
    if l = False. then d else
    if l = True. then True. else
    if d = True. then True. else D l (takeoutL l d)"

theorem sD_sem [simp]:
  shows "\<parallel>sD l d\<parallel>\<^sub>E = \<parallel>D l d\<parallel>\<^sub>E"
  unfolding sD_def by auto (rule sbeI, auto)+

interpretation sD_props:
  props "sD l d" "D l d"
  unfolding sD_def by unfold_locales (auto elim!: order_trans[rotated])


(*----------*)
(*-  Conj  -*)
(*----------*)

fun disjunctOf :: "disj \<Rightarrow> lit \<Rightarrow> bool" where
  "disjunctOf (D l d) m \<longleftrightarrow> l = m \<or> disjunctOf d m"
| "disjunctOf DFlse   m \<longleftrightarrow> False"

fun subsumes :: "disj \<Rightarrow> disj \<Rightarrow> bool" where
  "subsumes e (D l d) \<longleftrightarrow> disjunctOf e l \<and> subsumes e d"
| "subsumes e DFlse   \<longleftrightarrow> True"

fun takeoutD :: "disj \<Rightarrow> conj \<Rightarrow> conj" where
  "takeoutD m CTrue   = CTrue"
| "takeoutD m (C d c) = (if subsumes d m then takeoutD m c else C d (takeoutD m c))"

theorem sem_disjunctOf_false [dest]:
  assumes "disjunctOf d l" and "\<parallel>d\<parallel>\<^sub>E = False"
  shows "\<parallel>l\<parallel>\<^sub>E = False"
  using assms by (induct d l rule: disjunctOf.induct) auto
 
theorem sem_disjunctOf_true [dest]:
  assumes "disjunctOf d l" and "\<parallel>l\<parallel>\<^sub>E = True"
  shows "\<parallel>d\<parallel>\<^sub>E = True"
  using assms by (induct d l rule: disjunctOf.induct) auto

theorem sem_subsumes_false [dest]:
  assumes "subsumes e d" and "\<parallel>e\<parallel>\<^sub>E = False"
  shows "\<parallel>d\<parallel>\<^sub>E = False"
  using assms by (induct e d rule: subsumes.induct) auto

theorem sem_subsumes_true [dest]:
  assumes "subsumes e d" and "\<parallel>d\<parallel>\<^sub>E = True"
  shows "\<parallel>e\<parallel>\<^sub>E = True"
  using assms by (induct e d rule: subsumes.induct) auto

theorem sem_takeoutD_eq_true [simp]:
  assumes "\<parallel>c\<parallel>\<^sub>E = True"
  shows "\<parallel>takeoutD d c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by (induct d c rule: takeoutD.induct) auto

theorem sem_takeoutD_neutral [simp]:
  assumes "\<parallel>d\<parallel>\<^sub>E = True"
  shows "\<parallel>takeoutD d c\<parallel>\<^sub>E = \<parallel>c\<parallel>\<^sub>E"
  using assms by (induct d c rule: takeoutD.induct) auto

theorem sem_takeoutD_eq_false [simp]:
  assumes "\<parallel>c\<parallel>\<^sub>E = False" and "\<parallel>d\<parallel>\<^sub>E \<noteq> False"
  shows "\<parallel>takeoutD d c\<parallel>\<^sub>E = False"
  using assms by (induct d c rule: takeoutD.induct) auto

theorem sem_takeoutD_false [dest]:
  assumes "\<parallel>takeoutD d c\<parallel>\<^sub>E = False"
  shows "\<parallel>c\<parallel>\<^sub>E = False"
  using assms by (induct d c rule: takeoutD.induct) (auto, split if_splits, auto)

interpretation takeoutD_props:
  props "takeoutD d c" c
  by (unfold_locales, auto elim!: order_trans[rotated]) (induct d c rule: takeoutD.induct, auto)+


definition sC :: "disj \<Rightarrow> conj \<Rightarrow> conj" where
  "sC d c \<equiv>
    if d = True. then c else
    if d = False. then False. else
    if c = False. then False. else C d (takeoutD d c)"

theorem sC_sem [simp]:
  shows "\<parallel>sC d c\<parallel>\<^sub>E = \<parallel>C d c\<parallel>\<^sub>E"
  unfolding sC_def by auto (rule sbeI, auto)+

interpretation sC_props:
  props "sC d c" "C d c"
  unfolding sC_def by unfold_locales (auto elim!: order_trans[rotated])

end

