header {* Value, Integer and Path Multisets *}

theory VIPSet imports Multisets Documents begin

declare count_cong [intro!]  (* we need to break the abstraction again *)


(*-------------------------------------*)
(*- Value/Integer Multiset Conversion -*)
(*-------------------------------------*)

abbreviation asIntM :: "value mset \<Rightarrow> int mset" where
  "asIntM \<equiv> mapM asInt"

theorem asInt_inject_inj [simp]:
  shows "partial_inj asInt"
  unfolding partial_inj by auto

theorem asInt_inject_inv [simp]:
  shows "partial_inv asInt = VInt"
  unfolding partial_inv by auto

theorem asInt_ran [simp]:
  shows "ran asInt = UNIV"
  unfolding ran_def by auto

theorem asInt_dom [simp]:
  shows "dom asInt = ints"
  by auto


abbreviation VIntM :: "int mset \<Rightarrow> value mset" where
  "VIntM \<equiv> mapM (\<lambda>x. x\<^sub>i\<^sub>!)"  (* mapM (Some \<circ> VInt) *)

theorem VInt_inject_inj [simp]:
  shows "partial_inj (\<lambda>x. x\<^sub>i\<^sub>!)"
  unfolding partial_inj by auto

theorem VInt_inject_inv [simp]:
  shows "partial_inv (\<lambda>x. x\<^sub>i\<^sub>!) = inv VInt"
  unfolding partial_inv and inv_def by auto

theorem VInt_ran [simp]:
  shows "ran (\<lambda>x. x\<^sub>i\<^sub>!) = dom asInt"
  unfolding ran_def and dom_def by auto

theorem VInt_dom [simp]:
  shows "dom (\<lambda>x. x\<^sub>i\<^sub>!) = UNIV"
  by auto


abbreviation singleIntValue :: "int \<Rightarrow> value mset" ("{_}\<^sub>i" [0] 1000) where
  "{i}\<^sub>i \<equiv> {i\<^sub>i}\<^sub>*"

abbreviation hasTypeM :: "type \<Rightarrow> value mset \<Rightarrow> bool" where
  "hasTypeM t m \<equiv> \<forall>e. e \<in>\<^sub>* m \<longrightarrow> typeOf e = t"


(* Auxiliary Lemmas *)

lemma asIntM_size [simp]:
  shows "sizeM (asIntM y) = sizeM y\<down>ints"
  by (subst mapM_size_inject) simp+

lemma inv_VInt_apply [simp]:
  shows "inv VInt e\<^sub>i = e"
  unfolding inv_def by simp

theorem asInt_VInt_cancel [simp]:
  shows "asIntM (VIntM m) = m"
  by auto

theorem VInt_asInt_cancel [simp]:
  assumes "hasTypeM TInt m"
  shows "VIntM (asIntM m) = m"
  using assms by auto (case_tac e, auto)

lemma VInt_reduce_isInt [simp]:
  shows "(VIntM m)\<down>ints = VIntM m"
  by (rule reduceM_superset) auto

lemma VInt_size [simp]:
  shows "sizeM (VIntM m) = sizeM m"
  by (metis VInt_reduce_isInt asInt_VInt_cancel asInt_dom asInt_inject_inj mapM_size_inject)

lemma asInt_size [simp]:
  assumes "hasTypeM TInt m"
  shows "sizeM (asIntM m) = sizeM m"
  using assms and inj_on_def[of VInt] image_def[of VInt] unfolding sizeM_def
  by auto (rule setsum_reindex_cong[of VInt, symmetric], auto)


(*--------------*)
(*- Sum & Scal -*)
(*--------------*)

definition sumM :: "int mset \<Rightarrow> int" where  (* sound exactly for finite multisets *)
  "sumM m \<equiv> setsum (\<lambda>e. e * int (count m e)) (elemset m)"

abbreviation scalM :: "int \<Rightarrow> int mset \<Rightarrow> int mset" ("_ \<odot> _" [101, 895] 895) where
  "n \<odot> m \<equiv> mapM (\<lambda>x. (n * x)\<^sub>!) m"  (* mapM (Some \<circ> op * n) *)

theorem scal_inject_inj [simp]:
  fixes n :: int assumes "n \<noteq> 0"
  shows "partial_inj (\<lambda>x. (n * x)\<^sub>!)"
  using assms unfolding partial_inj by auto

theorem scal_inject_inv [simp]:
  fixes n :: int assumes "n \<noteq> 0"
  shows "partial_inv (\<lambda>x. (n * x)\<^sub>!) = inv (op * n)"
  using assms unfolding partial_inv and inv_def by auto

lemma scal_inject_inv_general [simp]:
  fixes n :: int assumes "n \<noteq> 0"
  shows "inv (op * n) (n * x) = x"
  using assms unfolding inv_def by simp

lemma scal_inject_inv_zero [simp]:
  fixes n :: int assumes "n \<noteq> 0"
  shows "inv (op * n) 0 = 0"
  using assms unfolding inv_def by simp

theorem scal_ran [simp]:
  shows "ran (\<lambda>x. (n * x)\<^sub>!) = range (op * n)"
  unfolding ran_def and dom_def by auto

theorem scal_constant [simp]:
  shows "partial_con (\<lambda>x. (0 :: int)\<^sub>!)"
  unfolding partial_con by simp

theorem scal_constant_the [simp]:
  shows "partial_the (\<lambda>x. (0 :: int)\<^sub>!) = 0"
  unfolding partial_the and ran_def by simp


theorem scal_dom_zero [simp]:
  fixes n :: int
  shows "dom (\<lambda>x. (n * x)\<^sub>!) = UNIV"
  by auto


(* Theorems *)

theorem eleM_scal [simp]:
  assumes "finiteM m"
  shows "e \<in>\<^sub>* 0 \<odot> m \<longleftrightarrow> e = 0 \<and> m \<noteq> {}\<^sub>*"
  using assms by auto

theorem zeroM_scal_result [simp]:
  assumes "finiteM m"
  shows "n \<odot> m = {}\<^sub>* \<longleftrightarrow> m = {}\<^sub>*"
  using assms unfolding mapM_def
  by auto (drule_tac f=count in arg_cong, auto, drule_tac x="n * e" in fun_cong, auto)

theorem elemset_scal [simp]:
  assumes "finiteM m"
  shows "elemset (n \<odot> m) = (op * n) ` elemset m"
  using assms unfolding image_def by (auto simp: ran_def restrict_map_def)

theorem sizeM_scal [simp]:
  shows "sizeM (n \<odot> m) = sizeM m"
  by (cases "n = 0", auto) (subst mapM_size_inject, simp+)

theorem scalM_one [simp]:
  shows "1 \<odot> m = m"
  unfolding mapM_def by auto (case_tac "e \<in>\<^sub>* m", subst setsum_diff1'[of _ e], auto)


theorem scalM_assoc [simp]:
  "a \<odot> (b \<odot> m) = (a * b) \<odot> m"
proof (cases "a = 0")
  case True thus ?thesis by clarify (case_tac "e = 0", simp+) next
  case False show ?thesis
  proof (cases "b = 0")
    case True thus ?thesis using `a \<noteq> 0` 
      by clarify (case_tac "e = 0", simp+, case_tac "e \<in> ran (\<lambda>x. (a * x)\<^sub>!)", auto) next
    case False show ?thesis
    proof clarify fix e show "count (a \<odot> (b \<odot> m)) e = count ((a * b) \<odot> m) e"
      proof (cases "e \<in> ran (\<lambda>x. (a * x)\<^sub>!)")
        case True thus ?thesis using `a \<noteq> 0` and `b \<noteq> 0`
          by (auto, case_tac "x \<in> ran (\<lambda>x. (b * x)\<^sub>!)")
             (auto simp: image_def mult_assoc[symmetric]) next
        case False hence "e \<notin> ran (\<lambda>x. (a * b * x)\<^sub>!)" by auto thus ?thesis using False by simp
      qed
    qed
  qed
qed


theorem zeroM_sum [simp]:
  shows "sumM {}\<^sub>* = 0"
  unfolding sumM_def by simp

theorem plusM_sum_homomorphism [simp]:
  assumes "finiteM a" and "finiteM b"
  shows "sumM (a \<oplus> b) = sumM a + sumM b"
proof -
  have "setsum (\<lambda>e. e * int (count a e)) {e. e \<in>\<^sub>* a \<or> e \<in>\<^sub>* b} =
        setsum (\<lambda>e. e * int (count a e)) (elemset a)"
   and "setsum (\<lambda>e. e * int (count b e)) {e. e \<in>\<^sub>* a \<or> e \<in>\<^sub>* b} =
        setsum (\<lambda>e. e * int (count b e)) (elemset b)" using assms
    by - (rule setsum_mono_zero_right, auto)+
  thus ?thesis using assms unfolding sumM_def
    by (simp, subst ring_distribs) (subst setsum_addf, auto)
qed

theorem minusM_sum_homomorphism [simp]:
  assumes "b \<subseteq>\<^sub>* a" and "finiteM a"
  shows "sumM (a \<ominus> b) = sumM a - sumM b"
proof -
  have "setsum (\<lambda>e. e * int (count a e - count b e)) {e. count b e < count a e} =
        setsum (\<lambda>e. e * int (count a e - count b e)) (elemset a)"
   and "setsum (\<lambda>e. e * int (count b e)) (elemset b) = setsum (\<lambda>e. e * int (count b e)) (elemset a)"
  using assms by auto (rule setsum_mono_zero_left, auto, (metis nat_less_le)?)+
  thus ?thesis using assms unfolding sumM_def
    by (simp, subst right_diff_distrib) (subst setsum_subtractf, simp)
qed


theorem singleM_sum [simp]:
  shows "sumM {e}\<^sub>* = e"
  unfolding sumM_def by simp


theorem scalM_sum_homomorphism [simp]:
  shows "sumM (n \<odot> m) = n * sumM m"
proof (cases "n = 0")
  case True thus ?thesis unfolding sumM_def by (auto intro!: setsum_0') next
  case False thus ?thesis unfolding sumM_def and setsum_right_distrib
    by auto (rule setsum_reindex_cong[of "op * n"], auto simp: inj_on_def ran_def restrict_map_def)
qed

theorem repM_sum_homomorphism [simp]:
  shows "sumM (n \<otimes> m) = int n * sumM m"
proof (cases "n = 0")
  case True thus ?thesis by simp next
  case False thus ?thesis unfolding sumM_def
    by (subst setsum_right_distrib) (rule setsum_cong, auto, metis int_mult)
qed


(*---------------------------*)
(*- Key Multiset Conversion -*)
(*---------------------------*)

abbreviation asKeyM :: "value mset \<Rightarrow> key mset" where
  "asKeyM \<equiv> mapM asKey"

theorem asKey_inject_inj [simp]:
  shows "partial_inj asKey"
  unfolding partial_inj by auto

theorem asKey_inject_inv [simp]:
  shows "partial_inv asKey = VKey"
  unfolding partial_inv by simp

theorem asKey_ran [simp]:
  shows "ran asKey = UNIV"
  unfolding ran_def by simp

theorem asKey_dom [simp]:
  shows "dom asKey = keys"
  by auto


lemma asKey_from_int [simp]:
  shows "asKeyM (VIntM m) = {}\<^sub>*"
  by auto

lemma asKey_size [simp]:
  assumes "hasTypeM TKey m"
  shows "sizeM (asKeyM m) = sizeM m"
  using assms and inj_on_def[of VKey] image_def[of VKey] unfolding sizeM_def
  by auto (rule setsum_reindex_cong[of VKey, symmetric], auto)


(*------------*)
(*- Children -*)
(*------------*)

no_notation divide (infixl "'/" 70)

definition children :: "path mset \<Rightarrow> label \<Rightarrow> key mset \<Rightarrow> path mset" ("_'/_[_]\<^sub>*" [1000, 0, 0] 1000) where
  "p/l[k]\<^sub>* \<equiv> collect (\<lambda>x. case x of
      Root \<Rightarrow> 0 | q/m[r] \<Rightarrow> if l = m then count p q * count k r else 0)"

theorem count_children [simp]:
  shows "Root \<notin>\<^sub>* p/l[k]\<^sub>*"
    and  "l \<noteq> m \<Longrightarrow> q/m[r] \<notin>\<^sub>* p/l[k]\<^sub>*"
    and "count p/l[k]\<^sub>* q/l[r] = count p q * count k r"
  using assms unfolding children_def by simp+


(* Theorems *)

theorem eleM_children [simp]:
  shows "p/l[k] \<in>\<^sub>* q/m[r]\<^sub>* \<longleftrightarrow> p \<in>\<^sub>* q \<and> l = m \<and> k \<in>\<^sub>* r"
  unfolding children_def by (case_tac "l = m") auto

theorem elemset_children [simp]:
  shows "elemset p/l[k]\<^sub>* = {q/l[r] | q r. q \<in>\<^sub>* p \<and> r \<in>\<^sub>* k}"
  unfolding children_def by (auto, case_tac x, auto) (split if_splits, simp+)+

theorem subsetM_children [simp]:
  assumes "p \<subseteq>\<^sub>* q" and "k \<subseteq>\<^sub>* r"
  shows "p/l[k]\<^sub>* \<subseteq>\<^sub>* q/l[r]\<^sub>*"
  using assms unfolding children_def by (auto, case_tac e, auto) (rule mult_le_mono, simp+)

theorem zeroM_children_left [simp]:
  shows "{}\<^sub>*/l[k]\<^sub>* = {}\<^sub>*"
  unfolding children_def by auto (case_tac e, auto)

theorem zeroM_children_right [simp]:
  shows "p/l[{}\<^sub>*]\<^sub>* = {}\<^sub>*"
  unfolding children_def by auto (case_tac e, auto)

theorem plusM_children_homomorphism_left [simp]:
  shows "(a \<oplus> b)/l[k]\<^sub>* = a/l[k]\<^sub>* \<oplus> b/l[k]\<^sub>*"
  unfolding children_def by auto (case_tac e, auto, rule add_mult_distrib)

theorem plusM_children_homomorphism_right [simp]:
  shows "p/l[a \<oplus> b]\<^sub>* = p/l[a]\<^sub>* \<oplus> p/l[b]\<^sub>*"
  unfolding children_def by auto (case_tac e, auto, rule add_mult_distrib2)

theorem minusM_children_homomorphism_left [simp]:
  shows "(a \<ominus> b)/l[k]\<^sub>* = a/l[k]\<^sub>* \<ominus> b/l[k]\<^sub>*"
  unfolding children_def by auto (case_tac e, auto, rule diff_mult_distrib)

theorem minusM_children_homomorphism_right [simp]:
  shows "p/l[a \<ominus> b]\<^sub>* = p/l[a]\<^sub>* \<ominus> p/l[b]\<^sub>*"
  unfolding children_def by auto (case_tac e, auto, rule diff_mult_distrib2)


theorem reduceM_isKey_size [simp]:
  shows "sizeM m\<down>keys = sizeM (asKeyM m)"
  using inj_on_def unfolding sizeM_def by (auto intro!: setsum_reindex_cong)

theorem reduceM_isKey_asKey [simp]:
  shows "asKeyM m\<down>keys = asKeyM m"
  by clarify (case_tac "e \<in> ran asKey", auto)


theorem sizeM_children [simp]:
  shows "sizeM p/l[k]\<^sub>* = sizeM p * sizeM k" 
proof -
  have "sizeM p * sizeM k = (\<Sum>i\<in>elemset p. \<Sum>j\<in>elemset k. count p i * count k j)"
    unfolding sizeM_def by (rule setsum_product) also
  have "... = (\<Sum>(x, y)\<in>elemset p \<times> elemset k. count p x * count k y)"
    by (rule setsum_cartesian_product) finally
  show ?thesis unfolding sizeM_def 
    by simp (rule setsum_reindex_cong[of "\<lambda>(x, y). x/l[y]"], auto simp: inj_on_def)
qed


theorem singleM_children [simp]:
  shows "{p}\<^sub>*/l[{k}\<^sub>*]\<^sub>* = {p/l[k]}\<^sub>*"
  unfolding children_def and singleM_def
  by auto (case_tac e, auto)

theorem singleM_children_result_matched [dest!]:
  assumes "p/l[k]\<^sub>* = {q/m[r]}\<^sub>*"
  shows "p = {q}\<^sub>* \<and> l = m \<and> k = {r}\<^sub>*"
proof -
  from assms have elem: "count p/l[k]\<^sub>* q/m[r] = 1" by simp
  hence "l = m" by - (rule classical, simp) moreover
  hence "count p q = 1" and "count k r = 1" using elem unfolding children_def by simp+ moreover
  hence "\<forall>x. x \<noteq> q \<longrightarrow> x \<notin>\<^sub>* p" and "\<forall>x. x \<noteq> r \<longrightarrow> x \<notin>\<^sub>* k"
    using assms unfolding children_def and singleM_def
    by auto (drule_tac x="x/l[r]" in fun_cong, simp, drule_tac x="q/l[x]" in fun_cong, simp)
  ultimately show ?thesis unfolding singleM_def by auto
qed

theorem singleM_children_result_root [simp]:
  shows "p/l[k]\<^sub>* \<noteq> {Root}\<^sub>*"
  unfolding children_def and singleM_def by auto (drule_tac x=Root in fun_cong, simp)

corollary singleM_children_result [dest!]:
  assumes "p/l[k]\<^sub>* = {x}\<^sub>*"
  shows "\<exists>q r. p = {q}\<^sub>* \<and> k = {r}\<^sub>* \<and> x = q/l[r]"
  using assms by (cases x) auto


theorem repM_children_homomorphism_left [simp]:
  shows "(n \<otimes> p)/l[k]\<^sub>* = n \<otimes> p/l[k]\<^sub>*"
  unfolding children_def by auto (case_tac e, auto)

theorem repM_children_homomorphism_right [simp]:
  shows "p/l[n \<otimes> k]\<^sub>* = n \<otimes> p/l[k]\<^sub>*"
  unfolding children_def by auto (case_tac e, auto)


(* Reductions & Kinds/Path/Children *)

abbreviation hasKindM where
  "hasKindM k m \<equiv> hasKind k (elemset m)"

theorem reduceM_desc_kind_eq [simp]:
  assumes "kindOf p = k\<^sub>!" and "hasKindM k q"
  shows "q\<down>(desc p) = q\<down>{p}"
proof
  fix e show "count q\<down>(desc p) e = count q\<down>{p} e"
  proof (cases "e \<in>\<^sub>* q \<and> e \<noteq> p")
    case False thus ?thesis by auto (metis eleM_reduce_not) next
    case True have "e \<notin> desc p" proof (rule notI)
      assume "e \<in> desc p" hence "kindOf e \<noteq> k\<^sub>!" and "kindOf e = k\<^sub>!" using assms and True by auto (metis anc_kind)
      thus False ..
    qed
    thus ?thesis using True by simp
  qed
qed

theorem reduceM_desc_kind_neq [simp]:
  assumes "kindOf p = j\<^sub>!" and "hasKindM i q" and "Step l i \<noteq> j"
  shows "q/l[k]\<^sub>*\<down>(desc p) = (q\<down>(desc p))/l[k]\<^sub>*"
proof
  fix e show "count q/l[k]\<^sub>*\<down>(desc p) e = count (q\<down>(desc p))/l[k]\<^sub>* e"
  proof (cases "e \<in>\<^sub>* q/l[k]\<^sub>*")
    case False thus ?thesis unfolding count_reduce
      by auto (erule contrapos_pp, auto, case_tac e, auto) next
    case True then obtain r and m where "e = r/l[m]" and "r \<in>\<^sub>* q" and "m \<in>\<^sub>* k" by (case_tac e) auto
    hence "count q/l[k]\<^sub>*\<down>(desc p) r/l[m] = count q\<down>(desc p) r * count k m"
    proof (cases "r/l[m] \<in> desc p")
      case False hence "r \<notin> desc p" by clarsimp thus ?thesis using False by simp next
      case True hence "r/l[m] = p \<or> r \<in> desc p" by clarsimp thus ?thesis proof
        assume "r/l[m] = p"
          hence "kindOf r/l[m] = j\<^sub>!" using assms by simp moreover
          have "kindOf r/l[m] \<noteq> j\<^sub>!" using assms and `r \<in>\<^sub>* q` by simp ultimately
          show ?thesis by simp next
        assume "r \<in> desc p" thus ?thesis by simp
      qed
    qed thus ?thesis using `e = r/l[m]` by simp
  qed
qed


(*-----------------*)
(*- Multiset Read -*)
(*-----------------*)

abbreviation readM :: "doc \<Rightarrow> path mset \<Rightarrow> value mset" ("_'(_')\<^sub>*" [1000, 0] 1000) where
  "d(m)\<^sub>* \<equiv> mapM (doc_read d) m"


(* Theorems *)

theorem singleM_read_empty [simp]:
  shows "d({p}\<^sub>*)\<^sub>* = {}\<^sub>* \<longleftrightarrow> p \<notin>\<^sub>1 d"
  by auto

theorem reduceM_single_elem [simp]:
  assumes "e \<notin>\<^sub>* m"
  shows "m\<down>{e} = {}\<^sub>*"
  using assms unfolding mapM_def
  by auto (rule setsum.neutral, auto, rule map_sound_finite_dom[simplified], simp)

theorem mapM_mono_none [simp]:
  assumes "f p = None"
  shows "mapM f m\<down>{p} = {}\<^sub>*"
  using assms by (fold singleM_rep_count_reduce) (subst mapM_rep_homomorphism, simp)

theorem mapM_mono_some [simp]:
  assumes "f p = x\<^sub>!"
  shows "mapM f m\<down>{p} = count m p \<otimes> {x}\<^sub>*"
  using assms by (fold singleM_rep_count_reduce) (subst mapM_rep_homomorphism, simp)


(* Updates *)

theorem readM_upd_unchanged [simp]:
  assumes "finiteD d" and "d(p)\<^sub>1 = v\<^sub>!"
  shows "d[p \<rightarrow> v](q)\<^sub>* = d(q)\<^sub>*"
  using assms unfolding mapM_def and doc_upd_def by auto (rule setsum_cong, auto)

lemma readM_upd_excluded [simp]:
  assumes "finiteD d" and "p \<notin>\<^sub>* q"
  shows "d[p \<rightarrow> v](q)\<^sub>* = d(q)\<^sub>*"
  using assms unfolding mapM_def and doc_upd_def by auto (rule setsum_cong, auto)

theorem readM_upd_added [simplified, simp]:
  assumes "finiteD d[p \<rightarrow> v]" and "d(p)\<^sub>1 \<noteq> v\<^sub>!"
  shows "count d[p \<rightarrow> v](q)\<^sub>* v = count d(q)\<^sub>* v + count q p"
proof (cases "p \<in>\<^sub>* q")
  case False thus ?thesis using assms by simp next
  case True have "finite {x. x \<noteq> p \<longrightarrow> x \<in>\<^sub>* q \<and> d(x)\<^sub>1 = v\<^sub>!}" using assms
    by - (erule rev_finite_subset, auto, case_tac "x = p", auto)
  thus ?thesis using assms and True unfolding mapM_def and doc_upd_def
    by auto (subst setsum_diff1'[of _ p], auto, rule setsum_cong, auto)
qed

theorem readM_upd_removed [simp]:
  assumes "finiteD d" and "d(p)\<^sub>1 = e\<^sub>!" and "e \<noteq> v"
  shows "count d[p \<rightarrow> v](q)\<^sub>* e = count d(q)\<^sub>* e - count q p"
proof (cases "p \<in>\<^sub>* q")
  case False thus ?thesis using assms by simp next
  case True thus ?thesis using assms and True unfolding mapM_def and doc_upd_def
    by auto (subst (2) setsum_diff1'[of _ p], auto, rule setsum_cong, auto)
qed

theorem readM_upd_other [simp]:
  assumes "finiteD d" and "d(p)\<^sub>1 \<noteq> e\<^sub>!" and "e \<noteq> v"
  shows "count d[p \<rightarrow> v](q)\<^sub>* e = count d(q)\<^sub>* e"
  using assms unfolding mapM_def and doc_upd_def by auto (rule setsum_cong, auto)

corollary readM_upd:
  assumes "finiteD d"
  shows "d[p \<rightarrow> v](q)\<^sub>* = d(q)\<^sub>* \<ominus> d(q\<down>{p})\<^sub>* \<oplus> (count q p \<otimes> {v}\<^sub>*)"
  using assms by (auto, cases "d(p)\<^sub>1", auto)
    (case_tac "v = e", auto, case_tac "v = a", auto, case_tac "a = e", auto, case_tac "e = v", auto)

corollary readM_upd_natural:
  assumes "finiteD d"
  shows "d[p \<rightarrow> v](q)\<^sub>* = d(q)\<^sub>* \<ominus> d(q\<down>{p})\<^sub>* \<oplus> (sizeM q\<down>{p} \<otimes> {v}\<^sub>*)"
  using assms by (fold singleM_rep_count_reduce, subst sizeM_rep_homomorphism, simp) (erule readM_upd)


(* Deletion *)

theorem readM_del:
  assumes "finiteD d"
  shows "d[p \<rightarrow> ](q)\<^sub>* = d(q)\<^sub>* \<ominus> d(q\<down>(desc p))\<^sub>*"
proof
  fix e
  have "count d[p \<rightarrow> ](q)\<^sub>* e =
        setsum (count q) ({x. x \<in>\<^sub>* q \<and> d(x)\<^sub>1 = e\<^sub>!} - {x. x \<in>\<^sub>* q \<and> p \<in> anc x \<and> d(x)\<^sub>1 = e\<^sub>!})"
    unfolding mapM_def and doc_del_def by auto (rule setsum_cong, auto) also
  have "... = setsum (count q) {x. x \<in>\<^sub>* q \<and> d(x)\<^sub>1 = e\<^sub>!} 
            - setsum (count q) {x. x \<in>\<^sub>* q \<and> p \<in> anc x \<and> d(x)\<^sub>1 = e\<^sub>!}"
    using assms by (subst setsum_diff_nat) auto finally
  show "count d[p \<rightarrow> ](q)\<^sub>* e = count (d(q)\<^sub>* \<ominus> d(q\<down>(desc p))\<^sub>*) e"
    by simp (subst mapM_def, subst mapM_def, auto)
qed


declare count_cong [rule del]  (* re-establish multisets as abstraction *)

end

